Package: distributional
Title: Vectorised Probability Distributions
Version: 0.7.0
Authors@R: 
    c(person(given = "Mitchell",
             family = "O'Hara-Wild",
             role = c("aut", "cre"),
             email = "mail@mitchelloharawild.com",
             comment = c(ORCID = "0000-0001-6729-7695")),
      person(given = "Matthew",
             family = "Kay",
             role = c("aut"),
             comment = c(ORCID = "0000-0001-9446-0419")),
      person(given = "Alex",
             family = "Hayes",
             role = c("aut"),
             comment = c(ORCID = "0000-0002-4985-5160")),
      person(given = "Rob",
             family = "Hyndman",
             role = c("aut"),
             comment = c(ORCID = "0000-0002-2140-5352")),
      person(given = "Earo",
             family = "Wang",
             role = c("ctb"),
             comment = c(ORCID = "0000-0001-6448-5260")),
      person(given = "Vencislav",
             family = "Popov",
             role = c("ctb"),
             comment = c(ORCID = "0000-0002-8073-4199")))
Description: Vectorised distribution objects with tools for manipulating, 
    visualising, and using probability distributions. Designed to allow model
    prediction outputs to return distributions rather than their parameters, 
    allowing users to directly interact with predictive distributions in a
    data-oriented workflow. In addition to providing generic replacements for
    p/d/q/r functions, other useful statistics can be computed including means,
    variances, intervals, and highest density regions.
License: GPL-3
Depends: R (>= 4.0.0)
Imports: vctrs (>= 0.3.0), rlang (>= 0.4.5), generics, stats, numDeriv,
        utils, lifecycle, pillar
Suggests: testthat (>= 2.1.0), covr, mvtnorm, actuar (>= 2.0.0), evd,
        ggdist, ggplot2, gk, LaplacesDemon, pkgdown
RdMacros: lifecycle
URL: https://pkg.mitchelloharawild.com/distributional/,
        https://github.com/mitchelloharawild/distributional
BugReports: https://github.com/mitchelloharawild/distributional/issues
Encoding: UTF-8
Language: en-GB
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-03-17 07:17:03 UTC; mitchell
Author: Mitchell O'Hara-Wild [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6729-7695>),
  Matthew Kay [aut] (ORCID: <https://orcid.org/0000-0001-9446-0419>),
  Alex Hayes [aut] (ORCID: <https://orcid.org/0000-0002-4985-5160>),
  Rob Hyndman [aut] (ORCID: <https://orcid.org/0000-0002-2140-5352>),
  Earo Wang [ctb] (ORCID: <https://orcid.org/0000-0001-6448-5260>),
  Vencislav Popov [ctb] (ORCID: <https://orcid.org/0000-0002-8073-4199>)
Maintainer: Mitchell O'Hara-Wild <mail@mitchelloharawild.com>
Repository: CRAN
Date/Publication: 2026-03-17 08:30:44 UTC
Built: R 4.6.0; ; 2026-04-26 02:23:16 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("$",hilo)
S3method("[[",distribution)
S3method("dimnames<-",distribution)
S3method("names<-",hilo)
S3method(.DollarNames,hilo)
S3method(Math,dist_default)
S3method(Math,dist_lognormal)
S3method(Math,dist_na)
S3method(Math,dist_normal)
S3method(Math,dist_sample)
S3method(Math,dist_transformed)
S3method(Ops,dist_default)
S3method(Ops,dist_na)
S3method(Ops,dist_normal)
S3method(Ops,dist_sample)
S3method(Ops,dist_transformed)
S3method(cdf,dist_bernoulli)
S3method(cdf,dist_beta)
S3method(cdf,dist_binomial)
S3method(cdf,dist_burr)
S3method(cdf,dist_categorical)
S3method(cdf,dist_cauchy)
S3method(cdf,dist_chisq)
S3method(cdf,dist_default)
S3method(cdf,dist_degenerate)
S3method(cdf,dist_exponential)
S3method(cdf,dist_f)
S3method(cdf,dist_gamma)
S3method(cdf,dist_geometric)
S3method(cdf,dist_gh)
S3method(cdf,dist_gk)
S3method(cdf,dist_gumbel)
S3method(cdf,dist_hypergeometric)
S3method(cdf,dist_inflated)
S3method(cdf,dist_inverse_exponential)
S3method(cdf,dist_inverse_gamma)
S3method(cdf,dist_inverse_gaussian)
S3method(cdf,dist_laplace)
S3method(cdf,dist_logarithmic)
S3method(cdf,dist_logistic)
S3method(cdf,dist_lognormal)
S3method(cdf,dist_mixture)
S3method(cdf,dist_multinomial)
S3method(cdf,dist_mvnorm)
S3method(cdf,dist_mvt)
S3method(cdf,dist_na)
S3method(cdf,dist_negbin)
S3method(cdf,dist_normal)
S3method(cdf,dist_pareto)
S3method(cdf,dist_percentile)
S3method(cdf,dist_poisson)
S3method(cdf,dist_poisson_inverse_gaussian)
S3method(cdf,dist_sample)
S3method(cdf,dist_student_t)
S3method(cdf,dist_studentized_range)
S3method(cdf,dist_transformed)
S3method(cdf,dist_truncated)
S3method(cdf,dist_uniform)
S3method(cdf,dist_weibull)
S3method(cdf,dist_wrap)
S3method(cdf,distribution)
S3method(covariance,default)
S3method(covariance,dist_bernoulli)
S3method(covariance,dist_beta)
S3method(covariance,dist_binomial)
S3method(covariance,dist_burr)
S3method(covariance,dist_categorical)
S3method(covariance,dist_cauchy)
S3method(covariance,dist_chisq)
S3method(covariance,dist_default)
S3method(covariance,dist_degenerate)
S3method(covariance,dist_dirichlet)
S3method(covariance,dist_exponential)
S3method(covariance,dist_f)
S3method(covariance,dist_gamma)
S3method(covariance,dist_geometric)
S3method(covariance,dist_gumbel)
S3method(covariance,dist_hypergeometric)
S3method(covariance,dist_inflated)
S3method(covariance,dist_inverse_exponential)
S3method(covariance,dist_inverse_gamma)
S3method(covariance,dist_inverse_gaussian)
S3method(covariance,dist_logarithmic)
S3method(covariance,dist_logistic)
S3method(covariance,dist_lognormal)
S3method(covariance,dist_mixture)
S3method(covariance,dist_multinomial)
S3method(covariance,dist_mvnorm)
S3method(covariance,dist_mvt)
S3method(covariance,dist_na)
S3method(covariance,dist_negbin)
S3method(covariance,dist_normal)
S3method(covariance,dist_pareto)
S3method(covariance,dist_poisson)
S3method(covariance,dist_poisson_inverse_gaussian)
S3method(covariance,dist_sample)
S3method(covariance,dist_student_t)
S3method(covariance,dist_transformed)
S3method(covariance,dist_uniform)
S3method(covariance,dist_weibull)
S3method(covariance,distribution)
S3method(covariance,numeric)
S3method(density,dist_bernoulli)
S3method(density,dist_beta)
S3method(density,dist_binomial)
S3method(density,dist_burr)
S3method(density,dist_categorical)
S3method(density,dist_cauchy)
S3method(density,dist_chisq)
S3method(density,dist_default)
S3method(density,dist_degenerate)
S3method(density,dist_dirichlet)
S3method(density,dist_exponential)
S3method(density,dist_f)
S3method(density,dist_gamma)
S3method(density,dist_geometric)
S3method(density,dist_gh)
S3method(density,dist_gk)
S3method(density,dist_gumbel)
S3method(density,dist_horseshoe)
S3method(density,dist_hypergeometric)
S3method(density,dist_inflated)
S3method(density,dist_inverse_exponential)
S3method(density,dist_inverse_gamma)
S3method(density,dist_inverse_gaussian)
S3method(density,dist_laplace)
S3method(density,dist_logarithmic)
S3method(density,dist_logistic)
S3method(density,dist_lognormal)
S3method(density,dist_mixture)
S3method(density,dist_multinomial)
S3method(density,dist_mvnorm)
S3method(density,dist_mvt)
S3method(density,dist_na)
S3method(density,dist_negbin)
S3method(density,dist_normal)
S3method(density,dist_pareto)
S3method(density,dist_percentile)
S3method(density,dist_poisson)
S3method(density,dist_poisson_inverse_gaussian)
S3method(density,dist_sample)
S3method(density,dist_student_t)
S3method(density,dist_transformed)
S3method(density,dist_truncated)
S3method(density,dist_uniform)
S3method(density,dist_weibull)
S3method(density,dist_wrap)
S3method(density,distribution)
S3method(dim,dist_default)
S3method(dim,dist_dirichlet)
S3method(dim,dist_mixture)
S3method(dim,dist_multinomial)
S3method(dim,dist_mvnorm)
S3method(dim,dist_mvt)
S3method(dimnames,distribution)
S3method(distributional::cdf,dist_gev)
S3method(distributional::cdf,dist_gpd)
S3method(distributional::covariance,dist_gev)
S3method(distributional::covariance,dist_gpd)
S3method(distributional::generate,dist_gev)
S3method(distributional::generate,dist_gpd)
S3method(distributional::log_density,dist_gev)
S3method(distributional::log_density,dist_gpd)
S3method(family,dist_default)
S3method(family,distribution)
S3method(format,dist_bernoulli)
S3method(format,dist_beta)
S3method(format,dist_binomial)
S3method(format,dist_burr)
S3method(format,dist_categorical)
S3method(format,dist_cauchy)
S3method(format,dist_chisq)
S3method(format,dist_default)
S3method(format,dist_degenerate)
S3method(format,dist_dirichlet)
S3method(format,dist_exponential)
S3method(format,dist_f)
S3method(format,dist_gamma)
S3method(format,dist_geometric)
S3method(format,dist_gev)
S3method(format,dist_gh)
S3method(format,dist_gk)
S3method(format,dist_gpd)
S3method(format,dist_gumbel)
S3method(format,dist_horseshoe)
S3method(format,dist_hypergeometric)
S3method(format,dist_inflated)
S3method(format,dist_inverse_exponential)
S3method(format,dist_inverse_gamma)
S3method(format,dist_inverse_gaussian)
S3method(format,dist_laplace)
S3method(format,dist_logarithmic)
S3method(format,dist_logistic)
S3method(format,dist_lognormal)
S3method(format,dist_mixture)
S3method(format,dist_multinomial)
S3method(format,dist_mvnorm)
S3method(format,dist_mvt)
S3method(format,dist_na)
S3method(format,dist_negbin)
S3method(format,dist_normal)
S3method(format,dist_pareto)
S3method(format,dist_percentile)
S3method(format,dist_poisson)
S3method(format,dist_poisson_inverse_gaussian)
S3method(format,dist_sample)
S3method(format,dist_student_t)
S3method(format,dist_studentized_range)
S3method(format,dist_transformed)
S3method(format,dist_truncated)
S3method(format,dist_uniform)
S3method(format,dist_weibull)
S3method(format,dist_wrap)
S3method(format,distribution)
S3method(format,hdr)
S3method(format,hilo)
S3method(format,pillar_distribution)
S3method(format,support_region)
S3method(generate,dist_bernoulli)
S3method(generate,dist_beta)
S3method(generate,dist_binomial)
S3method(generate,dist_burr)
S3method(generate,dist_categorical)
S3method(generate,dist_cauchy)
S3method(generate,dist_chisq)
S3method(generate,dist_default)
S3method(generate,dist_degenerate)
S3method(generate,dist_dirichlet)
S3method(generate,dist_exponential)
S3method(generate,dist_f)
S3method(generate,dist_gamma)
S3method(generate,dist_geometric)
S3method(generate,dist_gh)
S3method(generate,dist_gk)
S3method(generate,dist_gumbel)
S3method(generate,dist_horseshoe)
S3method(generate,dist_hypergeometric)
S3method(generate,dist_inflated)
S3method(generate,dist_inverse_exponential)
S3method(generate,dist_inverse_gamma)
S3method(generate,dist_inverse_gaussian)
S3method(generate,dist_laplace)
S3method(generate,dist_logarithmic)
S3method(generate,dist_logistic)
S3method(generate,dist_lognormal)
S3method(generate,dist_mixture)
S3method(generate,dist_multinomial)
S3method(generate,dist_mvnorm)
S3method(generate,dist_mvt)
S3method(generate,dist_na)
S3method(generate,dist_negbin)
S3method(generate,dist_normal)
S3method(generate,dist_pareto)
S3method(generate,dist_percentile)
S3method(generate,dist_poisson)
S3method(generate,dist_poisson_inverse_gaussian)
S3method(generate,dist_sample)
S3method(generate,dist_student_t)
S3method(generate,dist_transformed)
S3method(generate,dist_uniform)
S3method(generate,dist_weibull)
S3method(generate,dist_wrap)
S3method(generate,distribution)
S3method(has_symmetry,dist_beta)
S3method(has_symmetry,dist_binomial)
S3method(has_symmetry,dist_cauchy)
S3method(has_symmetry,dist_default)
S3method(has_symmetry,dist_laplace)
S3method(has_symmetry,dist_logistic)
S3method(has_symmetry,dist_normal)
S3method(has_symmetry,dist_student_t)
S3method(has_symmetry,dist_uniform)
S3method(has_symmetry,distribution)
S3method(hdr,default)
S3method(hdr,dist_default)
S3method(hdr,distribution)
S3method(hilo,default)
S3method(hilo,dist_default)
S3method(hilo,distribution)
S3method(is.na,hilo)
S3method(kurtosis,dist_bernoulli)
S3method(kurtosis,dist_beta)
S3method(kurtosis,dist_binomial)
S3method(kurtosis,dist_categorical)
S3method(kurtosis,dist_cauchy)
S3method(kurtosis,dist_chisq)
S3method(kurtosis,dist_degenerate)
S3method(kurtosis,dist_exponential)
S3method(kurtosis,dist_f)
S3method(kurtosis,dist_gamma)
S3method(kurtosis,dist_geometric)
S3method(kurtosis,dist_gumbel)
S3method(kurtosis,dist_hypergeometric)
S3method(kurtosis,dist_laplace)
S3method(kurtosis,dist_logistic)
S3method(kurtosis,dist_lognormal)
S3method(kurtosis,dist_na)
S3method(kurtosis,dist_negbin)
S3method(kurtosis,dist_normal)
S3method(kurtosis,dist_poisson)
S3method(kurtosis,dist_uniform)
S3method(kurtosis,dist_weibull)
S3method(kurtosis,distribution)
S3method(likelihood,dist_default)
S3method(likelihood,distribution)
S3method(log_cdf,dist_default)
S3method(log_cdf,dist_lognormal)
S3method(log_cdf,dist_na)
S3method(log_cdf,dist_normal)
S3method(log_cdf,distribution)
S3method(log_density,dist_bernoulli)
S3method(log_density,dist_beta)
S3method(log_density,dist_binomial)
S3method(log_density,dist_burr)
S3method(log_density,dist_cauchy)
S3method(log_density,dist_chisq)
S3method(log_density,dist_default)
S3method(log_density,dist_dirichlet)
S3method(log_density,dist_exponential)
S3method(log_density,dist_f)
S3method(log_density,dist_gamma)
S3method(log_density,dist_geometric)
S3method(log_density,dist_gh)
S3method(log_density,dist_gk)
S3method(log_density,dist_gumbel)
S3method(log_density,dist_horseshoe)
S3method(log_density,dist_hypergeometric)
S3method(log_density,dist_inverse_exponential)
S3method(log_density,dist_inverse_gamma)
S3method(log_density,dist_inverse_gaussian)
S3method(log_density,dist_laplace)
S3method(log_density,dist_logarithmic)
S3method(log_density,dist_logistic)
S3method(log_density,dist_lognormal)
S3method(log_density,dist_multinomial)
S3method(log_density,dist_mvnorm)
S3method(log_density,dist_mvt)
S3method(log_density,dist_na)
S3method(log_density,dist_negbin)
S3method(log_density,dist_normal)
S3method(log_density,dist_pareto)
S3method(log_density,dist_poisson)
S3method(log_density,dist_poisson_inverse_gaussian)
S3method(log_density,dist_student_t)
S3method(log_density,dist_uniform)
S3method(log_density,dist_weibull)
S3method(log_density,dist_wrap)
S3method(log_density,distribution)
S3method(log_likelihood,dist_default)
S3method(log_likelihood,distribution)
S3method(log_quantile,dist_default)
S3method(log_quantile,dist_lognormal)
S3method(log_quantile,dist_na)
S3method(log_quantile,dist_normal)
S3method(log_quantile,distribution)
S3method(mean,dist_bernoulli)
S3method(mean,dist_beta)
S3method(mean,dist_binomial)
S3method(mean,dist_burr)
S3method(mean,dist_categorical)
S3method(mean,dist_cauchy)
S3method(mean,dist_chisq)
S3method(mean,dist_default)
S3method(mean,dist_degenerate)
S3method(mean,dist_dirichlet)
S3method(mean,dist_exponential)
S3method(mean,dist_f)
S3method(mean,dist_gamma)
S3method(mean,dist_geometric)
S3method(mean,dist_gev)
S3method(mean,dist_gpd)
S3method(mean,dist_gumbel)
S3method(mean,dist_hypergeometric)
S3method(mean,dist_inflated)
S3method(mean,dist_inverse_exponential)
S3method(mean,dist_inverse_gamma)
S3method(mean,dist_inverse_gaussian)
S3method(mean,dist_laplace)
S3method(mean,dist_logarithmic)
S3method(mean,dist_logistic)
S3method(mean,dist_lognormal)
S3method(mean,dist_mixture)
S3method(mean,dist_multinomial)
S3method(mean,dist_mvnorm)
S3method(mean,dist_mvt)
S3method(mean,dist_na)
S3method(mean,dist_negbin)
S3method(mean,dist_normal)
S3method(mean,dist_pareto)
S3method(mean,dist_percentile)
S3method(mean,dist_poisson)
S3method(mean,dist_poisson_inverse_gaussian)
S3method(mean,dist_sample)
S3method(mean,dist_student_t)
S3method(mean,dist_transformed)
S3method(mean,dist_truncated)
S3method(mean,dist_uniform)
S3method(mean,dist_weibull)
S3method(mean,distribution)
S3method(median,dist_default)
S3method(median,dist_mvnorm)
S3method(median,dist_sample)
S3method(median,distribution)
S3method(parameters,dist_default)
S3method(parameters,dist_wrap)
S3method(parameters,distribution)
S3method(pillar_shaft,distribution)
S3method(print,dist_default)
S3method(quantile,dist_bernoulli)
S3method(quantile,dist_beta)
S3method(quantile,dist_binomial)
S3method(quantile,dist_burr)
S3method(quantile,dist_categorical)
S3method(quantile,dist_cauchy)
S3method(quantile,dist_chisq)
S3method(quantile,dist_default)
S3method(quantile,dist_degenerate)
S3method(quantile,dist_exponential)
S3method(quantile,dist_f)
S3method(quantile,dist_gamma)
S3method(quantile,dist_geometric)
S3method(quantile,dist_gh)
S3method(quantile,dist_gk)
S3method(quantile,dist_gumbel)
S3method(quantile,dist_hypergeometric)
S3method(quantile,dist_inflated)
S3method(quantile,dist_inverse_exponential)
S3method(quantile,dist_inverse_gamma)
S3method(quantile,dist_inverse_gaussian)
S3method(quantile,dist_laplace)
S3method(quantile,dist_logarithmic)
S3method(quantile,dist_logistic)
S3method(quantile,dist_lognormal)
S3method(quantile,dist_mixture)
S3method(quantile,dist_mvnorm)
S3method(quantile,dist_mvt)
S3method(quantile,dist_na)
S3method(quantile,dist_negbin)
S3method(quantile,dist_normal)
S3method(quantile,dist_pareto)
S3method(quantile,dist_percentile)
S3method(quantile,dist_poisson)
S3method(quantile,dist_poisson_inverse_gaussian)
S3method(quantile,dist_sample)
S3method(quantile,dist_student_t)
S3method(quantile,dist_studentized_range)
S3method(quantile,dist_transformed)
S3method(quantile,dist_truncated)
S3method(quantile,dist_uniform)
S3method(quantile,dist_weibull)
S3method(quantile,dist_wrap)
S3method(quantile,distribution)
S3method(skewness,dist_bernoulli)
S3method(skewness,dist_beta)
S3method(skewness,dist_binomial)
S3method(skewness,dist_categorical)
S3method(skewness,dist_cauchy)
S3method(skewness,dist_chisq)
S3method(skewness,dist_degenerate)
S3method(skewness,dist_exponential)
S3method(skewness,dist_f)
S3method(skewness,dist_gamma)
S3method(skewness,dist_geometric)
S3method(skewness,dist_gumbel)
S3method(skewness,dist_hypergeometric)
S3method(skewness,dist_laplace)
S3method(skewness,dist_logistic)
S3method(skewness,dist_lognormal)
S3method(skewness,dist_na)
S3method(skewness,dist_negbin)
S3method(skewness,dist_normal)
S3method(skewness,dist_poisson)
S3method(skewness,dist_sample)
S3method(skewness,dist_uniform)
S3method(skewness,dist_weibull)
S3method(skewness,distribution)
S3method(stats::density,dist_gev)
S3method(stats::density,dist_gpd)
S3method(stats::median,dist_gev)
S3method(stats::median,dist_gpd)
S3method(stats::quantile,dist_gev)
S3method(stats::quantile,dist_gpd)
S3method(sum,distribution)
S3method(support,dist_categorical)
S3method(support,dist_default)
S3method(support,dist_dirichlet)
S3method(support,dist_gpd)
S3method(support,dist_horseshoe)
S3method(support,dist_percentile)
S3method(support,dist_sample)
S3method(support,dist_transformed)
S3method(support,distribution)
S3method(variance,default)
S3method(variance,dist_default)
S3method(variance,dist_laplace)
S3method(variance,distribution)
S3method(variance,matrix)
S3method(variance,numeric)
S3method(vec_arith,distribution)
S3method(vec_arith,hilo)
S3method(vec_arith.distribution,default)
S3method(vec_arith.numeric,distribution)
S3method(vec_arith.numeric,hilo)
S3method(vec_cast,character.distribution)
S3method(vec_cast,character.hilo)
S3method(vec_cast,distribution.distribution)
S3method(vec_cast,distribution.double)
S3method(vec_cast,distribution.integer)
S3method(vec_math,distribution)
S3method(vec_math,hilo)
S3method(vec_ptype2,distribution.distribution)
S3method(vec_ptype2,distribution.double)
S3method(vec_ptype2,distribution.integer)
S3method(vec_ptype2,double.distribution)
S3method(vec_ptype2,hilo.hilo)
S3method(vec_ptype2,integer.distribution)
S3method(vec_ptype_abbr,distribution)
S3method(vec_ptype_abbr,support_region)
export(cdf)
export(covariance)
export(dist_bernoulli)
export(dist_beta)
export(dist_binomial)
export(dist_burr)
export(dist_categorical)
export(dist_cauchy)
export(dist_chisq)
export(dist_degenerate)
export(dist_dirichlet)
export(dist_exponential)
export(dist_f)
export(dist_gamma)
export(dist_geometric)
export(dist_gev)
export(dist_gh)
export(dist_gk)
export(dist_gpd)
export(dist_gumbel)
export(dist_horseshoe)
export(dist_hypergeometric)
export(dist_inflated)
export(dist_inverse_exponential)
export(dist_inverse_gamma)
export(dist_inverse_gaussian)
export(dist_laplace)
export(dist_logarithmic)
export(dist_logistic)
export(dist_lognormal)
export(dist_missing)
export(dist_mixture)
export(dist_multinomial)
export(dist_multivariate_normal)
export(dist_multivariate_t)
export(dist_negative_binomial)
export(dist_normal)
export(dist_pareto)
export(dist_percentile)
export(dist_poisson)
export(dist_poisson_inverse_gaussian)
export(dist_sample)
export(dist_student_t)
export(dist_studentized_range)
export(dist_transformed)
export(dist_truncated)
export(dist_uniform)
export(dist_weibull)
export(dist_wrap)
export(generate)
export(has_symmetry)
export(hdr)
export(hilo)
export(is_distribution)
export(is_hdr)
export(is_hilo)
export(kurtosis)
export(likelihood)
export(log_likelihood)
export(new_dist)
export(new_hdr)
export(new_hilo)
export(parameters)
export(skewness)
export(support)
export(variance)
import(rlang)
import(vctrs)
importFrom(generics,generate)
importFrom(lifecycle,deprecate_soft)
importFrom(lifecycle,deprecated)
importFrom(pillar,get_max_extent)
importFrom(pillar,new_ornament)
importFrom(pillar,new_pillar_shaft)
importFrom(pillar,pillar_shaft)
importFrom(stats,density)
importFrom(stats,family)
importFrom(stats,median)
importFrom(stats,quantile)
importFrom(utils,.DollarNames)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# distributional

<!-- badges: start -->

[![Lifecycle:
stable](https://img.shields.io/badge/lifecycle-stable-brightgreen.svg)](https://lifecycle.r-lib.org/articles/stages.html#stable)
[![R-CMD-check](https://github.com/mitchelloharawild/distributional/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/mitchelloharawild/distributional/actions/workflows/R-CMD-check.yaml)
<!-- [![Coverage Status](https://codecov.io/gh/mitchelloharawild/distributional/branch/master/graph/badge.svg)](https://codecov.io/github/mitchelloharawild/distributional?branch=master) -->
[![CRAN
status](https://www.r-pkg.org/badges/version/distributional)](https://CRAN.R-project.org/package=distributional)
![Download
count](https://cranlogs.r-pkg.org/badges/last-month/distributional)
<!-- badges: end -->

The distributional package allows distributions to be used in a
vectorised context. It provides methods which are minimal wrappers to
the standard d, p, q, and r distribution functions which are applied to
each distribution in the vector. Additional distributional statistics
can be computed, including the `mean()`, `median()`, `variance()`, and
intervals with `hilo()`.

The distributional nature of a model’s predictions is often understated,
with default output of prediction methods usually only producing point
predictions. Some R packages (such as
[forecast](https://CRAN.R-project.org/package=forecast)) further
emphasise uncertainty by producing point forecasts and intervals by
default, however the user’s ability to interact with them is limited.
This package vectorises distributions and provides methods for working
with them, making distributions compatible with prediction outputs of
modelling functions. These vectorised distributions can be illustrated
with [ggplot2](https://CRAN.R-project.org/package=ggplot2) using the
[ggdist](https://CRAN.R-project.org/package=ggdist) package, providing
further opportunity to visualise the uncertainty of predictions and
teach distributional theory.

## Installation

You can install the released version of distributional from
[CRAN](https://CRAN.R-project.org/package=distributional) with:

``` r
install.packages("distributional")
```

The development version can be installed from
[GitHub](https://github.com/mitchelloharawild/distributional) with:

``` r
# install.packages("remotes")
remotes::install_github("mitchelloharawild/distributional")
```

## Examples

Distributions are created using `dist_*()` functions. A list of included
distribution shapes can be found here:
<https://pkg.mitchelloharawild.com/distributional/reference/>

``` r
library(distributional)
my_dist <- c(dist_normal(mu = 0, sigma = 1), dist_student_t(df = 10))
my_dist
#> <distribution[2]>
#> [1] N(0, 1)     t(10, 0, 1)
```

The standard four distribution functions in R are usable via these
generics:

``` r
density(my_dist, 0) # c(dnorm(0, mean = 0, sd = 1), dt(0, df = 10))
#> [1] 0.3989423 0.3891084
cdf(my_dist, 5) # c(pnorm(5, mean = 0, sd = 1), pt(5, df = 10))
#> [1] 0.9999997 0.9997313
quantile(my_dist, 0.1) # c(qnorm(0.1, mean = 0, sd = 1), qt(0.1, df = 10))
#> [1] -1.281552 -1.372184
generate(my_dist, 10) # list(rnorm(10, mean = 0, sd = 1), rt(10, df = 10))
#> [[1]]
#>  [1]  1.262954285 -0.326233361  1.329799263  1.272429321  0.414641434
#>  [6] -1.539950042 -0.928567035 -0.294720447 -0.005767173  2.404653389
#> 
#> [[2]]
#>  [1]  0.99165484 -1.36999677 -0.40943004 -0.85261144 -1.37728388  0.81020460
#>  [7] -1.82965813 -0.06142032 -1.33933588 -0.28491414
```

You can also compute intervals using `hilo()`

``` r
hilo(my_dist, 0.95)
#> <hilo[2]>
#> [1] [-0.01190677, 0.01190677]0.95 [-0.01220773, 0.01220773]0.95
```

Additionally, some distributions may support other methods such as
mathematical operations and summary measures. If the methods aren’t
supported, a transformed distribution will be created.

``` r
my_dist
#> <distribution[2]>
#> [1] N(0, 1)     t(10, 0, 1)
my_dist*3 + 2
#> <distribution[2]>
#> [1] N(2, 9)        t(t(10, 0, 1))
mean(my_dist)
#> [1] 0 0
variance(my_dist)
#> [1] 1.00 1.25
```

You can also visualise the distribution(s) using the
[ggdist](https://mjskay.github.io/ggdist/) package.

``` r
library(ggdist)
library(ggplot2)

df <- data.frame(
  name = c("Gamma(2,1)", "Normal(5,1)", "Mixture"),
  dist = c(dist_gamma(2,1), dist_normal(5,1),
           dist_mixture(dist_gamma(2,1), dist_normal(5, 1), weights = c(0.4, 0.6)))
)

ggplot(df, aes(y = factor(name, levels = rev(name)))) +
  stat_dist_halfeye(aes(dist = dist)) + 
  labs(title = "Density function for a mixture of distributions", y = NULL, x = NULL)
```

<img src="man/figures/README-plot-1.png" alt="Half-eye plot showing density functions for three distributions: a Gamma(2,1), a Normal(5,1), and a mixture of both with weights 0.4 and 0.6 respectively. The mixture distribution appears bimodal, with a smaller mode near 2 (inherited from the right-skewed Gamma component) and a larger, more prominent mode near 5 (from the Normal component), reflecting the 0.4 and 0.6 weights assigned to each component." width="100%" />

## Related work

There are several packages which unify interfaces for distributions in
R:

- stats provides functions to work with possibly multiple distributions
  (comparisons made below).
- [distributions3](https://cran.r-project.org/package=distributions3)
  represents singular distributions using S3, with particularly nice
  documentation.
- [distr](https://cran.r-project.org/package=distr) represents singular
  distributions using S4.
- [distr6](https://cran.r-project.org/package=distr6) represents
  singular distributions using R6.
- Many more in the [CRAN task
  view](https://cran.r-project.org/view=Distributions)

This package differs from the above libraries by storing the
distributions in a vectorised format. It does this using
[vctrs](https://vctrs.r-lib.org/), so it should play nicely with the
tidyverse (try putting distributions into a tibble!).

#' The Normal distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Normal distribution is ubiquitous in statistics, partially because
#' of the central limit theorem, which states that sums of i.i.d. random
#' variables eventually become Normal. Linear transformations of Normal
#' random variables result in new random variables that are also Normal. If
#' you are taking an intro stats course, you'll likely use the Normal
#' distribution for Z-tests and in simple linear regression. Under
#' regularity conditions, maximum likelihood estimators are
#' asymptotically Normal. The Normal distribution is also called the
#' gaussian distribution.
#'
#' @param mu,mean The mean (location parameter) of the distribution, which is also
#'   the mean of the distribution. Can be any real number.
#' @param sigma,sd The standard deviation (scale parameter) of the distribution.
#'   Can be any positive number. If you would like a Normal distribution with
#'   **variance** \eqn{\sigma^2}, be sure to take the square root, as this is a
#'   common source of errors.
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a Normal random variable with mean
#'   `mu` = \eqn{\mu} and standard deviation `sigma` = \eqn{\sigma}.
#'
#'   **Support**: \eqn{R}, the set of all real numbers
#'
#'   **Mean**: \eqn{\mu}
#'
#'   **Variance**: \eqn{\sigma^2}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{1}{\sqrt{2 \pi \sigma^2}} e^{-(x - \mu)^2 / 2 \sigma^2}
#'   }{
#'     f(x) = 1 / sqrt(2 \pi \sigma^2) exp(-(x - \mu)^2 / (2 \sigma^2))
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   The cumulative distribution function has the form
#'
#'   \deqn{
#'     F(t) = \int_{-\infty}^t \frac{1}{\sqrt{2 \pi \sigma^2}} e^{-(x - \mu)^2 / 2 \sigma^2} dx
#'   }{
#'     F(t) = integral_{-\infty}^t 1 / sqrt(2 \pi \sigma^2) exp(-(x - \mu)^2 / (2 \sigma^2)) dx
#'   }
#'
#'   but this integral does not have a closed form solution and must be
#'   approximated numerically. The c.d.f. of a standard Normal is sometimes
#'   called the "error function". The notation \eqn{\Phi(t)} also stands
#'   for the c.d.f. of a standard Normal evaluated at \eqn{t}. Z-tables
#'   list the value of \eqn{\Phi(t)} for various \eqn{t}.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = e^{\mu t + \sigma^2 t^2 / 2}
#'   }{
#'     E(e^(tX)) = e^(\mu t + \sigma^2 t^2 / 2)
#'   }
#'
#' @seealso [stats::Normal]
#'
#' @examples
#' dist <- dist_normal(mu = 1:5, sigma = 3)
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @export
dist_normal <- function(mu = 0, sigma = 1, mean = mu, sd = sigma){
  mean <- vec_cast(mean, double())
  sd <- vec_cast(sd, double())
  if(any(sd[!is.na(sd)] < 0)){
    abort("Standard deviation of a normal distribution must be non-negative")
  }
  new_dist(mu = mean, sigma = sd, class = "dist_normal")
}

#' @export
format.dist_normal <- function(x, digits = 2, ...){
  sprintf(
    "N(%s, %s)",
    format(x[["mu"]], digits = digits, ...),
    format(x[["sigma"]]^2, digits = digits, ...)
  )
}

#' @export
density.dist_normal <- function(x, at, ...){
  stats::dnorm(at, x[["mu"]], x[["sigma"]])
}

#' @export
log_density.dist_normal <- function(x, at, ...){
  stats::dnorm(at, x[["mu"]], x[["sigma"]], log = TRUE)
}

#' @export
quantile.dist_normal <- function(x, p, ...){
  stats::qnorm(p, x[["mu"]], x[["sigma"]])
}
#' @export
log_quantile.dist_normal <- function(x, p, ...){
  stats::qnorm(p, x[["mu"]], x[["sigma"]], log.p = TRUE)
}

#' @export
cdf.dist_normal <- function(x, q, ...){
  stats::pnorm(q, x[["mu"]], x[["sigma"]])
}
#' @export
log_cdf.dist_normal <- function(x, q, ...){
  stats::pnorm(q, x[["mu"]], x[["sigma"]], log.p = TRUE)
}

#' @export
generate.dist_normal <- function(x, times, ...){
  stats::rnorm(times, x[["mu"]], x[["sigma"]])
}

#' @export
mean.dist_normal <- function(x, ...){
  x[["mu"]]
}

#' @export
covariance.dist_normal <- function(x, ...){
  x[["sigma"]]^2
}

#' @export
skewness.dist_normal <- function(x, ...) 0

#' @export
kurtosis.dist_normal <- function(x, ...) 0

#' @export
Ops.dist_normal <- function(e1, e2){
  ok <- switch(.Generic, `+` = , `-` = , `*` = , `/` = TRUE, FALSE)
  if (!ok) {
    return(NextMethod())
  }
  if(.Generic == "/" && inherits(e2, "dist_normal")){
    abort(sprintf("Cannot divide by a normal distribution"))
  }
  if(.Generic %in% c("-", "+") && missing(e2)){
    e2 <- e1
    e1 <- if(.Generic == "+") 1 else -1
    .Generic <- "*"
  }
  if(.Generic == "-"){
    .Generic <- "+"
    e2 <- -e2
  }
  else if(.Generic == "/"){
    .Generic <- "*"
    e2 <- 1/e2
  }

  # Ops between two normals
  if(inherits(e1, "dist_normal") && inherits(e2, "dist_normal")){
    if(.Generic == "*"){
      abort(sprintf("Multiplying two normal distributions is not supported."))
    }

    e1$mu <- e1$mu + e2$mu
    e1$sigma <- sqrt(e1$sigma^2 + e2$sigma^2)
    return(e1)
  }

  # Ops between a normal and scalar
  if(inherits(e1, "dist_normal")){
    dist <- e1
    scalar <- e2
  } else {
    dist <- e2
    scalar <- e1
  }
  if(!is.numeric(scalar)){
    abort(sprintf("Cannot %s a `%s` with a normal distribution",
                  switch(.Generic, `+` = "add", `-` = "subtract", `*` = "multiply", `/` = "divide"),
                  class(scalar)))
  }

  if(.Generic == "+"){
    dist$mu <- dist$mu + scalar
  }
  else if(.Generic == "*"){
    dist$mu <- dist$mu * scalar
    dist$sigma <- dist$sigma * abs(scalar)
  }
  dist
}

#' @method Math dist_normal
#' @export
Math.dist_normal <- function(x, ...) {
  # Shortcut to get log-normal distribution from Normal.
  if(.Generic == "exp") return(vec_data(dist_lognormal(x[["mu"]], x[["sigma"]]))[[1]])
  NextMethod()
}

test_that("is.na() on <dist>[[1]] (#29)", {
  x <- c(dist_normal(0,1), NA)
  expect_equal(
    is.na(x),
    c(FALSE, TRUE)
  )

  expect_equal(
    is.na(x[[1]]),
    FALSE
  )

  expect_equal(
    is.na(x[[2]]),
    TRUE
  )
})

test_that("Negative Binomial distribution", {
  dist <- dist_normal(0, 1)
  percentiles <- seq(0.01, 0.99, by = 0.01)
  x <- vapply(percentiles, quantile, double(1L), x = dist)
  dist <- dist_percentile(list(x), list(percentiles*100))

  expect_equal(format(dist), "percentile[99]")

  # quantiles
  expect_equal(quantile(dist, 0.6), stats::qnorm(0.6, 0, 1))
  expect_equal(quantile(dist, 0.61), stats::qnorm(0.61, 0, 1))

  # pdf

  # cdf
  expect_equal(cdf(dist, 0), stats::pnorm(0, 0, 1))
  expect_equal(cdf(dist, 1), stats::pnorm(1, 0, 1), tolerance = 1e-3)

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.6)), 0.6, tolerance = 1e-3)

  # stats
})

test_that("support gives the correct bounds", {
  s <- support(c(dist_normal(),
               dist_gamma(1, 1),
               dist_gamma(2, 1),
               dist_lognormal(),
               dist_beta(1, 1),
               dist_beta(1, 2),
               dist_beta(2, 1),
               dist_beta(2, 2),
               exp(dist_wrap('norm')),
               2*atan(dist_normal())))
  out <- unname(format(s))
  expect_equal(out, c("R","[0,Inf)","(0,Inf)","(0,Inf)","[0,1]","[0,1)","(0,1]","(0,1)","(0,Inf)","(-pi,pi)"))
})

test_that("Recycling rules and output for applying multiple inputs over multiple univariate distributions", {
  # <dist> is a distribution vector of length 10
  dist <- dist_named <- dist_normal(1:10, 1:10)
  dimnames(dist_named) <- "name"

  # p = 0.5:          apply p across all elements of <dist> (recycling p onto <dist>)
  #                   Returns a vector of length 10
  expect_equal(
    quantile(dist, 0.5),
    qnorm(0.5, 1:10, 1:10)
  )
  expect_equal(
    quantile(dist_named, 0.5),
    qnorm(0.5, 1:10, 1:10)
  )

  # p = c(0.5, 0.9):  Cannot recycle p (length 2) onto <dist> (length 10)
  #                   Returns a list containing values for each p.
  expect_equal(
    quantile(dist, c(0.5, 0.9)),
    mapply({function(mean, sd) qnorm(c(0.5, 0.9), mean, sd)},
           mean = 1:10, sd = 1:10, SIMPLIFY = FALSE)
  )
  expect_equal(
    quantile(dist_named, c(0.5, 0.9)),
    mapply({function(mean, sd) qnorm(c(0.5, 0.9), mean, sd)},
           mean = 1:10, sd = 1:10, SIMPLIFY = FALSE)
  )

  # p = ppoints(10):  apply each p to each element of <dist> (no recycling)
  #                   Returns a list for each distribution with the 10 quantiles.
  expect_equal(
    quantile(dist, ppoints(10)),
    mapply({function(mean, sd) qnorm(ppoints(10), mean, sd)},
           mean = 1:10, sd = 1:10, SIMPLIFY = FALSE)
  )
  expect_equal(
    quantile(dist_named, ppoints(10)),
    mapply({function(mean, sd) qnorm(ppoints(10), mean, sd)},
           mean = 1:10, sd = 1:10, SIMPLIFY = FALSE)
  )

  # p = list(0.5):    equivalent to p = 0.5, but returns a list output
  #                   Returns a tibble with 1 vector column of length 10
  expect_equal(
    quantile(dist, list(a = 0.5)),
    new_data_frame(list(a = quantile(dist, 0.5)))
  )
  expect_equal(
    quantile(dist_named, list(a = 0.5)),
    new_data_frame(list(a = quantile(dist, 0.5)))
  )

  # p = list(c(0.5, 0.9)):
  #                   Cannot recycle p[[1]] (length 2) onto <dist> (length 10)
  #                   Returns an error.
  expect_error(
    quantile(dist, list(a = c(0.5, 0.9))),
    "Cannot recycle input"
  )
  expect_error(
    quantile(dist_named, list(a = c(0.5, 0.9))),
    "Cannot recycle input"
  )

  # p = list(p1, 0.5): equivalent to df(quantile(<dist>, p1), quantile(<dist>, 0.5)).
  #                   Returns a tibble with 2 vector columns of length 10
  #                   Names of p are used in output.
  expect_equal(
    quantile(dist, list(a=ppoints(10), b=0.5)),
    new_data_frame(list(a = qnorm(ppoints(10), 1:10, 1:10), b = quantile(dist, 0.5)))
  )
  expect_equal(
    quantile(dist_named, list(a=ppoints(10), b=0.5)),
    new_data_frame(list(a = qnorm(ppoints(10), 1:10, 1:10), b = quantile(dist, 0.5)))
  )
})

test_that("Recycling rules and output for applying multiple inputs over multiple multivariate distributions", {
  # <dist> is a bivariate distribution vector of length 2
  dist <- dist_named <- dist_multivariate_normal(mu = list(c(1,2), c(3,5)),
                                   sigma = list(matrix(c(4,2,2,3), ncol=2),
                                                matrix(c(5,1,1,4), ncol=2)))
  dimnames(dist_named) <- c("a", "b")

  expect_equal(
    quantile(dist, 0.5, type = "marginal"),
    matrix(c(1,3,2,5), nrow = 2)
  )
  expect_equal(
    quantile(dist_named, 0.5, type = "marginal"),
    matrix(c(1,3,2,5), nrow = 2, dimnames = list(NULL, c("a", "b")))
  )

  expect_equal(
    quantile(dist, c(0.5, 0.9), type = "marginal"),
    list(matrix(c(1, 3.5631031310892, 2, 4.21971242404268), ncol = 2),
         matrix(c(3, 5.86563641722901, 5, 7.5631031310892), ncol = 2))
  )
  expect_equal(
    quantile(dist_named, c(0.5, 0.9), type = "marginal"),
    list(matrix(c(1, 3.5631031310892, 2, 4.21971242404268), ncol = 2, dimnames = list(NULL, c("a", "b"))),
         matrix(c(3, 5.86563641722901, 5, 7.5631031310892), ncol = 2, dimnames = list(NULL, c("a", "b"))))
  )

  expect_equal(
    quantile(dist, c(0.5, 0.9, 0.95), type = "marginal"),
    list(
      matrix(c(1, 3.5631031310892, 4.28970725390294, 2, 4.21971242404268, 4.84897005289389),
             ncol = 2),
      matrix(c(3, 5.86563641722901, 6.67800452290057, 5, 7.5631031310892, 8.28970725390294),
             ncol = 2)
    )
  )
  expect_equal(
    quantile(dist_named, c(0.5, 0.9, 0.95), type = "marginal"),
    list(
      matrix(c(1, 3.5631031310892, 4.28970725390294, 2, 4.21971242404268, 4.84897005289389),
             ncol = 2, dimnames = list(NULL, c("a","b"))),
      matrix(c(3, 5.86563641722901, 6.67800452290057, 5, 7.5631031310892, 8.28970725390294),
             ncol = 2, dimnames = list(NULL, c("a", "b")))
    )
  )

  expect_equal(
    quantile(dist, list(single = 0.5, varied = c(0.8, 0.3))),
    new_data_frame(
      list(single = quantile(dist, 0.5), varied = rbind(quantile(dist[1], 0.8), quantile(dist[2], 0.3)))
    )
  )
  expect_equal(
    quantile(dist_named, list(single = 0.5, varied = c(0.8, 0.3))),
    new_data_frame(
      list(single = quantile(dist_named, 0.5), varied = rbind(quantile(dist_named[1], 0.8), quantile(dist_named[2], 0.3)))
    )
  )

  skip_if_not_installed("mvtnorm")
  expect_equal(
    density(dist, cbind(2, 3)),
    c(0.046649277604197, 0.0215708514518913)
  )
  expect_equal(
    density(dist_named, cbind(2, 3)),
    c(0.046649277604197, 0.0215708514518913)
  )

  expect_equal(
    mean(dist),
    matrix(c(1,3,2,5), nrow = 2)
  )
  expect_equal(
    mean(dist_named),
    matrix(c(1,3,2,5), nrow = 2, dimnames = list(NULL, c("a", "b")))
  )

  expect_equal(
    covariance(dist),
    list(
      matrix(c(4,2,2,3), nrow = 2),
      matrix(c(5,1,1,4), nrow = 2)
    )
  )
  expect_equal(
    covariance(dist_named),
    list(
      matrix(c(4,2,2,3), nrow = 2, dimnames = list(NULL, c("a", "b"))),
      matrix(c(5,1,1,4), nrow = 2, dimnames = list(NULL, c("a", "b")))
    )
  )
})

#' The Gumbel distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Gumbel distribution is a special case of the Generalized Extreme Value
#' distribution, obtained when the GEV shape parameter \eqn{\xi} is equal to 0.
#' It may be referred to as a type I extreme value distribution.
#'
#' @inheritParams actuar::dgumbel
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_gumbel")`
#'
#'   In the following, let \eqn{X} be a Gumbel random variable with location
#'   parameter  `alpha` = \eqn{\alpha} and scale parameter `scale` = \eqn{\sigma}.
#'
#'   **Support**: \eqn{R}, the set of all real numbers.
#'
#'   **Mean**: 
#'   
#'   \deqn{
#'     E(X) = \alpha + \sigma\gamma
#'   }{
#'     E(X) = alpha + sigma*gamma
#'   }
#'   
#'   where \eqn{\gamma} is the Euler-Mascheroni constant, 
#'   approximately equal to 0.5772157.
#'
#'   **Variance**: 
#'   
#'   \deqn{
#'     \textrm{Var}(X) = \frac{\pi^2 \sigma^2}{6}
#'   }{
#'     Var(X) = (pi^2 * sigma^2) / 6
#'   }
#'
#'   **Skewness**: 
#'   
#'   \deqn{
#'     \textrm{Skew}(X) = \frac{12\sqrt{6}\zeta(3)}{\pi^3} \approx 1.1395
#'   }{
#'     Skew(X) = (12*sqrt(6)*zeta(3)) / pi^3
#'   }
#'   
#'   where \eqn{\zeta(3)} is Apery's constant, 
#'   approximately equal to 1.2020569. Note that skewness is independent 
#'   of the distribution parameters.
#'
#'   **Kurtosis (excess)**: 
#'   
#'   \deqn{
#'     \textrm{Kurt}(X) = \frac{12}{5} = 2.4
#'   }{
#'     Kurt(X) = 12/5
#'   }
#'   
#'   Note that excess kurtosis is independent of the distribution parameters.
#'
#'   **Median**: 
#'   
#'   \deqn{
#'     \textrm{Median}(X) = \alpha - \sigma\ln(\ln 2)
#'   }{
#'     Median(X) = alpha - sigma*ln(ln(2))
#'   }
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{1}{\sigma} \exp\left[-\frac{x - \alpha}{\sigma}\right]
#'            \exp\left\{-\exp\left[-\frac{x - \alpha}{\sigma}\right]\right\}
#'   }{
#'     f(x) = (1/sigma) * exp[-(x - alpha)/sigma] * exp(-exp[-(x - alpha)/sigma])
#'   }
#'   
#'   for \eqn{x} in \eqn{R}, the set of all real numbers.
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = \exp\left\{-\exp\left[-\frac{x - \alpha}{\sigma}\right]\right\}
#'   }{
#'     F(x) = exp(-exp[-(x - alpha)/sigma])
#'   }
#'   
#'   for \eqn{x} in \eqn{R}, the set of all real numbers.
#'
#'   **Quantile function (inverse c.d.f)**:
#'
#'   \deqn{
#'     F^{-1}(p) = \alpha - \sigma \ln(-\ln p)
#'   }{
#'     F^(-1)(p) = alpha - sigma * ln(-ln(p))
#'   }
#'   
#'   for \eqn{p} in (0, 1).
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \Gamma(1 - \sigma t) e^{\alpha t}
#'   }{
#'     E(e^(tX)) = Gamma(1 - sigma*t) * e^(alpha*t)
#'   }
#'   
#'   for \eqn{\sigma t < 1}, where \eqn{\Gamma} is the gamma function.
#'
#' @seealso [actuar::Gumbel], [actuar::dgumbel()], [actuar::pgumbel()], 
#'   [actuar::qgumbel()], [actuar::rgumbel()], [actuar::mgumbel()]
#'
#' @examples
#' dist <- dist_gumbel(alpha = c(0.5, 1, 1.5, 3), scale = c(2, 2, 3, 4))
#' dist
#'
#' @examplesIf requireNamespace("actuar", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_gumbel
#' @export
dist_gumbel <- function(alpha, scale){
  alpha <- vec_cast(alpha, double())
  scale <- vec_cast(scale, double())
  if(any(scale <= 0)){
    abort("The scale parameter of a Gumbel distribution must be strictly positive.")
  }
  new_dist(a = alpha, s = scale, class = "dist_gumbel")
}

#' @export
format.dist_gumbel <- function(x, digits = 2, ...){
  sprintf(
    "Gumbel(%s, %s)",
    format(x[["a"]], digits = digits, ...),
    format(x[["s"]], digits = digits, ...)
  )
}

#' @export
density.dist_gumbel <- function(x, at, ...){
  require_package("actuar")
  actuar::dgumbel(at, x[["a"]], x[["s"]])
}

#' @export
log_density.dist_gumbel <- function(x, at, ...){
  require_package("actuar")
  actuar::dgumbel(at, x[["a"]], x[["s"]], log = TRUE)
}

#' @export
quantile.dist_gumbel <- function(x, p, ...){
  require_package("actuar")
  actuar::qgumbel(p, x[["a"]], x[["s"]])
}

#' @export
cdf.dist_gumbel <- function(x, q, ...){
  require_package("actuar")
  actuar::pgumbel(q, x[["a"]], x[["s"]])
}

#' @export
generate.dist_gumbel <- function(x, times, ...){
  require_package("actuar")
  actuar::rgumbel(times, x[["a"]], x[["s"]])
}

#' @export
mean.dist_gumbel <- function(x, ...){
  actuar::mgumbel(1, x[["a"]], x[["s"]])
}

#' @export
covariance.dist_gumbel <- function(x, ...){
  (pi*x[["s"]])^2/6
}

#' @export
skewness.dist_gumbel <- function(x, ...) {
  1.13954709940465
  # zeta3 <- 1.20205690315959401459612
  # (12 * sqrt(6) * zeta3) / pi^3
}

#' @export
kurtosis.dist_gumbel <- function(x, ...) 12/5

#' Missing distribution
#'
#' @description
#' `r lifecycle::badge('maturing')`
#'
#' A placeholder distribution for handling missing values in a vector of
#' distributions.
#'
#' @param length The number of missing distributions
#'
#' @name dist_missing
#'
#' @examples
#' dist <- dist_missing(3L)
#'
#' dist
#' mean(dist)
#' variance(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @export
dist_missing <- function(length = 1) {
  vctrs::vec_rep(NA_dist_, length)
}

NA_dist_ <- structure(list(NULL), class = c("distribution", "vctrs_vctr", "list"))

#' @export
format.dist_na <- function(x, ...) {
  "NA"
}

#' @export
density.dist_na <- function(x, at, ...){
  rep_len(NA_real_, length(at))
}

#' @export
log_density.dist_na <- density.dist_na

#' @export
quantile.dist_na <- function(x, p, ...){
  rep_len(NA_real_, length(p))
}
#' @export
log_quantile.dist_na <- quantile.dist_na

#' @export
cdf.dist_na <- function(x, q, ...){
  rep_len(NA_real_, length(q))
}
#' @export
log_cdf.dist_na <- cdf.dist_na

#' @export
generate.dist_na <- function(x, times, ...){
  rep(NA_real_, times)
}

#' @export
mean.dist_na <- function(x, ...) NA_real_

#' @export
covariance.dist_na <- function(x, ...) NA_real_

#' @export
skewness.dist_na <- function(x, ...) NA_real_

#' @export
kurtosis.dist_na <- function(x, ...) NA_real_

#' @export
Math.dist_na <- function(x, ...) {
  x
}

#' @export
Ops.dist_na <- function(e1, e2) {
  dist_missing(max(length(e1), length(e2)))
}

#' The (non-central) location-scale Student t Distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Student's T distribution is closely related to the [Normal()]
#' distribution, but has heavier tails. As \eqn{\nu} increases to \eqn{\infty},
#' the Student's T converges to a Normal. The T distribution appears
#' repeatedly throughout classic frequentist hypothesis testing when
#' comparing group means.
#'
#' @inheritParams stats::dt
#' @param mu The location parameter of the distribution.
#'   If `ncp == 0` (or `NULL`), this is the median.
#' @param sigma The scale parameter of the distribution.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_student_t")`
#'
#'   In the following, let \eqn{X} be a location-scale Student's T random variable with
#'   `df` = \eqn{\nu}, `mu` = \eqn{\mu}, `sigma` = \eqn{\sigma}, and
#'   `ncp` = \eqn{\delta} (non-centrality parameter).
#'
#'   If \eqn{Z} follows a standard Student's T distribution (with `df` = \eqn{\nu}
#'   and `ncp` = \eqn{\delta}), then \eqn{X = \mu + \sigma Z}.
#'
#'
#'   **Support**: \eqn{R}, the set of all real numbers
#'
#'   **Mean**:
#'
#'   For the central distribution (`ncp` = 0 or `NULL`):
#'
#'   \deqn{
#'     E(X) = \mu
#'   }{
#'     E(X) = \mu
#'   }
#'
#'   for \eqn{\nu > 1}, and undefined otherwise.
#'
#'   For the non-central distribution (`ncp` \eqn{\neq} 0):
#'
#'   \deqn{
#'     E(X) = \mu + \delta \sqrt{\frac{\nu}{2}} \frac{\Gamma((\nu-1)/2)}{\Gamma(\nu/2)} \sigma
#'   }{
#'     E(X) = \mu + \delta \sqrt(\nu / 2) \Gamma((\nu-1)/2) / \Gamma(\nu/2) \sigma
#'   }
#'
#'   for \eqn{\nu > 1}, and undefined otherwise.
#'
#'   **Variance**:
#'
#'   For the central distribution (`ncp` = 0 or `NULL`):
#'
#'   \deqn{
#'     \mathrm{Var}(X) = \frac{\nu}{\nu - 2} \sigma^2
#'   }{
#'     Var(X) = \nu / (\nu - 2) \sigma^2
#'   }
#'
#'   for \eqn{\nu > 2}. Undefined if \eqn{\nu \le 1}, infinite when \eqn{1 < \nu \le 2}.
#'
#'   For the non-central distribution (`ncp` \eqn{\neq} 0):
#'
#'   \deqn{
#'     \mathrm{Var}(X) = \left[\frac{\nu(1+\delta^2)}{\nu-2} - \left(\delta \sqrt{\frac{\nu}{2}} \frac{\Gamma((\nu-1)/2)}{\Gamma(\nu/2)}\right)^2\right] \sigma^2
#'   }{
#'     Var(X) = [(\nu(1+\delta^2))/(\nu-2) - (\delta \sqrt(\nu / 2) \Gamma((\nu-1)/2) / \Gamma(\nu/2))^2] \sigma^2
#'   }
#'
#'   for \eqn{\nu > 2}. Undefined if \eqn{\nu \le 1}, infinite when \eqn{1 < \nu \le 2}.
#'
#'   **Probability density function (p.d.f)**:
#'
#'   For the central distribution (`ncp` = 0 or `NULL`), the standard
#'   t distribution with `df` = \eqn{\nu} has density:
#'
#'   \deqn{
#'     f_Z(z) = \frac{\Gamma((\nu + 1)/2)}{\sqrt{\pi \nu} \Gamma(\nu/2)} \left(1 + \frac{z^2}{\nu} \right)^{- (\nu + 1)/2}
#'   }{
#'     f_Z(z) = \Gamma((\nu + 1) / 2) / (\sqrt(\pi \nu) \Gamma(\nu / 2)) (1 + z^2 / \nu)^(- (\nu + 1) / 2)
#'   }
#'
#'   The location-scale version with `mu` = \eqn{\mu} and `sigma` = \eqn{\sigma}
#'   has density:
#'
#'   \deqn{
#'     f(x) = \frac{1}{\sigma} f_Z\left(\frac{x - \mu}{\sigma}\right)
#'   }{
#'     f(x) = (1/\sigma) f_Z((x - \mu) / \sigma)
#'   }
#'
#'   For the non-central distribution (`ncp` \eqn{\neq} 0), the density is
#'   computed numerically via [stats::dt()].
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   For the central distribution (`ncp` = 0 or `NULL`), the cumulative
#'   distribution function is computed numerically via [stats::pt()], which
#'   uses the relationship to the incomplete beta function:
#'
#'   \deqn{
#'     F_\nu(t) = \frac{1}{2} I_x\left(\frac{\nu}{2}, \frac{1}{2}\right)
#'   }{
#'     F_\nu(t) = (1/2) I_x(\nu/2, 1/2)
#'   }
#'
#'   for \eqn{t \le 0}, where \eqn{x = \nu/(\nu + t^2)} and \eqn{I_x(a,b)} is
#'   the incomplete beta function ([stats::pbeta()]). For \eqn{t \ge 0}:
#'
#'   \deqn{
#'     F_\nu(t) = 1 - \frac{1}{2} I_x\left(\frac{\nu}{2}, \frac{1}{2}\right)
#'   }{
#'     F_\nu(t) = 1 - (1/2) I_x(\nu/2, 1/2)
#'   }
#'
#'   The location-scale version is: \eqn{F(x) = F_\nu((x - \mu)/\sigma)}.
#'
#'   For the non-central distribution (`ncp` \eqn{\neq} 0), the cumulative
#'   distribution function is computed numerically via [stats::pt()].
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   Does not exist in closed form. Moments are computed using the formulas
#'   for mean and variance above where available.
#'
#' @seealso [stats::TDist]
#'
#' @examples
#' dist <- dist_student_t(df = c(1,2,5), mu = c(0,1,2), sigma = c(1,2,3))
#'
#' dist
#' mean(dist)
#' variance(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_student_t
#' @export
dist_student_t <- function(df, mu = 0, sigma = 1, ncp = NULL){
  df <- vec_cast(df, numeric())
  if(any(df <= 0)){
    abort("The degrees of freedom parameter of a Student t distribution must be strictly positive.")
  }
  mu <- vec_cast(mu, double())
  sigma <- vec_cast(sigma, double())
  if(any(sigma[!is.na(sigma)] <= 0)){
    abort("The scale (sigma) parameter of a Student t distribution must be strictly positive.")
  }
  new_dist(df = df, mu = mu, sigma = sigma, ncp = ncp, class = "dist_student_t")
}

#' @export
format.dist_student_t <- function(x, digits = 2, ...){
  out <- sprintf(
    "t(%s, %s, %s%s)",
    format(x[["df"]], digits = digits, ...),
    format(x[["mu"]], digits = digits, ...),
    format(x[["sigma"]], digits = digits, ...),
    if(is.null(x[["ncp"]])) "" else paste(",", format(x[["ncp"]], digits = digits, ...))
  )
}

#' @export
density.dist_student_t <- function(x, at, ...){
  ncp <- x[["ncp"]] %||% missing_arg()
  sigma <- x[["sigma"]]

  stats::dt((at - x[["mu"]])/sigma, x[["df"]], ncp) / sigma
}

#' @export
log_density.dist_student_t <- function(x, at, ...){
  ncp <- x[["ncp"]] %||% missing_arg()
  sigma <- x[["sigma"]]

  stats::dt((at - x[["mu"]])/sigma, x[["df"]], ncp, log = TRUE) - log(sigma)
}


#' @export
quantile.dist_student_t <- function(x, p, ...){
  ncp <- x[["ncp"]] %||% missing_arg()

  stats::qt(p, x[["df"]], ncp) * x[["sigma"]] + x[["mu"]]
}

#' @export
cdf.dist_student_t <- function(x, q, ...){
  ncp <- x[["ncp"]] %||% missing_arg()

  stats::pt((q - x[["mu"]])/x[["sigma"]], x[["df"]], ncp)
}

#' @export
generate.dist_student_t <- function(x, times, ...){
  ncp <- x[["ncp"]] %||% missing_arg()

  stats::rt(times, x[["df"]], ncp) * x[["sigma"]] + x[["mu"]]
}

#' @export
mean.dist_student_t <- function(x, ...){
  df <- x[["df"]]
  if(df <= 1) return(NA_real_)
  if(is.null(x[["ncp"]])){
    x[["mu"]]
  } else {
    x[["mu"]] + x[["ncp"]] * sqrt(df/2) * (gamma((df-1)/2)/gamma(df/2)) * x[["sigma"]]
  }
}

#' @export
covariance.dist_student_t <- function(x, ...){
  df <- x[["df"]]
  ncp <- x[["ncp"]]
  if(df <= 1) return(NA_real_)
  if(df <= 2) return(Inf)
  if(is.null(ncp)){
    df / (df - 2) * x[["sigma"]]^2
  } else {
    ((df*(1+ncp^2))/(df-2) - (ncp * sqrt(df/2) * (gamma((df-1)/2)/gamma(df/2)))^2) * x[["sigma"]]^2
  }
}

#' @export
has_symmetry.dist_student_t <- function(x, ...) TRUE

#' Create a new distribution
#'
#' @description
#' `r lifecycle::badge('maturing')`
#'
#' Allows extension package developers to define a new distribution class
#' compatible with the distributional package.
#'
#' @param ... Parameters of the distribution (named).
#' @param class The class of the distribution for S3 dispatch.
#' @param dimnames The names of the variables in the distribution (optional).
#'
#' @export
new_dist <- function(..., class = NULL, dimnames = NULL){
  args <- transpose(vctrs::vec_recycle_common(...))
  wrap_dist(
    lapply(args, enclass_dist, class = class),
    dimnames = dimnames
  )
}

enclass_dist <- function(x, class) {
  structure(x, class = c(class, "dist_default"))
}


wrap_dist <- function(x, dimnames = NULL){
  vctrs::new_vctr(x, vars = dimnames, class = "distribution")
}

#' @export
vec_ptype_abbr.distribution <- function(x, ...){
  "dist"
}

#' @export
format.distribution <- function(x, ...){
  x <- vec_data(x)
  out <- vapply(x, format, character(1L), ...)
  out[vapply(x, is.null, logical(1L))] <- "NA"
  out
}

#' @importFrom pillar pillar_shaft new_pillar_shaft get_max_extent
#' @export
pillar_shaft.distribution <- function(x, ...) {
  dist = format(x)
  dist_min = format(x, width = 30)

  pillar::new_pillar_shaft(
    list(dist = dist,
         dist_min = dist_min),
    width = pillar::get_max_extent(dist),
    min_width = pillar::get_max_extent(dist_min),
    class = "pillar_distribution"
  )
}

#' @export
#' @importFrom pillar new_ornament
format.pillar_distribution <- function(x, width, ...) {
  if (get_max_extent(x$dist) <= width) {
    ornament <- x$dist
  } else {
    ornament <- x$dist_min
  }

  pillar::new_ornament(ornament, align = "right")
}

#' @export
`dimnames<-.distribution` <- function(x, value){
  attr(x, "vars") <- value
  x
}
#' @export
dimnames.distribution <- function(x){
  attr(x, "vars")
}

#' @export
`[[.distribution` <- `[`

#' The probability density/mass function
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Computes the probability density function for a continuous distribution, or
#' the probability mass function for a discrete distribution.
#'
#' @param x The distribution(s).
#' @param at The point at which to compute the density/mass.
#' @param ... Additional arguments passed to methods.
#' @param log If `TRUE`, probabilities will be given as log probabilities.
#'
#' @importFrom stats density
#' @export
density.distribution <- function(x, at, ..., log = FALSE){
  if(log) return(log_density(x, at, ...))
  at <- arg_listable(at, .ptype = NULL)
  dist_apply(x, density, at = at, ...)
}

log_density <- function(x, at, ...) {
  UseMethod("log_density")
}
#' @export
log_density.distribution <- function(x, at, ...){
  at <- arg_listable(at, .ptype = NULL)
  dist_apply(x, log_density, at = at, ...)
}

#' Distribution Quantiles
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Computes the quantiles of a distribution.
#'
#' @inheritParams density.distribution
#' @param p The probability of the quantile.
#' @param ... Additional arguments passed to methods.
#'
#' @importFrom stats quantile
#' @export
quantile.distribution <- function(x, p, ..., log = FALSE){
  if(log) return(log_quantile(x, p, ...))
  p <- arg_listable(p, .ptype = double())
  dist_apply(x, quantile, p = p, ...)
}
log_quantile <- function(x, p, ...) {
  UseMethod("log_quantile")
}
#' @export
log_quantile.distribution <- function(x, p, ...){
  vec_assert(q, double(), 1L)
  p <- arg_listable(p, .ptype = double())
  dist_apply(x, log_quantile, p = p, ...)
}

#' The cumulative distribution function
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' @inheritParams density.distribution
#' @param q The quantile at which the cdf is calculated.
#'
#' @name cdf
#' @export
cdf <- function (x, q, ..., log = FALSE){
  if(log) return(log_cdf(x, q, ...))
  UseMethod("cdf")
}
#' @rdname cdf
#' @export
cdf.distribution <- function(x, q, ...){
  q <- arg_listable(q, .ptype = NULL)
  dist_apply(x, cdf, q = q, ...)
}
log_cdf <- function(x, q, ...) {
  UseMethod("log_cdf")
}
#' @export
log_cdf.distribution <- function(x, q, ...){
  q <- arg_listable(q, .ptype = NULL)
  dist_apply(x, log_cdf, q = q, ...)
}

#' Randomly sample values from a distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Generate random samples from probability distributions.
#'
#' @param x The distribution(s).
#' @param times The number of samples.
#' @param ... Additional arguments used by methods.
#'
#' @export
generate.distribution <- function(x, times, ...){
  times <- vec_cast(times, integer())
  times <- vec_recycle(times, size = length(x))
  dn <- dimnames(x)
  x <- vec_data(x)
  dist_is_na <- vapply(x, is.null, logical(1L))
  x[dist_is_na] <- list(structure(list(), class = c("dist_na", "dist_default")))
  mapply(
    function(x, ...) {
      y <- generate(x, ...)
      if (is.matrix(y)) colnames(y) <- dn
      y
    }, x, times = times, ...,
    SIMPLIFY = FALSE
  )
  # dist_apply(x, generate, times = times, ...)
  # Needs work to structure MV appropriately.
}

#' The (log) likelihood of a sample matching a distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' @param x The distribution(s).
#' @param ... Additional arguments used by methods.
#'
#' @name likelihood
#' @export
likelihood <- function (x, ...){
  UseMethod("likelihood")
}

#' @rdname likelihood
#' @param sample A list of sampled values to compare to distribution(s).
#' @param log If `TRUE`, the log-likelihood will be computed.
#' @export
likelihood.distribution <- function(x, sample, ..., log = FALSE){
  if(vec_is(sample, numeric())) {
    warn("The `sample` argument of `likelihood()` should contain a list of numbers.
The same sample will be used for each distribution, i.e. `sample = list(sample)`.")
    sample <- list(sample)
  }
  if(log){
    dist_apply(x, log_likelihood, sample = sample, ...)
  } else {
    dist_apply(x, likelihood, sample = sample, ...)
  }
}

#' @rdname likelihood
#' @export
log_likelihood <- function(x, ...) {
  UseMethod("log_likelihood")
}
#' @export
log_likelihood.distribution <- function(x, sample, ...){
  dist_apply(x, log_likelihood, sample = sample, ...)
}

#' Extract the parameters of a distribution
#'
#' @description
#' `r lifecycle::badge('experimental')`
#'
#' @param x The distribution(s).
#' @param ... Additional arguments used by methods.
#'
#' @name parameters
#' @examples
#' dist <- c(
#'   dist_normal(1:2),
#'   dist_poisson(3),
#'   dist_multinomial(size = c(4, 3),
#'   prob = list(c(0.3, 0.5, 0.2), c(0.1, 0.5, 0.4)))
#'   )
#' parameters(dist)
#' @export
parameters <- function(x, ...) {
  UseMethod("parameters")
}

#' @rdname parameters
#' @export
parameters.distribution <- function(x, ...) {
  x <- lapply(vec_data(x), parameters)
  x <- lapply(x, function(z) data_frame(!!!z, .name_repair = "minimal"))
  vec_rbind(!!!x)
}

#' Extract the name of the distribution family
#'
#' @description
#' `r lifecycle::badge('experimental')`
#'
#' @param object The distribution(s).
#' @param ... Additional arguments used by methods.
#'
#' @examples
#' dist <- c(
#'   dist_normal(1:2),
#'   dist_poisson(3),
#'   dist_multinomial(size = c(4, 3),
#'   prob = list(c(0.3, 0.5, 0.2), c(0.1, 0.5, 0.4)))
#'   )
#' family(dist)
#'
#' @importFrom stats family
#' @export
family.distribution <- function(object, ...) {
  vapply(vec_data(object), family, character(1L))
}

#' Region of support of a distribution
#'
#' @description
#' `r lifecycle::badge('experimental')`
#'
#' @param x The distribution(s).
#' @param ... Additional arguments used by methods.
#'
#' @name support
#' @export
support <- function(x, ...) {
  UseMethod("support")
}

#' @rdname support
#' @export
support.distribution <- function(x, ...) {
  dist_apply(x, support, ...)
}

#' Mean of a probability distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Returns the empirical mean of the probability distribution. If the method
#' does not exist, the mean of a random sample will be returned.
#'
#' @param x The distribution(s).
#' @param ... Additional arguments used by methods.
#'
#' @export
mean.distribution <- function(x, ...){
  dist_apply(x, mean, ...)
}

#' Variance
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' A generic function for computing the variance of an object.
#'
#' @param x An object.
#' @param ... Additional arguments used by methods.
#'
#' @details
#'
#' The implementation of `variance()` for numeric variables coerces the input to
#' a vector then uses [`stats::var()`] to compute the variance. This means that,
#' unlike [`stats::var()`], if `variance()` is passed a matrix or a 2-dimensional
#' array, it will still return the variance ([`stats::var()`] returns the
#' covariance matrix in that case).
#'
#' @seealso [`variance.distribution()`], [`covariance()`]
#'
#' @export
variance <- function(x, ...){
  UseMethod("variance")
}
#' @export
variance.default <- function(x, ...){
  stop(
    "The variance() method is not supported for objects of type ",
    paste(deparse(class(x)), collapse = "")
  )
}
#' @rdname variance
#' @export
variance.numeric <- function(x, ...){
  stats::var(as.vector(x), ...)
}
#' @rdname variance
#' @export
variance.matrix <- function(x, ...){
  diag(stats::cov(x, ...))
}

#' Variance of a probability distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Returns the empirical variance of the probability distribution. If the method
#' does not exist, the variance of a random sample will be returned.
#'
#' @param x The distribution(s).
#' @param ... Additional arguments used by methods.
#'
#' @export
variance.distribution <- function(x, ...){
  dist_apply(x, variance, ...)
}

#' Covariance
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' A generic function for computing the covariance of an object.
#'
#' @param x An object.
#' @param ... Additional arguments used by methods.
#'
#' @seealso [`covariance.distribution()`], [`variance()`]
#'
#' @export
covariance <- function(x, ...){
  UseMethod("covariance")
}
#' @export
covariance.default <- function(x, ...){
  stop(
    "The covariance() method is not supported for objects of type ",
    paste(deparse(class(x)), collapse = "")
  )
}
#' @rdname variance
#' @export
covariance.numeric <- function(x, ...){
  stats::cov(x, ...)
}
#' Covariance of a probability distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Returns the empirical covariance of the probability distribution. If the
#' method does not exist, the covariance of a random sample will be returned.
#'
#' @param x The distribution(s).
#' @param ... Additional arguments used by methods.
#'
#' @export
covariance.distribution <- function(x, ...){
  dist_apply(x, covariance, ...)
}

#' Skewness of a probability distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' @param x The distribution(s).
#' @param ... Additional arguments used by methods.
#'
#' @export
skewness <- function(x, ...) {
  UseMethod("skewness")
}
#' @rdname skewness
#' @export
skewness.distribution <- function(x, ...){
  dist_apply(x, skewness, ...)
}

#' Kurtosis of a probability distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' @param x The distribution(s).
#' @param ... Additional arguments used by methods.
#'
#' @export
kurtosis <- function(x, ...) {
  UseMethod("kurtosis")
}
#' @rdname kurtosis
#' @export
kurtosis.distribution <- function(x, ...){
  dist_apply(x, kurtosis, ...)
}

#' Median of a probability distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Returns the median (50th percentile) of a probability distribution. This is
#' equivalent to `quantile(x, p=0.5)`.
#'
#' @param x The distribution(s).
#' @param na.rm Unused, included for consistency with the generic function.
#' @param ... Additional arguments used by methods.
#'
#' @importFrom stats median
#' @export
median.distribution <- function(x, na.rm = FALSE, ...){
  dist_apply(x, median, na.rm = na.rm, ...)
}

#' Probability intervals of a probability distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Returns a `hilo` central probability interval with probability coverage of
#' `size`. By default, the distribution's [`quantile()`] will be used to compute
#' the lower and upper bound for a centered interval
#'
#' @param x The distribution(s).
#' @param size The size of the interval (between 0 and 100).
#' @param ... Additional arguments used by methods.
#'
#' @seealso [`hdr.distribution()`]
#'
#' @importFrom stats median
#' @export
hilo.distribution <- function(x, size = 95, ...){
  size <- arg_listable(size, .ptype = double())
  dist_apply(x, hilo, size = size, ...)
}

#' Highest density regions of probability distributions
#'
#' @description
#' `r lifecycle::badge('maturing')`
#'
#' This function is highly experimental and will change in the future. In
#' particular, improved functionality for object classes and visualisation tools
#' will be added in a future release.
#'
#' Computes minimally sized probability intervals highest density regions.
#'
#' @param x The distribution(s).
#' @param size The size of the interval (between 0 and 100).
#' @param n The resolution used to estimate the distribution's density.
#' @param ... Additional arguments used by methods.
#'
#' @export
hdr.distribution <- function(x, size = 95, n = 512, ...){
  size <- arg_listable(size, .ptype = double())
  dist_apply(x, hdr, size = size, n = n, ...)
}

#' @export
sum.distribution <- function(x, ...){
  vec_restore(list(Reduce("+", x)), x)
}

#' @method vec_arith distribution
#' @export
vec_arith.distribution <- function(op, x, y, ...){
  UseMethod("vec_arith.distribution", y)
}
#' @method vec_arith.distribution default
#' @export
vec_arith.distribution.default <- function(op, x, y, ...){
  dist_is_na <- vapply(x, is.null, logical(1L))
  x[dist_is_na] <- list(structure(list(), class = c("dist_na", "dist_default")))
  if(is_empty(y)){
    out <- lapply(vec_data(x), get(op))
  }
  else {
    x <- vec_recycle_common(x = x, y = y)
    y <- x[["y"]]
    if(is_distribution(y)) y <- vec_data(y)
    x <- x[["x"]]
    out <- mapply(get(op), x = vec_data(x), y = y, SIMPLIFY = FALSE)
  }
  vec_restore(out, x)
}

#' @method vec_arith.numeric distribution
#' @export
vec_arith.numeric.distribution <- function(op, x, y, ...){
  x <- vec_recycle_common(x = x, y = y)
  y <- x[["y"]]
  x <- x[["x"]]
  out <- mapply(get(op), x = x, y = vec_data(y), SIMPLIFY = FALSE)
  vec_restore(out, y)
}

#' @method vec_math distribution
#' @export
vec_math.distribution <- function(.fn, .x, ...) {
  if(.fn %in% c("is.nan", "is.infinite")) return(rep_len(FALSE, length(.x)))
  if(.fn == "is.finite") return(rep_len(TRUE, length(.x)))
  out <- lapply(vec_data(.x), get(.fn), ...)
  vec_restore(out, .x)
}

#' @export
vec_ptype2.distribution.distribution <- function(x, y, ...){
  if(!identical(dimnames(x), dimnames(y))){
    abort("Distributions must have the same `dimnames` to be combined.")
  }
  x
}
#' @export
vec_ptype2.double.distribution <- function(x, y, ...) new_dist()
#' @export
vec_ptype2.distribution.double <- function(x, y, ...) new_dist()
#' @export
vec_ptype2.integer.distribution <- function(x, y, ...) new_dist()
#' @export
vec_ptype2.distribution.integer <- function(x, y, ...) new_dist()

#' @export
vec_cast.distribution.distribution <- function(x, to, ...){
  dimnames(x) <- dimnames(to)
  x
}
#' @export
vec_cast.distribution.double <- function(x, to, ...){
  x <- dist_degenerate(x)
  dimnames(x) <- dimnames(to)
  x
}
#' @export
vec_cast.distribution.integer <- vec_cast.distribution.double
#' @export
vec_cast.character.distribution <- function(x, to, ...){
  format(x)
}

#' Test if the object is a distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' This function returns `TRUE` for distributions and `FALSE` for all other objects.
#'
#' @param x An object.
#'
#' @return TRUE if the object inherits from the distribution class.
#' @rdname is-distribution
#' @examples
#' dist <- dist_normal()
#' is_distribution(dist)
#' is_distribution("distributional")
#' @export
is_distribution <- function(x) {
    inherits(x, "distribution")
}

test_that("Inverse Gamma distribution", {
  dist <- dist_inverse_gamma(3, 2)

  expect_equal(format(dist), "InvGamma(3, 0.5)")

  # Require package installed
  skip_if_not_installed("actuar", "2.0.0")

  # quantiles
  expect_equal(quantile(dist, 0.1), actuar::qinvgamma(0.1, 3, 2))
  expect_equal(quantile(dist, 0.5), actuar::qinvgamma(0.5, 3, 2))

  # pdf
  expect_equal(density(dist, 0), actuar::dinvgamma(0, 3, 2))
  expect_equal(density(dist, 3), actuar::dinvgamma(3, 3, 2))

  # cdf
  expect_equal(cdf(dist, 0), actuar::pinvgamma(0, 3, 2))
  expect_equal(cdf(dist, 3), actuar::pinvgamma(3, 3, 2))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4)), 0.4, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), (1/2) / (3 - 1))
  expect_equal(median(dist), actuar::qinvgamma(0.5, 3, 2))
  expect_equal(median(dist[[1]]), actuar::qinvgamma(0.5, 3, 2))
  expect_equal(variance(dist), (1/2)^2/((3-1)^2*(3-2)))
})

test_that("Studentized Range distribution", {
  dist <- dist_studentized_range(6, 5, 1)

  expect_equal(format(dist), "StudentizedRange(6, 5, 1)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qtukey(0.1, 6, 5, 1))
  expect_equal(quantile(dist, 0.5), stats::qtukey(0.5, 6, 5, 1))

  # pdf

  # cdf
  expect_equal(cdf(dist, 0), stats::ptukey(0, 6, 5, 1))
  expect_equal(cdf(dist, 3), stats::ptukey(3, 6, 5, 1))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)
})

test_that("Emperical/sample distribution", {
  x <- generate(dist_normal(0, 1), 100)
  dist <- dist_sample(x)

  expect_equal(format(dist), "sample[100]")

  # quantiles
  expect_equal(quantile(dist, 0.6), unname(quantile(x[[1]], 0.6)))
  expect_equal(quantile(dist, 0.24), unname(quantile(x[[1]], 0.24)))

  # pdf

  # cdf

  # F(Finv(a)) ~= a

  # stats
  expect_equal(mean(dist), mean(x[[1]]))
  expect_equal(median(dist), median(x[[1]]))
  expect_equal(median(dist[[1]]), median(x[[1]]))
  expect_equal(variance(dist), var(x[[1]]))

  # transform
  expect_equal(
    dist,
    dist + 1 - 1
  )
})

test_that("CDF of degenerate dist_sample() is correct", {
  expect_equal(cdf(dist_sample(list(2)), 2)[[1]], 1)
})

test_that("Degenerate distribution", {
  dist <- dist_degenerate(1)

  expect_equal(
    dist,
    vec_cast(1, new_dist())
  )

  expect_equal(format(dist), "1")

  # quantiles
  expect_equal(quantile(dist, 0), 1)
  expect_equal(quantile(dist, 0.5), 1)
  expect_equal(quantile(dist, 1), 1)

  # pdf
  expect_equal(density(dist, 1), 1)
  expect_equal(density(dist, 0.5), 0)
  expect_equal(density(dist, 0.99999), 0)

  # cdf
  expect_equal(cdf(dist, 0), 0)
  expect_equal(cdf(dist, 1), 1)
  expect_equal(cdf(dist, 0.9999), 0)

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 1)), 1, tolerance = 1e-3)
  expect_equal(cdf(dist, quantile(dist, 0)), 1, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 1)
  expect_equal(variance(dist), 0)
})

test_that("Multinomial distribution", {
  p <- c(0.3, 0.5, 0.2)
  dist <- dist_multinomial(size = 4, prob = list(p))
  dimnames(dist) <- c("a", "b", "c")

  expect_equal(format(dist), "Multinomial(4)[3]")

  # quantiles

  # pdf
  expect_equal(density(dist, cbind(1, 2, 1)), dmultinom(c(1, 2, 1), 4, p))

  # cdf
  # For multinomial, CDF is the probability that each category count is <= given values
  # This requires summing probabilities over all valid outcomes within the bounds
  
  # CDF should be 1 when bounds allow all possible outcomes
  expect_equal(cdf(dist, cbind(4, 4, 4)), 1)
  
  # CDF at a specific point should be >= PDF at that point
  test_point <- cbind(1, 2, 1)
  expect_gte(cdf(dist, test_point), density(dist, test_point))
  
  # CDF should be monotonically increasing
  expect_gte(cdf(dist, cbind(2, 2, 2)), cdf(dist, cbind(1, 1, 1)))
  
  # CDF at tight bounds should match the single point probability if only one outcome is possible
  # When bounds are (0, 0, 4), only the outcome (0, 0, 4) is valid
  expect_equal(cdf(dist, cbind(0, 0, 4)), dmultinom(c(0, 0, 4), 4, p))
  
  # CDF at bounds (1, 1, 4) should include multiple valid outcomes:
  # (0,0,4), (0,1,3), (1,0,3), (1,1,2)
  valid_outcomes <- rbind(c(0,0,4), c(0,1,3), c(1,0,3), c(1,1,2))
  expected_cdf <- sum(apply(valid_outcomes, 1, function(x) dmultinom(x, 4, p)))
  expect_equal(cdf(dist, cbind(1, 1, 4)), expected_cdf)
  
  # CDF at impossible values (where sum of bounds < size) should be 0
  expect_equal(cdf(dist, cbind(0, 0, 0)), 0)

  # F(Finv(a)) ~= a

  # stats
  expect_equal(
    mean(dist),
    matrix(c(1.2, 2, 0.8), nrow = 1, dimnames = list(NULL, c("a", "b", "c")))
  )

  expect_equal(
    covariance(dist),
    list(
      matrix(
        c(0.84, -0.6, -0.24, -0.6, 1, -0.4, -0.24, -0.4, 0.64),
        nrow = 3, dimnames = list(NULL, c("a", "b", "c"))
      )
    )
  )
})

test_that("Logarithmic distribution", {
  dist <- dist_logarithmic(0.66)

  expect_equal(format(dist), "Logarithmic(0.66)")

  # Require package installed
  skip_if_not_installed("actuar", "2.0.0")

  # quantiles
  expect_equal(quantile(dist, 0.5), actuar::qlogarithmic(0.5, 0.66))
  expect_equal(quantile(dist, 0.99), actuar::qlogarithmic(0.99, 0.66))

  # pdf
  expect_equal(density(dist, 1), actuar::dlogarithmic(1, 0.66))
  expect_equal(density(dist, 9), actuar::dlogarithmic(9, 0.66))

  # cdf
  expect_equal(cdf(dist, 3), actuar::plogarithmic(3, 0.66))
  expect_equal(cdf(dist, 12), actuar::plogarithmic(12, 0.66))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.9963064)), 0.9963064, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), -1/log(1-0.66)*(0.66/(1-0.66)))
  expect_equal(variance(dist), -(0.66^2 + 0.66*log(1-0.66))/((1-0.66)^2*log(1-0.66)^2))
})

test_that("is_distribution", {
  expect_false(is_distribution(iris))
  expect_true(is_distribution(dist_normal()))
  expect_false(is_distribution(NULL))
  expect_false(is_distribution(0))

  df <- data.frame(a = 1:10, b = dist_poisson(1:10), c = dist_normal(1:10))
  expect_true(all(sapply(df, is_distribution) == c(FALSE, TRUE, TRUE)))
})

test_that("variance() works correctly on vectors/matrices of different dimension", {
  x = 1:8

  expect_equal(variance(x), 6)
  expect_equal(variance(matrix(x, nrow = 2)), rep(0.5, 4))
})

test_that("variance() throws an error on non-numeric objects", {
  expect_error(variance("foo"))
})

test_that("Multivariate normal distribution", {
  mu <- c(1,2)
  sigma <- matrix(c(4,2,2,3), ncol=2)
  dist <- dist_multivariate_normal(mu = list(mu), sigma = list(sigma))
  dimnames(dist) <- c("a", "b")

  expect_equal(format(dist), "MVN[2]")

  # stats
  expect_equal(
    mean(dist),
    matrix(c(1,2), nrow = 1, dimnames = list(NULL, c("a", "b")))
  )
  expect_equal(covariance(dist), list(`colnames<-`(sigma, dimnames(dist))))

  # quantiles
  expect_equal(
    quantile(dist, 0.1, type = "marginal"),
    matrix(c(qnorm(0.1, mu[1], sqrt(sigma[1,1])), qnorm(0.1, mu[2], sqrt(sigma[2,2]))),
           nrow = 1, dimnames = list(NULL, c("a", "b")))
  )

  skip_if_not_installed("mvtnorm")
  expect_equivalent(
    quantile(dist, 0.1, type = "equicoordinate"),
    rep(mvtnorm::qmvnorm(0.1, mean = mu, sigma = sigma)$quantile, 2)
  )

  # pdf
  expect_equal(density(dist, cbind(1, 2)), mvtnorm::dmvnorm(c(1, 2), mean = mu, sigma = sigma))
  expect_equal(density(dist, cbind(-3, 4)), mvtnorm::dmvnorm(c(-3, 4), mean = mu, sigma = sigma))

  # cdf
  expect_equivalent(cdf(dist, cbind(1, 2)), mvtnorm::pmvnorm(upper = c(1,2), mean = mu, sigma = sigma))
  expect_equivalent(cdf(dist, cbind(-3, 4)), mvtnorm::pmvnorm(upper = c(-3, 4), mean = mu, sigma = sigma))

  # F(Finv(a)) ~= a
  # expect_equal(cdf(dist, list(as.numeric(quantile(dist, 0.53)))), 0.53, tolerance = 1e-3)
})

test_that("F distribution", {
  dist <- dist_f(5, 2)

  expect_equal(format(dist), "F(5, 2)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qf(0.1, 5, 2))
  expect_equal(quantile(dist, 0.5), stats::qf(0.5, 5, 2))

  # pdf
  expect_equal(density(dist, 0), stats::df(0, 5, 2))
  expect_equal(density(dist, 3), stats::df(3, 5, 2))

  # cdf
  expect_equal(cdf(dist, 0), stats::pf(0, 5, 2))
  expect_equal(cdf(dist, 3), stats::pf(3, 5, 2))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), NA_real_)
  expect_equal(variance(dist), NA_real_)
  dist <- dist_f(5, 5)
  expect_equal(mean(dist), 5/(5-2))
  expect_equal(variance(dist), 2*5^2*(5+5-2)/(5*(5-2)^2*(5-4)))
})

test_that("Negative Binomial distribution", {
  dist <- dist_negative_binomial(10, 0.4)

  expect_equal(format(dist), "NB(10, 0.4)")

  # quantiles
  expect_equal(quantile(dist, 0.6), stats::qnbinom(0.6, 10, 0.4))
  expect_equal(quantile(dist, 0.61), stats::qnbinom(0.61, 10, 0.4))

  # pdf
  expect_equal(density(dist, 0), stats::dnbinom(0, 10, 0.4))
  expect_equal(density(dist, 1), stats::dnbinom(1, 10, 0.4))

  # cdf
  expect_equal(cdf(dist, 0), stats::pnbinom(0, 10, 0.4))
  expect_equal(cdf(dist, 1), stats::pnbinom(1, 10, 0.4))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.6358)), 0.6358, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 0.6*10/(1-0.6))
  expect_equal(variance(dist), 0.6*10/(1-0.6)^2)
})

test_that("Burr distribution", {

  dist <- dist_burr(2, 3)

  expect_equal(format(dist), "Burr12(2, 3, 1)")

  # Require package installed
  skip_if_not_installed("actuar", "2.0.0")

  # quantiles
  expect_equal(quantile(dist, 0.1), actuar::qburr(0.1, 2, 3))
  expect_equal(quantile(dist, 0.5), actuar::qburr(0.5, 2, 3))

  # pdf
  expect_equal(density(dist, 0), actuar::dburr(0, 2, 3))
  expect_equal(density(dist, 3), actuar::dburr(3, 2, 3))

  # cdf
  expect_equal(cdf(dist, 0), actuar::pburr(0, 2, 3))
  expect_equal(cdf(dist, 3), actuar::pburr(3, 2, 3))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4)), 0.4, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), actuar::mburr(1, 2, 3))
  expect_equal(variance(dist), actuar::mburr(2, 2, 3) - actuar::mburr(1, 2, 3)^2)
})

test_that("Beta distribution", {
  dist <- dist_beta(3, 4)

  expect_equal(format(dist), "Beta(3, 4)")

  # quantiles
  expect_equal(quantile(dist, 0.1), qbeta(0.1, 3, 4))
  expect_equal(quantile(dist, 0.5), qbeta(0.5, 3, 4))

  # pdf
  expect_equal(density(dist, 0), dbeta(0, 3, 4))
  expect_equal(density(dist, 3), dbeta(3, 3, 4))

  # cdf
  expect_equal(cdf(dist, 0), pbeta(0, 3, 4))
  expect_equal(cdf(dist, 3), pbeta(3, 3, 4))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4)), 0.4, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 3/(3+4))
  expect_equal(variance(dist), 3*4/((3+4)^2*(3+4+1)))
})

test_that("Chisq distribution", {
  dist <- dist_chisq(9)

  expect_equal(format(dist), "x2(9)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qchisq(0.1, 9))
  expect_equal(quantile(dist, 0.5), stats::qchisq(0.5, 9))

  # pdf
  expect_equal(density(dist, 0), stats::dchisq(0, 9))
  expect_equal(density(dist, 3), stats::dchisq(3, 9))

  # cdf
  expect_equal(cdf(dist, 0), stats::pchisq(0, 9))
  expect_equal(cdf(dist, 3), stats::pchisq(3, 9))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 9)
  expect_equal(variance(dist), 2*9)
})

test_that("Horseshoe distribution - format and structure", {
  dist <- dist_horseshoe(lambda = c(0.5, 1, 2), tau = 1)

  # length
  expect_length(dist, 3L)

  # format
  expect_equal(format(dist), c("HS(0.5, 1)", "HS(1, 1)", "HS(2, 1)"))
})

test_that("Horseshoe distribution - density", {
  lambda <- 1
  tau <- 1
  dist <- dist_horseshoe(lambda = lambda, tau = tau)

  x_vals <- c(-2, -1, 0, 1, 2)

  expect_equal(
    density(dist, x_vals)[[1]],
    LaplacesDemon::dhs(x_vals, lambda, tau)
  )
  expect_equal(
    density(dist, x_vals, log = TRUE)[[1]],
    LaplacesDemon::dhs(x_vals, lambda, tau, log = TRUE)
  )
})

test_that("Horseshoe distribution - support", {
  dist <- dist_horseshoe(lambda = 1, tau = 1)
  s <- support(dist)

  expect_s3_class(s, "support_region")
  expect_equal(format(s), "R")
})

test_that("Horseshoe distribution - generate", {
  dist <- dist_horseshoe(lambda = 1, tau = 1)

  set.seed(42)
  samples <- generate(dist, 100)[[1]]

  expect_length(samples, 100L)
  expect_type(samples, "double")
  # Horseshoe is symmetric around 0 -- large sample mean should be near 0
  expect_lt(abs(mean(samples)), 1)
})

test_that("Horseshoe distribution - input validation", {
  expect_error(
    dist_horseshoe(lambda = -1, tau = 1),
    "lambda"
  )
  expect_error(
    dist_horseshoe(lambda = 0, tau = 1),
    "lambda"
  )
  expect_error(
    dist_horseshoe(lambda = 1, tau = -1),
    "tau"
  )
  expect_error(
    dist_horseshoe(lambda = 1, tau = 0),
    "tau"
  )
})

test_that("Horseshoe distribution - heavier tails with larger lambda", {
  # With larger lambda the density at x=0 is lower (more spread),
  # but the tails are heavier
  dist_narrow <- dist_horseshoe(lambda = 0.1, tau = 1)
  dist_wide   <- dist_horseshoe(lambda = 5,   tau = 1)

  # Larger lambda => lower peak at 0
  expect_gt(density(dist_narrow, 0), density(dist_wide, 0))
})

test_that("Horseshoe distribution - larger tau scales distribution", {
  dist_small <- dist_horseshoe(lambda = 1, tau = 0.5)
  dist_large <- dist_horseshoe(lambda = 1, tau = 2)

  # Larger tau => more spread => lower peak at 0
  expect_gt(density(dist_small, 0), density(dist_large, 0))
})

test_that("Logistic distribution", {
  dist <- dist_logistic(5, 2)

  expect_equal(format(dist), "Logistic(5, 2)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qlogis(0.1, 5, 2))
  expect_equal(quantile(dist, 0.5), stats::qlogis(0.5, 5, 2))

  # pdf
  expect_equal(density(dist, 0), stats::dlogis(0, 5, 2))
  expect_equal(density(dist, 3), stats::dlogis(3, 5, 2))

  # cdf
  expect_equal(cdf(dist, 0), stats::plogis(0, 5, 2))
  expect_equal(cdf(dist, 3), stats::plogis(3, 5, 2))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 5)
  expect_equal(variance(dist), (2*pi)^2/3)
})

test_that("Gumbel distribution", {
  dist <- dist_gumbel(1, 2)

  expect_equal(format(dist), "Gumbel(1, 2)")

  # Require package installed
  skip_if_not_installed("actuar", "2.0.0")

  # quantiles
  expect_equal(quantile(dist, 0.1), actuar::qgumbel(0.1, 1, 2))
  expect_equal(quantile(dist, 0.5), actuar::qgumbel(0.5, 1, 2))

  # pdf
  expect_equal(density(dist, 0), actuar::dgumbel(0, 1, 2))
  expect_equal(density(dist, 3), actuar::dgumbel(3, 1, 2))

  # cdf
  expect_equal(cdf(dist, 0), actuar::pgumbel(0, 1, 2))
  expect_equal(cdf(dist, 3), actuar::pgumbel(3, 1, 2))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4)), 0.4, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), actuar::mgumbel(1, 1, 2))
  expect_equal(variance(dist), actuar::mgumbel(2, 1, 2) - actuar::mgumbel(1, 1, 2)^2)
})

test_that("Check zero inflation", {
  dist <- dist_inflated(dist_poisson(6), 0.33)

  expect_equal(format(dist), "0+Pois(6)")

  # quantiles
  expect_equal(quantile(dist, 0.1), 0)
  expect_equal(quantile(dist, 0.5), 4)

  # pdf
  expect_equal(density(dist, 0), 0.33 + 0.67*dpois(0, 6))
  expect_equal(density(dist, 3), 0.67*dpois(3, 6))

  # cdf
  expect_equal(cdf(dist, 0), 0.33 + 0.67*ppois(0, 6))
  expect_equal(cdf(dist, 3), 0.33 + 0.67*ppois(3, 6))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.52)), 0.52, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 0.67*6)
  expect_equal(variance(dist), 0.67*6 + (0.33/0.67)*(0.67*6)^2)
})

test_that("Check non-zero inflation", {
  dist <- dist_inflated(dist_poisson(6), 0.33, 2)

  expect_equal(format(dist), "2+Pois(6)")

  # quantiles
  expect_equal(quantile(dist, 0), 0)
  expect_equal(quantile(dist, 0.1), 2)
  expect_equal(quantile(dist, 0.33), 2)
  expect_equal(quantile(dist, 0.5), 4)

  # pdf
  expect_equal(density(dist, 0), 0.67*dpois(0, 6))
  expect_equal(density(dist, 2), 0.33 + 0.67*dpois(2, 6))
  expect_equal(density(dist, 3), 0.67*dpois(3, 6))

  # cdf
  expect_equal(cdf(dist, 0), 0.67*ppois(0, 6))
  expect_equal(cdf(dist, 2), 0.33 + 0.67*ppois(2, 6))
  expect_equal(cdf(dist, 3), 0.33 + 0.67*ppois(3, 6))

  # stats
  expect_equal(mean(dist), 0.33*2 + 0.67*6)
})

test_that("continuous distribution inflated at 0 has mass below inflated point", {
  # Normal(0,1) inflated at 0: base dist has mass below 0, so quantile(p=0) = -Inf
  dist <- dist_inflated(dist_normal(0, 1), prob = 0.3, x = 0)

  # p=0 should be -Inf (base dist has support below 0)
  expect_equal(quantile(dist, 0), -Inf)

  # p=1 should be +Inf
  expect_equal(quantile(dist, 1), Inf)

  # F_inflated(0) = 0.3 + 0.7 * 0.5 = 0.65, so p just below 0.65 maps below 0
  # and p just above maps to 0 or above
  # F_base(0) = 0.5, so the jump in F_inflated covers (0.7*0.5, 0.3 + 0.7*0.5] = (0.35, 0.65]
  # p < 0.35 maps to base dist below 0; 0.35 < p <= 0.65 maps to inflated point; p > 0.65 maps above
  expect_lt(quantile(dist, 0.34), 0)
  expect_equal(quantile(dist, 0.35), 0)  # right at jump start: inflated point
  expect_gt(quantile(dist, 0.66), 0)

  # Scalar / vector consistency
  probs <- c(0, 0.1, 0.35, 0.50, 0.65, 0.80, 1)
  result <- quantile(dist, probs)[[1]]
  scalar <- vapply(probs, function(p) quantile(dist, p), numeric(1))
  expect_equal(result, scalar)
})

test_that("quantile.dist_inflated is vectorised over p (zero-inflated)", {
  # Reproduces the bug in issue #151
  dist <- dist_inflated(dist_poisson(4), prob = 0.30, x = 0)
  probs <- c(0.01, 0.10, 0.25, 0.50, 0.75, 0.90, 0.99)

  # Must not error and return a length-1 list containing all quantile values
  result <- quantile(dist, probs)
  expect_type(result, "list")
  expect_length(result, 1L)
  expect_length(result[[1]], length(probs))

  # Each element must equal the scalar result for the same probability
  scalar_results <- vapply(probs, function(p) quantile(dist, p), numeric(1))
  expect_equal(result[[1]], scalar_results)

  # Boundary: p = 0 and p = 1
  boundary <- quantile(dist, c(0, 1))
  expect_length(boundary[[1]], 2)

  # Values below the inflation mass return the inflated point (0)
  expect_equal(quantile(dist, 0.10), 0)   # 0.10 < prob = 0.30, so inflated point
  expect_true(quantile(dist, 0.99) > 0)   # well above inflation mass
})

test_that("quantile.dist_inflated is vectorised over p (non-zero-inflated)", {
  dist <- dist_inflated(dist_poisson(6), prob = 0.33, x = 2)
  probs <- c(0.05, 0.20, 0.40, 0.60, 0.80, 0.95)

  result <- quantile(dist, probs)
  expect_type(result, "list")
  expect_length(result, 1L)
  expect_length(result[[1]], length(probs))

  scalar_results <- vapply(probs, function(p) quantile(dist, p), numeric(1))
  expect_equal(result[[1]], scalar_results)

  # CDF round-trip: F(Q(p)) >= p for discrete distributions
  cdfs <- vapply(result[[1]], function(q) cdf(dist, q), numeric(1))
  expect_true(all(cdfs >= probs - 1e-9))
})

test_that("quantile.dist_inflated scalar and length-1 vector agree", {
  dist <- dist_inflated(dist_poisson(4), prob = 0.30, x = 0)
  expect_equal(quantile(dist, 0.5), quantile(dist, c(0.5))[[1]])
})

test_that("Categorical distribution", {
  dist <- dist_categorical(list(c(0.4, 0.2, 0.3, 0.1)))

  expect_equal(format(dist), "Categorical[4]")

  # quantiles
  expect_true(all(is.na(quantile(dist, 0.5))))
  expect_true(all(is.na(quantile(dist, 0.2))))
  expect_equal(quantile(dist, c(0.1, 0.9)), list(c(NA_real_, NA_real_)))

  # pdf
  expect_equal(density(dist, -1), NA_real_)
  expect_equal(density(dist, 0), NA_real_)
  expect_equal(density(dist, 1), 0.4)
  expect_equal(density(dist, 2), 0.2)
  expect_equal(density(dist, 5), NA_real_)
  expect_equal(density(dist, 3:5), list(c(0.3, 0.1, NA_real_)))

  # cdf
  expect_true(all(is.na(cdf(dist, 1))))
  expect_equal(cdf(dist, 1:2), list(c(NA_real_, NA_real_)))

  # stats
  expect_true(all(is.na(mean(dist))))
  expect_true(all(is.na(variance(dist))))
})

test_that("Bernoulli distribution", {
  dist <- dist_bernoulli(0.4)

  expect_equal(format(dist), "Bernoulli(0.4)")

  # quantiles
  expect_equal(quantile(dist, 0.6), FALSE)
  expect_equal(quantile(dist, 0.61), TRUE)

  # pdf
  expect_equal(density(dist, 0), stats::dbinom(0, 1, 0.4))
  expect_equal(density(dist, 1), stats::dbinom(1, 1, 0.4))

  # cdf
  expect_equal(cdf(dist, 0), stats::pbinom(0, 1, 0.4))
  expect_equal(cdf(dist, 1), stats::pbinom(1, 1, 0.4))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.6)), 0.6, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 0.4)
  expect_equal(variance(dist), 0.4*(1-0.4))
})

test_that("Normal distribution", {
  # defaults
  dist <- dist_normal()
  expect_equal(mean(dist), 0)
  expect_equal(variance(dist), 1)

  # display
  expect_s3_class(dist, "distribution")
  expect_output(print(dist), "N\\(0, 1\\)")
  expect_output(print(dist_normal(numeric())), "<distribution\\[0\\]>")

  # error checking
  expect_error(
    dist_normal(0, -1),
    "non-negative"
  )
  expect_silent(
    dist_normal(mu = 0L, sigma = 1L)
  )

  mu <- rnorm(10)
  sigma <- abs(rnorm(10))
  dist <- dist_normal(mu, sigma)

  # summary statistics
  expect_equal(mean(dist), mu)
  expect_equal(median(dist), mu)
  expect_equal(median(dist[[1]]), mu[[1]])
  expect_equal(variance(dist), sigma^2)

  # math
  expect_equal(mean(dist*3+1), mu*3+1)
  expect_equal(variance(dist*3+1), (sigma*3)^2)
  expect_equal(mean(dist + dist), mean(dist) + mean(dist))
  expect_equal(variance(dist + dist), variance(dist) + variance(dist))

  # density
  expect_equal(
    density(dist, 0),
    dnorm(0, mean = mu, sd = sigma)
  )
  # cdf
  expect_equal(
    cdf(dist, 5),
    pnorm(5, mean = mu, sd = sigma)
  )
  # quantile
  expect_equal(
    quantile(dist, 0.1),
    qnorm(0.1, mean = mu, sd = sigma)
  )
  # generate
  expect_equal(
    {
      set.seed(0)
      generate(dist, 10)
    },
    {
      set.seed(0)
      mapply(function(m, s) rnorm(10, m, s), m = mu, s = sigma, SIMPLIFY = FALSE)
    }
  )
})

test_that("Geometric distribution", {
  dist <- dist_geometric(0.4)

  expect_equal(format(dist), "Geometric(0.4)")

  # quantiles
  expect_equal(quantile(dist, 0.6), stats::qgeom(0.6, 0.4))
  expect_equal(quantile(dist, 0.9), stats::qgeom(0.9, 0.4))

  # pdf
  expect_equal(density(dist, 0), stats::dgeom(0, 0.4))
  expect_equal(density(dist, 5), stats::dgeom(5, 0.4))

  # cdf
  expect_equal(cdf(dist, 0), stats::pgeom(0, 0.4))
  expect_equal(cdf(dist, 10), stats::pgeom(10, 0.4))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.64)), 0.64, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 1/0.4 - 1)
  expect_equal(variance(dist), 1/0.4^2 - 1/0.4)
})

test_that("Poisson Inverse Gaussian distribution", {
  dist <- dist_poisson_inverse_gaussian(0.1, 0.8)

  expect_equal(format(dist), "PIG(0.1, 0.8)")

  # Require package installed
  skip_if_not_installed("actuar", "2.0.0")

  # quantiles
  expect_equal(quantile(dist, 0.1), actuar::qpig(0.1, 0.1, 0.8))
  expect_equal(quantile(dist, 0.5), actuar::qpig(0.5, 0.1, 0.8))

  # pdf
  expect_equal(density(dist, 0), actuar::dpig(0, 0.1, 0.8))
  expect_equal(density(dist, 3), actuar::dpig(3, 0.1, 0.8))

  # cdf
  expect_equal(cdf(dist, 0), actuar::ppig(0, 0.1, 0.8))
  expect_equal(cdf(dist, 3), actuar::ppig(3, 0.1, 0.8))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.994)), 0.994, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 0.1)
  expect_equal(variance(dist), 0.1/0.8*(0.1^2 + 0.8))
})

test_that("Truncated Normal distributions", {
  dist <- dist_truncated(dist_normal(0, 1), -5, 5)

  # format
  expect_equal(format(dist), sprintf("%s[-5,5]", format(dist_normal(0,1))))

  # quantiles
  expect_equal(
    quantile(dist, 0.1),
    -1.28155025885944 #dput(extraDistr::qtnorm(0.1, 0, 1, -5, 5))
  )
  expect_equal(
    quantile(dist, 0.5),
    -1.39145821233588e-16 #dput(extraDistr::qtnorm(0.5, 0, 1, -5, 5))
  )

  # pdf
  expect_equal(
    density(dist, 0),
    0.398942509116427, #dput(extraDistr::dtnorm(0, 0, 1, -5, 5))
  )
  expect_equal(
    density(dist, 3),
    0.00443185095273209, #dput(extraDistr::dtnorm(3, 0, 1, -5, 5))
  )

  # cdf
  expect_equal(cdf(dist, 0), 0.5)
  expect_equal(
    cdf(dist, 3),
    0.998650387846205, #dput(extraDistr::ptnorm(3, 0, 1, -5, 5))
  )

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.3)), 0.3, tolerance = 1e-6)

  # stats
  # expect_equal(mean(dist), ???)
  # expect_equal(variance(dist), ???)
})

test_that("Truncated Binomial distributions", {
  dist <- dist_truncated(dist_binomial(100, 0.83), 76, 86)

  # format
  expect_equal(format(dist), sprintf("%s[76,86]", format(dist_binomial(100,0.83))))

  # quantiles
  expect_equal(
    quantile(dist, 0.1),
    79 #dput(extraDistr::qtbinom(0.1, 100, 0.83, 76, 86))
  )
  expect_equal(
    quantile(dist, 0.5),
    82 #dput(extraDistr::qtbinom(0.5, 100, 0.83, 76, 86))
  )

  # pdf
  expect_equal(density(dist, 75), 0)
  expect_equal(density(dist, 87), 0)
  expect_equal(
    density(dist, 80),
    0.094154977726162, #dput(extraDistr::dtbinom(80, 100, 0.83, 76, 86))
  )
  expect_equal(
    density(dist, 85),
    0.123463609708811, #dput(extraDistr::dtbinom(85, 100, 0.83, 76, 86))
  )

  # cdf
  expect_equal(cdf(dist, 0), 0)
  expect_equal(cdf(dist, 76), 0)
  expect_equal(cdf(dist, 86), 1)
  expect_equal(
    cdf(dist, 80),
    0.259185477677455, #dput(extraDistr::ptbinom(80, 100, 0.83, 76, 86))
  )

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.372)), 0.372, tolerance = 1e-3)

  # stats
  # expect_equal(mean(dist), ???)
  # expect_equal(variance(dist), ???)
})

test_that("Cauchy distribution", {
  dist <- dist_cauchy(-2, 1)

  expect_equal(format(dist), "Cauchy(-2, 1)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qcauchy(0.1, -2, 1))
  expect_equal(quantile(dist, 0.5), stats::qcauchy(0.5, -2, 1))

  # pdf
  expect_equal(density(dist, 0), stats::dcauchy(0, -2, 1))
  expect_equal(density(dist, 3), stats::dcauchy(3, -2, 1))

  # cdf
  expect_equal(cdf(dist, 0), stats::pcauchy(0, -2, 1))
  expect_equal(cdf(dist, 3), stats::pcauchy(3, -2, 1))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), NA_real_)
  expect_equal(variance(dist), NA_real_)
})

test_that("Inverse Exponential distribution", {
  dist <- dist_inverse_exponential(5)

  expect_equal(format(dist), "InvExp(5)")

  # Require package installed
  skip_if_not_installed("actuar", "2.0.0")

  # quantiles
  expect_equal(quantile(dist, 0.1), actuar::qinvexp(0.1, 5))
  expect_equal(quantile(dist, 0.5), actuar::qinvexp(0.5, 5))

  # pdf
  expect_equal(density(dist, 0), actuar::dinvexp(0, 5))
  expect_equal(density(dist, 3), actuar::dinvexp(3, 5))

  # cdf
  expect_equal(cdf(dist, 0), actuar::pinvexp(0, 5))
  expect_equal(cdf(dist, 3), actuar::pinvexp(3, 5))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4)), 0.4, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), NA_real_) # dput(actuar::minvexp(1, 5))
  expect_equal(variance(dist), NA_real_) # dput(actuar::minvexp(2, 5) - actuar::minvexp(1, 5)^2)
})

test_that("Exponential distribution", {
  dist <- dist_exponential(2)

  expect_equal(format(dist), "Exp(2)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qexp(0.1, 2))
  expect_equal(quantile(dist, 0.5), stats::qexp(0.5, 2))

  # pdf
  expect_equal(density(dist, 0), stats::dexp(0, 2))
  expect_equal(density(dist, 3), stats::dexp(3, 2))

  # cdf
  expect_equal(cdf(dist, 0), stats::pexp(0, 2))
  expect_equal(cdf(dist, 3), stats::pexp(3, 2))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 1/2)
  expect_equal(variance(dist), 1/2^2)
})

test_that("g-and-h distribution", {
  # Define g-and-h distribution parameters
  A <- 0
  B <- 1
  g <- 0
  h <- 0.5
  c <- 0.8
  dist <- dist_gh(A, B, g, h, c)

  # Check formatting
  expect_equal(format(dist), "gh(A = 0, B = 1, g = 0, h = 0.5)")

  # Require package installed
  skip_if_not_installed("gk", "0.1.0")

  # quantiles
  expect_equal(quantile(dist, 0.1), gk::qgh(0.1, A, B, g, h, c))
  expect_equal(quantile(dist, 0.5), gk::qgh(0.5, A, B, g, h, c))

  # pdf
  expect_equal(density(dist, 0), gk::dgh(0, A, B, g, h, c))
  expect_equal(density(dist, 3), gk::dgh(3, A, B, g, h, c))

  # cdf
  expect_equal(cdf(dist, 0), gk::pgh(0, A, B, g, h, c))
  expect_equal(cdf(dist, 3), gk::pgh(3, A, B, g, h, c))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4)), 0.4, tolerance = 1e-3)

  # Generate random samples
  set.seed(123)
  samples <- generate(dist, 10)
  set.seed(123)
  expect_equal(samples[[1L]], gk::rgh(10, A, B, g, h, c))
})

test_that("Uniform distribution", {
  dist <- dist_uniform(-2, 4)

  expect_equal(format(dist), "U(-2, 4)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qunif(0.1, -2, 4))
  expect_equal(quantile(dist, 0.5), stats::qunif(0.5, -2, 4))

  # pdf
  expect_equal(density(dist, 0), stats::dunif(0, -2, 4))
  expect_equal(density(dist, 3), stats::dunif(3, -2, 4))

  # cdf
  expect_equal(cdf(dist, 0), stats::punif(0, -2, 4))
  expect_equal(cdf(dist, 3), stats::punif(3, -2, 4))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 0.5*(-2 + 4))
  expect_equal(variance(dist), (4+2)^2/12)
})

test_that("Missing distribution", {
  dist <- dist_missing()

  expect_equal(format(dist), "NA")

  # quantiles
  expect_equal(quantile(dist, c(0.1, 0.9)), list(c(NA_real_, NA_real_)))

  # pdf
  expect_equal(density(dist, 1:2), list(c(NA_real_, NA_real_)))

  # cdf
  expect_equal(cdf(dist, 1:2), list(c(NA_real_, NA_real_)))

  # stats
  expect_equal(mean(dist), NA_real_)
  expect_equal(variance(dist), NA_real_)
})

test_that("Mixture of Normals", {
  dist <- dist_mixture(dist_normal(0, 1), dist_normal(10, 4), weights = c(0.5, 0.5))

  # format
  expect_equal(format(dist), "mixture(0.5*N(0, 1), 0.5*N(10, 16))")

  # quantiles
  expect_equal(quantile(dist, 0.5), 2, tolerance = 1e-5)
  expect_equal(quantile(dist, 0.1), -0.854, tolerance = 1e-3)

  # pdf
  expect_equal(density(dist, 0), 0.5 * dnorm(0) + 0.5 * dnorm(0, 10, 4))
  expect_equal(density(dist, 3), 0.5 * dnorm(3) + 0.5 * dnorm(3, 10, 4))

  # cdf
  expect_equal(cdf(dist, 0), 0.5 * pnorm(0) + 0.5 * pnorm(0, 10, 4))
  expect_equal(cdf(dist, 3), 0.5 * pnorm(3) + 0.5 * pnorm(3, 10, 4))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.5)), 0.5, tolerance = 1e-6)

  expect_equal(mean(dist), 5)
  expect_equal(variance(dist), 33.5)
})

test_that("Mixture of different distributions", {
  dist <- dist_mixture(dist_normal(0, 1), dist_student_t(10), weights = c(0.3, 0.7))

  # format
  expect_equal(format(dist), "mixture(0.3*N(0, 1), 0.7*t(10, 0, 1))")

  # quantiles
  expect_equal(quantile(dist, 0.5), 0, tolerance = 1e-5)
  expect_equal(quantile(dist, 0.1), -1.343, tolerance = 1e-3)

  # pdf
  expect_equal(density(dist, 0), 0.3 * dnorm(0) + 0.7 * dt(0, 10))
  expect_equal(density(dist, 3), 0.3 * dnorm(3) + 0.7 * dt(3, 10))

  # cdf
  expect_equal(cdf(dist, 0), 0.3 * pnorm(0) + 0.7 * pt(0, 10))
  expect_equal(cdf(dist, 3), 0.3 * pnorm(3) + 0.7 * pt(3, 10))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.5)), 0.5, tolerance = 1e-6)

  expect_equal(mean(dist), 0)
  expect_equal(variance(dist), 1.175)
})

test_that("Mixture of point masses", {
  dist <- dist_mixture(dist_degenerate(1), dist_degenerate(2), dist_degenerate(3), weights = c(0.1, 0.2, 0.7))

  # format
  expect_equal(format(dist, width = 10), "mixture(n=3)")
  expect_equal(format(dist), "mixture(0.1*1, 0.2*2, 0.7*3)")

  # quantiles
  expect_equal(quantile(dist, c(0, 0.1, 0.3, 1))[[1]], c(1, 1:3), tolerance = .Machine$double.eps^0.25)

  # pmf
  expect_equal(density(dist, 1:3)[[1]], c(0.1, 0.2, 0.7))

  # cdf
  expect_equal(cdf(dist, 1:3)[[1]], c(0.1, 0.3, 1))

  # mean
  expect_equal(mean(dist), 2.6)
})

test_that("Mixture of multivariate distributions", {
  mu1 <- c(0, 0)
  mu2 <- c(1, 2)
  mu3 <- c(-10, -10)
  mu4 <- c(10, 10)
  sigma1 <- diag(2)
  sigma2 <- matrix(c(4, 2, 2, 3), 2, 2)
  w1 <- 0.3
  w2 <- 1 - w1
  w3 <- 0.5
  w4 <- 1 - w3
  # Two mixtures of bivariate normals
  dist <- c(
    dist1 = dist_mixture(
      dist_multivariate_normal(mu = list(mu1), sigma = list(sigma1)),
      dist_multivariate_normal(mu = list(mu2), sigma = list(sigma2)),
      weights = c(w1, w2)
    ),
    dist2 = dist_mixture(
      dist_multivariate_normal(mu = list(mu3), sigma = list(sigma1)),
      dist_multivariate_normal(mu = list(mu4), sigma = list(sigma1)),
      weights = c(w3, w4)
    )
  )
  # Mean
  expect_equal(mean(dist), rbind(w1 * mu1 + w2 * mu2, w3 * mu3 + w4 * mu4))
  # Quantile
  expect_error(quantile(dist, 0.5))
  # CDF
  at <- matrix(rnorm(4), 2, 2)
  cdf_dist <- cdf(dist, q = at)
  expect_equal(length(cdf_dist), 2L)
  expect_equal(lengths(cdf_dist), c(2L, 2L))
  # Density
  skip_if_not_installed("mvtnorm")
  expect_equal(
    density(dist, at),
    list(
      c(w1 * mvtnorm::dmvnorm(at, mean = mu1, sigma = sigma1) +
        w2 * mvtnorm::dmvnorm(at, mean = mu2, sigma = sigma2)),
      c(w3 * mvtnorm::dmvnorm(at, mean = mu3, sigma = sigma1) +
        w4 * mvtnorm::dmvnorm(at, mean = mu4, sigma = sigma1))
    )
  )
  # Mixture equivalent to multivariate normal
  dist <- dist_multivariate_normal(mu = list(mu2), sigma = list(sigma2))
  mdist <- dist_mixture(dist, dist, weights = c(w1, w2))
  expect_equal(mean(dist), mean(mdist))
  expect_equal(covariance(dist), covariance(mdist))
  expect_equal(unname(density(dist, rbind(mu1))), density(mdist, rbind(mu1)))
  set.seed(1)
  expect_equal(cdf(dist, rbind(c(0, 0))), cdf(mdist, rbind(c(0, 0))),
    tolerance = 0.001
  )
})

test_that("Multivariate t distribution", {
  df <- 5
  mu <- c(1, 2)
  sigma <- matrix(c(4, 2, 2, 3), ncol = 2)
  dist <- dist_multivariate_t(df = df, mu = list(mu), sigma = list(sigma))
  dimnames(dist) <- c("a", "b")

  expect_equal(format(dist), "MVT[2](5)")

  # stats
  expect_equal(
    mean(dist),
    matrix(c(1, 2), nrow = 1, dimnames = list(NULL, c("a", "b")))
  )
  
  # Covariance for multivariate t is df/(df-2) * sigma
  expected_cov <- df / (df - 2) * sigma
  expect_equal(covariance(dist), list(`colnames<-`(expected_cov, dimnames(dist))))

  # quantiles - marginal
  # For marginal t distribution: scale = sqrt(diag(sigma))
  expect_equal(
    quantile(dist, 0.1, kind = "marginal"),
    matrix(c(
      qt(0.1, df) * sqrt(sigma[1, 1]) + mu[1],
      qt(0.1, df) * sqrt(sigma[2, 2]) + mu[2]
    ),
    nrow = 1, dimnames = list(NULL, c("a", "b")))
  )

  skip_if_not_installed("mvtnorm")
  
  # quantiles - equicoordinate
  expect_equivalent(
    quantile(dist, 0.1, kind = "equicoordinate"),
    rep(mvtnorm::qmvt(0.1, delta = mu, sigma = sigma, df = df)$quantile, 2),
    tolerance = 1e-3
  )

  # pdf
  expect_equal(
    density(dist, cbind(1, 2)),
    mvtnorm::dmvt(c(1, 2), delta = mu, sigma = sigma, df = df, log = FALSE)
  )
  expect_equal(
    density(dist, cbind(-3, 4)),
    mvtnorm::dmvt(c(-3, 4), delta = mu, sigma = sigma, df = df, log = FALSE)
  )

  # cdf
  expect_equivalent(
    cdf(dist, cbind(1, 2)),
    as.numeric(mvtnorm::pmvt(upper = c(1, 2), delta = mu, sigma = sigma, df = df)),
    tolerance = 1e-3
  )
  expect_equivalent(
    cdf(dist, cbind(-3, 4)),
    as.numeric(mvtnorm::pmvt(upper = c(-3, 4), delta = mu, sigma = sigma, df = df)),
    tolerance = 1e-3
  )
})

test_that("Multivariate t with df = 1 (Cauchy-like)", {
  df <- 1
  mu <- c(0, 0)
  sigma <- diag(2)
  dist <- dist_multivariate_t(df = df, mu = list(mu), sigma = list(sigma))

  # Mean is undefined for df <= 1
  expect_warning(mean(dist), "undefined")
  expect_true(all(is.na(mean(dist))))
})

test_that("Multivariate t with df = 2", {
  df <- 2
  mu <- c(1, 2)
  sigma <- matrix(c(1, 0.5, 0.5, 1), ncol = 2)
  dist <- dist_multivariate_t(df = df, mu = list(mu), sigma = list(sigma))

  # Mean exists for df > 1
  expect_equal(mean(dist), matrix(mu, nrow = 1))
  
  # Covariance is undefined for df <= 2
  expect_warning(covariance(dist), "undefined")
  expect_true(all(is.na(covariance(dist)[[1]])))
})

test_that("Multivariate t with larger df approaches normal", {
  skip_if_not_installed("mvtnorm")
  
  df <- 100  # Large df
  mu <- c(1, 2)
  sigma <- matrix(c(4, 2, 2, 3), ncol = 2)
  
  dist_t <- dist_multivariate_t(df = df, mu = list(mu), sigma = list(sigma))
  dist_normal <- dist_multivariate_normal(mu = list(mu), sigma = list(sigma))
  
  # Densities should be similar for large df
  test_point <- cbind(1, 2)
  expect_equal(
    density(dist_t, test_point),
    density(dist_normal, test_point),
    tolerance = 0.01
  )
})

test_that("Multivariate t parameter validation", {
  expect_error(
    dist_multivariate_t(df = 0, mu = list(c(0, 0)), sigma = list(diag(2))),
    "positive"
  )
  
  expect_error(
    dist_multivariate_t(df = -5, mu = list(c(0, 0)), sigma = list(diag(2))),
    "positive"
  )
  
  # Valid inputs should not error
  expect_silent(
    dist_multivariate_t(df = 5, mu = list(c(0, 0)), sigma = list(diag(2)))
  )
})

test_that("Multivariate t generation", {
  skip_if_not_installed("mvtnorm")
  
  df <- 5
  mu <- c(1, 2)
  sigma <- matrix(c(4, 2, 2, 3), ncol = 2)
  dist <- dist_multivariate_t(df = df, mu = list(mu), sigma = list(sigma))
  
  set.seed(123)
  samples <- generate(dist, 100)
  
  # Check output structure - generate returns a list
  expect_true(is.list(samples))
  expect_equal(length(samples), 1)
  expect_true(is.matrix(samples[[1]]))
  expect_equal(ncol(samples[[1]]), 2)
  expect_equal(nrow(samples[[1]]), 100)
  
  # Sample mean should be close to theoretical mean (for df > 1)
  expect_equal(colMeans(samples[[1]]), mu, tolerance = 0.5)
})

test_that("Multivariate t with different dimensions", {
  # 3-dimensional
  df <- 10
  mu <- c(1, 2, 3)
  sigma <- diag(3)
  dist <- dist_multivariate_t(df = df, mu = list(mu), sigma = list(sigma))
  
  # Test the dim method on the unwrapped element
  expect_equal(dim.dist_mvt(vctrs::vec_data(dist)[[1]]), 3)
  expect_equal(format(dist), "MVT[3](10)")
  
  # 1-dimensional (should still work)
  dist1d <- dist_multivariate_t(df = 5, mu = list(0), sigma = list(matrix(1)))
  expect_equal(dim.dist_mvt(vctrs::vec_data(dist1d)[[1]]), 1)
  expect_equal(format(dist1d), "MVT[1](5)")
})

test_that("GPD", {
  dist <- dist_gpd(location = c(0, .5, 0), scale = c(1, 2, 3), shape = c(0, 0.1, 1.1))
  # Mean
  expect_equal(mean(dist), c(
    1,
    0.5 + 2 / 0.9, # location + scale/(1 - shape)
    Inf # Since shape >= 1
  ), tolerance = 0.0001)
  # Median
  expect_equal(median(dist), c(
    -log(0.5),
    0.5 + 2 * (2^0.1 - 1) / 0.1, # location + scale * (2^shape - 1) / shape
    3 * (2^1.1 - 1) / 1.1 # location + scale * (2^shape - 1) / shape
  ), tolerance = 0.0001)
  expect_equal(median(dist), quantile(dist, 0.5))
  # Variance
  expect_equal(distributional::variance(dist), c(
    1,
    2^2 / 0.9^2 / (1 - 2 * 0.1), # scale^2 / (1 - shape)^2 / (1 - 2 * shape)
    Inf # since shape >= 0.5
  ), tolerance = 0.0001)
  # Density
  at <- (0:20) / 10 + 1e-8 # Avoiding being on the boundary where evd gives wrong result
  expect_equal(density(dist, at), list(
    evd::dgpd(at, loc = 0, scale = 1, shape = 0),
    evd::dgpd(at, loc = 0.5, scale = 2, shape = 0.1),
    evd::dgpd(at, loc = 0, scale = 3, shape = 1.1)
  ))
  # CDF
  expect_equal(distributional::cdf(dist, at), list(
    evd::pgpd(at, loc = 0, scale = 1, shape = 0),
    evd::pgpd(at, loc = 0.5, scale = 2, shape = 0.1),
    evd::pgpd(at, loc = 0, scale = 3, shape = 1.1)
  ))
  # Quantiles
  p <- (1:19) / 20
  expect_equal(quantile(dist, p = p), list(
    evd::qgpd(p = p, loc = 0, scale = 1, shape = 0),
    evd::qgpd(p = p, loc = 0.5, scale = 2, shape = 0.1),
    evd::qgpd(p = p, loc = 0, scale = 3, shape = 1.1)
  ))
  # Generate
  set.seed(123)
  rand_dist <- distributional::generate(dist, times = 1e6)
  expect_equal(lapply(rand_dist[1:2], mean) |> unlist(),
    mean(dist)[1:2],
    tolerance = 0.01
  )
  expect_equal(lapply(rand_dist[1:2], var) |> unlist(),
    distributional::variance(dist)[1:2],
    tolerance = 0.01
  )
  expect_equal(lapply(rand_dist, median) |> unlist(),
    median(dist),
    tolerance = 0.01
  )
})

test_that("Weibull distribution", {
  dist <- dist_weibull(1.5, 1)

  expect_equal(format(dist), "Weibull(1.5, 1)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qweibull(0.1, 1.5, 1))
  expect_equal(quantile(dist, 0.5), stats::qweibull(0.5, 1.5, 1))

  # pdf
  expect_equal(density(dist, 0), stats::dweibull(0, 1.5, 1))
  expect_equal(density(dist, 3), stats::dweibull(3, 1.5, 1))

  # cdf
  expect_equal(cdf(dist, 0), stats::pweibull(0, 1.5, 1))
  expect_equal(cdf(dist, 3), stats::pweibull(3, 1.5, 1))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 1 * gamma(1 + 1/1.5))
  expect_equal(variance(dist), 1^2 * (gamma(1 + 2/1.5) - gamma(1 + 1/1.5)^2))
})

test_that("Dirichlet distribution", {
  alpha <- c(2, 5, 3)
  dist <- dist_dirichlet(alpha = list(alpha))

  # format
  expect_equal(format(dist), "Dirichlet[3]")

  # stats
  alpha_0 <- sum(alpha)
  expect_equal(
    mean(dist),
    matrix(alpha / alpha_0, nrow = 1)
  )

  expect_equal(
    variance(dist),
    matrix(alpha * (alpha_0 - alpha) / (alpha_0^2 * (alpha_0 + 1)), nrow = 1)
  )

  expected_cov <- matrix(NA_real_, nrow = 3, ncol = 3)
  denom <- alpha_0^2 * (alpha_0 + 1)
  for (i in 1:3) {
    for (j in 1:3) {
      expected_cov[i, j] <- if (i == j) {
        alpha[i] * (alpha_0 - alpha[i]) / denom
      } else {
        -alpha[i] * alpha[j] / denom
      }
    }
  }
  expect_equal(covariance(dist), list(expected_cov))

  # pdf
  x <- c(0.2, 0.5, 0.3)
  expect_equal(
    density(dist, cbind(x[1], x[2], x[3])),
    LaplacesDemon::ddirichlet(x, alpha)
  )
  expect_equal(
    density(dist, cbind(x[1], x[2], x[3]), log = TRUE),
    LaplacesDemon::ddirichlet(x, alpha, log = TRUE)
  )

  # support
  s <- support(dist)
  expect_s3_class(s, "support_region")
  expect_equal(format(s), "[0,1]^3")

  # generate
  set.seed(42)
  samples <- generate(dist, 10)[[1]]
  expect_equal(nrow(samples), 10L)
  expect_equal(ncol(samples), 3L)
  # rows should sum to 1
  expect_equal(rowSums(samples), rep(1, 10), tolerance = 1e-10)
  # all values non-negative
  expect_true(all(samples >= 0))
})

test_that("Dirichlet distribution - symmetric case", {
  alpha <- c(1, 1, 1)
  dist <- dist_dirichlet(alpha = list(alpha))
  alpha_0 <- sum(alpha)

  # uniform Dirichlet: mean is (1/3, 1/3, 1/3)
  expect_equal(mean(dist), matrix(rep(1 / 3, 3), nrow = 1), tolerance = 1e-10)

  # variance should be equal across components
  vars <- as.numeric(variance(dist))
  expect_true(all(abs(vars - vars[1]) < 1e-10))
})

test_that("Dirichlet distribution - two-component reduces to Beta", {
  # Dir(a, b) marginal X1 has mean a/(a+b) and variance matching Beta(a,b)
  a <- 3
  b <- 4
  dist_dir <- dist_dirichlet(alpha = list(c(a, b)))
  alpha_0 <- a + b

  expect_equal(
    as.numeric(mean(dist_dir)),
    c(a / alpha_0, b / alpha_0)
  )

  expected_var_x1 <- a * b / (alpha_0^2 * (alpha_0 + 1))
  expect_equal(as.numeric(variance(dist_dir))[1], expected_var_x1)
})

test_that("Gamma distribution", {
  dist <- dist_gamma(7.5, 2)

  expect_equal(format(dist), "Gamma(7.5, 2)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qgamma(0.1, 7.5, 2))
  expect_equal(quantile(dist, 0.5), stats::qgamma(0.5, 7.5, 2))

  # pdf
  expect_equal(density(dist, 0), stats::dgamma(0, 7.5, 2))
  expect_equal(density(dist, 3), stats::dgamma(3, 7.5, 2))

  # cdf
  expect_equal(cdf(dist, 0), stats::pgamma(0, 7.5, 2))
  expect_equal(cdf(dist, 3), stats::pgamma(3, 7.5, 2))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 7.5/2)
  expect_equal(variance(dist), 7.5/2^2)
})

test_that("Hypergeometric distribution", {
  dist <- dist_hypergeometric(500, 50, 100)

  expect_equal(format(dist), "Hypergeometric(500, 50, 100)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qhyper(0.1, 500, 50, 100))
  expect_equal(quantile(dist, 0.5), stats::qhyper(0.5, 500, 50, 100))

  # pdf
  expect_equal(density(dist, 0), stats::dhyper(0, 500, 50, 100))
  expect_equal(density(dist, 3), stats::dhyper(3, 500, 50, 100))

  # cdf
  expect_equal(cdf(dist, 0), stats::phyper(0, 500, 50, 100))
  expect_equal(cdf(dist, 3), stats::phyper(3, 500, 50, 100))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  p <- 500/(500+50)
  expect_equal(mean(dist), 100*p)
  expect_equal(variance(dist), 100*p*(1-p)*(500+50-100)/(500+50-1))
})

test_that("Student T distribution", {
  dist <- dist_student_t(5)

  expect_equal(format(dist), "t(5, 0, 1)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qt(0.1, 5))
  expect_equal(quantile(dist, 0.5), stats::qt(0.5, 5))

  # pdf
  expect_equal(density(dist, 0), stats::dt(0, 5))
  expect_equal(density(dist, 3), stats::dt(3, 5))

  # cdf
  expect_equal(cdf(dist, 0), stats::pt(0, 5))
  expect_equal(cdf(dist, 3), stats::pt(3, 5))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 0)
  expect_equal(variance(dist), 5 / (5-2))
})

test_that("Noncentral Student t distribution", {
  dist <- dist_student_t(8, ncp = 6)

  expect_equal(format(dist), "t(8, 0, 1, 6)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qt(0.1, 8, ncp = 6))
  expect_equal(quantile(dist, 0.5), stats::qt(0.5, 8, ncp = 6))

  # pdf
  expect_equal(density(dist, 0), stats::dt(0, 8, ncp = 6))
  expect_equal(density(dist, 3), stats::dt(3, 8, ncp = 6))

  # cdf
  expect_equal(cdf(dist, 0), stats::pt(0, 8, ncp = 6))
  expect_equal(cdf(dist, 3), stats::pt(3, 8, ncp = 6))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 2 * gamma(7/2))
  expect_equal(variance(dist), 148/3 - 4 * gamma(7/2)^2)
})

test_that("Location-scale Student t distribution", {
  dist <- dist_student_t(5, 2, 3)

  expect_equal(format(dist), "t(5, 2, 3)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qt(0.1, 5)*3 + 2)
  expect_equal(quantile(dist, 0.5), stats::qt(0.5, 5)*3 + 2)

  # pdf
  expect_equal(density(dist, 0), stats::dt(-2/3, 5)/3)
  expect_equal(density(dist, 3), stats::dt((3-2)/3, 5)/3)

  # cdf
  expect_equal(cdf(dist, 0), stats::pt(-2/3, 5))
  expect_equal(cdf(dist, 3), stats::pt((3-2)/3, 5))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 2)
  expect_equal(variance(dist), 5 / (5-2) * 3^2)
})

test_that("Noncentral location-scale Student t distribution", {
  dist <- dist_student_t(5, 2, 3, ncp = 6)

  expect_equal(format(dist), "t(5, 2, 3, 6)")

  # quantiles
  expect_equal(quantile(dist, 0.1), stats::qt(0.1, 5, 6)*3 + 2)
  expect_equal(quantile(dist, 0.5), stats::qt(0.5, 5, 6)*3 + 2)

  # pdf
  expect_equal(density(dist, 0), stats::dt(-2/3, 5, 6)/3)
  expect_equal(density(dist, 3), stats::dt((3-2)/3, 5, 6)/3)

  # cdf
  expect_equal(cdf(dist, 0), stats::pt(-2/3, 5, 6))
  expect_equal(cdf(dist, 3), stats::pt((3-2)/3, 5, 6))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4246)), 0.4246, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), 2 + 6 * sqrt(5/2) * gamma(2)/gamma(5/2) * 3)
  expect_equal(variance(dist), ((5*(1+6^2))/(5-2) - (6 * sqrt(5/2) * (gamma((5-1)/2)/gamma(5/2)))^2)*3^2)
})

test_that("Pareto distribution", {
  dist <- dist_pareto(10, 1)

  expect_equal(format(dist), "Pareto(10, 1)")

  # Require package installed
  skip_if_not_installed("actuar", "2.0.0")

  # quantiles
  expect_equal(quantile(dist, 0.1), actuar::qpareto(0.1, 10, 1))
  expect_equal(quantile(dist, 0.5), actuar::qpareto(0.5, 10, 1))

  # pdf
  expect_equal(density(dist, 0), actuar::dpareto(0, 10, 1))
  expect_equal(density(dist, 3), actuar::dpareto(3, 10, 1))

  # cdf
  expect_equal(cdf(dist, 0), actuar::ppareto(0, 10, 1))
  expect_equal(cdf(dist, 3), actuar::ppareto(3, 10, 1))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4)), 0.4, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), actuar::mpareto(1, 10, 1))
  expect_equal(variance(dist), actuar::mpareto(2, 10, 1) - actuar::mpareto(1, 10, 1)^2)
})

test_that("Log-normal distribution", {
  # defaults
  dist <- dist_lognormal()
  expect_equal(mean(dist), exp(1/2))
  expect_equal(variance(dist), (exp(1)-1)*exp(1))

  # display
  expect_s3_class(dist, "distribution")
  expect_output(print(dist), "lN\\(0, 1\\)")
  expect_output(print(dist_lognormal(numeric())), "<distribution\\[0\\]>")

  # error checking
  expect_error(
    dist_lognormal(0, -1),
    "non-negative"
  )
  expect_silent(
    dist_lognormal(mu = 0, sigma = 1)
  )

  # density
  expect_equal(
    density(dist, 0),
    dlnorm(0, mean = 0, sd = 1)
  )
  # cdf
  expect_equal(
    cdf(dist, 5),
    plnorm(5, mean = 0, sd = 1)
  )
  # quantile
  expect_equal(
    quantile(dist, 0.1),
    qlnorm(0.1, mean = 0, sd = 1)
  )
  # generate
  expect_equal(
    {
      set.seed(0)
      generate(dist, 10)
    },
    {
      set.seed(0)
      mapply(function(m, s) rlnorm(10, m, s), m = 0, s = 1, SIMPLIFY = FALSE)
    }
  )
})

test_that("Inverse Gaussian distribution", {
  dist <- dist_inverse_gaussian(3, .2)

  expect_equal(format(dist), "IG(3, 0.2)")

  # Require package installed
  skip_if_not_installed("actuar", "2.0.0")

  # quantiles
  expect_equal(quantile(dist, 0.1), actuar::qinvgauss(0.1, 3, .2))
  expect_equal(quantile(dist, 0.5), actuar::qinvgauss(0.5, 3, .2))

  # pdf
  expect_equal(density(dist, 0), actuar::dinvgauss(0, 3, .2))
  expect_equal(density(dist, 3), actuar::dinvgauss(3, 3, .2))

  # cdf
  expect_equal(cdf(dist, 0), actuar::pinvgauss(0, 3, .2))
  expect_equal(cdf(dist, 3), actuar::pinvgauss(3, 3, .2))

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.4)), 0.4, tolerance = 1e-3)

  # stats
  expect_equal(mean(dist), actuar::minvgauss(1, 3, .2))
  expect_equal(variance(dist), actuar::minvgauss(2, 3, .2) - actuar::minvgauss(1, 3, .2)^2)
})

test_that("GEV", {
  dist <- dist_gev(location = c(0, .5, 0), scale = c(1, 2, 3), shape = c(0, 0.1, 1.1))
  euler <- 0.57721566490153286 # Euler's constant
  # Mean
  expect_equal(mean(dist), c(
    euler,
    0.5 + 2 * (gamma(0.9) - 1) / 0.1, # location + scale*(gamma(1 - shape) - 1)/shape
    Inf # Since shape >= 1
  ), tolerance = 0.0001)
  # Median
  expect_equal(median(dist), c(
    -log(log(2)),
    0.5 + 2 * (log(2)^(-0.1) - 1) / 0.1, # location + scale*(log(2)^(-shape) - 1)/shape
    3 * (log(2)^(-1.1) - 1) / 1.1 # location + scale*(log(2)^(-shape) - 1)/shape
  ), tolerance = 0.0001)
  expect_equal(median(dist), quantile(dist, 0.5))
  # Variance
  expect_equal(distributional::variance(dist), c(
    pi^2 / 6,
    2^2 * (gamma(1 - 2 * 0.1) - gamma(1 - 0.1)^2) / 0.1^2, # scale^2 * (g2 - g1^2)/shape^2
    Inf # since shape >= 0.5
  ), tolerance = 0.0001)
  # Density
  at <- (0:20) / 10
  expect_equal(density(dist, at), list(
    evd::dgev(at, loc = 0, scale = 1, shape = 0),
    evd::dgev(at, loc = 0.5, scale = 2, shape = 0.1),
    evd::dgev(at, loc = 0, scale = 3, shape = 1.1)
  ))
  # CDF
  expect_equal(distributional::cdf(dist, at), list(
    evd::pgev(at, loc = 0, scale = 1, shape = 0),
    evd::pgev(at, loc = 0.5, scale = 2, shape = 0.1),
    evd::pgev(at, loc = 0, scale = 3, shape = 1.1)
  ))
  # Quantiles
  p <- (1:19) / 20
  expect_equal(quantile(dist, p = p), list(
    evd::qgev(p = p, loc = 0, scale = 1, shape = 0),
    evd::qgev(p = p, loc = 0.5, scale = 2, shape = 0.1),
    evd::qgev(p = p, loc = 0, scale = 3, shape = 1.1)
  ))
  # Generate
  set.seed(123)
  rand_dist <- distributional::generate(dist, times = 1e6)
  expect_equal(unlist(lapply(rand_dist[1:2], mean)),
    mean(dist)[1:2],
    tolerance = 0.01
  )
  expect_equal(unlist(lapply(rand_dist[1:2], var)),
    distributional::variance(dist)[1:2],
    tolerance = 0.01
  )
  expect_equal(unlist(lapply(rand_dist, median)),
    median(dist),
    tolerance = 0.01
  )
})

test_that("dist_convolved: basic construction and format", {
  d <- dist_convolved(dist_gamma(2, 2), dist_gamma(2, 2))
  expect_s3_class(d, "distribution")
  expect_match(format(d), "Gamma\\(2, 2\\) \\+ Gamma\\(2, 2\\)")
})

test_that("dist_convolved: density approximates the analytical result", {
  # Gamma(2,2) + Gamma(2,2) = Gamma(4,2) analytically
  d_conv  <- dist_convolved(dist_gamma(2, 2), dist_gamma(2, 2))
  d_exact <- dist_gamma(4, 2)

  x <- c(0.5, 1, 2, 3, 4)
  approx_dens <- density(d_conv, x)
  exact_dens  <- density(d_exact, x)

  expect_equal(approx_dens, exact_dens, tolerance = 1e-3)
})

test_that("dist_convolved: CDF approximates the analytical result", {
  d_conv  <- dist_convolved(dist_gamma(2, 2), dist_gamma(2, 2))
  d_exact <- dist_gamma(4, 2)

  q <- c(0.5, 1, 2, 3, 4)
  expect_equal(cdf(d_conv, q), cdf(d_exact, q), tolerance = 1e-3)
})

test_that("dist_convolved: quantile approximates the analytical result", {
  d_conv  <- dist_convolved(dist_gamma(2, 2), dist_gamma(2, 2))
  d_exact <- dist_gamma(4, 2)

  p <- c(0.1, 0.25, 0.5, 0.75, 0.9)
  expect_equal(quantile(d_conv, p), quantile(d_exact, p), tolerance = 1e-2)
})

test_that("dist_convolved: generate returns correct length and type", {
  d <- dist_convolved(dist_gamma(2, 2), dist_exponential(1))
  # generate() on a distribution vector returns a list of length = length(d)
  samp <- generate(d, 100)
  expect_length(samp, 1L)       # 1 distribution in the vector
  expect_length(samp[[1]], 100L) # 100 samples from it
  expect_true(is.numeric(samp[[1]]))
})

test_that("dist_convolved: mean is exact sum of component means", {
  d <- dist_convolved(dist_gamma(2, 2), dist_exponential(3))
  # E[Gamma(2,2)] = 2/2 = 1; E[Exp(3)] = 1/3
  expect_equal(mean(d), 1 + 1/3, tolerance = 1e-10)
})

test_that("dist_convolved: variance is exact sum of component variances", {
  d <- dist_convolved(dist_gamma(2, 2), dist_exponential(3))
  # Var[Gamma(2,2)] = 2/4 = 0.5; Var[Exp(3)] = 1/9
  expect_equal(variance(d), 0.5 + 1/9, tolerance = 1e-10)
})

test_that("dist_convolved: normal sum via + uses closed-form, not convolution", {
  # dist_normal has Ops.dist_normal which returns an exact dist_normal
  d <- dist_normal(0, 1) + dist_normal(0, 1)
  expect_s3_class(vec_data(d)[[1]], "dist_normal")
})

test_that("dist_convolved: + operator creates dist_convolved for other families", {
  d <- dist_gamma(2, 2) + dist_gamma(2, 2)
  expect_s3_class(vec_data(d)[[1]], "dist_convolved")
})

test_that("dist_convolved: subtraction via - works", {
  # X - Y = X + (-Y); test that it constructs without error
  d <- dist_normal(5, 1) - dist_normal(0, 1)
  # E[N(5,1) - N(0,1)] = 5; Var = 1 + 1 = 2
  # Using analytical result (normal - normal = normal)
  expect_s3_class(vec_data(d)[[1]], "dist_normal")
})

test_that("dist_convolved: subtraction of non-normal distributions", {
  # Gamma - Gamma is not analytically simple; should give a dist_convolved
  d <- dist_gamma(4, 1) - dist_gamma(2, 1)
  expect_s3_class(vec_data(d)[[1]], "dist_convolved")

  # Mean: E[Gamma(4,1)] - E[Gamma(2,1)] = 4 - 2 = 2
  expect_equal(mean(d), 2, tolerance = 1e-10)
})

test_that("dist_convolved: works with mixed distribution families", {
  skip_unless_r(">= 4.3.0")

  d <- dist_normal(0, 1) + dist_exponential(1)
  expect_s3_class(vec_data(d)[[1]], "dist_convolved")
  # Mean = 0 + 1 = 1
  expect_equal(mean(d), 1, tolerance = 1e-10)
})

test_that("dist_convolved: vectorises over pairs", {
  d <- dist_convolved(
    dist_gamma(c(2, 3), c(2, 1)),
    dist_gamma(c(2, 1), c(2, 1))
  )
  expect_length(d, 2L)
  expect_s3_class(d, "distribution")
})

test_that("dist_convolved: invalid arguments are rejected", {
  expect_error(dist_convolved(1, dist_normal()))
})

test_that("dist_convolved: family returns 'convolved'", {
  d <- dist_convolved(dist_gamma(2, 2), dist_gamma(2, 2))
  expect_equal(family(d), "convolved")
})

test_that("dist_convolved: transformed distribution + distribution works", {
  skip_unless_r(">= 4.3.0")

  # dist_lognormal is a dist_transformed internally; adding another dist
  # should produce a dist_convolved
  d <- dist_lognormal(0, 1) + dist_exponential(1)
  expect_s3_class(vec_data(d)[[1]], "dist_convolved")
  # Mean of lognormal(0,1) = exp(0.5); mean of Exp(1) = 1
  expect_equal(mean(d), exp(0.5) + 1, tolerance = 1e-10)
})

# ---------------------------------------------------------------------------
# k-way convolution tests
# ---------------------------------------------------------------------------

test_that("dist_convolved: explicit 3-arg construction works", {
  d3 <- dist_convolved(dist_gamma(2, 2), dist_gamma(2, 2), dist_gamma(2, 2))
  raw <- vec_data(d3)[[1L]]
  expect_s3_class(raw, "dist_convolved")
  expect_length(raw[["dist"]], 3L)
})

test_that("dist_convolved: requires at least two distributions", {
  expect_error(dist_convolved(dist_gamma(2, 2)), "at least two")
})

test_that("dist_convolved: k-way chaining produces a flat dist_convolved", {
  d3 <- dist_gamma(2, 2) + dist_gamma(2, 2) + dist_gamma(2, 2)
  raw <- vec_data(d3)[[1L]]
  expect_s3_class(raw, "dist_convolved")
  # Flattening happens at construction: 3 components, not a nested 2+1
  expect_length(raw[["dist"]], 3L)
})

test_that("dist_convolved: k-way format lists all components", {
  d3 <- dist_gamma(2, 2) + dist_gamma(2, 2) + dist_gamma(2, 2)
  expect_match(format(d3), "Gamma\\(2, 2\\) \\+ Gamma\\(2, 2\\) \\+ Gamma\\(2, 2\\)")
})

test_that("dist_convolved: 3-way density matches Gamma(6,2) analytically", {
  # Gamma(2,2) + Gamma(2,2) + Gamma(2,2) = Gamma(6,2)
  d_conv  <- dist_gamma(2, 2) + dist_gamma(2, 2) + dist_gamma(2, 2)
  d_exact <- dist_gamma(6, 2)

  x <- c(1, 2, 3, 4, 5)
  expect_equal(density(d_conv, x), density(d_exact, x), tolerance = 1e-2)
})

test_that("dist_convolved: 3-way CDF matches Gamma(6,2) analytically", {
  d_conv  <- dist_gamma(2, 2) + dist_gamma(2, 2) + dist_gamma(2, 2)
  d_exact <- dist_gamma(6, 2)

  q <- c(1, 2, 3, 4, 5)
  expect_equal(cdf(d_conv, q), cdf(d_exact, q), tolerance = 1e-2)
})

test_that("dist_convolved: k-way mean is exact sum of all component means", {
  # E[Gamma(2,2)] = 1; three of them => mean = 3
  d3 <- dist_gamma(2, 2) + dist_gamma(2, 2) + dist_gamma(2, 2)
  expect_equal(mean(d3), 3, tolerance = 1e-10)

  skip_unless_r(">= 4.3.0")

  # 4-way with mixed families
  d4 <- dist_gamma(2, 2) + dist_exponential(2) + dist_normal(1, 1) + dist_gamma(3, 1)
  expected_mean <- 1 + 0.5 + 1 + 3  # 1 + 1/2 + 1 + 3/1
  expect_equal(mean(d4), expected_mean, tolerance = 1e-10)
})

test_that("dist_convolved: k-way variance is exact sum of component variances", {
  # Var[Gamma(2,2)] = 2/4 = 0.5; three of them => var = 1.5
  d3 <- dist_gamma(2, 2) + dist_gamma(2, 2) + dist_gamma(2, 2)
  expect_equal(variance(d3), 1.5, tolerance = 1e-10)
})

test_that("dist_convolved: chaining is associative in output", {
  # (d1 + d2) + d3 vs d1 + (d2 + d3) should give the same mean and density
  d_left  <- (dist_gamma(2, 2) + dist_gamma(2, 2)) + dist_gamma(2, 2)
  d_right <- dist_gamma(2, 2) + (dist_gamma(2, 2) + dist_gamma(2, 2))

  expect_equal(mean(d_left), mean(d_right), tolerance = 1e-10)
  expect_equal(density(d_left, 3), density(d_right, 3), tolerance = 1e-3)
})

# ---------------------------------------------------------------------------
# Analytical convolution identities: FFT approximation accuracy
# ---------------------------------------------------------------------------

test_that("dist_convolved: Normal + Normal FFT approximates Normal(mu1+mu2, sqrt(s1^2+s2^2))", {
  # Force the FFT path by calling dist_convolved() directly.
  # N(1, 1) + N(2, 1.5) = N(3, sqrt(1 + 2.25)) = N(3, sqrt(3.25))
  d_conv  <- dist_convolved(dist_normal(1, 1), dist_normal(2, 1.5))
  d_exact <- dist_normal(3, sqrt(3.25))

  x <- c(0, 1, 3, 5, 7)
  expect_equal(density(d_conv, x), density(d_exact, x), tolerance = 1e-3)
  expect_equal(cdf(d_conv, x),     cdf(d_exact, x),     tolerance = 1e-3)

  p <- c(0.1, 0.25, 0.5, 0.75, 0.9)
  expect_equal(quantile(d_conv, p), quantile(d_exact, p), tolerance = 1e-2)
})

test_that("dist_convolved: Exponential + Exponential approximates Gamma(2, rate)", {
  # Exp(lambda) + Exp(lambda) = Gamma(shape = 2, rate = lambda)
  lambda  <- 3
  d_conv  <- dist_convolved(dist_exponential(lambda), dist_exponential(lambda))
  d_exact <- dist_gamma(2, lambda)

  # Avoid x values very close to 0 where density is small and relative error inflates
  x <- c(0.2, 0.4, 0.7, 1.0, 1.5)
  expect_equal(density(d_conv, x), density(d_exact, x), tolerance = 1e-2)
  expect_equal(cdf(d_conv, x),     cdf(d_exact, x),     tolerance = 1e-2)

  p <- c(0.1, 0.25, 0.5, 0.75, 0.9)
  expect_equal(quantile(d_conv, p), quantile(d_exact, p), tolerance = 1e-2)
})

test_that("dist_convolved: Chi-squared + Chi-squared approximates Chi-squared(k1 + k2)", {
  # chi^2(k1) + chi^2(k2) = chi^2(k1 + k2)
  d_conv  <- dist_convolved(dist_chisq(4), dist_chisq(6))
  d_exact <- dist_chisq(10)

  x <- c(2, 5, 8, 12, 18)
  expect_equal(density(d_conv, x), density(d_exact, x), tolerance = 1e-3)
  expect_equal(cdf(d_conv, x),     cdf(d_exact, x),     tolerance = 1e-3)

  p <- c(0.1, 0.25, 0.5, 0.75, 0.9)
  expect_equal(quantile(d_conv, p), quantile(d_exact, p), tolerance = 1e-2)
})

#' The Beta distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Beta distribution is a continuous probability distribution defined on the
#' interval \[0, 1\], commonly used to model probabilities and proportions.
#'
#' @param shape1,shape2 The non-negative shape parameters of the Beta distribution.
#'
#' @details
#'
#'   `r pkgdown_doc_link("dist_beta")`
#'
#'   In the following, let \eqn{X} be a Beta random variable with parameters
#'   `shape1` = \eqn{\alpha} and `shape2` = \eqn{\beta}.
#'
#'   **Support**: \eqn{x \in [0, 1]}
#'
#'   **Mean**: \eqn{\frac{\alpha}{\alpha + \beta}}
#'
#'   **Variance**: \eqn{\frac{\alpha\beta}{(\alpha + \beta)^2(\alpha + \beta + 1)}}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{x^{\alpha - 1}(1-x)^{\beta - 1}}{B(\alpha, \beta)} = 
#'     \frac{\Gamma(\alpha + \beta)}{\Gamma(\alpha)\Gamma(\beta)} x^{\alpha - 1}(1-x)^{\beta - 1}
#'   }{
#'     f(x) = x^(alpha - 1) * (1-x)^(beta - 1) / B(alpha, beta)
#'   }
#'
#'   where \eqn{B(\alpha, \beta) = \frac{\Gamma(\alpha)\Gamma(\beta)}{\Gamma(\alpha + \beta)}}
#'   is the Beta function.
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = I_x(alpha, beta) = \frac{B(x; \alpha, \beta)}{B(\alpha, \beta)}
#'   }{
#'     F(x) = I_x(\alpha, \beta)
#'   }
#'
#'   where \eqn{I_x(\alpha, \beta)} is the regularized incomplete beta function and
#'   \eqn{B(x; \alpha, \beta) = \int_0^x t^{\alpha-1}(1-t)^{\beta-1} dt}.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   The moment generating function does not have a simple closed form, but the
#'   moments can be calculated as:
#'
#'   \deqn{
#'     E(X^k) = \prod_{r=0}^{k-1} \frac{\alpha + r}{\alpha + \beta + r}
#'   }{
#'     E(X^k) = prod((alpha + r) / (alpha + beta + r)) for r = 0, ..., k-1
#'   }
#'
#' @seealso [stats::Beta]
#'
#' @examples
#' dist <- dist_beta(shape1 = c(0.5, 5, 1, 2, 2), shape2 = c(0.5, 1, 3, 2, 5))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_beta
#' @export
dist_beta <- function(shape1, shape2){
  shape1 <- vec_cast(shape1, double())
  shape2 <- vec_cast(shape2, double())
  if(any((shape1 < 0) | shape2 < 0)){
    abort("Shape parameters of a Beta distribution must be non-negative.")
  }
  new_dist(shape1 = shape1, shape2 = shape2, class = "dist_beta")
}

#' @export
format.dist_beta <- function(x, digits = 2, ...){
  sprintf(
    "Beta(%s, %s)",
    format(x[["shape1"]], digits = digits, ...),
    format(x[["shape2"]], digits = digits, ...)
  )
}

#' @export
density.dist_beta <- function(x, at, ...){
  stats::dbeta(at, x[["shape1"]], x[["shape2"]])
}

#' @export
log_density.dist_beta <- function(x, at, ...){
  stats::dbeta(at, x[["shape1"]], x[["shape2"]], log = TRUE)
}

#' @export
quantile.dist_beta <- function(x, p, ...){
  stats::qbeta(p, x[["shape1"]], x[["shape2"]])
}

#' @export
cdf.dist_beta <- function(x, q, ...){
  stats::pbeta(q, x[["shape1"]], x[["shape2"]])
}

#' @export
generate.dist_beta <- function(x, times, ...){
  stats::rbeta(times, x[["shape1"]], x[["shape2"]])
}

#' @export
mean.dist_beta <- function(x, ...){
  x[["shape1"]]/(x[["shape1"]] + x[["shape2"]])
}

#' @export
covariance.dist_beta <- function(x, ...){
  a <- x[["shape1"]]
  b <- x[["shape2"]]
  a*b/((a+b)^2*(a+b+1))
}

#' @export
skewness.dist_beta <- function(x, ...) {
  a <- x[["shape1"]]
  b <- x[["shape2"]]
  2 * (b - a) * sqrt(a + b + 1) / (a + b + 2) * sqrt(a * b)
}

#' @export
kurtosis.dist_beta <- function(x, ...) {
  a <- x[["shape1"]]
  b <- x[["shape2"]]
  num <- 6 * ((a - b)^2 * (a + b + 1) - (a * b) * (a + b + 2))
  denom <- a * b * (a + b + 2) * (a + b + 3)
  num / denom
}

#' @export
has_symmetry.dist_beta <- function(x, ...) {
  a <- x[["shape1"]]
  b <- x[["shape2"]]
  a == b
}

#' The Weibull distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Generalization of the gamma distribution. Often used in survival and
#' time-to-event analyses.
#'
#' @inheritParams stats::dweibull
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_weibull")`
#'
#'   In the following, let \eqn{X} be a Weibull random variable with
#'   shape parameter `shape` = \eqn{k} and scale parameter `scale` = \eqn{\lambda}.
#'
#'   **Support**: \eqn{[0, \infty)}
#'
#'   **Mean**: 
#'
#'   \deqn{
#'     E(X) = \lambda \Gamma\left(1 + \frac{1}{k}\right)
#'   }{
#'     E(X) = \lambda \Gamma(1 + 1/k)
#'   }
#'
#'   where \eqn{\Gamma} is the gamma function.
#'
#'   **Variance**: 
#'
#'   \deqn{
#'     \text{Var}(X) = \lambda^2 \left[\Gamma\left(1 + \frac{2}{k}\right) - \left(\Gamma\left(1 + \frac{1}{k}\right)\right)^2\right]
#'   }{
#'     Var(X) = \lambda^2 [\Gamma(1 + 2/k) - (\Gamma(1 + 1/k))^2]
#'   }
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{k}{\lambda}\left(\frac{x}{\lambda}\right)^{k-1}e^{-(x/\lambda)^k}, \quad x \ge 0
#'   }{
#'     f(x) = (k/\lambda)(x/\lambda)^(k-1) e^(-(x/\lambda)^k), x >= 0
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = 1 - e^{-(x/\lambda)^k}, \quad x \ge 0
#'   }{
#'     F(x) = 1 - e^(-(x/\lambda)^k), x >= 0
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \sum_{n=0}^\infty \frac{t^n\lambda^n}{n!} \Gamma\left(1+\frac{n}{k}\right)
#'   }{
#'     E(e^(tX)) = \sum_(n=0)^\infty (t^n \lambda^n / n!) \Gamma(1 + n/k)
#'   }
#'
#'   **Skewness**:
#'
#'   \deqn{
#'     \gamma_1 = \frac{\mu^3 - 3\mu\sigma^2 - \mu^3}{\sigma^3}
#'   }{
#'     \gamma_1 = (\mu_3 - 3\mu\sigma^2 - \mu^3) / \sigma^3
#'   }
#'
#'   where \eqn{\mu = E(X)}, \eqn{\sigma^2 = \text{Var}(X)}, and the third 
#'   raw moment is
#'
#'   \deqn{
#'     \mu^3 = \lambda^3 \Gamma\left(1 + \frac{3}{k}\right)
#'   }{
#'     \mu_3 = \lambda^3 \Gamma(1 + 3/k)
#'   }
#'
#'   **Excess Kurtosis**:
#'
#'   \deqn{
#'     \gamma_2 = \frac{\mu^4 - 4\gamma_1\mu\sigma^3 - 6\mu^2\sigma^2 - \mu^4}{\sigma^4} - 3
#'   }{
#'     \gamma_2 = (\mu_4 - 4\gamma_1\mu\sigma^3 - 6\mu^2\sigma^2 - \mu^4) / \sigma^4 - 3
#'   }
#'
#'   where the fourth raw moment is
#'
#'   \deqn{
#'     \mu^4 = \lambda^4 \Gamma\left(1 + \frac{4}{k}\right)
#'   }{
#'     \mu_4 = \lambda^4 \Gamma(1 + 4/k)
#'   }
#'
#' @seealso [stats::Weibull]
#'
#' @examples
#' dist <- dist_weibull(shape = c(0.5, 1, 1.5, 5), scale = rep(1, 4))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_weibull
#' @export
dist_weibull <- function(shape, scale){
  shape <- vec_cast(shape, double())
  scale <- vec_cast(scale, double())
  if(any(shape[!is.na(shape)] < 0)){
    abort("The shape parameter of a Weibull distribution must be non-negative.")
  }
  if(any(scale[!is.na(scale)] <= 0)){
    abort("The scale parameter of a Weibull distribution must be strictly positive.")
  }
  new_dist(shape = shape, scale = scale, class = "dist_weibull")
}

#' @export
format.dist_weibull <- function(x, digits = 2, ...){
  sprintf(
    "Weibull(%s, %s)",
    format(x[["shape"]], digits = digits, ...),
    format(x[["scale"]], digits = digits, ...)
  )
}

#' @export
density.dist_weibull <- function(x, at, ...){
  stats::dweibull(at, x[["shape"]], x[["scale"]])
}

#' @export
log_density.dist_weibull <- function(x, at, ...){
  stats::dweibull(at, x[["shape"]], x[["scale"]], log = TRUE)
}

#' @export
quantile.dist_weibull <- function(x, p, ...){
  stats::qweibull(p, x[["shape"]], x[["scale"]])
}

#' @export
cdf.dist_weibull <- function(x, q, ...){
  stats::pweibull(q, x[["shape"]], x[["scale"]])
}

#' @export
generate.dist_weibull <- function(x, times, ...){
  stats::rweibull(times, x[["shape"]], x[["scale"]])
}

#' @export
mean.dist_weibull <- function(x, ...){
  x[["scale"]] * gamma(1 + 1/x[["shape"]])
}

#' @export
covariance.dist_weibull <- function(x, ...){
  x[["scale"]]^2 * (gamma(1 + 2/x[["shape"]]) - gamma(1 + 1/x[["shape"]])^2)
}

#' @export
skewness.dist_weibull <- function(x, ...) {
  mu <- mean(x)
  sigma <- sqrt(variance(x))
  r <- mu / sigma
  gamma(1 + 3/x[["shape"]]) * (x[["scale"]]/sigma)^3 - 3*r - r^3
}

#' @export
kurtosis.dist_weibull <- function(x, ...) {
  mu <- mean(x)
  sigma <- sqrt(variance(x))
  gamma <- skewness(x)
  r <- mu / sigma
  (x[["scale"]]/sigma)^4 * gamma(1 + 4/x[["shape"]]) - 4*gamma*r -6*r^2 - r^4 - 3
}

#' The Inverse Exponential distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' @inheritParams actuar::dinvexp
#'
#' @seealso [actuar::InverseExponential]
#'
#' @examples
#' dist <- dist_inverse_exponential(rate = 1:5)
#' dist
#'
#' @examplesIf requireNamespace("actuar", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_inverse_exponential
#' @export
dist_inverse_exponential <- function(rate){
  rate <- vec_cast(rate, double())
  if(any(rate <= 0)){
    abort("The rate parameter of a Inverse Exponential distribution must be strictly positive.")
  }
  new_dist(r = rate, class = "dist_inverse_exponential")
}

#' @export
format.dist_inverse_exponential <- function(x, digits = 2, ...){
  sprintf(
    "InvExp(%s)",
    format(x[["r"]], digits = digits, ...)
  )
}

#' @export
density.dist_inverse_exponential <- function(x, at, ...){
  require_package("actuar")
  actuar::dinvexp(at, x[["r"]])
}

#' @export
log_density.dist_inverse_exponential <- function(x, at, ...){
  require_package("actuar")
  actuar::dinvexp(at, x[["r"]], log = TRUE)
}

#' @export
quantile.dist_inverse_exponential <- function(x, p, ...){
  require_package("actuar")
  actuar::qinvexp(p, x[["r"]])
}

#' @export
cdf.dist_inverse_exponential <- function(x, q, ...){
  require_package("actuar")
  actuar::pinvexp(q, x[["r"]])
}

#' @export
generate.dist_inverse_exponential <- function(x, times, ...){
  require_package("actuar")
  actuar::rinvexp(times, x[["r"]])
}

#' @export
mean.dist_inverse_exponential <- function(x, ...){
  NA_real_
}

#' @export
covariance.dist_inverse_exponential <- function(x, ...){
  NA_real_
}

#' @keywords internal
"_PACKAGE"

# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
#' @importFrom lifecycle deprecate_soft
#' @importFrom lifecycle deprecated
## usethis namespace: end
#' @import vctrs
#' @import rlang
NULL

# Used for generating transformation function expressions
globalVariables("x")

#' Sampling distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The sampling distribution represents an empirical distribution based on
#' observed samples. It is useful for bootstrapping, representing posterior
#' distributions from Markov Chain Monte Carlo (MCMC) algorithms, or working
#' with any empirical data where the parametric form is unknown. Unlike
#' parametric distributions, the sampling distribution makes no assumptions
#' about the underlying data-generating process and instead uses the sample
#' itself to estimate distributional properties. The distribution can handle
#' both univariate and multivariate samples.
#'
#' @param x A list of sampled values. For univariate distributions, each
#'   element should be a numeric vector. For multivariate distributions, each
#'   element should be a matrix where columns represent variables and rows
#'   represent observations.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_sample")`
#'
#'   In the following, let \eqn{X} be a random variable with sample
#'   \eqn{x_1, x_2, \ldots, x_n} of size \eqn{n}.
#'
#'   **Support**: The observed range of the sample
#'
#'   **Mean** (univariate):
#'
#'   \deqn{
#'     \bar{x} = \frac{1}{n} \sum_{i=1}^{n} x_i
#'   }{
#'     mean(x) = (1/n) sum(x_i)
#'   }
#'
#'   **Mean** (multivariate): Computed independently for each variable.
#'
#'   **Variance** (univariate):
#'
#'   \deqn{
#'     s^2 = \frac{1}{n-1} \sum_{i=1}^{n} (x_i - \bar{x})^2
#'   }{
#'     s^2 = (1/(n-1)) sum((x_i - mean(x))^2)
#'   }
#'
#'   **Covariance** (multivariate): The sample covariance matrix.
#'
#'   **Skewness** (univariate):
#'
#'   \deqn{
#'     g_1 = \frac{\sqrt{n} \sum_{i=1}^{n} (x_i - \bar{x})^3}{\left(\sum_{i=1}^{n} (x_i - \bar{x})^2\right)^{3/2}} \left(1 - \frac{1}{n}\right)^{3/2}
#'   }{
#'     g_1 = (sqrt(n) sum((x_i - mean(x))^3) / (sum((x_i - mean(x))^2)^(3/2))) * (1 - 1/n)^(3/2)
#'   }
#'
#'   **Probability density function**: Approximated numerically using
#'   kernel density estimation.
#'
#'   **Cumulative distribution function** (univariate):
#'
#'   \deqn{
#'     F(q) = \frac{1}{n} \sum_{i=1}^{n} I(x_i \leq q)
#'   }{
#'     F(q) = (1/n) sum(I(x_i <= q))
#'   }
#'
#'   where \eqn{I(\cdot)} is the indicator function.
#'
#'   **Cumulative distribution function** (multivariate):
#'
#'   \deqn{
#'     F(\mathbf{q}) = \frac{1}{n} \sum_{i=1}^{n} I(\mathbf{x}_i \leq \mathbf{q})
#'   }{
#'     F(q) = (1/n) sum(I(x_i <= q))
#'   }
#'
#'   where the inequality is applied element-wise.
#'
#'   **Quantile function** (univariate): The sample quantile, computed using
#'   the specified quantile type (see [stats::quantile()]).
#'
#'   **Quantile function** (multivariate): Marginal quantiles are computed
#'   independently for each variable.
#'
#'   **Random generation**: Bootstrap sampling with replacement from the
#'   empirical sample.
#'
#' @seealso [stats::density()], [stats::quantile()], [stats::cov()]
#'
#' @examples
#' # Univariate numeric samples
#' dist <- dist_sample(x = list(rnorm(100), rnorm(100, 10)))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' generate(dist, 10)
#'
#' density(dist, 1)
#'
#' # Multivariate numeric samples
#' dist <- dist_sample(x = list(cbind(rnorm(100), rnorm(100, 10))))
#' dimnames(dist) <- c("x", "y")
#'
#' dist
#' mean(dist)
#' variance(dist)
#' generate(dist, 10)
#' quantile(dist, 0.4) # Returns the marginal quantiles
#' cdf(dist, matrix(c(0.3,9), nrow = 1))
#'
#' @name dist_sample
#' @export
dist_sample <- function(x){
  vec_assert(x, list())
  x <- as_list_of(x, .ptype = vec_ptype(x[[1]]))
  new_dist(x = x, class = "dist_sample")
}

#' @export
format.dist_sample <- function(x, ...){
  sprintf(
    "sample[%s]",
    vapply(x, vec_size, integer(1L))
  )
}

#' @export
density.dist_sample <- function(x, at, ..., na.rm = TRUE){
  # Apply independently over sample variates
  if(is.matrix(x$x)) {
    abort("Multivariate sample density is not yet implemented.")
  }
  z <- numeric(length(at))
  zi <- is.finite(at)
  at <- at[zi]
  zl <- vec_size(at)

  # Shortcut if only one point in density is needed
  if(zl == 1){
    z[zi] <- density(x[["x"]], from = at, to = at, n = 1)$y
  } else if (zl > 1) {
    d <- density(x[["x"]], from = min(at), to = max(at), ..., na.rm=na.rm)
    z[zi] <- stats::approx(d$x, d$y, xout = at)$y
  }
  z
}


#' @export
quantile.dist_sample <- function(x, p, type = "marginal", ..., na.rm = TRUE){
  type <- match.arg(type)
  # Apply independently over sample variates
  if(is.matrix(x$x)) {
    # Marginal quantiles
    return(
      matrix(apply(x$x, 2, quantile, p = p, ..., na.rm = na.rm), nrow = length(p))
    )
  }
  quantile(x$x, probs = p, ..., na.rm = na.rm, names = FALSE)
}

#' @export
cdf.dist_sample <- function(x, q, ..., na.rm = TRUE){
  if(vec_size(q) > 1) return(vapply(q, cdf, numeric(1L), x = x, ...))
  if(is.matrix(x$x)) {
    return(mean(x$x <= vec_recycle(q, vec_size(x$x))))
  }
  mean(x <= q, ..., na.rm = na.rm)
  # vapply(x, function(x, q) mean(x <= q, ..., na.rm = na.rm), numeric(1L), q = q)
}

#' @export
generate.dist_sample <- function(x, times, ...){
  i <- sample.int(vec_size(x[["x"]]), size = times, replace = TRUE)
  if(is.matrix(x$x)) x$x[i,,drop = FALSE] else x$x[i]
}

#' @export
mean.dist_sample <- function(x, ...){
  if(is.matrix(x$x)) {
    matrix(colMeans(x$x, ...), nrow = 1L)
  } else {
    mean(x$x, ...)
  }
}

#' @export
median.dist_sample <- function(x, na.rm = FALSE, ...){
  if(is.matrix(x$x))
    matrix(apply(x$x, 2, median, na.rm = na.rm, ...), nrow = 1L)
  else
    median(x$x, na.rm = na.rm, ...)
}

#' @export
covariance.dist_sample <- function(x, ...){
  if(is.matrix(x$x)) stats::cov(x$x, ...) else stats::var(x$x, ...)
}

#' @export
skewness.dist_sample <- function(x, ..., na.rm = FALSE) {
  if(is.matrix(x$x)) {abort("Multivariate sample skewness is not yet implemented.")}
  n <- lengths(x, use.names = FALSE)
  x <- lapply(x, function(.) . - mean(., na.rm = na.rm))
  sum_x2 <- vapply(x, function(.) sum(.^2, na.rm = na.rm), numeric(1L), USE.NAMES = FALSE)
  sum_x3 <- vapply(x, function(.) sum(.^3, na.rm = na.rm), numeric(1L), USE.NAMES = FALSE)
  y <- sqrt(n) * sum_x3/(sum_x2^(3/2))
  y * ((1 - 1/n))^(3/2)
}

#' @export
support.dist_sample <- function(x, ...) {
  new_support_region(
    list(vctrs::vec_init(x$x, n = 0L)),
    list(range(x$x)),
    list(rep(TRUE, 2))
  )
}

#' @method Math dist_sample
#' @export
Math.dist_sample <- function(x, ...) {
  x <- .mapply(.Generic, list(parameters(x)$x, ...), MoreArgs = NULL)
  names(x) <- "x"
  enclass_dist(x, "dist_sample")
}

#' @method Ops dist_sample
#' @export
Ops.dist_sample <- function(e1, e2) {
  is_dist <- c(inherits(e1, "dist_sample"), inherits(e2, "dist_sample"))
  if(is_dist[1]) {
    e1 <- parameters(e1)$x
  }
  if(is_dist[2]) {
    e2 <- parameters(e2)$x
  }

  x <- .mapply(.Generic, list(e1, e2), MoreArgs = NULL)
  names(x) <- "x"
  enclass_dist(x, "dist_sample")
}

#' Modify a distribution with a transformation
#'
#' @description
#' `r lifecycle::badge('maturing')`
#'
#' The [`density()`], [`mean()`], and [`variance()`] methods are approximate as
#' they are based on numerical derivatives.
#'
#' @param dist A univariate distribution vector.
#' @param transform A function used to transform the distribution. This
#' transformation should be monotonic over appropriate domain.
#' @param inverse The inverse of the `transform` function.
#'
#' @examples
#' # Create a log normal distribution
#' dist <- dist_transformed(dist_normal(0, 0.5), exp, log)
#' density(dist, 1) # dlnorm(1, 0, 0.5)
#' cdf(dist, 4) # plnorm(4, 0, 0.5)
#' quantile(dist, 0.1) # qlnorm(0.1, 0, 0.5)
#' generate(dist, 10) # rlnorm(10, 0, 0.5)
#'
#' @export
dist_transformed <- function(dist, transform, inverse){
  vec_is(dist, new_dist())
  if(is.function(transform)) transform <- list(transform)
  if(is.function(inverse)) inverse <- list(inverse)
  new_dist(dist = vec_data(dist),
           transform = transform, inverse = inverse,
           dimnames = dimnames(dist), class = "dist_transformed")
}

#' @export
format.dist_transformed <- function(x, ...){
  sprintf(
    "t(%s)",
    format(x[["dist"]])
  )
}

#' @export
support.dist_transformed <- function(x, ...) {
  support <- support(x[["dist"]])
  lim <- field(support, "lim")[[1]]
  lim <- suppressWarnings(x[['transform']](lim))
  if (all(!is.na(lim))) {
    lim <- sort(lim)
  }
  field(support, "lim")[[1]] <- lim
  support
}

#' @export
density.dist_transformed <- function(x, at, ...){
  inv <- function(v) suppressWarnings(x[["inverse"]](v))
  jacobian <- vapply(at, numDeriv::jacobian, numeric(1L), func = inv)
  d <- density(x[["dist"]], inv(at)) * abs(jacobian)
  limits <- field(support(x), "lim")[[1]]
  closed <- field(support(x), "closed")[[1]]
  if (!any(is.na(limits))) {
    `%less_than%` <- if (closed[1]) `<` else `<=`
    `%greater_than%` <- if (closed[2]) `>` else `>=`
    d[which(at %less_than% limits[1] | at %greater_than% limits[2])] <- 0
  }
  d
}

#' @export
cdf.dist_transformed <- function(x, q, ...){
  inv <- function(v) suppressWarnings(x[["inverse"]](v))
  p <- cdf(x[["dist"]], inv(q), ...)
  # TODO - remove null dist check when dist_na is structured correctly (revdep temp fix)
  if(!is.null(x[["dist"]]) && !monotonic_increasing(x[["transform"]], support(x[["dist"]]))) p <- 1 - p

  # TODO: Rework for support of closed limits and prevent computation
  x_sup <- support(x)
  x_lim <- field(x_sup, "lim")[[1]]
  x_cls <- field(x_sup, "closed")[[1]]
  if (!any(is.na(x_lim))) {
    p[q <= x_lim[1] & !x_cls[1]] <- 0
    p[q >= x_lim[2] & !x_cls[2]] <- 1
  }
  p
}

#' @export
quantile.dist_transformed <- function(x, p, ...){
  # TODO - remove null dist check when dist_na is structured correctly (revdep temp fix)
  if(!is.null(x[["dist"]]) && !monotonic_increasing(x[["transform"]], support(x[["dist"]]))) p <- 1 - p
  x[["transform"]](quantile(x[["dist"]], p, ...))
}

#' @export
generate.dist_transformed <- function(x, ...){
  x[["transform"]](generate(x[["dist"]], ...))
}

#' @export
mean.dist_transformed <- function(x, ...){
  mu <- mean(x[["dist"]])
  sigma2 <- variance(x[["dist"]])
  if(is.na(sigma2)){
    # warning("Could not compute the transformed distribution's mean as the base distribution's variance is unknown. The transformed distribution's median has been returned instead.")
    return(x[["transform"]](mu))
  }
  drop(
    x[["transform"]](mu) + numDeriv::hessian(x[["transform"]], mu, method.args=list(d = 0.01))/2*sigma2
  )
}

#' @export
covariance.dist_transformed <- function(x, ...){
  mu <- mean(x[["dist"]])
  sigma2 <- variance(x[["dist"]])
  if(is.na(sigma2)) return(NA_real_)
  drop(
    numDeriv::jacobian(x[["transform"]], mu)^2*sigma2 + (numDeriv::hessian(x[["transform"]], mu, method.args=list(d = 0.01))*sigma2)^2/2
  )
}

#' @method Math dist_transformed
#' @export
Math.dist_transformed <- function(x, ...) {
  trans <- new_function(exprs(x = ), body = expr((!!sym(.Generic))((!!x$transform)(x), !!!dots_list(...))))

  inverse_fun <- get_unary_inverse(.Generic)
  inverse <- new_function(exprs(x = ), body = expr((!!x$inverse)((!!inverse_fun)(x, !!!dots_list(...)))))

  vec_data(dist_transformed(wrap_dist(list(x[["dist"]])), trans, inverse))[[1]]
}

#' @method Ops dist_transformed
#' @export
Ops.dist_transformed <- function(e1, e2) {
  if(.Generic %in% c("-", "+") && missing(e2)){
    e2 <- e1
    e1 <- if(.Generic == "+") 1 else -1
    .Generic <- "*"
  }
  is_dist <- c(inherits(e1, "dist_default"), inherits(e2, "dist_default"))
  trans <- if(all(is_dist)) {
    if(identical(e1$dist, e2$dist)){
      new_function(exprs(x = ), expr((!!sym(.Generic))((!!e1$transform)(x), (!!e2$transform)(x))))
    } else {
      stop(sprintf("The %s operation is not supported for <%s> and <%s>", .Generic, class(e1)[1], class(e2)[1]))
    }
  } else if(is_dist[1]){
    new_function(exprs(x = ), body = expr((!!sym(.Generic))((!!e1$transform)(x), !!e2)))
  } else {
    new_function(exprs(x = ), body = expr((!!sym(.Generic))(!!e1, (!!e2$transform)(x))))
  }

  inverse <- if(all(is_dist)) {
    invert_fail
  } else if(is_dist[1]){
    inverse_fun <- get_binary_inverse_1(.Generic, e2)
    new_function(exprs(x = ), body = expr((!!e1$inverse)((!!inverse_fun)(x))))
  } else {
    inverse_fun <- get_binary_inverse_2(.Generic, e1)
    new_function(exprs(x = ), body = expr((!!e2$inverse)((!!inverse_fun)(x))))
  }

  vec_data(dist_transformed(wrap_dist(list(list(e1,e2)[[which(is_dist)[1]]][["dist"]])), trans, inverse))[[1]]
}

monotonic_increasing <- function(f, support) {
  # Shortcut for identity function (used widely in ggdist)
  if(!is.primitive(f) && identical(body(f), as.name(names(formals(f))))) {
    return(TRUE)
  }

  # Currently assumes (without checking, #9) monotonicity of f over the domain
  x <- f(field(support, "lim")[[1]])
  isTRUE(x[[2L]] > x[[1L]])
}

#' Create a new support region vector
#'
#' @param x A list of prototype vectors defining the distribution type.
#' @param limits A list of value limits for the distribution.
#' @param closed A list of logical(2L) indicating whether the limits are closed.
#'
new_support_region <- function(x = numeric(), limits = list(), closed = list()) {
  vctrs::new_rcrd(list(x = x, lim = limits, closed = closed), class = "support_region")
}

#' @export
format.support_region <- function(x, digits = 3, ...) {
  type <- vapply(field(x, "x"), function(z) {
    out <- if(is.integer(z)) "Z"
    else if(is.numeric(z)) "R"
    else if(is.complex(z)) "C"
    else vec_ptype_abbr(z)
    if(is.matrix(z)) {
      if(ncol(z) > 1) {
        out <- paste(out, ncol(z), sep = "^")
      }
    }
    out
  }, FUN.VALUE = character(1L))
  brackets <- list(c("(","["), c(")","]"))
  mapply(function(type, z, closed) {
    br1 <- brackets[[1]][closed[1] + 1L]
    br2 <- brackets[[2]][closed[2] + 1L]
    fz <- sapply(z, function(x) format(x, digits = digits))
    fz <- gsub("3.14", "pi", fz, fixed = TRUE)
    if (any(is.na(z)) || all(is.infinite(z))) type
    else if (type == "Z") {
      if (identical(z, c(0L, Inf))) "N0"
      else if (identical(z, c(1L, Inf))) "N+"
      else paste0("{", z[1], ",", z[1]+1L, ",...,", z[2], "}")
    }
    else if (type == "R") paste0(br1, fz[1], ",", fz[2], br2)
    else type
  }, type, field(x, "lim"), field(x, "closed"))
}

#' @export
vec_ptype_abbr.support_region <- function(x, ...){
  "support"
}

#' The Categorical distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Categorical distributions are used to represent events with multiple
#' outcomes, such as what number appears on the roll of a dice. This is also
#' referred to as the 'generalised Bernoulli' or 'multinoulli' distribution.
#' The Categorical distribution is a special case of the [Multinomial()]
#' distribution with `n = 1`.
#'
#' @param prob A list of probabilities of observing each outcome category.
#' @param outcomes The list of vectors where each value represents each outcome.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_categorical")`
#'
#'   In the following, let \eqn{X} be a Categorical random variable with
#'   probability parameters `prob` = \eqn{\{p_1, p_2, \ldots, p_k\}}.
#'
#'   The Categorical probability distribution is widely used to model the
#'   occurance of multiple events. A simple example is the roll of a dice, where
#'   \eqn{p = \{1/6, 1/6, 1/6, 1/6, 1/6, 1/6\}} giving equal chance of observing
#'   each number on a 6 sided dice.
#'
#'   **Support**: \eqn{\{1, \ldots, k\}}{{1, ..., k}}
#'
#'   **Mean**: Not defined for unordered categories. For ordered categories with
#'   integer outcomes \eqn{\{1, 2, \ldots, k\}}, the mean is:
#'
#'   \deqn{
#'     E(X) = \sum_{i=1}^{k} i \cdot p_i
#'   }{
#'     E(X) = sum(i * p_i) for i = 1 to k
#'   }
#'
#'   **Variance**: Not defined for unordered categories. For ordered categories
#'   with integer outcomes \eqn{\{1, 2, \ldots, k\}}, the variance is:
#'
#'   \deqn{
#'     \text{Var}(X) = \sum_{i=1}^{k} i^2 \cdot p_i - \left(\sum_{i=1}^{k} i \cdot p_i\right)^2
#'   }{
#'     Var(X) = sum(i^2 * p_i) - [sum(i * p_i)]^2
#'   }
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     P(X = i) = p_i
#'   }{
#'     P(X = i) = p_i
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   The c.d.f is undefined for unordered categories. For ordered categories
#'   with outcomes \eqn{x_1 < x_2 < \ldots < x_k}, the c.d.f is:
#'
#'   \deqn{
#'     P(X \le x_j) = \sum_{i=1}^{j} p_i
#'   }{
#'     P(X <= x_j) = sum(p_i) for i = 1 to j
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \sum_{i=1}^{k} e^{tx_i} \cdot p_i
#'   }{
#'     E(e^(tX)) = sum(e^(t * x_i) * p_i) for i = 1 to k
#'   }
#'
#'   **Skewness**: Approximated numerically for ordered categories.
#'
#'   **Kurtosis**: Approximated numerically for ordered categories.
#'
#' @seealso [stats::Multinomial]
#'
#' @examples
#' dist <- dist_categorical(prob = list(c(0.05, 0.5, 0.15, 0.2, 0.1), c(0.3, 0.1, 0.6)))
#'
#' dist
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' # The outcomes aren't ordered, so many statistics are not applicable.
#' cdf(dist, 0.6)
#' quantile(dist, 0.7)
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#' 
#' # Some of these statistics are meaningful for ordered outcomes
#' dist <- dist_categorical(list(rpois(26, 3)), list(ordered(letters)))
#' dist
#' cdf(dist, "m")
#' quantile(dist, 0.5)
#'
#' dist <- dist_categorical(
#'   prob = list(c(0.05, 0.5, 0.15, 0.2, 0.1), c(0.3, 0.1, 0.6)),
#'   outcomes = list(letters[1:5], letters[24:26])
#' )
#'
#' generate(dist, 10)
#'
#' density(dist, "a")
#' density(dist, "z", log = TRUE)
#'
#' @name dist_categorical
#' @export
dist_categorical <- function(prob, outcomes = NULL){
  prob <- lapply(prob, function(x) x/sum(x))
  prob <- as_list_of(prob, .ptype = double())
  if(is.null(outcomes)) {
    new_dist(p = prob, class = "dist_categorical")
  } else {
    new_dist(p = prob, x = outcomes, class = "dist_categorical")
  }
}

#' @export
format.dist_categorical <- function(x, digits = 2, ...){
  sprintf(
    "Categorical[%s]",
    format(length(x[["p"]]), digits = digits, ...)
  )
}

#' @export
density.dist_categorical <- function(x, at, ...){
  if(!is.null(x[["x"]])) at <- match(at, x[["x"]])
  at[at <= 0] <- NA_real_
  x[["p"]][at]
}

#' @export
quantile.dist_categorical <- function(x, p, ...){
  if(is.ordered(x[["x"]])) {
    return(x[["x"]][findInterval(p, cumsum(x[["p"]]))])
  }
  rep_len(NA_real_, length(p))
}

#' @export
cdf.dist_categorical <- function(x, q, ...){
  if(is.ordered(x[["x"]])) {
    return(cumsum(x[["p"]])[match(q, x[["x"]])])
  }
  rep_len(NA_real_, length(q))
}

#' @export
generate.dist_categorical <- function(x, times, ...){
  z <- sample(
    x = seq_along(x[["p"]]), size = times, prob = x[["p"]], replace = TRUE
  )
  if(is.null(x[["x"]])) return(z)
  x[["x"]][z]
}

#' @export
support.dist_categorical <- function(x, ...) {
  region <- if(is.null(x[["p"]])) seq_along(x[["p"]]) else x[["x"]]
  new_support_region(
    list(vctrs::vec_init(region, n = 0L)),
    list(region),
    list(c(TRUE, TRUE))
  )
}

#' @export
mean.dist_categorical <- function(x, ...){
  NA_real_
}

#' @export
covariance.dist_categorical <- function(x, ...){
  NA_real_
}

#' @export
skewness.dist_categorical <- function(x, ...) {
  NA_real_
}

#' @export
kurtosis.dist_categorical <- function(x, ...) {
  NA_real_
}

#' The Uniform distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' A distribution with constant density on an interval.
#'
#' @inheritParams stats::dunif
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a Poisson random variable with parameter
#'   `lambda` = \eqn{\lambda}.
#'
#'   **Support**: \eqn{[a,b]}{[a,b]}
#'
#'   **Mean**: \eqn{\frac{1}{2}(a+b)}
#'
#'   **Variance**: \eqn{\frac{1}{12}(b-a)^2}
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{1}{b-a} for x \in [a,b]
#'   }{
#'     f(x) = \frac{1}{b-a} for x in [a,b]
#'   }
#'   \deqn{
#'     f(x) = 0 otherwise
#'   }{
#'     f(x) = 0 otherwise
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = 0 for x < a
#'   }{
#'     F(x) = 0 for x < a
#'   }
#'   \deqn{
#'     F(x) = \frac{x - a}{b-a} for x \in [a,b]
#'   }{
#'     F(x) = \frac{x - a}{b-a} for x in [a,b]
#'   }
#'   \deqn{
#'     F(x) = 1 for x > b
#'   }{
#'     F(x) = 1 for x > b
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \frac{e^{tb} - e^{ta}}{t(b-a)} for t \neq 0
#'   }{
#'     E(e^(tX)) = \frac{e^{tb} - e^{ta}}{t(b-a)} for t \neq 0
#'   }
#'   \deqn{
#'     E(e^{tX}) = 1 for t = 0
#'   }{
#'     E(e^(tX)) = 1 for t = 0
#'   }
#'
#' @seealso [stats::Uniform]
#'
#' @examples
#' dist <- dist_uniform(min = c(3, -2), max = c(5, 4))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_uniform
#' @export
dist_uniform <- function(min, max){
  min <- vec_cast(min, double())
  max <- vec_cast(max, double())
  if(any(min > max)){
    abort("The min of a Uniform distribution must be less than max.")
  }
  new_dist(l = min, u = max, class = "dist_uniform")
}

#' @export
format.dist_uniform <- function(x, digits = 2, ...){
  sprintf(
    "U(%s, %s)",
    format(x[["l"]], digits = digits, ...),
    format(x[["u"]], digits = digits, ...)
  )
}

#' @export
density.dist_uniform <- function(x, at, ...){
  stats::dunif(at, x[["l"]], x[["u"]])
}

#' @export
log_density.dist_uniform <- function(x, at, ...){
  stats::dunif(at, x[["l"]], x[["u"]], log = TRUE)
}

#' @export
quantile.dist_uniform <- function(x, p, ...){
  stats::qunif(p, x[["l"]], x[["u"]])
}

#' @export
cdf.dist_uniform <- function(x, q, ...){
  stats::punif(q, x[["l"]], x[["u"]])
}

#' @export
generate.dist_uniform <- function(x, times, ...){
  stats::runif(times, x[["l"]], x[["u"]])
}

#' @export
mean.dist_uniform <- function(x, ...){
  (x[["u"]]+x[["l"]])/2
}

#' @export
covariance.dist_uniform <- function(x, ...){
  (x[["u"]]-x[["l"]])^2/12
}

#' @export
skewness.dist_uniform <- function(x, ...) 0

#' @export
kurtosis.dist_uniform <- function(x, ...) -6/5

#' Inflate a value of a probability distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Inflated distributions add extra probability mass at a specific value,
#' most commonly zero (zero-inflation). These distributions are useful for
#' modeling data with excess observations at a particular value compared to
#' what the base distribution would predict. Common applications include
#' zero-inflated Poisson or negative binomial models for count data with
#' many zeros.
#'
#' @param dist The distribution(s) to inflate.
#' @param prob The added probability of observing `x`.
#' @param x The value to inflate. The default of `x = 0` is for zero-inflation.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_inflated")`
#'
#'   In the following, let \eqn{Y} be an inflated random variable based on
#'   a base distribution \eqn{X}, with inflation value `x` = \eqn{c} and
#'   inflation probability `prob` = \eqn{p}.
#'
#'   **Support**: Same as the base distribution, but with additional
#'   probability mass at \eqn{c}
#'
#'   **Mean**: (when `x` is numeric)
#'
#'   \deqn{
#'     E(Y) = p \cdot c + (1-p) \cdot E(X)
#'   }{
#'     E(Y) = p * c + (1-p) * E(X)
#'   }
#'
#'   **Variance**: (when `x = 0`)
#'
#'   \deqn{
#'     \text{Var}(Y) = (1-p) \cdot \text{Var}(X) + p(1-p) \cdot [E(X)]^2
#'   }{
#'     Var(Y) = (1-p) * Var(X) + p(1-p) * [E(X)]^2
#'   }
#'
#'   For non-zero inflation values, the variance is not computed in closed form.
#'
#'   **Probability mass/density function (p.m.f/p.d.f)**:
#'
#'   For discrete distributions:
#'   \deqn{
#'     f_Y(y) = \begin{cases}
#'       p + (1-p) \cdot f_X(c) & \text{if } y = c \\
#'       (1-p) \cdot f_X(y) & \text{if } y \neq c
#'     \end{cases}
#'   }{
#'     f_Y(y) = p + (1-p) * f_X(c) if y = c, (1-p) * f_X(y) if y != c
#'   }
#'
#'   For continuous distributions:
#'   \deqn{
#'     f_Y(y) = \begin{cases}
#'       p & \text{if } y = c \\
#'       (1-p) \cdot f_X(y) & \text{if } y \neq c
#'     \end{cases}
#'   }{
#'     f_Y(y) = p if y = c, (1-p) * f_X(y) if y != c
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F_Y(q) = \begin{cases}
#'       (1-p) \cdot F_X(q) & \text{if } q < c \\
#'       p + (1-p) \cdot F_X(q) & \text{if } q \geq c
#'     \end{cases}
#'   }{
#'     F_Y(q) = (1-p) * F_X(q) if q < c, p + (1-p) * F_X(q) if q >= c
#'   }
#'
#'   **Quantile function**:
#'
#'   The quantile function is computed numerically by inverting the
#'   inflated CDF, accounting for the jump in probability at the
#'   inflation point.
#'
#' @examples
#' # Zero-inflated Poisson
#' dist <- dist_inflated(dist_poisson(lambda = 2), prob = 0.3, x = 0)
#'
#' dist
#' mean(dist)
#' variance(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 0)
#' density(dist, 1)
#'
#' cdf(dist, 2)
#'
#' quantile(dist, 0.5)
#'
#' @name dist_inflated
#' @export
dist_inflated <- function(dist, prob, x = 0){
  vec_is(dist, new_dist())
  if(prob < 0 || prob > 1){
    abort("The inflation probability must be between 0 and 1.")
  }
  new_dist(dist = dist, x = x, p = prob,
           dimnames = dimnames(dist), class = "dist_inflated")
}

#' @export
format.dist_inflated <- function(x, ...){
  sprintf(
    "%s+%s",
    format(x[["x"]]),
    format(x[["dist"]])
  )
}

#' @export
density.dist_inflated <- function(x, at, ...){
  x[["p"]]*(at==x[["x"]]) + (1-x[["p"]])*density(x[["dist"]], at, ...)
}

#' @export
quantile.dist_inflated <- function(x, p, ...){
  qt <- quantile(x[["dist"]], pmax(0, (p - x[["p"]]) / (1-x[["p"]])), ...)
  if(qt >= x[["x"]]) return(qt)
  qt <- quantile(x[["dist"]], p, ...)
  if(qt < x[["x"]]) qt else x[["x"]]
}

#' @export
cdf.dist_inflated <- function(x, q, ...){
  x[["p"]]*(q>=x[["x"]]) + (1-x[["p"]])*cdf(x[["dist"]], q, ...)
}

#' @export
generate.dist_inflated <- function(x, times, ...){
  p <- x[["p"]]
  inf <- stats::runif(times) < p
  r <- vec_init(x[["x"]], times)
  r[inf] <- x[["x"]]
  r[!inf] <- generate(x[["dist"]], sum(!inf))
  r
}

#' @export
mean.dist_inflated <- function(x, ...){
  # Can't compute if inflation value is not numeric
  if(!vec_is(x[["x"]], numeric())) return(NA_real_)

  p <- x[["p"]]
  p*x[["x"]] + (1-p)*mean(x[["dist"]])
}

#' @export
covariance.dist_inflated <- function(x, ...){
  # Can't compute if inflation value is not numeric
  if(!vec_is(x[["x"]], numeric())) return(NA_real_)
  # Can't (easily) compute if inflation value is not zero
  if(x[["x"]] != 0) return(NA_real_)

  m1 <- mean(x[["dist"]])
  v <- variance(x[["dist"]])
  m2 <- v + m1^2
  p <- x[["p"]]
  (1-p)*v + p*(1-p)*m1^2
}

#' The Bernoulli distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Bernoulli distributions are used to represent events like coin flips
#' when there is single trial that is either successful or unsuccessful.
#' The Bernoulli distribution is a special case of the [Binomial()]
#' distribution with `n = 1`.
#'
#' @inheritParams dist_binomial
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a Bernoulli random variable with parameter
#'   `p` = \eqn{p}. Some textbooks also define \eqn{q = 1 - p}, or use
#'   \eqn{\pi} instead of \eqn{p}.
#'
#'   The Bernoulli probability  distribution is widely used to model
#'   binary variables, such as 'failure' and 'success'. The most
#'   typical example is the flip of a coin, when  \eqn{p} is thought as the
#'   probability of flipping a head, and \eqn{q = 1 - p} is the
#'   probability of flipping a tail.
#'
#'   **Support**: \eqn{\{0, 1\}}{{0, 1}}
#'
#'   **Mean**: \eqn{p}
#'
#'   **Variance**: \eqn{p \cdot (1 - p) = p \cdot q}{p (1 - p)}
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     P(X = x) = p^x (1 - p)^{1-x} = p^x q^{1-x}
#'   }{
#'     P(X = x) = p^x (1 - p)^(1-x)
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     P(X \le x) =
#'     \left \{
#'       \begin{array}{ll}
#'         0 & x < 0 \\
#'         1 - p & 0 \leq x < 1 \\
#'         1 & x \geq 1
#'       \end{array}
#'     \right.
#'   }{
#'     P(X \le x) = (1 - p) 1_{[0, 1)}(x) + 1_{1}(x)
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = (1 - p) + p e^t
#'   }{
#'     E(e^(tX)) = (1 - p) + p e^t
#'   }
#'
#' @examples
#' dist <- dist_bernoulli(prob = c(0.05, 0.5, 0.3, 0.9, 0.1))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @export
dist_bernoulli <- function(prob){
  prob <- vec_cast(prob, double())
  if(any((prob < 0) | (prob > 1))){
    abort("The probability of success must be between 0 and 1.")
  }
  new_dist(p = prob, class = "dist_bernoulli")
}

#' @export
format.dist_bernoulli <- function(x, digits = 2, ...){
  sprintf(
    "Bernoulli(%s)",
    format(x[["p"]], digits = digits, ...)
  )
}

#' @export
density.dist_bernoulli <- function(x, at, ...){
  stats::dbinom(at, 1, x[["p"]])
}

#' @export
log_density.dist_bernoulli <- function(x, at, ...){
  stats::dbinom(at, 1, x[["p"]], log = TRUE)
}

#' @export
quantile.dist_bernoulli <- function(x, p, ...){
  as.logical(stats::qbinom(p, 1, x[["p"]]))
}

#' @export
cdf.dist_bernoulli <- function(x, q, ...){
  stats::pbinom(q, 1, x[["p"]])
}

#' @export
generate.dist_bernoulli <- function(x, times, ...){
  as.logical(stats::rbinom(times, 1, x[["p"]]))
}

#' @export
mean.dist_bernoulli <- function(x, ...){
  x[["p"]]
}

#' @export
covariance.dist_bernoulli <- function(x, ...){
  x[["p"]]*(1-x[["p"]])
}

#' @export
skewness.dist_bernoulli <- function(x, ...) {
  p <- x[["p"]]
  q <- 1 - x[["p"]]
  (1 - (2 * p)) / sqrt(p * q)
}

#' @export
kurtosis.dist_bernoulli <- function(x, ...) {
  p <- x[["p"]]
  q <- 1 - x[["p"]]
  (1 - (6 * p * q)) / (p * q)
}

#' Create a mixture of distributions
#'
#' @description
#' `r lifecycle::badge('maturing')`
#'
#' @param ... Distributions to be used in the mixture.
#' @param weights The weight of each distribution passed to `...`.
#'
#' @examples
#' dist_mixture(dist_normal(0, 1), dist_normal(5, 2), weights = c(0.3, 0.7))
#'
#' @export
dist_mixture <- function(..., weights = numeric()){
  dist <- dots_list(...)
  dn <- unique(lapply(dist, dimnames))
  dn <- if(length(dn) == 1) dn[[1]] else NULL

  vec_is(weights, numeric(), length(dist))
  if(sum(weights) != 1){
    abort("Weights of a mixture model must sum to 1.")
  }
  if(any(weights < 0)){
    abort("All weights in a mixtue model must be non-negative.")
  }
  new_dist(dist = transpose(dist), w = list(weights),
           class = "dist_mixture", dimnames = dn)
}

#' @export
format.dist_mixture <- function(x, width = getOption("width"), ...){
  dists <- lapply(x[["dist"]], format) |>
    unlist()

  dist_info <- paste0(x[["w"]], "*", dists) |>
    paste0(collapse = ", ")

  long_dist <- paste0("mixture(", dist_info, ")")
  short_dist <- paste0("mixture(n=", length(dists), ")")
  ifelse(nchar(long_dist) <= width, long_dist, short_dist)
}

#' @export
density.dist_mixture <- function(x, at, ...){
  if(NROW(at) > 1) return(vapply(at, density, numeric(1L), x = x, ...))
  sum(x[["w"]]*vapply(x[["dist"]], density, numeric(1L), at = at, ...))
}

#' @export
quantile.dist_mixture <- function(x, p, ...){
  d <- dim(x)
  if(d > 1)
    stop("quantile is not implemented for multivariate mixtures.")
  if(length(p) > 1) return(vapply(p, quantile, numeric(1L), x = x, ...))

  # Find bounds for optimisation based on range of each quantile
  dist_q <- vapply(x[["dist"]], quantile, numeric(1L), p, ..., USE.NAMES = FALSE)
  if(vctrs::vec_unique_count(dist_q) == 1) return(dist_q[1])
  if(p == 0) return(min(dist_q))
  if(p == 1) return(max(dist_q))

  # Search the cdf() for appropriate quantile
  stats::uniroot(
    function(pos) p - cdf(x, pos, ...),
    interval = c(min(dist_q), max(dist_q)),
    extendInt = "yes"
  )$root
}

#' @export
cdf.dist_mixture <- function(x, q, times = 1e5, ...){
  d <- dim(x)
  if(d == 1L) {
    if(length(q) > 1) return(vapply(q, cdf, numeric(1L), x = x, ...))
    sum(x[["w"]]*vapply(x[["dist"]], cdf, numeric(1L), q = q, ...))
  } else {
    NextMethod()
  }
}

#' @export
generate.dist_mixture <- function(x, times, ...){
  dist_idx <- .bincode(stats::runif(times), breaks = c(0, cumsum(x[["w"]])))
  r <- matrix(nrow = times, ncol = dim(x))
  for(i in seq_along(x[["dist"]])){
    r_pos <- dist_idx == i
    if(any(r_pos)) {
      r[r_pos,] <- generate(x[["dist"]][[i]], sum(r_pos), ...)
    }
  }
  r[,seq(NCOL(r)), drop = TRUE]
}

#' @export
mean.dist_mixture <- function(x, ...){
  d <- dim(x)
  m <- vapply(x[["dist"]], mean, numeric(d), ...)
  if(d == 1L) {
    sum(x[["w"]] * m)
  } else {
    matrix(x[["w"]], ncol = d, nrow = 1) %*% t(m)
  }
}

#' @export
covariance.dist_mixture <- function(x, ...){
  d <- dim(x)
  if(d == 1L) {
    m <- vapply(x[["dist"]], mean, numeric(1L), ...)
    v <- vapply(x[["dist"]], variance, numeric(1L), ...)
    m1 <- sum(x[["w"]]*m)
    m2 <- sum(x[["w"]]*(m^2 + v))
    m2 - m1^2
  } else {
    m <- lapply(x[["dist"]], mean)
    w <- as.list(x[["w"]])
    mbar <- mapply("*", m, w, SIMPLIFY = FALSE)
    mbar <- do.call("+", mbar)
    m <- lapply(m, function(u){u - mbar})
    v <- lapply(x[["dist"]], function(u){covariance(u)[[1]]})
    cov <- mapply(function(m,v,w) {w * ( t(m) %*% m + v ) }, m, v, w, SIMPLIFY = FALSE)
    list(do.call("+", cov))
  }
}

#' @export
dim.dist_mixture <- function(x){
  dim(x[["dist"]][[1]])
}

#' The F Distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The F distribution is commonly used in statistical inference, particularly
#' in the analysis of variance (ANOVA), testing the equality of variances,
#' and in regression analysis. It arises as the ratio of two scaled
#' chi-squared distributions divided by their respective degrees of freedom.
#'
#' @param df1 Degrees of freedom for the numerator. Can be any positive number.
#' @param df2 Degrees of freedom for the denominator. Can be any positive number.
#' @param ncp Non-centrality parameter. If `NULL` (default), the central F
#'   distribution is used. If specified, must be non-negative.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_f")`
#'
#'   In the following, let \eqn{X} be an F random variable with numerator
#'   degrees of freedom `df1` = \eqn{d_1} and denominator degrees of freedom
#'   `df2` = \eqn{d_2}.
#'
#'   **Support**: \eqn{x \in (0, \infty)}
#'
#'   **Mean**: 
#'   
#'   For the central F distribution (\code{ncp = NULL}):
#'   
#'   \deqn{
#'     E(X) = \frac{d_2}{d_2 - 2}
#'   }{
#'     E(X) = d_2 / (d_2 - 2)
#'   }
#'   
#'   for \eqn{d_2 > 2}, otherwise undefined.
#'   
#'   For the non-central F distribution with non-centrality parameter 
#'   `ncp` = \eqn{\lambda}:
#'   
#'   \deqn{
#'     E(X) = \frac{d_2 (d_1 + \lambda)}{d_1 (d_2 - 2)}
#'   }{
#'     E(X) = d_2 (d_1 + \lambda) / (d_1 (d_2 - 2))
#'   }
#'   
#'   for \eqn{d_2 > 2}, otherwise undefined.
#'
#'   **Variance**: 
#'   
#'   For the central F distribution (\code{ncp = NULL}):
#'   
#'   \deqn{
#'     \text{Var}(X) = \frac{2 d_2^2 (d_1 + d_2 - 2)}{d_1 (d_2 - 2)^2 (d_2 - 4)}
#'   }{
#'     Var(X) = 2 d_2^2 (d_1 + d_2 - 2) / (d_1 (d_2 - 2)^2 (d_2 - 4))
#'   }
#'   
#'   for \eqn{d_2 > 4}, otherwise undefined.
#'   
#'   For the non-central F distribution with non-centrality parameter 
#'   `ncp` = \eqn{\lambda}:
#'   
#'   \deqn{
#'     \text{Var}(X) = \frac{2 d_2^2}{d_1^2} \cdot \frac{(d_1 + \lambda)^2 + (d_1 + 2\lambda)(d_2 - 2)}{(d_2 - 2)^2 (d_2 - 4)}
#'   }{
#'     Var(X) = 2 d_2^2 / d_1^2 * ((d_1 + lambda)^2 + (d_1 + 2*lambda)(d_2 - 2)) / ((d_2 - 2)^2 (d_2 - 4))
#'   }
#'   
#'   for \eqn{d_2 > 4}, otherwise undefined.
#'
#'   **Skewness**: 
#'   
#'   For the central F distribution (\code{ncp = NULL}):
#'   
#'   \deqn{
#'     \text{Skew}(X) = \frac{(2 d_1 + d_2 - 2) \sqrt{8 (d_2 - 4)}}{(d_2 - 6) \sqrt{d_1 (d_1 + d_2 - 2)}}
#'   }{
#'     Skew(X) = (2 d_1 + d_2 - 2) sqrt(8 (d_2 - 4)) / ((d_2 - 6) sqrt(d_1 (d_1 + d_2 - 2)))
#'   }
#'   
#'   for \eqn{d_2 > 6}, otherwise undefined.
#'   
#'   For the non-central F distribution, skewness has no simple closed form
#'   and is not computed.
#'
#'   **Excess Kurtosis**: 
#'   
#'   For the central F distribution (\code{ncp = NULL}):
#'   
#'   \deqn{
#'     \text{Kurt}(X) = \frac{12[d_1 (5 d_2 - 22)(d_1 + d_2 - 2) + (d_2 - 4)(d_2 - 2)^2]}{d_1 (d_2 - 6)(d_2 - 8)(d_1 + d_2 - 2)}
#'   }{
#'     Kurt(X) = 12[d_1 (5 d_2 - 22)(d_1 + d_2 - 2) + (d_2 - 4)(d_2 - 2)^2] / (d_1 (d_2 - 6)(d_2 - 8)(d_1 + d_2 - 2))
#'   }
#'   
#'   for \eqn{d_2 > 8}, otherwise undefined.
#'   
#'   For the non-central F distribution, kurtosis has no simple closed form
#'   and is not computed.
#'
#'   **Probability density function (p.d.f)**:
#'
#'   For the central F distribution (\code{ncp = NULL}):
#'
#'   \deqn{
#'     f(x) = \frac{\sqrt{\frac{(d_1 x)^{d_1} d_2^{d_2}}{(d_1 x + d_2)^{d_1 + d_2}}}}{x \, B(d_1/2, d_2/2)}
#'   }{
#'     f(x) = sqrt((d_1 x)^d_1 d_2^d_2 / (d_1 x + d_2)^(d_1 + d_2)) / (x B(d_1/2, d_2/2))
#'   }
#'   
#'   where \eqn{B(\cdot, \cdot)} is the beta function.
#'   
#'   For the non-central F distribution, the density involves an infinite
#'   series and is approximated numerically.
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   The c.d.f. does not have a simple closed form expression and is
#'   approximated numerically using regularized incomplete beta functions
#'   and related special functions.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   The moment generating function for the F distribution does not exist
#'   in general (it diverges for \eqn{t > 0}).
#'
#' @seealso [stats::FDist]
#'
#' @examples
#' dist <- dist_f(df1 = c(1,2,5,10,100), df2 = c(1,1,2,1,100))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_f
#' @export
dist_f <- function(df1, df2, ncp = NULL){
  df1 <- vec_cast(df1, double())
  df2 <- vec_cast(df2, double())
  ncp <- vec_cast(ncp, double())
  if(any((df1 < 0) | (df2 < 0))){
    abort("The degrees of freedom parameters of the F distribution must be non-negative.")
  }
  if(is.null(ncp)){
    new_dist(df1 = df1, df2 = df2, class = "dist_f")
  } else {
    new_dist(df1 = df1, df2 = df2, ncp = ncp, class = "dist_f")
  }
}

#' @export
format.dist_f <- function(x, digits = 2, ...){
  sprintf(
    "F(%s, %s)",
    format(x[["df1"]], digits = digits, ...),
    format(x[["df2"]], digits = digits, ...)
  )
}

#' @export
density.dist_f <- function(x, at, ...){
  if(is.null(x[["ncp"]])) {
    stats::df(at, x[["df1"]], x[["df2"]])
  } else {
    stats::df(at, x[["df1"]], x[["df2"]], x[["ncp"]])
  }
}

#' @export
log_density.dist_f <- function(x, at, ...){
  if(is.null(x[["ncp"]])) {
    stats::df(at, x[["df1"]], x[["df2"]], log = TRUE)
  } else {
    stats::df(at, x[["df1"]], x[["df2"]], x[["ncp"]], log = TRUE)
  }
}

#' @export
quantile.dist_f <- function(x, p, ...){
  if(is.null(x[["ncp"]])) {
    stats::qf(p, x[["df1"]], x[["df2"]])
  } else {
    stats::qf(p, x[["df1"]], x[["df2"]], x[["ncp"]])
  }
}

#' @export
cdf.dist_f <- function(x, q, ...){
  if(is.null(x[["ncp"]])) {
    stats::pf(q, x[["df1"]], x[["df2"]])
  } else {
    stats::pf(q, x[["df1"]], x[["df2"]], x[["ncp"]])
  }
}

#' @export
generate.dist_f <- function(x, times, ...){
  if(is.null(x[["ncp"]])) {
    stats::rf(times, x[["df1"]], x[["df2"]])
  } else {
    stats::rf(times, x[["df1"]], x[["df2"]], x[["ncp"]])
  }
}

#' @export
mean.dist_f <- function(x, ...){
  df1 <- x[["df1"]]
  df2 <- x[["df2"]]
  if(df2 > 2) {
    if(is.null(x[["ncp"]])){
      df2 / (df2 - 2)
    } else {
      (df2 * (df1 + x[["ncp"]])) / (df1 * (df2 - 2))
    }
  } else {
    NA_real_
  }
}

#' @export
covariance.dist_f <- function(x, ...){
  df1 <- x[["df1"]]
  df2 <- x[["df2"]]
  if(df2 > 4) {
    if(is.null(x[["ncp"]])){
      (2 * df2^2 * (df1 + df2 - 2))/(df1*(df2-2)^2*(df2-4))
    } else {
      2*((df1 + x[["ncp"]])^2 + (df1 + 2*x[["ncp"]])*(df2 - 2))/((df2-2)^2*(df2-4)) * (df2^2/df1^2)
    }
  } else {
    NA_real_
  }
}

#' @export
skewness.dist_f <- function(x, ...) {
  df1 <- x[["df1"]]
  df2 <- x[["df2"]]
  if(!is.null(x[["ncp"]])) return(NA_real_)
  if (df2 > 6) {
    a <- (2 * df1 + df2 - 2) * sqrt(8 * (df2 - 4))
    b <- (df2 - 6) * sqrt(df1 * (df1 + df2 - 2))
    a / b
  } else {
    NA_real_
  }
}

#' @export
kurtosis.dist_f <- function(x, ...) {
  df1 <- x[["df1"]]
  df2 <- x[["df2"]]
  if(!is.null(x[["ncp"]])) return(NA_real_)
  if (df2 > 8) {
    a <- df1 * (5 * df2 - 22) * (df1 + df2 - 2) + (df2 - 4) * (df2 - 2)^2
    b <- df1 * (df2 - 6) * (df2 - 8) * (df1 + df2 - 2)
    12 * a / b
  } else {
    NA_real_
  }
}

transpose <- function(.l) {
  if(is_empty(.l)) return(.l)
  inner_names <- names(.l[[1L]])
  result <- lapply(seq_along(.l[[1L]]), function(i) {
    lapply(.l, .subset2, i)
  })
  set_names(result, inner_names)
}

transpose_c <- function(.l) {
  stopifnot(is_list_of(.l))
  .ptype <- vec_init(attr(.l, "ptype"), 1L)
  if(is_empty(.l)) return(.l)
  inner_names <- names(.l[[1L]])
  .l <- vec_recycle_common(!!!.l)
  result <- lapply(seq_along(.l[[1L]]), function(i) {
    unname(vec_c(!!!lapply(.l, vec_slice, i), .ptype = .ptype))
  })
  set_names(result, inner_names)
}

split_matrix_rows <- function(x) {
  lapply(seq_len(nrow(x)), function(i) x[i,,drop=FALSE])
}

# Declare a function's argument as allowing list inputs for mapping values
arg_listable <- function(x, .ptype) {
  if(is.list(x)) {
    x <- as_list_of(as.list(x), .ptype)
    if(is.matrix(attr(x, "ptype"))) {
      x <- lapply(x, split_matrix_rows)
      x <- as_list_of(x, .ptype)
    }
    if(is.null(names(x))) {
      names(x) <- vec_as_names(character(vec_size(x)), repair = "unique")
    }
  } else if(is.matrix(x)) {
    x <- split_matrix_rows(x)
  } else {
    vec_assert(x, .ptype)
  }
  # Declares list arguments to be unpacked for dist_apply()
  class(x) <- c("arg_listable", class(x))
  x
}

validate_recycling <- function(x, arg) {
  if(is_list_of(arg)) return(lapply(arg, validate_recycling, x = x))
  if(!any(vec_size(arg) == c(1, vec_size(x)))) {
    abort(
      sprintf("Cannot recycle input of size %i to match the distributions (size %i).",
              vec_size(arg), vec_size(x)
      )
    )
  }
}

dist_apply <- function(x, .f, ...){
  dn <- dimnames(x)
  x <- vec_data(x)
  dist_is_na <- vapply(x, is.null, logical(1L))
  x[dist_is_na] <- list(structure(list(), class = c("dist_na", "dist_default")))

  args <- dots_list(...)
  is_arg_listable <- vapply(args, inherits, FUN.VALUE = logical(1L), "arg_listable")
  unpack_listable <- multi_arg <- FALSE
  if(any(is_arg_listable)) {
    if(sum(is_arg_listable) > 1) abort("Only distribution argument can be unpacked at a time.\nThis shouldn't happen, please report a bug at https://github.com/mitchelloharawild/distributional/issues/")
    arg_pos <- which(is_arg_listable)

    if(unpack_listable <- is_list_of(args[[arg_pos]])) {
      validate_recycling(x, args[[arg_pos]])
      .unpack_names <- names(args[[arg_pos]])
      args[[arg_pos]] <- transpose_c(args[[arg_pos]])
    } else if (multi_arg <- (length(args[[arg_pos]]) > 1)){
      args[[arg_pos]] <- list(unclass(args[[arg_pos]]))
    }
  }

  out <- do.call(mapply, c(.f, list(x), args, SIMPLIFY = FALSE, USE.NAMES = FALSE))
  # out <- mapply(.f, x, ..., SIMPLIFY = FALSE, USE.NAMES = FALSE)

  if(unpack_listable) {
    # TODO - update and repair multivariate distribution i/o with unpacking
    out <- as_list_of(out)
    if (rbind_mat <- is.matrix(attr(out, "ptype"))) {
      out <- as_list_of(lapply(out, split_matrix_rows))
    }
    out <- transpose_c(out)
    if(rbind_mat) {
      out <- lapply(out, function(x) `colnames<-`(do.call(rbind, x), dn))
    }
    names(out) <- .unpack_names
    out <- new_data_frame(out, n = vec_size(x))
  # } else if(length(out[[1]]) > 1) {
  #   out <- suppressMessages(vctrs::vec_rbind(!!!out))
  } else if(multi_arg) {
    if(length(dn) > 1) out <- lapply(out, `colnames<-`, dn)
  } else {
    out <- vec_c(!!!out)
    out <- if(vec_is_list(out))
      lapply(out, set_matrix_dimnames, dn = dn)
    else
      set_matrix_dimnames(out, dn)
  }
  out
}

set_matrix_dimnames <- function(x, dn) {
  if((is.matrix(x) || is.data.frame(x)) && !is.null(dn)){
    # Set dimension names
    colnames(x) <- dn
  }
  x
}

# inlined from https://github.com/r-lib/cli/blob/master/R/utf8.R
is_utf8_output <- function() {
  opt <- getOption("cli.unicode", NULL)
  if (!is_null(opt)) {
    isTRUE(opt)
  } else {
    l10n_info()$`UTF-8` && !is_latex_output()
  }
}

is_latex_output <- function() {
  if (!("knitr" %in% loadedNamespaces())) {
    return(FALSE)
  }
  get("is_latex_output", asNamespace("knitr"))()
}

require_package <- function(pkg){
  if(!requireNamespace(pkg, quietly = TRUE)){
    abort(
      sprintf('The `%s` package must be installed to use this functionality. It can be installed with install.packages("%s")', pkg, pkg)
    )
  }
}

restore_rng <- function(expr, seed = NULL) {
  old_seed <- .GlobalEnv$.Random.seed
  # Set new temporary seed
  set.seed(seed)
  # Restore previous seed
  on.exit(.GlobalEnv$.Random.seed <- old_seed)

  expr
}

near <- function(x, y) {
  tol <- .Machine$double.eps^0.5
  abs(x - y) < tol
}

#' Construct hdr intervals
#'
#' @param lower,upper A list of numeric vectors specifying the region's lower
#' and upper bounds.
#' @param size A numeric vector specifying the coverage size of the region.
#'
#' @return A "hdr" vector
#'
#' @author Mitchell O'Hara-Wild
#'
#' @examples
#'
#' new_hdr(lower = list(1, c(3,6)), upper = list(10, c(5, 8)), size = c(80, 95))
#'
#' @export
new_hdr <- function(lower = list_of(.ptype = double()),
                    upper = list_of(.ptype = double()),
                    size = double()) {
  lower <- as_list_of(lower)
  upper <- as_list_of(upper)
  vec_assert(lower, list_of(.ptype = double()))
  vec_assert(upper, list_of(.ptype = double()))
  vec_assert(size, double())
  if (any(size < 0 | size > 100, na.rm = TRUE))
    abort("'size' must be between [0, 100].")


  out <- vec_recycle_common(lower = lower, upper = upper)
  mapply(
    function(l,u) if (any(u<l, na.rm = TRUE)) abort("`upper` can't be lower than `lower`."),
    l = out[["lower"]], u = out[["upper"]]
  )
  out[["level"]] <- vctrs::vec_recycle(size, vec_size(out[[1]]))

  vctrs::new_rcrd(out, class = "hdr")
}

#' Compute highest density regions
#'
#' Used to extract a specified prediction interval at a particular confidence
#' level from a distribution.
#'
#' @param x Object to create hilo from.
#' @param ... Additional arguments used by methods.
#'
#' @export
hdr <- function(x, ...){
  UseMethod("hdr")
}

#' @export
hdr.default <- function(x, ...){
  abort(sprintf(
    "Objects of type `%s` are not supported by `hdr()`, you can create a custom `hdr` with `new_hdr()`",
    class(x)
  ))
}

#' Is the object a hdr
#'
#' @param x An object.
#'
#' @export
is_hdr <- function(x) {
  inherits(x, "hdr")
}

#' @export
format.hdr <- function(x, justify = "right", ...) {
  out <- mapply(function(l,u,s) {
    limit <- paste(
      format(l, justify = justify, ...),
      format(u, justify = justify, ...),
      sep = ", "
    )
    limit <- paste0("[", limit, "]", collapse = "")
    paste0(limit, s)
  }, l = field(x, "lower"), u = field(x, "upper"), s = field(x, "level"))
  as.vector(out, "character")
}

#' The Inverse Gaussian distribution
#'
#' The inverse Gaussian distribution (also known as the Wald distribution) is
#' commonly used to model positive-valued data, particularly in contexts
#' involving first passage times and reliability analysis.
#' 
#' @description
#' `r lifecycle::badge('stable')`
#'
#' @inheritParams actuar::dinvgauss
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_inverse_gaussian")`
#'
#'   In the following, let \eqn{X} be an Inverse Gaussian random variable with
#'   parameters `mean` = \eqn{\mu} and `shape` = \eqn{\lambda}.
#'
#'   **Support**: \eqn{(0, \infty)}
#'
#'   **Mean**: \eqn{\mu}
#'
#'   **Variance**: \eqn{\frac{\mu^3}{\lambda}}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \sqrt{\frac{\lambda}{2\pi x^3}}
#'     \exp\left(-\frac{\lambda(x - \mu)^2}{2\mu^2 x}\right)
#'   }{
#'     f(x) = sqrt(\lambda / (2 \pi x^3))
#'     exp(-\lambda (x - \mu)^2 / (2 \mu^2 x))
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = \Phi\left(\sqrt{\frac{\lambda}{x}}
#'     \left(\frac{x}{\mu} - 1\right)\right) +
#'     \exp\left(\frac{2\lambda}{\mu}\right)
#'     \Phi\left(-\sqrt{\frac{\lambda}{x}}
#'     \left(\frac{x}{\mu} + 1\right)\right)
#'   }{
#'     F(x) = \Phi(sqrt(\lambda/x) (x/\mu - 1)) +
#'     exp(2\lambda/\mu) \Phi(-sqrt(\lambda/x) (x/\mu + 1))
#'   }
#'
#'   where \eqn{\Phi} is the standard normal c.d.f.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \exp\left(\frac{\lambda}{\mu}
#'     \left(1 - \sqrt{1 - \frac{2\mu^2 t}{\lambda}}\right)\right)
#'   }{
#'     E(e^(tX)) = exp((\lambda/\mu) (1 - sqrt(1 - 2\mu^2 t/\lambda)))
#'   }
#'
#'   for \eqn{t < \frac{\lambda}{2\mu^2}}.
#'
#'   **Skewness**: \eqn{3\sqrt{\frac{\mu}{\lambda}}}
#'
#'   **Excess Kurtosis**: \eqn{\frac{15\mu}{\lambda}}
#'
#'   **Quantiles**: No closed-form expression, approximated numerically.
#'
#' @seealso [actuar::InverseGaussian]
#'
#' @examples
#' dist <- dist_inverse_gaussian(mean = c(1,1,1,3,3), shape = c(0.2, 1, 3, 0.2, 1))
#' dist
#'
#' @examplesIf requireNamespace("actuar", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_inverse_gaussian
#' @export
dist_inverse_gaussian <- function(mean, shape){
  mean <- vec_cast(mean, double())
  shape <- vec_cast(shape, double())
  if(any(mean[!is.na(mean)] <= 0)){
    abort("The mean parameter of a Inverse Gaussian distribution must be strictly positive.")
  }
  if(any(shape[!is.na(shape)] <= 0)){
    abort("The shape parameter of a Inverse Gaussian distribution must be strictly positive.")
  }
  new_dist(m = mean, s = shape, class = "dist_inverse_gaussian")
}

#' @export
format.dist_inverse_gaussian <- function(x, digits = 2, ...){
  sprintf(
    "IG(%s, %s)",
    format(x[["m"]], digits = digits, ...),
    format(x[["s"]], digits = digits, ...)
  )
}

#' @export
density.dist_inverse_gaussian <- function(x, at, ...){
  require_package("actuar")
  actuar::dinvgauss(at, x[["m"]], x[["s"]])
}

#' @export
log_density.dist_inverse_gaussian <- function(x, at, ...){
  require_package("actuar")
  actuar::dinvgauss(at, x[["m"]], x[["s"]], log = TRUE)
}

#' @export
quantile.dist_inverse_gaussian <- function(x, p, ...){
  require_package("actuar")
  actuar::qinvgauss(p, x[["m"]], x[["s"]])
}

#' @export
cdf.dist_inverse_gaussian <- function(x, q, ...){
  require_package("actuar")
  actuar::pinvgauss(q, x[["m"]], x[["s"]])
}

#' @export
generate.dist_inverse_gaussian <- function(x, times, ...){
  require_package("actuar")
  actuar::rinvgauss(times, x[["m"]], x[["s"]])
}

#' @export
mean.dist_inverse_gaussian <- function(x, ...){
  x[["m"]]
}

#' @export
covariance.dist_inverse_gaussian <- function(x, ...){
  x[["m"]]^3/x[["s"]]
}

#' The Hypergeometric distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' To understand the HyperGeometric distribution, consider a set of
#' \eqn{r} objects, of which \eqn{m} are of the type I and
#' \eqn{n} are of the type II. A sample with size \eqn{k} (\eqn{k<r})
#' with no replacement is randomly chosen. The number of observed
#' type I elements observed in this sample is set to be our random
#' variable \eqn{X}.
#'
#' @param m The number of type I elements available.
#' @param n The number of type II elements available.
#' @param k The size of the sample taken.
#'
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a HyperGeometric random variable with
#'   success probability `p` = \eqn{p = m/(m+n)}.
#'
#'   **Support**: \eqn{x \in { \{\max{(0, k-n)}, \dots, \min{(k,m)}}\}}
#'
#'   **Mean**: \eqn{\frac{km}{n+m} = kp}
#'
#'   **Variance**: \eqn{\frac{km(n)(n+m-k)}{(n+m)^2 (n+m-1)} =
#'   kp(1-p)(1 - \frac{k-1}{m+n-1})}
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     P(X = x) = \frac{{m \choose x}{n \choose k-x}}{{m+n \choose k}}
#'   }{
#'     P(X = x) = \frac{{m \choose x}{n \choose k-x}}{{m+n \choose k}}
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     P(X \le k) \approx \Phi\Big(\frac{x - kp}{\sqrt{kp(1-p)}}\Big)
#'  }
#'
#' @seealso [stats::Hypergeometric]
#'
#' @examples
#' dist <- dist_hypergeometric(m = rep(500, 3), n = c(50, 60, 70), k = c(100, 200, 300))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_hypergeometric
#' @export
dist_hypergeometric <- function(m, n, k){
  m <- vec_cast(m, integer())
  n <- vec_cast(n, integer())
  k <- vec_cast(k, integer())
  new_dist(m = m, n = n, k = k, class = "dist_hypergeometric")
}

#' @export
format.dist_hypergeometric <- function(x, digits = 2, ...){
  sprintf(
    "Hypergeometric(%s, %s, %s)",
    format(x[["m"]], digits = digits, ...),
    format(x[["n"]], digits = digits, ...),
    format(x[["k"]], digits = digits, ...)
  )
}

#' @export
density.dist_hypergeometric <- function(x, at, ...){
  stats::dhyper(at, x[["m"]], x[["n"]], x[["k"]])
}

#' @export
log_density.dist_hypergeometric <- function(x, at, ...){
  stats::dhyper(at, x[["m"]], x[["n"]], x[["k"]], log = TRUE)
}

#' @export
quantile.dist_hypergeometric <- function(x, p, ...){
  stats::qhyper(p, x[["m"]], x[["n"]], x[["k"]])
}

#' @export
cdf.dist_hypergeometric <- function(x, q, ...){
  stats::phyper(q, x[["m"]], x[["n"]], x[["k"]])
}

#' @export
generate.dist_hypergeometric <- function(x, times, ...){
  stats::rhyper(times, x[["m"]], x[["n"]], x[["k"]])
}

#' @export
mean.dist_hypergeometric <- function(x, ...){
  p <- x[["m"]]/(x[["m"]] + x[["n"]])
  x[["k"]] * p
}

#' @export
covariance.dist_hypergeometric <- function(x, ...){
  m <- x[["m"]]
  n <- x[["n"]]
  k <- x[["k"]]
  p <- m/(m + n)
  k * p * (1 - p) * ((m + n - k) / (m + n - 1))
}

#' @export
skewness.dist_hypergeometric <- function(x, ...) {
  N <- x[["n"]] + x[["m"]]
  K <- x[["m"]]
  n <- x[["k"]]

  a <- (N - 2 * K) * (N - 1)^0.5 * (N - 2 * n)
  b <- (n * K * (N - K) * (N - n))^0.5 * (N - 2)
  a / b
}

#' @export
kurtosis.dist_hypergeometric <- function(x, ...) {
  N <- x[["n"]] + x[["m"]]
  K <- x[["m"]]
  n <- x[["k"]]

  1 / (n * K * (N - K) * (N - n) * (N - 2) * (N - 3))
}

#' The Generalized Extreme Value Distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The GEV distribution is widely used in extreme value theory to model
#' the distribution of maxima (or minima) of samples. The parametric form
#' encompasses the Gumbel, Frechet, and reverse Weibull distributions.
#' 
#' @param location the location parameter \eqn{\mu} of the GEV distribution.
#' @param scale the scale parameter \eqn{\sigma} of the GEV distribution.
#' Must be strictly positive.
#' @param shape the shape parameter \eqn{\xi} of the GEV distribution.
#' Determines the tail behavior: \eqn{\xi = 0} gives Gumbel,
#' \eqn{\xi > 0} gives Frechet, \eqn{\xi < 0} gives reverse Weibull.
#' 
#' @details
#'
#' `r pkgdown_doc_link("dist_gev")`
#'
#' In the following, let \eqn{X} be a GEV random variable with parameters
#' `location` = \eqn{\mu}, `scale` = \eqn{\sigma}, and `shape` = \eqn{\xi}.
#'
#' **Support**:
#' * \eqn{x \in \mathbb{R}} (all real numbers) if \eqn{\xi = 0}
#' * \eqn{x \geq \mu - \sigma/\xi} if \eqn{\xi > 0}
#' * \eqn{x \leq \mu - \sigma/\xi} if \eqn{\xi < 0}
#'
#' **Mean**: 
#' \deqn{
#'   E(X) = \begin{cases}
#'     \mu + \sigma \gamma & \text{if } \xi = 0 \\
#'     \mu + \sigma \frac{\Gamma(1-\xi) - 1}{\xi} & \text{if } \xi < 1 \\
#'     \infty & \text{if } \xi \geq 1
#'   \end{cases}
#' }{
#'   E(X) = \mu + \sigma \gamma  (if \xi = 0)
#'         \mu + \sigma (\Gamma(1-\xi) - 1) / \xi  (if \xi < 1)
#'         Inf  (if \xi >= 1)
#' }
#' where \eqn{\gamma \approx 0.5772} is the Euler-Mascheroni constant and
#' \eqn{\Gamma(\cdot)} is the gamma function.
#'
#' **Median**:
#' \deqn{
#'   \text{Median}(X) = \begin{cases}
#'     \mu - \sigma \log(\log 2) & \text{if } \xi = 0 \\
#'     \mu + \sigma \frac{(\log 2)^{-\xi} - 1}{\xi} & \text{if } \xi \neq 0
#'   \end{cases}
#' }{
#'   Median(X) = \mu - \sigma log(log 2)  (if \xi = 0)
#'               \mu + \sigma (log(2)^(-\xi) - 1) / \xi  (if \xi != 0)
#' }
#'
#' **Variance**:
#' \deqn{
#'   \text{Var}(X) = \begin{cases}
#'     \frac{\pi^2 \sigma^2}{6} & \text{if } \xi = 0 \\
#'     \frac{\sigma^2}{\xi^2} [\Gamma(1-2\xi) - \Gamma(1-\xi)^2] & \text{if } \xi < 0.5 \\
#'     \infty & \text{if } \xi \geq 0.5
#'   \end{cases}
#' }{
#'   Var(X) = \pi^2 \sigma^2 / 6  (if \xi = 0)
#'            \sigma^2 [\Gamma(1-2\xi) - \Gamma(1-\xi)^2] / \xi^2  (if \xi < 0.5)
#'            Inf  (if \xi >= 0.5)
#' }
#'
#' **Probability density function (p.d.f)**:
#'
#' For \eqn{\xi = 0} (Gumbel):
#' \deqn{
#'   f(x) = \frac{1}{\sigma} \exp\left(-\frac{x-\mu}{\sigma}\right)
#'          \exp\left[-\exp\left(-\frac{x-\mu}{\sigma}\right)\right]
#' }{
#'   f(x) = (1/\sigma) exp(-(x-\mu)/\sigma) exp(-exp(-(x-\mu)/\sigma))
#' }
#'
#' For \eqn{\xi \neq 0}:
#' \deqn{
#'   f(x) = \frac{1}{\sigma} \left[1 + \xi\left(\frac{x-\mu}{\sigma}\right)\right]^{-1/\xi-1}
#'          \exp\left\{-\left[1 + \xi\left(\frac{x-\mu}{\sigma}\right)\right]^{-1/\xi}\right\}
#' }{
#'   f(x) = (1/\sigma) [1 + \xi(x-\mu)/\sigma]^(-1/\xi-1) exp(-[1 + \xi(x-\mu)/\sigma]^(-1/\xi))
#' }
#' where \eqn{1 + \xi(x-\mu)/\sigma > 0}.
#'
#' **Cumulative distribution function (c.d.f)**:
#'
#' For \eqn{\xi = 0} (Gumbel):
#' \deqn{
#'   F(x) = \exp\left[-\exp\left(-\frac{x-\mu}{\sigma}\right)\right]
#' }{
#'   F(x) = exp(-exp(-(x-\mu)/\sigma))
#' }
#'
#' For \eqn{\xi \neq 0}:
#' \deqn{
#'   F(x) = \exp\left\{-\left[1+\xi\left(\frac{x-\mu}{\sigma}\right)\right]^{-1/\xi}\right\}
#' }{
#'   F(x) = exp(-[1 + \xi(x-\mu)/\sigma]^(-1/\xi))
#' }
#' where \eqn{1 + \xi(x-\mu)/\sigma > 0}.
#'
#' **Quantile function**:
#'
#' For \eqn{\xi = 0} (Gumbel):
#' \deqn{
#'   Q(p) = \mu - \sigma \log(-\log p)
#' }{
#'   Q(p) = \mu - \sigma log(-log p)
#' }
#'
#' For \eqn{\xi \neq 0}:
#' \deqn{
#'   Q(p) = \mu + \frac{\sigma}{\xi}\left[(-\log p)^{-\xi} - 1\right]
#' }{
#'   Q(p) = \mu + \sigma [(-log p)^(-\xi) - 1] / \xi
#' }
#'
#' @references Jenkinson, A. F. (1955) The frequency distribution of the annual
#' maximum (or minimum) of meteorological elements. \emph{Quart. J. R. Met. Soc.},
#' \bold{81}, 158–171.
#' 
#' @seealso [evd::dgev()]
#' 
#' @examples
#' # Create GEV distributions with different shape parameters
#' 
#' # Gumbel distribution (shape = 0)
#' gumbel <- dist_gev(location = 0, scale = 1, shape = 0)
#' 
#' # Frechet distribution (shape > 0, heavy-tailed)
#' frechet <- dist_gev(location = 0, scale = 1, shape = 0.3)
#' 
#' # Reverse Weibull distribution (shape < 0, bounded above)
#' weibull <- dist_gev(location = 0, scale = 1, shape = -0.2)
#' 
#' dist <- c(gumbel, frechet, weibull)
#' dist
#' 
#' # Statistical properties
#' mean(dist)
#' median(dist)
#' variance(dist)
#' 
#' # Generate random samples
#' generate(dist, 10)
#' 
#' # Evaluate density
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#' 
#' # Evaluate cumulative distribution
#' cdf(dist, 4)
#' 
#' # Calculate quantiles
#' quantile(dist, 0.95)
#' 
#' @export
dist_gev <- function(location, scale, shape) {
  location <- vctrs::vec_cast(unname(location), double())
  shape <- vctrs::vec_cast(unname(shape), double())
  scale <- vctrs::vec_cast(unname(scale), double())
  if (any(scale <= 0)) {
    stop("The scale parameter of a GEV distribution must be strictly positive")
  }
  distributional::new_dist(location = location, scale = scale, shape = shape, class = "dist_gev")
}

#' @export
format.dist_gev <- function(x, digits = 2, ...) {
  sprintf(
    "GEV(%s, %s, %s)",
    format(x[["location"]], digits = digits, ...),
    format(x[["scale"]], digits = digits, ...),
    format(x[["shape"]], digits = digits, ...)
  )
}

#' @exportS3Method distributional::log_density
log_density.dist_gev <- function(x, at, ...) {
  z <- (at - x[["location"]]) / x[["scale"]]
  if (x[["shape"]] == 0) {
    pdf <- -z - exp(-z)
  } else {
    xx <- 1 + x[["shape"]] * z
    xx[xx <= 0] <- NA_real_
    pdf <- -xx^(-1 / x[["shape"]]) - (1 / x[["shape"]] + 1) * log(xx)
    pdf[is.na(pdf)] <- -Inf
  }
  pdf - log(x[["scale"]])
}

#' @exportS3Method stats::density
density.dist_gev <- function(x, at, ...) {
  exp(log_density.dist_gev(x, at, ...))
}

#' @exportS3Method distributional::cdf
cdf.dist_gev <- function(x, q, ...) {
  z <- (q - x[["location"]]) / x[["scale"]]
  if (x[["shape"]] == 0) {
    exp(-exp(-z))
  } else {
    exp(-pmax(1 + x[["shape"]] * z, 0)^(-1 / x[["shape"]]))
  }
}

#' @exportS3Method stats::quantile
quantile.dist_gev <- function(x, p, ...) {
  if (x[["shape"]] == 0) {
    x[["location"]] - x[["scale"]] * log(-log(p))
  } else {
    x[["location"]] + x[["scale"]] * ((-log(p))^(-x[["shape"]]) - 1) / x[["shape"]]
  }
}

#' @exportS3Method distributional::generate
generate.dist_gev <- function(x, times, ...) {
  z <- stats::rexp(times)
  if (x[["shape"]] == 0) {
    x[["location"]] - x[["scale"]] * log(z)
  } else {
    x[["location"]] + x[["scale"]] * (z^(-x[["shape"]]) - 1) / x[["shape"]]
  }
}

#' @export
mean.dist_gev <- function(x, ...) {
  if (x[["shape"]] == 0) {
    x[["location"]] + x[["scale"]] * 0.57721566490153286
  } else if (x[["shape"]] < 1) {
    x[["location"]] + x[["scale"]] * (gamma(1 - x[["shape"]]) - 1) / x[["shape"]]
  } else {
    Inf
  }
}

#' @exportS3Method stats::median
median.dist_gev <- function(x, ...) {
  if (x[["shape"]] == 0) {
    x[["location"]] - x[["scale"]] * log(log(2))
  } else {
    x[["location"]] + x[["scale"]] * (log(2)^(-x[["shape"]]) - 1) / x[["shape"]]
  }
}

#' @exportS3Method distributional::covariance
covariance.dist_gev <- function(x, ...) {
  if (x[["shape"]] == 0) {
    x[["scale"]]^2 * pi^2 / 6
  } else if (x[["shape"]] < 0.5) {
    g2 <- gamma(1 - 2 * x[["shape"]])
    g1 <- gamma(1 - x[["shape"]])
    x[["scale"]]^2 * (g2 - g1^2) / x[["shape"]]^2
  } else {
    Inf
  }
}

#' The Laplace distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Laplace distribution, also known as the double exponential distribution,
#' is a continuous probability distribution that is symmetric around its location
#' parameter.
#'
#' @param mu The location parameter (mean) of the Laplace distribution.
#' @param sigma The positive scale parameter of the Laplace distribution.
#'
#' @details
#' 
#' `r pkgdown_doc_link("dist_laplace")`
#'
#'   In the following, let \eqn{X} be a Laplace random variable with location
#'   parameter `mu` = \eqn{\mu} and scale parameter `sigma` = \eqn{\sigma}.
#'
#'   **Support**: \eqn{R}, the set of all real numbers
#'
#'   **Mean**: \eqn{\mu}
#'
#'   **Variance**: \eqn{2\sigma^2}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{1}{2\sigma} \exp\left(-\frac{|x - \mu|}{\sigma}\right)
#'   }{
#'     f(x) = 1 / (2 * sigma) * exp(-|x - mu| / sigma)
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = \begin{cases}
#'       \frac{1}{2} \exp\left(\frac{x - \mu}{\sigma}\right) & \text{if } x < \mu \\
#'       1 - \frac{1}{2} \exp\left(-\frac{x - \mu}{\sigma}\right) & \text{if } x \geq \mu
#'     \end{cases}
#'   }{
#'     F(x) = 1/2 * exp((x - mu) / sigma) if x < mu,
#'            1 - 1/2 * exp(-(x - mu) / sigma) if x >= mu
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \frac{\exp(\mu t)}{1 - \sigma^2 t^2} \text{ for } |t| < \frac{1}{\sigma}
#'   }{
#'     E(e^(tX)) = exp(mu * t) / (1 - sigma^2 * t^2) for |t| < 1/sigma
#'   }
#'
#' @seealso [extraDistr::Laplace]
#'
#' @examples
#' dist <- dist_laplace(mu = c(0, 2, -1), sigma = c(1, 2, 0.5))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 0)
#' density(dist, 0, log = TRUE)
#'
#' cdf(dist, 1)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_laplace
#' @export
dist_laplace <- function(mu, sigma) {
  mu <- vec_cast(mu, double())
  sigma <- vec_cast(sigma, double())
  if (any(sigma <= 0)) {
    abort("The scale parameter of a Laplace distribution must be positive.")
  }
  new_dist(mu = mu, sigma = sigma, class = "dist_laplace")
}

#' @export
format.dist_laplace <- function(x, digits = 2, ...) {
  sprintf(
    "Laplace(%s, %s)",
    format(x[["mu"]], digits = digits, ...),
    format(x[["sigma"]], digits = digits, ...)
  )
}

#' @export
density.dist_laplace <- function(x, at, ...) {
  mu <- x[["mu"]]
  sigma <- x[["sigma"]]
  (1 / (2 * sigma)) * exp(-abs(at - mu) / sigma)
}

#' @export
log_density.dist_laplace <- function(x, at, ...) {
  mu <- x[["mu"]]
  sigma <- x[["sigma"]]
  -log(2 * sigma) - abs(at - mu) / sigma
}

#' @export
quantile.dist_laplace <- function(x, p, ...) {
  mu <- x[["mu"]]
  sigma <- x[["sigma"]]
  mu - sigma * sign(p - 0.5) * log(1 - 2 * abs(p - 0.5))
}

#' @export
cdf.dist_laplace <- function(x, q, ...) {
  mu <- x[["mu"]]
  sigma <- x[["sigma"]]
  ifelse(q < mu,
    0.5 * exp((q - mu) / sigma),
    1 - 0.5 * exp(-(q - mu) / sigma)
  )
}

#' @export
generate.dist_laplace <- function(x, times, ...) {
  mu <- x[["mu"]]
  sigma <- x[["sigma"]]
  u <- stats::runif(times) - 0.5
  mu - sigma * sign(u) * log(1 - 2 * abs(u))
}

#' @export
mean.dist_laplace <- function(x, ...) {
  x[["mu"]]
}

#' @export
variance.dist_laplace <- function(x, ...) {
  2 * x[["sigma"]]^2
}

#' @export
skewness.dist_laplace <- function(x, ...) {
  0
}

#' @export
kurtosis.dist_laplace <- function(x, ...) {
  3
}

#' @export
has_symmetry.dist_laplace <- function(x, ...) {
  TRUE
}

#' The Geometric Distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Geometric distribution can be thought of as a generalization
#' of the [dist_bernoulli()] distribution where we ask: "if I keep flipping a
#' coin with probability `p` of heads, what is the probability I need
#' \eqn{k} flips before I get my first heads?" The Geometric
#' distribution is a special case of Negative Binomial distribution.
#'
#' @inheritParams stats::dgeom
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a Geometric random variable with
#'   success probability `p` = \eqn{p}. Note that there are multiple
#'   parameterizations of the Geometric distribution.
#'
#'   **Support**: 0 < p < 1, \eqn{x = 0, 1, \dots}
#'
#'   **Mean**: \eqn{\frac{1-p}{p}}
#'
#'   **Variance**: \eqn{\frac{1-p}{p^2}}
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     P(X = x) = p(1-p)^x,
#'    }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     P(X \le x) = 1 - (1-p)^{x+1}
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \frac{pe^t}{1 - (1-p)e^t}
#'   }{
#'     E(e^{tX}) = \frac{pe^t}{1 - (1-p)e^t}
#'   }
#'
#' @seealso [stats::Geometric]
#'
#' @examples
#' dist <- dist_geometric(prob = c(0.2, 0.5, 0.8))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#' @name dist_geometric
#' @export
dist_geometric <- function(prob){
  prob <- vec_cast(prob, double())
  if(any((prob < 0) | (prob > 1))){
    abort("The prob parameter of an Geometric distribution must be between 0 and 1.")
  }
  new_dist(p = prob, class = "dist_geometric")
}

#' @export
format.dist_geometric <- function(x, digits = 2, ...){
  sprintf(
    "Geometric(%s)",
    format(x[["p"]], digits = digits, ...)
  )
}

#' @export
density.dist_geometric <- function(x, at, ...){
  stats::dgeom(at, x[["p"]])
}

#' @export
log_density.dist_geometric <- function(x, at, ...){
  stats::dgeom(at, x[["p"]], log = TRUE)
}

#' @export
quantile.dist_geometric <- function(x, p, ...){
  stats::qgeom(p, x[["p"]])
}

#' @export
cdf.dist_geometric <- function(x, q, ...){
  stats::pgeom(q, x[["p"]])
}

#' @export
generate.dist_geometric <- function(x, times, ...){
  stats::rgeom(times, x[["p"]])
}

#' @export
mean.dist_geometric <- function(x, ...){
  1/x[["p"]] - 1
}

#' @export
covariance.dist_geometric <- function(x, ...){
  (1 - x[["p"]])/x[["p"]]^2
}

#' @export
skewness.dist_geometric <- function(x, ...) (2 - x[["p"]]) / sqrt(1 - x[["p"]])

#' @export
kurtosis.dist_geometric <- function(x, ...) 6 + (x[["p"]]^2 / (1 - x[["p"]]))

#' The multivariate t-distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The multivariate t-distribution is a generalization of the univariate
#' Student's t-distribution to multiple dimensions. It is commonly used for
#' modeling heavy-tailed multivariate data and in robust statistics.
#'
#' @param df A numeric vector of degrees of freedom (must be positive).
#' @param mu A list of numeric vectors for the distribution location parameter.
#' @param sigma A list of matrices for the distribution scale matrix.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_multivariate_t")`
#'
#'   In the following, let \eqn{\mathbf{X}} be a multivariate t random vector
#'   with degrees of freedom `df` = \eqn{\nu}, location parameter
#'   `mu` = \eqn{\boldsymbol{\mu}}, and scale matrix
#'   `sigma` = \eqn{\boldsymbol{\Sigma}}.
#'
#'   **Support**: \eqn{\mathbf{x} \in \mathbb{R}^k}, where \eqn{k} is the
#'   dimension of the distribution
#'
#'   **Mean**: \eqn{\boldsymbol{\mu}} for \eqn{\nu > 1}, undefined otherwise
#'
#'   **Covariance matrix**: 
#'   
#'   \deqn{
#'     \text{Cov}(\mathbf{X}) = \frac{\nu}{\nu - 2} \boldsymbol{\Sigma}
#'   }{
#'     Cov(X) = \nu / (\nu - 2) \Sigma
#'   }
#'   
#'   for \eqn{\nu > 2}, undefined otherwise
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(\mathbf{x}) = \frac{\Gamma\left(\frac{\nu + k}{2}\right)}
#'     {\Gamma\left(\frac{\nu}{2}\right) \nu^{k/2} \pi^{k/2}
#'     |\boldsymbol{\Sigma}|^{1/2}}
#'     \left[1 + \frac{1}{\nu}(\mathbf{x} - \boldsymbol{\mu})^T
#'     \boldsymbol{\Sigma}^{-1} (\mathbf{x} - \boldsymbol{\mu})\right]^{-\frac{\nu + k}{2}}
#'   }{
#'     f(x) = \Gamma((\nu + k)/2) / (\Gamma(\nu/2) \nu^(k/2) \pi^(k/2) |\Sigma|^(1/2))
#'     [1 + 1/\nu (x - \mu)^T \Sigma^(-1) (x - \mu)]^(-(\nu + k)/2)
#'   }
#'
#'   where \eqn{k} is the dimension of the distribution and \eqn{\Gamma(\cdot)} is
#'   the gamma function.
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(\mathbf{t}) = \int_{-\infty}^{t_1} \cdots \int_{-\infty}^{t_k} f(\mathbf{x}) \, d\mathbf{x}
#'   }{
#'     F(t) = integral_{-\infty}^{t_1} ... integral_{-\infty}^{t_k} f(x) dx
#'   }
#'
#'   This integral does not have a closed form solution and is approximated numerically.
#'
#'   **Quantile function**:
#'
#'   The equicoordinate quantile function finds \eqn{q} such that:
#'
#'   \deqn{
#'     P(X_1 \leq q, \ldots, X_k \leq q) = p
#'   }{
#'     P(X_1 <= q, ..., X_k <= q) = p
#'   }
#'
#'   This does not have a closed form solution and is approximated numerically.
#'
#'   The marginal quantile function for each dimension \eqn{i} is:
#'
#'   \deqn{
#'     Q_i(p) = \mu_i + \sqrt{\Sigma_{ii}} \cdot t_{\nu}^{-1}(p)
#'   }{
#'     Q_i(p) = \mu_i + sqrt(\Sigma_{ii}) * t_\nu^(-1)(p)
#'   }
#'
#'   where \eqn{t_{\nu}^{-1}(p)} is the quantile function of the univariate
#'   Student's t-distribution with \eqn{\nu} degrees of freedom, and
#'   \eqn{\Sigma_{ii}} is the \eqn{i}-th diagonal element of `sigma`.
#'
#' @seealso [mvtnorm::dmvt], [mvtnorm::pmvt], [mvtnorm::qmvt], [mvtnorm::rmvt]
#'
#' @examples
#' dist <- dist_multivariate_t(
#'   df = 5,
#'   mu = list(c(1, 2)),
#'   sigma = list(matrix(c(4, 2, 2, 3), ncol = 2))
#' )
#' dimnames(dist) <- c("x", "y")
#' dist
#'
#' @examplesIf requireNamespace("mvtnorm", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, cbind(2, 1))
#' density(dist, cbind(2, 1), log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#' quantile(dist, 0.7, kind = "marginal")
#'
#' @name dist_multivariate_t
#' @export
dist_multivariate_t <- function(df = 1, mu = 0, sigma = diag(1)) {
  df <- vec_cast(df, double())
  if (any(df <= 0, na.rm = TRUE)) {
    abort("Degrees of freedom must be positive.")
  }
  
  new_dist(
    df = df,
    mu = mu,
    sigma = sigma,
    dimnames = colnames(sigma[[1]]),
    class = "dist_mvt"
  )
}

#' @export
format.dist_mvt <- function(x, digits = 2, ...) {
  sprintf(
    "MVT[%i](%s)",
    length(x[["mu"]]),
    format(x[["df"]], digits = digits, ...)
  )
}

#' @export
density.dist_mvt <- function(x, at, ..., na.rm = FALSE) {
  require_package("mvtnorm")
  if (is.list(at)) return(vapply(at, density, numeric(1L), x = x, ...))
  mvtnorm::dmvt(at, delta = x[["mu"]], sigma = x[["sigma"]], df = x[["df"]], log = FALSE, ...)
}

#' @export
log_density.dist_mvt <- function(x, at, ..., na.rm = FALSE) {
  require_package("mvtnorm")
  if (is.list(at)) return(vapply(at, log_density, numeric(1L), x = x, ...))
  mvtnorm::dmvt(at, delta = x[["mu"]], sigma = x[["sigma"]], df = x[["df"]], log = TRUE, ...)
}

#' @export
quantile.dist_mvt <- function(x, p, kind = c("equicoordinate", "marginal"),
                               ..., na.rm = FALSE) {
  kind <- match.arg(kind)
  q <- if (kind == "marginal") {
    scale <- sqrt(diag(x[["sigma"]]))
    stats::qt(p, df = rep(x[["df"]], each = length(p))) * 
      rep(scale, each = length(p)) + 
      rep(x[["mu"]], each = length(p))
  } else {
    require_package("mvtnorm")
    vapply(p, function(p, ...) {
      if (p == 0) return(-Inf) else if (p == 1) return(Inf)
      mvtnorm::qmvt(p, ...)$quantile
    }, numeric(1L), delta = x[["mu"]], sigma = x[["sigma"]], df = x[["df"]], ...)
  }
  
  matrix(q, nrow = length(p), ncol = dim(x))
}

#' @export
cdf.dist_mvt <- function(x, q, ..., na.rm = FALSE) {
  if (is.list(q)) return(vapply(q, cdf, numeric(1L), x = x, ...))
  require_package("mvtnorm")
  mvtnorm::pmvt(upper = as.numeric(q), delta = x[["mu"]], sigma = x[["sigma"]], df = x[["df"]], ...)[1]
}

#' @export
generate.dist_mvt <- function(x, times, ..., na.rm = FALSE) {
  require_package("mvtnorm")
  mvtnorm::rmvt(times, delta = x[["mu"]], sigma = x[["sigma"]], df = x[["df"]], ...)
}

#' @export
mean.dist_mvt <- function(x, ...) {
  if (x[["df"]] <= 1) {
    warning("Mean is undefined for df <= 1")
    return(matrix(NA_real_, nrow = 1, ncol = length(x[["mu"]])))
  }
  matrix(x[["mu"]], nrow = 1)
}

#' @export
covariance.dist_mvt <- function(x, ...) {
  if (x[["df"]] <= 2) {
    warning("Covariance is undefined for df <= 2")
    return(list(matrix(NA_real_, nrow = length(x[["mu"]]), ncol = length(x[["mu"]]))))
  }
  list(x[["df"]] / (x[["df"]] - 2) * x[["sigma"]])
}

#' @export
dim.dist_mvt <- function(x) {
  length(x[["mu"]])
}

#' The Studentized Range distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Tukey's studentized range distribution, used for Tukey's
#' honestly significant differences test in ANOVA.
#'
#' @inheritParams stats::qtukey
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_studentized_range")`
#'
#'   In the following, let \eqn{Q} be a Studentized Range random variable with
#'   parameters `nmeans` = \eqn{k} (number of groups), `df` = \eqn{\nu} (degrees
#'   of freedom), and `nranges` = \eqn{n} (number of ranges).
#'
#'   **Support**: \eqn{R^+}, the set of positive real numbers.
#'
#'   **Mean**: Approximated numerically.
#'
#'   **Variance**: Approximated numerically.
#'
#'   **Probability density function (p.d.f)**: The density does not have a
#'   closed-form expression and is computed numerically.
#'
#'   **Cumulative distribution function (c.d.f)**: The c.d.f does not have a
#'   simple closed-form expression. For \eqn{n = 1} (single range), it involves
#'   integration over the joint distribution of the sample range and an
#'   independent chi-square variable. The general form is computed numerically
#'   using algorithms described in the references for [stats::ptukey()].
#'
#'   **Moment generating function (m.g.f)**: Does not exist in closed form.
#'
#' @seealso [stats::Tukey]
#'
#' @examples
#' dist <- dist_studentized_range(nmeans = c(6, 2), df = c(5, 4), nranges = c(1, 1))
#'
#' dist
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_studentized_range
#' @export
dist_studentized_range <- function(nmeans, df, nranges){
  nmeans <- vec_cast(nmeans, double())
  df <- vec_cast(df, double())
  new_dist(nm = nmeans, df = df, nr = nranges, class = "dist_studentized_range")
}

#' @export
format.dist_studentized_range <- function(x, digits = 2, ...){
  sprintf(
    "StudentizedRange(%s, %s, %s)",
    format(x[["nm"]], digits = digits, ...),
    format(x[["df"]], digits = digits, ...),
    format(x[["nr"]], digits = digits, ...)
  )
}

#' @export
quantile.dist_studentized_range <- function(x, p, ...){
  stats::qtukey(p, x[["nm"]], x[["df"]], x[["nr"]])
}

#' @export
cdf.dist_studentized_range <- function(x, q, ...){
  stats::ptukey(q, x[["nm"]], x[["df"]], x[["nr"]])
}

#' The Gamma distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Several important distributions are special cases of the Gamma
#' distribution. When the shape parameter is `1`, the Gamma is an
#' exponential distribution with parameter \eqn{1/\beta}. When the
#' \eqn{shape = n/2} and \eqn{rate = 1/2}, the Gamma is a equivalent to
#' a chi squared distribution with n degrees of freedom. Moreover, if
#' we have \eqn{X_1} is \eqn{Gamma(\alpha_1, \beta)} and
#' \eqn{X_2} is \eqn{Gamma(\alpha_2, \beta)}, a function of these two variables
#' of the form \eqn{\frac{X_1}{X_1 + X_2}} \eqn{Beta(\alpha_1, \alpha_2)}.
#' This last property frequently appears in another distributions, and it
#' has extensively been used in multivariate methods. More about the Gamma
#' distribution will be added soon.
#'
#' @inheritParams stats::dgamma
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a Gamma random variable
#'   with parameters
#'   `shape` = \eqn{\alpha} and
#'   `rate` = \eqn{\beta}.
#'
#'   **Support**: \eqn{x \in (0, \infty)}
#'
#'   **Mean**: \eqn{\frac{\alpha}{\beta}}
#'
#'   **Variance**: \eqn{\frac{\alpha}{\beta^2}}
#'
#'   **Probability density function (p.m.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{\beta^{\alpha}}{\Gamma(\alpha)} x^{\alpha - 1} e^{-\beta x}
#'   }{
#'     f(x) = \frac{\beta^{\alpha}}{\Gamma(\alpha)} x^{\alpha - 1} e^{-\beta x}
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{\Gamma(\alpha, \beta x)}{\Gamma{\alpha}}
#'   }{
#'     f(x) = \frac{\Gamma(\alpha, \beta x)}{\Gamma{\alpha}}
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \Big(\frac{\beta}{ \beta - t}\Big)^{\alpha}, \thinspace t < \beta
#'   }{
#'     E(e^(tX)) = \Big(\frac{\beta}{ \beta - t}\Big)^{\alpha}, \thinspace t < \beta
#'   }
#'
#' @seealso [stats::GammaDist]
#'
#' @examples
#' dist <- dist_gamma(shape = c(1,2,3,5,9,7.5,0.5), rate = c(0.5,0.5,0.5,1,2,1,1))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_gamma
#' @export
dist_gamma <- function(shape, rate, scale = 1/rate){
  shape <- vec_cast(shape, double())
  rate <- vec_cast(rate, double())
  if(any(shape[!is.na(shape)] < 0)){
    abort("The shape parameter of a Gamma distribution must be non-negative.")
  }
  if(any(rate[!is.na(rate)] <= 0)){
    abort("The rate parameter of a Gamma distribution must be strictly positive.")
  }
  new_dist(shape = shape, rate = 1/scale, class = "dist_gamma")
}

#' @export
format.dist_gamma <- function(x, digits = 2, ...){
  sprintf(
    if (is_utf8_output()) "\u0393(%s, %s)" else "Gamma(%s, %s)",
    format(x[["shape"]], digits = digits, ...),
    format(x[["rate"]], digits = digits, ...)
  )
}

#' @export
density.dist_gamma <- function(x, at, ...){
  stats::dgamma(at, x[["shape"]], x[["rate"]])
}

#' @export
log_density.dist_gamma <- function(x, at, ...){
  stats::dgamma(at, x[["shape"]], x[["rate"]], log = TRUE)
}

#' @export
quantile.dist_gamma <- function(x, p, ...){
  stats::qgamma(p, x[["shape"]], x[["rate"]])
}

#' @export
cdf.dist_gamma <- function(x, q, ...){
  stats::pgamma(q, x[["shape"]], x[["rate"]])
}

#' @export
generate.dist_gamma <- function(x, times, ...){
  stats::rgamma(times, x[["shape"]], x[["rate"]])
}

#' @export
mean.dist_gamma <- function(x, ...){
  x[["shape"]] / x[["rate"]]
}

#' @export
covariance.dist_gamma <- function(x, ...){
  x[["shape"]] / x[["rate"]]^2
}

#' @export
skewness.dist_gamma <- function(x, ...) 2 / sqrt(x[["shape"]])

#' @export
kurtosis.dist_gamma <- function(x, ...) 6 / x[["shape"]]

#' The log-normal distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The log-normal distribution is a commonly used transformation of the Normal
#' distribution. If \eqn{X} follows a log-normal distribution, then \eqn{\ln{X}}
#' would be characteristed by a Normal distribution.
#'
#' @param mu The mean (location parameter) of the distribution, which is the
#'   mean of the associated Normal distribution. Can be any real number.
#' @param sigma The standard deviation (scale parameter) of the distribution.
#'   Can be any positive number.
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{Y} be a Normal random variable with mean
#'   `mu` = \eqn{\mu} and standard deviation `sigma` = \eqn{\sigma}. The
#'   log-normal distribution \eqn{X = exp(Y)} is characterised by:
#'
#'   **Support**: \eqn{R+}, the set of all real numbers greater than or equal to 0.
#'
#'   **Mean**: \eqn{e^(\mu + \sigma^2/2}
#'
#'   **Variance**: \eqn{(e^(\sigma^2)-1) e^(2\mu + \sigma^2}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{1}{x\sqrt{2 \pi \sigma^2}} e^{-(\ln{x} - \mu)^2 / 2 \sigma^2}
#'   }{
#'     f(x) = 1 / (x * sqrt(2 \pi \sigma^2)) exp(-(log(x) - \mu)^2 / (2 \sigma^2))
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   The cumulative distribution function has the form
#'
#'   \deqn{
#'     F(x) = \Phi((\ln{x} - \mu)/\sigma)
#'   }{
#'     F(x) = Phi((log(x) - \mu)/\sigma)
#'   }
#'
#'   Where \eqn{Phi}{Phi} is the CDF of a standard Normal distribution, N(0,1).
#'
#' @seealso [stats::Lognormal]
#'
#' @examples
#' dist <- dist_lognormal(mu = 1:5, sigma = 0.1)
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' # A log-normal distribution X is exp(Y), where Y is a Normal distribution of
#' # the same parameters. So log(X) will produce the Normal distribution Y.
#' log(dist)
#' @export
dist_lognormal <- function(mu = 0, sigma = 1){
  mu <- vec_cast(mu, double())
  sigma <- vec_cast(sigma, double())
  if(any(sigma[!is.na(sigma)] < 0)){
    abort("Standard deviation of a log-normal distribution must be non-negative")
  }
  new_dist(mu = mu, sigma = sigma, class = "dist_lognormal")
}

#' @export
format.dist_lognormal <- function(x, digits = 2, ...){
  sprintf(
    "lN(%s, %s)",
    format(x[["mu"]], digits = digits, ...),
    format(x[["sigma"]]^2, digits = digits, ...)
  )
}

#' @export
density.dist_lognormal <- function(x, at, ...){
  stats::dlnorm(at, x[["mu"]], x[["sigma"]])
}

#' @export
log_density.dist_lognormal <- function(x, at, ...){
  stats::dlnorm(at, x[["mu"]], x[["sigma"]], log = TRUE)
}

#' @export
quantile.dist_lognormal <- function(x, p, ...){
  stats::qlnorm(p, x[["mu"]], x[["sigma"]])
}
#' @export
log_quantile.dist_lognormal <- function(x, p, ...){
  stats::qlnorm(p, x[["mu"]], x[["sigma"]], log.p = TRUE)
}

#' @export
cdf.dist_lognormal <- function(x, q, ...){
  stats::plnorm(q, x[["mu"]], x[["sigma"]])
}
#' @export
log_cdf.dist_lognormal <- function(x, q, ...){
  stats::plnorm(q, x[["mu"]], x[["sigma"]], log.p = TRUE)
}

#' @export
generate.dist_lognormal <- function(x, times, ...){
  stats::rlnorm(times, x[["mu"]], x[["sigma"]])
}

#' @export
mean.dist_lognormal <- function(x, ...){
  exp(x[["mu"]] + x[["sigma"]]^2/2)
}

#' @export
covariance.dist_lognormal <- function(x, ...){
  s2 <- x[["sigma"]]^2
  (exp(s2)-1)*exp(2*x[["mu"]] + s2)
}

#' @export
skewness.dist_lognormal <- function(x, ...) {
  es2 <- exp(x[["sigma"]]^2)
  (es2+2)*sqrt(es2-1)
}

#' @export
kurtosis.dist_lognormal <- function(x, ...) {
  s2 <- x[["sigma"]]^2
  exp(4*s2) + 2*exp(3*s2) + 3*exp(2*s2) - 6
}

# make a normal distribution from a lognormal distribution using the
# specified base
normal_dist_with_base <- function(x, base = exp(1)) {
  vec_data(dist_normal(x[["mu"]], x[["sigma"]]) / log(base))[[1]]
}

#' @method Math dist_lognormal
#' @export
Math.dist_lognormal <- function(x, ...) {
  switch(.Generic,
    # Shortcuts to get Normal distribution from log-normal.
    log = normal_dist_with_base(x, ...),
    log2 = normal_dist_with_base(x, 2),
    log10 = normal_dist_with_base(x, 10),

    NextMethod()
  )
}

#' The Inverse Gamma distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Inverse Gamma distribution is commonly used as a prior distribution 
#' in Bayesian statistics, particularly for variance parameters.
#'
#' @inheritParams actuar::dinvgamma
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_inverse_gamma")`
#'
#'   In the following, let \eqn{X} be an Inverse Gamma random variable with
#'   shape parameter `shape` = \eqn{\alpha} and rate parameter 
#'   `rate` = \eqn{\beta} (equivalently, scale = \eqn{1/\beta}).
#'
#'   **Support**: \eqn{x \in (0, \infty)}
#'
#'   **Mean**: \eqn{\frac{\beta}{\alpha - 1}} for \eqn{\alpha > 1}, 
#'   otherwise undefined
#'
#'   **Variance**: \eqn{\frac{\beta^2}{(\alpha - 1)^2 (\alpha - 2)}} 
#'   for \eqn{\alpha > 2}, otherwise undefined
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{\beta^\alpha}{\Gamma(\alpha)} x^{-\alpha - 1} 
#'     e^{-\beta/x}
#'   }{
#'     f(x) = \beta^\alpha / \Gamma(\alpha) x^(-\alpha - 1) e^(-\beta/x)
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = \frac{\Gamma(\alpha, \beta/x)}{\Gamma(\alpha)} = 
#'     Q(\alpha, \beta/x)
#'   }{
#'     F(x) = \Gamma(\alpha, \beta/x) / \Gamma(\alpha) = Q(\alpha, \beta/x)
#'   }
#'
#'   where \eqn{\Gamma(\alpha, z)} is the upper incomplete gamma function and 
#'   \eqn{Q} is the regularized incomplete gamma function.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     M_X(t) = \frac{2 (-\beta t)^{\alpha/2}}{\Gamma(\alpha)} 
#'     K_\alpha\left(\sqrt{-4\beta t}\right)
#'   }{
#'     M_X(t) = 2 (-\beta t)^(\alpha/2) / \Gamma(\alpha) K_\alpha(sqrt(-4\beta t))
#'   }
#'
#'   for \eqn{t < 0}, where \eqn{K_\alpha} is the modified Bessel function 
#'   of the second kind. The MGF does not exist for \eqn{t \ge 0}.
#'
#' @seealso [actuar::InverseGamma]
#'
#' @examples
#' dist <- dist_inverse_gamma(shape = c(1,2,3,3), rate = c(1,1,1,2))
#' dist
#'
#' @examplesIf requireNamespace("actuar", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_inverse_gamma
#' @export
dist_inverse_gamma <- function(shape, rate = 1/scale, scale){
  shape <- vec_cast(shape, double())
  rate <- vec_cast(rate, double())
  if(any(shape <= 0)){
    abort("The shape parameter of a Inverse Gamma distribution must be strictly positive.")
  }
  if(any(rate <= 0)){
    abort("The rate/scale parameter of a Inverse Gamma distribution must be strictly positive.")
  }
  new_dist(s = shape, r = rate, class = "dist_inverse_gamma")
}

#' @export
format.dist_inverse_gamma <- function(x, digits = 2, ...){
  sprintf(
    "InvGamma(%s, %s)",
    format(x[["s"]], digits = digits, ...),
    format(1/x[["r"]], digits = digits, ...)
  )
}

#' @export
density.dist_inverse_gamma <- function(x, at, ...){
  require_package("actuar")
  actuar::dinvgamma(at, x[["s"]], x[["r"]])
}

#' @export
log_density.dist_inverse_gamma <- function(x, at, ...){
  require_package("actuar")
  actuar::dinvgamma(at, x[["s"]], x[["r"]], log = TRUE)
}

#' @export
quantile.dist_inverse_gamma <- function(x, p, ...){
  require_package("actuar")
  actuar::qinvgamma(p, x[["s"]], x[["r"]])
}

#' @export
cdf.dist_inverse_gamma <- function(x, q, ...){
  require_package("actuar")
  actuar::pinvgamma(q, x[["s"]], x[["r"]])
}

#' @export
generate.dist_inverse_gamma <- function(x, times, ...){
  require_package("actuar")
  actuar::rinvgamma(times, x[["s"]], x[["r"]])
}

#' @export
mean.dist_inverse_gamma <- function(x, ...){
  if(x[["s"]] <= 1) return(NA_real_)
  1/(x[["r"]]*(x[["s"]]-1))
}

#' @export
covariance.dist_inverse_gamma <- function(x, ...){
  if(x[["s"]] <= 2) return(NA_real_)
  1/(x[["r"]]^2*(x[["s"]]-1)^2*(x[["s"]]-2))
}

#' Ribbon plots for hilo intervals
#'
#' @description
#' `r lifecycle::badge('deprecated')`
#'
#' This function is deprecated in favour of the ggdist package and will removed
#' in a future release of this package. Consider using [ggdist::stat_lineribbon()] or
#' [ggdist::geom_lineribbon()] as an appropriate alternative.
#'
#' `geom_hilo_ribbon()` displays the interval defined by a hilo object. The
#' luminance of the shaded area indicates its confidence level. The shade colour
#' can be controlled by the `fill` aesthetic, however the luminance will be
#' overwritten to represent the confidence level.
#'
#' @inheritParams ggplot2::geom_ribbon
#'
#' @seealso
#'   [`geom_hilo_linerange()`] for discrete hilo intervals (vertical lines)
#'
#' @examples
#' dist <- dist_normal(1:3, 1:3)
#' library(ggplot2)
#' ggplot(
#'   data.frame(x = rep(1:3, 2), interval = c(hilo(dist, 80), hilo(dist, 95)))
#' ) +
#'   geom_hilo_ribbon(aes(x = x, hilo = interval))
#'
#' @export
geom_hilo_ribbon <- function(mapping = NULL, data = NULL, stat = "identity",
                             position = "identity", na.rm = FALSE,
                             show.legend = NA, inherit.aes = TRUE, ...) {
  if(!is.null(mapping$hilo)){
    mapping$ymin <- expr(ggplot2::after_stat(vctrs::vec_data(hilo)$lower))
    mapping$ymax <- expr(ggplot2::after_stat(vctrs::vec_data(hilo)$upper))
    mapping$level <- expr(ggplot2::after_stat(vctrs::vec_data(hilo)$level))
  }
  ggplot2::layer(
    geom = GeomHiloRibbon, mapping = mapping, data = data, stat = stat,
    position = position, show.legend = show.legend, inherit.aes = inherit.aes,
    params = list(na.rm = na.rm, ...)
  )
}

GeomHiloRibbon <- ggplot2::ggproto(
  "GeomHiloRibbon", ggplot2::GeomRibbon,
  required_aes = c(ggplot2::GeomRibbon$required_aes),
  optional_aes = c("hilo", "level"),
  default_aes = ggplot2::aes(
    colour = "grey20", fill = "grey60", size = .5,
    linetype = 1, weight = 1, alpha = 1, level = NA
  ),
  draw_key = function(data, params, size) {
    lwd <- min(data$size, min(size) / 4)
    # Calculate and set colour
    fillcol <- darken_fill(data$fill, 80)

    grid::rectGrob(
      width = grid::unit(1, "npc") - grid::unit(lwd, "mm"),
      height = grid::unit(1, "npc") - grid::unit(lwd, "mm"),
      gp = grid::gpar(
        col = fillcol,
        fill = scales::alpha(fillcol, data$alpha),
        lty = data$linetype,
        lwd = lwd * ggplot2::.pt,
        linejoin = "mitre"
      )
    )
  },

  draw_group = function(data, panel_scales, coord) {
    # Calculate colour
    data$fill <- darken_fill(data$fill, data$level)
    data$colour <- NA
    # Compute alpha transparency
    data$alpha <- grDevices::col2rgb(data$fill, alpha = TRUE)[4, ] / 255 * data$alpha

    # Create grobs
    grobs <- lapply(
      split(data, -data$level),
      ggplot2::GeomRibbon$draw_group,
      panel_scales, coord
    )
    ggplot2:::ggname("geom_hilo_ribbon", do.call(grid::grobTree, grobs))
  }
)

#' Line ranges for hilo intervals
#'
#' @description
#' `r lifecycle::badge('deprecated')`
#'
#' This function is deprecated in favour of the ggdist package and will removed
#' in a future release of this package. Consider using [ggdist::stat_slabinterval()] or
#' [ggdist::geom_slabinterval()] as an appropriate alternative.
#'
#' `geom_hilo_linerange()` displays the interval defined by a hilo object. The
#' luminance of the shaded area indicates its confidence level. The shade colour
#' can be controlled by the `fill` aesthetic, however the luminance will be
#' overwritten to represent the confidence level.
#'
#' @inheritParams ggplot2::geom_linerange
#'
#' @seealso
#'   [`geom_hilo_ribbon()`] for continuous hilo intervals (ribbons)
#'
#' @examples
#' dist <- dist_normal(1:3, 1:3)
#' library(ggplot2)
#' ggplot(
#'   data.frame(x = rep(1:3, 2), interval = c(hilo(dist, 80), hilo(dist, 95)))
#' ) +
#'   geom_hilo_linerange(aes(x = x, hilo = interval))
#'
#' @export
geom_hilo_linerange <- function(mapping = NULL, data = NULL, stat = "identity",
                             position = "identity", na.rm = FALSE,
                             show.legend = NA, inherit.aes = TRUE, ...) {
  if(!is.null(mapping$hilo)){
    mapping$ymin <- expr(ggplot2::after_stat(vctrs::vec_data(hilo)$lower))
    mapping$ymax <- expr(ggplot2::after_stat(vctrs::vec_data(hilo)$upper))
    mapping$level <- expr(ggplot2::after_stat(vctrs::vec_data(hilo)$level))
  }
  ggplot2::layer(
    geom = GeomHiloLinerange, mapping = mapping, data = data, stat = stat,
    position = position, show.legend = show.legend, inherit.aes = inherit.aes,
    params = list(na.rm = na.rm, ...)
  )
}

GeomHiloLinerange <- ggplot2::ggproto(
  "GeomHiloLinerange", ggplot2::GeomLinerange,
  required_aes = c(ggplot2::GeomLinerange$required_aes),
  optional_aes = c("hilo", "level"),
  default_aes = ggplot2::aes(
    colour = "grey20", fill = "grey60", size = 2,
    linetype = 1, weight = 1, alpha = 1, level = NA
  ),
  draw_key = function(data, params, size) {
    lwd <- min(data$size, min(size) / 4)
    # Calculate and set colour
    fillcol <- darken_fill(data$colour, 80)

    grid::rectGrob(
      width = grid::unit(1, "npc") - grid::unit(lwd, "mm"),
      height = grid::unit(1, "npc") - grid::unit(lwd, "mm"),
      gp = grid::gpar(
        col = fillcol,
        fill = scales::alpha(fillcol, data$alpha),
        lty = data$linetype,
        lwd = lwd * ggplot2::.pt,
        linejoin = "mitre"
      )
    )
  },

  draw_panel = function(data, panel_scales, coord) {
    # Calculate colour
    data$colour <- darken_fill(data$colour, data$level)
    data$fill <- NA
    # Compute alpha transparency
    data$alpha <- grDevices::col2rgb(data$colour, alpha = TRUE)[4, ] / 255 * data$alpha

    # Create grobs
    grobs <- lapply(
      split(data, -data$level),
      ggplot2::GeomLinerange$draw_panel,
      panel_scales, coord
    )
    ggplot2:::ggname("geom_hilo_linerange", do.call(grid::grobTree, grobs))
  }
)

#' @importFrom farver convert_colour
#' @importFrom grDevices col2rgb
darken_fill <- function(col, prob) {
  col <- farver::convert_colour(t(grDevices::col2rgb(col)), "RGB", "HSL")
  n_prob <- length(unique(prob))
  col[,3] <- seq(90 - pmin((n_prob - 1)*10, 30), 90, length.out = n_prob)[match(prob, sort(unique(prob)))]
  col <- farver::convert_colour(col, "HSL", "RGB")
  col2hex(col)
}

#' @importFrom grDevices rgb
col2hex <- function(col){
  grDevices::rgb(col,  maxColorValue = 255)
}

dist <- c(dist_normal(0, 1), dist_beta(5, 1))

test_that("geom_hilo_ribbon()", {
  dist <- dist_normal(1:3, 1:3)
  p <- ggplot2::ggplot(
    data.frame(x = rep(1:3, 2), interval = c(hilo(dist, 80), hilo(dist, 95)))
  ) +
    geom_hilo_ribbon(ggplot2::aes(x = x, hilo = interval))
  expect_silent(print(p))
  expect_length(
    ggplot2::layer_data(p)$hilo,
    3*2
  )
  expect_equal(
    ggplot2::layer_data(p)$hilo,
    c(hilo(dist, 80), hilo(dist, 95))[c(1,4,2,5,3,6)]
  )
})
test_that("geom_hilo_linerange()", {
  dist <- dist_normal(1:3, 1:3)
  p <- ggplot2::ggplot(
    data.frame(x = rep(1:3, 2), interval = c(hilo(dist, 80), hilo(dist, 95)))
  ) +
    geom_hilo_linerange(ggplot2::aes(x = x, hilo = interval))
  expect_silent(print(p))
  expect_length(
    ggplot2::layer_data(p)$hilo,
    3*2
  )
  expect_equal(
    ggplot2::layer_data(p)$hilo,
    c(hilo(dist, 80), hilo(dist, 95))
  )
})

#' The Multinomial distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The multinomial distribution is a generalization of the binomial
#' distribution to multiple categories. It is perhaps easiest to think
#' that we first extend a [dist_bernoulli()] distribution to include more
#' than two categories, resulting in a [dist_categorical()] distribution.
#' We then extend repeat the Categorical experiment several (\eqn{n})
#' times.
#'
#' @param size The number of draws from the Categorical distribution.
#' @param prob The probability of an event occurring from each draw.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_multinomial")`
#'
#'   In the following, let \eqn{X = (X_1, ..., X_k)} be a Multinomial
#'   random variable with success probability `prob` = \eqn{p}. Note that
#'   \eqn{p} is vector with \eqn{k} elements that sum to one. Assume
#'   that we repeat the Categorical experiment `size` = \eqn{n} times.
#'
#'   **Support**: Each \eqn{X_i} is in \eqn{\{0, 1, 2, ..., n\}}{{0, 1, 2, ..., n}}.
#'
#'   **Mean**: The mean of \eqn{X_i} is \eqn{n p_i}.
#'
#'   **Variance**: The variance of \eqn{X_i} is \eqn{n p_i (1 - p_i)}.
#'     For \eqn{i \neq j}, the covariance of \eqn{X_i} and \eqn{X_j}
#'     is \eqn{-n p_i p_j}.
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     P(X_1 = x_1, ..., X_k = x_k) = \frac{n!}{x_1! x_2! \cdots x_k!} p_1^{x_1} \cdot p_2^{x_2} \cdot \ldots \cdot p_k^{x_k}
#'   }{
#'     P(X_1 = x_1, ..., X_k = x_k) = n! / (x_1! x_2! ... x_k!) p_1^x_1 p_2^x_2 ... p_k^x_k
#'   }
#'
#'   where \eqn{\sum_{i=1}^k x_i = n} and \eqn{\sum_{i=1}^k p_i = 1}.
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     P(X_1 \le q_1, ..., X_k \le q_k) = \sum_{\substack{x_1, \ldots, x_k \ge 0 \\ x_i \le q_i \text{ for all } i \\ \sum_{i=1}^k x_i = n}} \frac{n!}{x_1! x_2! \cdots x_k!} p_1^{x_1} \cdot p_2^{x_2} \cdot \ldots \cdot p_k^{x_k}
#'   }{
#'     P(X_1 <= q_1, ..., X_k <= q_k) = sum over all (x_1, ..., x_k) with x_i <= q_i and sum x_i = n of n! / (x_1! ... x_k!) p_1^x_1 ... p_k^x_k
#'   }
#'
#'   The c.d.f. is computed as a finite sum of the p.m.f. over all integer vectors
#'   in the support that satisfy the componentwise inequalities.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{t'X}) = \left(\sum_{i=1}^k p_i e^{t_i}\right)^n
#'   }{
#'     E(e^(t'X)) = (p_1 e^t_1 + p_2 e^t_2 + ... + p_k e^t_k)^n
#'   }
#'
#'   where \eqn{t = (t_1, ..., t_k)} is a vector of the same dimension as \eqn{X}.
#'
#'   **Skewness**: The skewness of \eqn{X_i} is
#'
#'   \deqn{
#'     \frac{1 - 2p_i}{\sqrt{n p_i (1 - p_i)}}
#'   }{
#'     (1 - 2p_i) / sqrt(n p_i (1 - p_i))
#'   }
#'
#'   **Excess Kurtosis**: The excess kurtosis of \eqn{X_i} is
#'
#'   \deqn{
#'     \frac{1 - 6p_i(1 - p_i)}{n p_i (1 - p_i)}
#'   }{
#'     (1 - 6p_i(1 - p_i)) / (n p_i (1 - p_i))
#'   }
#'
#' @seealso [stats::dmultinom()], [stats::rmultinom()]
#'
#' @examples
#' dist <- dist_multinomial(size = c(4, 3), prob = list(c(0.3, 0.5, 0.2), c(0.1, 0.5, 0.4)))
#'
#' dist
#' mean(dist)
#' variance(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, list(d = rbind(cbind(1,2,1), cbind(0,2,1))))
#' density(dist, list(d = rbind(cbind(1,2,1), cbind(0,2,1))), log = TRUE)
#' 
#' cdf(dist, cbind(1,2,1))
#'
#' @name dist_multinomial
#' @export
dist_multinomial <- function(size, prob){
  size <- vec_cast(size, double())
  prob <- lapply(prob, function(x) x/sum(x))
  prob <- as_list_of(prob, .ptype = double())
  new_dist(s = size, p = prob, class = "dist_multinomial")
}

#' @export
format.dist_multinomial <- function(x, digits = 2, ...){
  sprintf(
    "Multinomial(%s)[%s]",
    format(x[["s"]], digits = digits, ...),
    format(length(x[["p"]]), digits = digits, ...)
  )
}

#' @export
density.dist_multinomial <- function(x, at, ...){
  if(is.list(at)) return(vapply(at, density, numeric(1L), x = x, ...))
  stats::dmultinom(at, x[["s"]], x[["p"]])
}

#' @export
log_density.dist_multinomial <- function(x, at, ...){
  if(is.list(at)) return(vapply(at, log_density, numeric(1L), x = x, ...))
  stats::dmultinom(at, x[["s"]], x[["p"]], log = TRUE)
}

#' @export
generate.dist_multinomial <- function(x, times, ...){
  t(stats::rmultinom(times, x[["s"]], x[["p"]]))
}

#' @export
mean.dist_multinomial <- function(x, ...){
  matrix(x[["s"]]*x[["p"]], nrow = 1)
}

#' @export
covariance.dist_multinomial <- function(x, ...){
  s <- x[["s"]]
  p <- x[["p"]]
  v <- numeric(length(p)^2)
  for(i in seq_along(p)){
    for(j in seq_along(p)){
      v[(i-1)*length(p) + j] <- if(i == j) s*p[i]*(1-p[j]) else -s*p[i]*p[j]
    }
  }
  list(matrix(v, nrow = length(p)))
}

#' @export
dim.dist_multinomial <- function(x){
  length(x[["p"]])
}

#' @export
cdf.dist_multinomial <- function(x, q, ...) {
  # Extract parameters
  size <- x$s
  prob <- x$p
  k <- length(prob)

  # Validate dimensions
  if (ncol(q) != k) {
    stop("Number of columns in q must match the number of categories in the multinomial distribution")
  }

  n_queries <- nrow(q)
  result <- numeric(n_queries)

  for (i in seq_len(n_queries)) {
    q_i <- q[i, ]

    # Any negative quantile => CDF = 0
    if (any(q_i < 0)) {
      result[i] <- 0
      next
    }

    # Discrete CDF: floor input values
    q_i <- floor(q_i)

    # Max counts for the first (k-1) categories
    max_vec <- pmin(q_i[1:(k - 1)], size)

    # Build ranges for first (k-1) categories
    ranges <- lapply(max_vec, function(m) 0:m)

    # Enumerate only the first (k-1) coordinates
    partial <- expand.grid(ranges, KEEP.OUT.ATTRS = FALSE, stringsAsFactors = FALSE)

    # Compute the last category as the remainder
    sums <- rowSums(partial)
    xk <- size - sums

    # Valid if 0 <= xk <= q_k
    valid_idx <- which(xk >= 0 & xk <= q_i[k])

    if (!length(valid_idx)) {
      result[i] <- 0
      next
    }

    # Construct full count vectors
    counts <- cbind(as.matrix(partial[valid_idx, , drop = FALSE]), xk[valid_idx])

    # Sum multinomial pmf over valid outcomes
    probs <- apply(counts, 1L, stats::dmultinom, size = size, prob = prob)
    result[i] <- sum(probs)
  }

  result
}

#' The Exponential Distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Exponential distributions are frequently used to model waiting times and the
#' time between events in a Poisson process.
#'
#' @inheritParams stats::dexp
#'
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_exponential")`
#' 
#'
#'   In the following, let \eqn{X} be an Exponential random variable with
#'   parameter `rate` = \eqn{\lambda}.
#'
#'   **Support**: \eqn{x \in [0, \infty)}{x >= 0}
#'
#'   **Mean**: \eqn{\frac{1}{\lambda}}{1 / \lambda}
#'
#'   **Variance**: \eqn{\frac{1}{\lambda^2}}{1 / \lambda^2}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \lambda e^{-\lambda x}
#'   }{
#'     f(x) = \lambda e^(-\lambda x)
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = 1 - e^{-\lambda x}
#'   }{
#'     F(x) = 1 - e^(-\lambda x)
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \frac{\lambda}{\lambda - t}, \quad t < \lambda
#'   }{
#'     E(e^(tX)) = \lambda / (\lambda - t), for t < \lambda
#'   }
#'
#' @seealso [stats::Exponential]
#'
#' @examples
#' dist <- dist_exponential(rate = c(2, 1, 2/3))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_exponential
#' @export
dist_exponential <- function(rate){
  rate <- vec_cast(rate, double())
  if(any(rate < 0)){
    abort("The rate parameter of an Exponential distribution must be non-negative.")
  }
  new_dist(rate = rate, class = "dist_exponential")
}

#' @export
format.dist_exponential <- function(x, digits = 2, ...){
  sprintf(
    "Exp(%s)",
    format(x[["rate"]], digits = digits, ...)
  )
}

#' @export
density.dist_exponential <- function(x, at, ...){
  stats::dexp(at, x[["rate"]])
}

#' @export
log_density.dist_exponential <- function(x, at, ...){
  stats::dexp(at, x[["rate"]], log = TRUE)
}

#' @export
quantile.dist_exponential <- function(x, p, ...){
  stats::qexp(p, x[["rate"]])
}

#' @export
cdf.dist_exponential <- function(x, q, ...){
  stats::pexp(q, x[["rate"]])
}

#' @export
generate.dist_exponential <- function(x, times, ...){
  stats::rexp(times, x[["rate"]])
}

#' @export
mean.dist_exponential <- function(x, ...){
  1/x[["rate"]]
}

#' @export
covariance.dist_exponential <- function(x, ...){
  1/x[["rate"]]^2
}

#' @export
skewness.dist_exponential <- function(x, ...) 2

#' @export
kurtosis.dist_exponential <- function(x, ...) 6

#' Create a distribution from p/d/q/r style functions
#'
#' @description
#' `r lifecycle::badge('maturing')`
#'
#' If a distribution is not yet supported, you can vectorise p/d/q/r functions
#' using this function. `dist_wrap()` stores the distributions parameters, and
#' provides wrappers which call the appropriate p/d/q/r functions.
#'
#' Using this function to wrap a distribution should only be done if the
#' distribution is not yet available in this package. If you need a distribution
#' which isn't in the package yet, consider making a request at
#' https://github.com/mitchelloharawild/distributional/issues.
#'
#' @param dist The name of the distribution used in the functions (name that is
#' prefixed by p/d/q/r)
#' @param ... Named arguments used to parameterise the distribution.
#' @param package The package from which the distribution is provided. If NULL,
#' the calling environment's search path is used to find the distribution
#' functions. Alternatively, an arbitrary environment can also be provided here.
# #' @param p,d,q,r The functions used to compute the p/d/q/r
# #' (pdf/cdf/quantile/generate)
#'
#' @examples
#' dist <- dist_wrap("norm", mean = 1:3, sd = c(3, 9, 2))
#'
#' density(dist, 1) # dnorm()
#' cdf(dist, 4) # pnorm()
#' quantile(dist, 0.975) # qnorm()
#' generate(dist, 10) # rnorm()
#'
#' library(actuar)
#' dist <- dist_wrap("invparalogis", package = "actuar", shape = 2, rate = 2)
#' density(dist, 1) # actuar::dinvparalogis()
#' cdf(dist, 4) # actuar::pinvparalogis()
#' quantile(dist, 0.975) # actuar::qinvparalogis()
#' generate(dist, 10) # actuar::rinvparalogis()
#'
#' @export
dist_wrap <- function(dist, ..., package = NULL){
  vec_assert(dist, character(), 1L)
  if(is.null(package)) {
    env <- rlang::caller_env()
  } else if (is.character(package)) {
    env <- rlang::pkg_env(package)
  } else {
    env <- as.environment(package)
  }
  par <- vec_recycle_common(dist = dist, env = list(env), ...)
  new_dist(!!!par, class = "dist_wrap")
}

#' @export
format.dist_wrap <- function(x, ...){
  sprintf(
    "%s(%s)",
    x[["dist"]],
    paste0(x[-(1:2)], collapse = ", ")
  )
}

#' @export
density.dist_wrap <- function(x, at, ...){
  fn <- get(paste0("d", x[["dist"]][[1]]), envir = x$env, mode = "function")

  # Remove distribution name and environment from parameters
  par <- x[-(1:2)]

  do.call(fn, c(list(at), par))
}

#' @export
log_density.dist_wrap <- function(x, at, ...){
  fn <- get(paste0("d", x[["dist"]][[1]]), envir = x$env, mode = "function")

  # Remove distribution name and environment from parameters
  par <- x[-(1:2)]

  # Use density(log = TRUE) if supported
  if(is.null(formals(fn)$log)){
    log(do.call(fn, c(list(at), par)))
  } else {
    do.call(fn, c(list(at), par, log = TRUE))
  }
}

#' @export
cdf.dist_wrap <- function(x, q, ...){
  fn <- get(paste0("p", x[["dist"]][[1]]), envir = x$env, mode = "function")

  # Remove distribution name and environment from parameters
  par <- x[-(1:2)]

  do.call(fn, c(list(q), par))
}

#' @export
quantile.dist_wrap <- function(x, p, ...){
  fn <- get(paste0("q", x[["dist"]][[1]]), envir = x$env, mode = "function")

  # Remove distribution name and environment from parameters
  par <- x[-(1:2)]

  do.call(fn, c(list(p), par))
}

#' @export
generate.dist_wrap <- function(x, times, ...){
  fn <- get(paste0("r", x[["dist"]][[1]]), envir = x$env, mode = "function")

  # Remove distribution name and environment from parameters
  par <- x[-(1:2)]

  do.call(fn, c(list(times), par))
}

#' @export
parameters.dist_wrap <- function(x, ...) {
  # All parameters except distribution environment
  x[-2L]
}

#' A convolved distribution
#'
#' @description
#' `r lifecycle::badge('experimental')`
#'
#' Creates the distribution of the sum of two or more independent random
#' variables using numerical convolution of their densities. The density of
#' the result is approximated on a grid using a Fast Fourier Transform (FFT),
#' and then interpolated with [stats::approxfun()].
#'
#' The cumulative distribution function and quantile function are derived from
#' the approximated density via numerical integration and inversion respectively.
#'
#' This is primarily intended to be used via arithmetic on distributions:
#' `dist1 + dist2` or `dist1 + dist2 + dist3 + ...`. Distributions with known
#' closed-form sums (e.g. two [dist_normal()]) will use the exact result rather
#' than this approximation. Chaining `+` automatically performs a single k-way
#' FFT convolution rather than nested binary convolutions, avoiding compounding
#' approximation errors.
#'
#' @param ... Two or more distribution vectors to sum. All will be recycled to
#'   a common length.
#'
#' @details
#'
#'   Let \eqn{Z = X_1 + X_2 + \cdots + X_k} where the \eqn{X_i} are
#'   independent random variables.
#'
#'   **Mean**:
#'   \deqn{E(Z) = \sum_{i=1}^{k} E(X_i)}
#'
#'   **Variance**:
#'   \deqn{\mathrm{Var}(Z) = \sum_{i=1}^{k} \mathrm{Var}(X_i)}
#'
#'   **Probability density function (p.d.f)**:
#'   \deqn{f_Z(z) = (f_{X_1} * f_{X_2} * \cdots * f_{X_k})(z)}
#'
#'   The k-way convolution is computed in a single FFT pass: each component
#'   density is evaluated on a common grid, transformed via FFT, all transforms
#'   are multiplied element-wise, and a single inverse FFT yields the result.
#'   This avoids compounding approximation errors from nested binary convolutions.
#'
#'   The accuracy of the FFT approximation can be controlled by passing `n`
#'   (number of grid points, default `2^12`) and `tail_p` (tail probability used
#'   to find finite grid bounds for distributions with infinite support, default
#'   `1e-6`) to `density()`, `cdf()`, `quantile()`, or `generate()`.
#'
#' @examples
#' # Sum of a lognormal and an exponential (no closed-form result)
#' d <- dist_convolved(dist_lognormal(0, 1), dist_exponential(1))
#' d
#'
#' density(d, 2)
#' cdf(d, 2)
#' quantile(d, 0.5)
#' generate(d, 5)
#'
#' # Three distributions from different families
#' d3 <- dist_convolved(dist_lognormal(0, 0.5), dist_gamma(2, 1), dist_exponential(2))
#' d3
#'
#' # Via arithmetic
#' d2 <- dist_lognormal(0, 1) + dist_exponential(1)
#' density(d2, 2)
#'
#' # Mean and variance are computed exactly from components
#' mean(d)
#' variance(d)
#'
#' @seealso [stats::convolve()], [stats::approxfun()]
#'
#' @name dist_convolved
#' @importFrom stats fft approxfun
#' @export
dist_convolved <- function(...) {
  dists <- dots_list(...)

  if (length(dists) < 2L) {
    abort("`dist_convolved()` requires at least two distributions.")
  }
  lapply(dists, vec_assert, ptype = new_dist())

  # Recycle inputs to a common length, then build a flat component list for
  # each output position, unpacking any nested dist_convolved inputs.
  recycled <- do.call(vctrs::vec_recycle_common, dists)
  n <- vctrs::vec_size(recycled[[1L]])

  components_by_pos <- lapply(seq_len(n), function(i) {
    unlist(
      lapply(recycled, function(d) {
        r <- vec_data(d)[[i]]
        if (inherits(r, "dist_convolved")) r[["dist"]] else list(r)
      }),
      recursive = FALSE
    )
  })

  new_dist(dist = components_by_pos, class = "dist_convolved")
}


#' @export
format.dist_convolved <- function(x, ...) {
  parts <- vapply(x[["dist"]], format, character(1L))
  paste(parts, collapse = " + ")
}

#' @export
density.dist_convolved <- function(x, at, ..., n = 2^12, tail_p = 1e-6) {
  convolve_approx(x, n = n, tail_p = tail_p)$dens_fn(at)
}

#' @export
cdf.dist_convolved <- function(x, q, ..., n = 2^12, tail_p = 1e-6) {
  convolve_approx(x, n = n, tail_p = tail_p)$cdf_fn(q)
}

#' @export
quantile.dist_convolved <- function(x, p, ..., n = 2^12, tail_p = 1e-6) {
  convolve_approx(x, n = n, tail_p = tail_p)$qtl_fn(p)
}

#' @export
generate.dist_convolved <- function(x, times, ..., n = 2^12, tail_p = 1e-6) {
  convolve_approx(x, n = n, tail_p = tail_p)$qtl_fn(stats::runif(times))
}

#' @export
mean.dist_convolved <- function(x, ...) {
  Reduce(`+`, lapply(x[["dist"]], mean))
}

#' @export
covariance.dist_convolved <- function(x, ...) {
  Reduce(`+`, lapply(x[["dist"]], covariance))
}

#' @export
family.dist_convolved <- function(object, ...) {
  "convolved"
}

# Compute the FFT-based k-way convolution approximation for a single
# dist_convolved object. Components are already flat (no nested dist_convolved).
convolve_approx <- function(x, n, tail_p) {
  dists <- x[["dist"]]
  k <- length(dists)

  # ---- build evaluation grid ------------------------------------------------

  # Support bounds per component; use tail_p quantiles to bound infinite tails
  bounds <- lapply(dists, function(d) {
    lo <- quantile(d, 0)
    if (is.infinite(lo)) lo <- quantile(d, tail_p)
    hi <- quantile(d, 1)
    if (is.infinite(hi)) hi <- quantile(d, 1 - tail_p)
    c(lo, hi)
  })

  # Common input grid spanning the union of all component supports
  x_lo <- min(vapply(bounds, `[`, numeric(1L), 1L))
  x_hi <- max(vapply(bounds, `[`, numeric(1L), 2L))

  x_grid <- seq(x_lo, x_hi, length.out = n)
  dx     <- (x_hi - x_lo) / n   # step size for integral scaling

  # ---- evaluate densities on common grid ------------------------------------
  dens_vecs <- lapply(dists, function(d) {
    pmax(density(d, x_grid), 0)
  })

  # ---- k-way FFT convolution ------------------------------------------------
  # Linear convolution of k vectors each of length n has length k*(n-1)+1.
  L <- k * (n - 1L) + 1L

  # Pad to next power of 2 for efficient FFT
  L_fft <- 2L ^ ceiling(log2(L))

  # FFT each zero-padded component, then multiply all element-wise
  ffts <- lapply(dens_vecs, function(dv) {
    fft(c(dv, rep(0, L_fft - n)))
  })
  combined_fft <- Reduce(`*`, ffts)

  # Inverse FFT; normalization convention: divide by L_fft
  d_conv <- Re(fft(combined_fft, inverse = TRUE)) / L_fft

  # Discard circular-wrap artefacts; keep only the L valid output points
  d_conv <- d_conv[seq_len(L)]

  # Scale: each of the (k-1) pairwise convolution steps contributes one dx
  d_conv <- d_conv * dx ^ (k - 1L)

  # Clamp numerical artefacts (tiny negative values)
  d_conv <- pmax(d_conv, 0)

  # ---- output grid ----------------------------------------------------------
  x_conv <- seq(k * x_lo, k * x_hi, length.out = L)

  # ---- density approxfun ----------------------------------------------------
  dens_fn <- approxfun(x_conv, d_conv, rule = 2, yleft = 0, yright = 0)

  # ---- CDF via cumulative trapezoidal integration ---------------------------
  dx_conv  <- x_conv[2L] - x_conv[1L]
  trap_mid <- (d_conv[-L] + d_conv[-1L]) / 2 * dx_conv
  cdf_vals <- c(0, cumsum(trap_mid))

  mx <- max(cdf_vals)
  if (mx > 0) cdf_vals <- cdf_vals / mx
  cdf_vals <- pmin(pmax(cdf_vals, 0), 1)

  cdf_fn <- approxfun(x_conv, cdf_vals, rule = 2, yleft = 0, yright = 1)

  # ---- quantile function (inverse CDF) --------------------------------------
  unique_idx <- !duplicated(cdf_vals)
  qtl_fn <- approxfun(
    cdf_vals[unique_idx], x_conv[unique_idx],
    rule = 2
  )

  list(dens_fn = dens_fn, cdf_fn = cdf_fn, qtl_fn = qtl_fn)
}

#' The Logistic distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' A continuous distribution on the real line. For binary outcomes
#' the model given by \eqn{P(Y = 1 | X) = F(X \beta)} where
#' \eqn{F} is the Logistic [cdf()] is called *logistic regression*.
#'
#' @inheritParams stats::dlogis
#'
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a Logistic random variable with
#'   `location` = \eqn{\mu} and `scale` = \eqn{s}.
#'
#'   **Support**: \eqn{R}, the set of all real numbers
#'
#'   **Mean**: \eqn{\mu}
#'
#'   **Variance**: \eqn{s^2 \pi^2 / 3}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{e^{-(\frac{x - \mu}{s})}}{s [1 + \exp(-(\frac{x - \mu}{s})) ]^2}
#'   }{
#'     f(x) = e^(-(t - \mu) / s) / (s (1 + e^(-(t - \mu) / s))^2)
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(t) = \frac{1}{1 + e^{-(\frac{t - \mu}{s})}}
#'   }{
#'     F(t) = 1 / (1 +  e^(-(t - \mu) / s))
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = e^{\mu t} \beta(1 - st, 1 + st)
#'   }{
#'     E(e^(tX)) = = e^(\mu t) \beta(1 - st, 1 + st)
#'   }
#'
#'   where \eqn{\beta(x, y)} is the Beta function.
#'
#' @seealso [stats::Logistic]
#'
#' @examples
#' dist <- dist_logistic(location = c(5,9,9,6,2), scale = c(2,3,4,2,1))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_logistic
#' @export
dist_logistic <- function(location, scale){
  location <- vec_cast(location, double())
  scale <- vec_cast(scale, double())
  new_dist(l = location, s = scale, class = "dist_logistic")
}

#' @export
format.dist_logistic <- function(x, digits = 2, ...){
  sprintf(
    "Logistic(%s, %s)",
    format(x[["l"]], digits = digits, ...),
    format(x[["s"]], digits = digits, ...)
  )
}

#' @export
density.dist_logistic <- function(x, at, ...){
  stats::dlogis(at, x[["l"]], x[["s"]])
}

#' @export
log_density.dist_logistic <- function(x, at, ...){
  stats::dlogis(at, x[["l"]], x[["s"]], log = TRUE)
}

#' @export
quantile.dist_logistic <- function(x, p, ...){
  stats::qlogis(p, x[["l"]], x[["s"]])
}

#' @export
cdf.dist_logistic <- function(x, q, ...){
  stats::plogis(q, x[["l"]], x[["s"]])
}

#' @export
generate.dist_logistic <- function(x, times, ...){
  stats::rlogis(times, x[["l"]], x[["s"]])
}

#' @export
mean.dist_logistic <- function(x, ...){
  x[["l"]]
}

#' @export
covariance.dist_logistic <- function(x, ...){
  (x[["s"]]*pi)^2/3
}

#' @export
skewness.dist_logistic <- function(x, ...) 0

#' @export
kurtosis.dist_logistic <- function(x, ...) 6 / 5

#' The Dirichlet distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Dirichlet distribution is a multivariate generalisation of the Beta
#' distribution. It is the conjugate prior of the Categorical and Multinomial
#' distributions, and describes a probability distribution over the
#' \eqn{(k-1)}-simplex — the set of \eqn{k}-dimensional vectors whose
#' components are non-negative and sum to one.
#'
#' @param alpha A list of positive numeric concentration vectors.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_dirichlet")`
#'
#'   In the following, let \eqn{\mathbf{X} = (X_1, \ldots, X_k)} be a
#'   Dirichlet random variable with concentration parameter
#'   `alpha` = \eqn{\boldsymbol{\alpha} = (\alpha_1, \ldots, \alpha_k)},
#'   where each \eqn{\alpha_i > 0}.
#'
#'   **Support**: \eqn{\mathbf{x}} on the \eqn{(k-1)}-simplex,
#'   i.e. \eqn{x_i \geq 0} and \eqn{\sum_{i=1}^k x_i = 1}.
#'
#'   **Mean**: \eqn{E(X_i) = \frac{\alpha_i}{\alpha_0}} where
#'   \eqn{\alpha_0 = \sum_{i=1}^k \alpha_i}.
#'
#'   **Variance**:
#'
#'   \deqn{
#'     \mathrm{Var}(X_i) = \frac{\alpha_i(\alpha_0 - \alpha_i)}{\alpha_0^2(\alpha_0 + 1)}
#'   }{
#'     Var(X_i) = alpha_i * (alpha_0 - alpha_i) / (alpha_0^2 * (alpha_0 + 1))
#'   }
#'
#'   **Covariance**:
#'
#'   \deqn{
#'     \mathrm{Cov}(X_i, X_j) = \frac{-\alpha_i \alpha_j}{\alpha_0^2(\alpha_0 + 1)}, \quad i \neq j
#'   }{
#'     Cov(X_i, X_j) = -alpha_i * alpha_j / (alpha_0^2 * (alpha_0 + 1)), i != j
#'   }
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(\mathbf{x}) = \frac{1}{B(\boldsymbol{\alpha})}
#'     \prod_{i=1}^k x_i^{\alpha_i - 1}
#'   }{
#'     f(x) = 1 / B(alpha) * prod(x_i^(alpha_i - 1))
#'   }
#'
#'   where \eqn{B(\boldsymbol{\alpha}) = \frac{\prod_{i=1}^k \Gamma(\alpha_i)}{\Gamma(\alpha_0)}}
#'   is the multivariate Beta function.
#'
#' @seealso [LaplacesDemon::ddirichlet()], [LaplacesDemon::rdirichlet()]
#'
#' @examples
#' dist <- dist_dirichlet(alpha = list(c(2, 5, 3)))
#' dist
#'
#' @examplesIf requireNamespace("LaplacesDemon", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, cbind(0.2, 0.5, 0.3))
#' density(dist, cbind(0.2, 0.5, 0.3), log = TRUE)
#'
#' @name dist_dirichlet
#' @export
dist_dirichlet <- function(alpha) {
  alpha <- as_list_of(as.list(alpha), .ptype = double())
  new_dist(alpha = alpha, class = "dist_dirichlet")
}

#' @export
format.dist_dirichlet <- function(x, digits = 2, ...) {
  sprintf(
    "Dirichlet[%i]",
    length(x[["alpha"]])
  )
}

#' @export
density.dist_dirichlet <- function(x, at, ..., na.rm = FALSE) {
  require_package("LaplacesDemon")
  if (is.list(at)) return(vapply(at, density, numeric(1L), x = x, ...))
  LaplacesDemon::ddirichlet(as.numeric(at), x[["alpha"]])
}

#' @export
log_density.dist_dirichlet <- function(x, at, ..., na.rm = FALSE) {
  require_package("LaplacesDemon")
  if (is.list(at)) return(vapply(at, log_density, numeric(1L), x = x, ...))
  LaplacesDemon::ddirichlet(as.numeric(at), x[["alpha"]], log = TRUE)
}

#' @export
generate.dist_dirichlet <- function(x, times, ..., na.rm = FALSE) {
  require_package("LaplacesDemon")
  LaplacesDemon::rdirichlet(times, x[["alpha"]])
}

#' @export
mean.dist_dirichlet <- function(x, ...) {
  alpha <- x[["alpha"]]
  alpha_0 <- sum(alpha)
  matrix(alpha / alpha_0, nrow = 1)
}

#' @export
covariance.dist_dirichlet <- function(x, ...) {
  alpha <- x[["alpha"]]
  k <- length(alpha)
  alpha_0 <- sum(alpha)
  denom <- alpha_0^2 * (alpha_0 + 1)
  sigma <- matrix(NA_real_, nrow = k, ncol = k)
  for (i in seq_len(k)) {
    for (j in seq_len(k)) {
      sigma[i, j] <- if (i == j) {
        alpha[i] * (alpha_0 - alpha[i]) / denom
      } else {
        -alpha[i] * alpha[j] / denom
      }
    }
  }
  list(sigma)
}

#' @export
support.dist_dirichlet <- function(x, ...) {
  k <- length(x[["alpha"]])
  proto <- matrix(rep(0, k), nrow = 1)
  new_support_region(
    list(proto),
    list(c(0, 1)),
    list(c(TRUE, TRUE))
  )
}

#' @export
dim.dist_dirichlet <- function(x) {
  length(x[["alpha"]])
}

#' Create a mixture of distributions
#'
#' @description
#' `r lifecycle::badge('maturing')`
#'
#' A mixture distribution combines multiple component distributions with
#' specified weights. Two types of mixture are supported:
#'
#' - **Probability mixture** (`type = "probability"`, the default): the CDF is
#'   a weighted average of the component CDFs. Also known as *linear pooling*.
#'
#' - **Quantile mixture** (`type = "quantile"`): the quantile function is a
#'   weighted average of the component quantile functions. Also known as
#'   *Vincentization*.
#'
#' @param ... Distributions to be used in the mixture. Can be any
#'   distributional objects.
#' @param weights A numeric vector of non-negative weights that sum to 1.
#'   The length must match the number of distributions passed to `...`.
#'   Each weight \eqn{w_i} represents the probability that a random draw
#'   comes from the \eqn{i}-th component distribution.
#' @param type `"probability"` (default) for a probability mixture or
#'   `"quantile"` for a quantile mixture (Vincentization). Quantile mixtures
#'   only support univariate component distributions.
#'
#' @details
#'
#'   In the following, let \eqn{X} be a mixture random variable composed
#'   of \eqn{K} component distributions \eqn{F_1, F_2, \ldots, F_K} with
#'   corresponding weights \eqn{w_1, w_2, \ldots, w_K} where
#'   \eqn{\sum_{i=1}^K w_i = 1} and \eqn{w_i \geq 0} for all \eqn{i}.
#'
#'   ## Probability mixture (`type = "probability"`)
#'
#'   **Support**: The union of the supports of all component distributions
#'
#'   **Mean**:
#'
#'   For univariate mixtures:
#'   \deqn{
#'     E(X) = \sum_{i=1}^K w_i \mu_i
#'   }{
#'     E(X) = sum_{i=1}^K w_i * mu_i
#'   }
#'
#'   where \eqn{\mu_i} is the mean of the \eqn{i}-th component distribution.
#'
#'   For multivariate mixtures:
#'   \deqn{
#'     E(\mathbf{X}) = \sum_{i=1}^K w_i \boldsymbol{\mu}_i
#'   }{
#'     E(X) = sum_{i=1}^K w_i * mu_i
#'   }
#'
#'   where \eqn{\boldsymbol{\mu}_i} is the mean vector of the \eqn{i}-th
#'   component distribution.
#'
#'   **Variance**:
#'
#'   For univariate mixtures:
#'   \deqn{
#'     \text{Var}(X) = \sum_{i=1}^K w_i (\mu_i^2 + \sigma_i^2) - \left(\sum_{i=1}^K w_i \mu_i\right)^2
#'   }{
#'     Var(X) = sum_{i=1}^K w_i * (mu_i^2 + sigma_i^2) - (sum_{i=1}^K w_i * mu_i)^2
#'   }
#'
#'   where \eqn{\sigma_i^2} is the variance of the \eqn{i}-th component
#'   distribution.
#'
#'   **Covariance**:
#'
#'   For multivariate mixtures:
#'   \deqn{
#'     \text{Cov}(\mathbf{X}) = \sum_{i=1}^K w_i \left[ (\boldsymbol{\mu}_i - \bar{\boldsymbol{\mu}})(\boldsymbol{\mu}_i - \bar{\boldsymbol{\mu}})^T + \boldsymbol{\Sigma}_i \right]
#'   }{
#'     Cov(X) = sum_{i=1}^K w_i * [ (mu_i - mu_bar)(mu_i - mu_bar)^T + Sigma_i ]
#'   }
#'
#'   where \eqn{\bar{\boldsymbol{\mu}} = \sum_{i=1}^K w_i \boldsymbol{\mu}_i}
#'   is the overall mean vector and \eqn{\boldsymbol{\Sigma}_i} is the
#'   covariance matrix of the \eqn{i}-th component distribution.
#'
#'   **Probability density/mass function (p.d.f/p.m.f)**:
#'
#'   \deqn{
#'     f(x) = \sum_{i=1}^K w_i f_i(x)
#'   }{
#'     f(x) = sum_{i=1}^K w_i * f_i(x)
#'   }
#'
#'   where \eqn{f_i(x)} is the density or mass function of the \eqn{i}-th
#'   component distribution.
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   For univariate mixtures:
#'   \deqn{
#'     F(x) = \sum_{i=1}^K w_i F_i(x)
#'   }{
#'     F(x) = sum_{i=1}^K w_i * F_i(x)
#'   }
#'
#'   where \eqn{F_i(x)} is the c.d.f. of the \eqn{i}-th component
#'   distribution.
#'
#'   For multivariate mixtures, the c.d.f. is approximated numerically.
#'
#'   **Quantile function**:
#'
#'   For univariate probability mixtures, the quantile function has no closed
#'   form and is computed numerically by inverting the c.d.f. using root-finding
#'   ([stats::uniroot()]).
#'
#'   For multivariate mixtures, quantiles are not yet implemented.
#'
#'   ## Quantile mixture (`type = "quantile"`)
#'
#'   Also known as a *Vincent average* or *Vincentization*, only univariate
#'   component distributions are supported.
#'
#'   **Quantile function** (closed form):
#'   \deqn{
#'     Q(p) = \sum_{i=1}^K w_i Q_i(p)
#'   }{
#'     Q(p) = sum_{i=1}^K w_i * Q_i(p)
#'   }
#'
#'   where \eqn{Q_i(p)} is the quantile function of the \eqn{i}-th component
#'   distribution.
#'
#'   **Cumulative distribution function**: computed numerically by inverting
#'   \eqn{Q(p)} via [stats::uniroot()].
#'
#'   **Probability density function**: derived analytically from the quantile
#'   function. For \eqn{p = F(x)}:
#'   \deqn{
#'     f(x) = \frac{1}{Q'(p)} = \frac{1}{\sum_{i=1}^K w_i / f_i(Q_i(p))}
#'   }{
#'     f(x) = 1 / sum_{i=1}^K w_i / f_i(Q_i(p))
#'   }
#'
#'   **Mean**: \eqn{E(X) = \sum_{i=1}^K w_i \mu_i} (identical to the
#'   probability mixture mean).
#'
#'   **Variance**: computed numerically as
#'   \eqn{\int_0^1 Q(p)^2 \, dp - \left(E(X)\right)^2}.
#'
#' @seealso [stats::uniroot()], [vctrs::vec_unique_count()]
#'
#' @examples
#' # Probability mixture of two normal distributions (default)
#' dist <- dist_mixture(dist_normal(0, 1), dist_normal(5, 2), weights = c(0.3, 0.7))
#' dist
#'
#' mean(dist)
#' variance(dist)
#'
#' density(dist, 2)
#' cdf(dist, 2)
#' quantile(dist, 0.5)
#'
#' generate(dist, 10)
#'
#' # Quantile mixture (Vincentization) of two normal distributions
#' vdist <- dist_mixture(
#'   dist_normal(0, 1), dist_normal(5, 2),
#'   weights = c(0.3, 0.7), type = "quantile"
#' )
#' vdist
#'
#' mean(vdist)
#' variance(vdist)
#'
#' density(vdist, 2)
#' cdf(vdist, 2)
#' quantile(vdist, 0.5)
#'
#' @name dist_mixture
#' @export
dist_mixture <- function(..., weights = numeric(), type = c("probability", "quantile")) {
  type <- match.arg(type)
  dist <- dots_list(...)

  vec_is(weights, numeric(), length(dist))

  if (!abs(sum(weights) - 1) < sqrt(.Machine$double.eps)) {
    abort("Weights of a mixture model must sum to 1.")
  }
  if (any(weights < 0)) {
    abort("All weights in a mixtue model must be non-negative.")
  }

  if (type == "quantile") {
    # Quantile mixtures require univariate components
    dims <- vapply(dist, function(d) dim(vctrs::vec_data(d)[[1L]]), integer(1L))
    if (any(dims > 1L)) {
      abort("Quantile mixtures only support univariate distributions.")
    }
    return(new_dist(
      dist = transpose(dist), w = list(weights),
      class = "dist_q_mixture"
    ))
  }

  dn <- unique(lapply(dist, dimnames))
  dn <- if (length(dn) == 1) dn[[1]] else NULL

  new_dist(
    dist = transpose(dist), w = list(weights),
    class = "dist_mixture", dimnames = dn
  )
}

# ---- Probability mixture -----------------------------------------------------

#' @export
format.dist_mixture <- function(x, width = getOption("width"), ...) {
  dists <- unlist(lapply(x[["dist"]], format))

  dist_info <- paste0(x[["w"]], "*", dists, collapse = ", ")

  long_dist <- paste0("mixture(", dist_info, ")")
  short_dist <- paste0("mixture(n=", length(dists), ")")
  ifelse(nchar(long_dist) <= width, long_dist, short_dist)
}

#' @export
density.dist_mixture <- function(x, at, ...) {
  if (NROW(at) > 1) return(vapply(at, density, numeric(1L), x = x, ...))
  sum(x[["w"]] * vapply(x[["dist"]], density, numeric(1L), at = at, ...))
}

#' @export
quantile.dist_mixture <- function(x, p, ...) {
  d <- dim(x)
  if (d > 1) {
    stop("quantile is not implemented for multivariate mixtures.")
  }
  if (length(p) > 1) return(vapply(p, quantile, numeric(1L), x = x, ...))

  # Find bounds for optimisation based on range of each quantile
  dist_q <- vapply(x[["dist"]], quantile, numeric(1L), p, ..., USE.NAMES = FALSE)
  if (vctrs::vec_unique_count(dist_q) == 1) return(dist_q[1])
  if (p == 0) return(min(dist_q))
  if (p == 1) return(max(dist_q))

  # Search the cdf() for appropriate quantile
  stats::uniroot(
    function(pos) p - cdf(x, pos, ...),
    interval = c(min(dist_q), max(dist_q)),
    extendInt = "yes"
  )$root
}

#' @export
cdf.dist_mixture <- function(x, q, ...) {
  d <- dim(x)
  if (d == 1L) {
    if (length(q) > 1) return(vapply(q, cdf, numeric(1L), x = x, ...))
    sum(x[["w"]] * vapply(x[["dist"]], cdf, numeric(1L), q = q, ...))
  } else {
    NextMethod()
  }
}

#' @export
generate.dist_mixture <- function(x, times, ...) {
  dist_idx <- .bincode(stats::runif(times), breaks = c(0, cumsum(x[["w"]])))
  r <- matrix(nrow = times, ncol = dim(x))
  for (i in seq_along(x[["dist"]])) {
    r_pos <- dist_idx == i
    if (any(r_pos)) {
      r[r_pos, ] <- generate(x[["dist"]][[i]], sum(r_pos), ...)
    }
  }
  r[, seq(NCOL(r)), drop = TRUE]
}

#' @export
mean.dist_mixture <- function(x, ...) {
  d <- dim(x)
  m <- vapply(x[["dist"]], mean, numeric(d), ...)
  if (d == 1L) {
    sum(x[["w"]] * m)
  } else {
    matrix(x[["w"]], ncol = d, nrow = 1) %*% t(m)
  }
}

#' @export
covariance.dist_mixture <- function(x, ...) {
  d <- dim(x)
  if (d == 1L) {
    m <- vapply(x[["dist"]], mean, numeric(1L), ...)
    v <- vapply(x[["dist"]], variance, numeric(1L), ...)
    m1 <- sum(x[["w"]] * m)
    m2 <- sum(x[["w"]] * (m^2 + v))
    m2 - m1^2
  } else {
    m <- lapply(x[["dist"]], mean)
    w <- as.list(x[["w"]])
    mbar <- mapply("*", m, w, SIMPLIFY = FALSE)
    mbar <- do.call("+", mbar)
    m <- lapply(m, function(u) {
      u - mbar
    })
    v <- lapply(x[["dist"]], function(u) {
      covariance(u)[[1]]
    })
    cov <- mapply(function(m, v, w) {
      w * (t(m) %*% m + v)
    }, m, v, w, SIMPLIFY = FALSE)
    list(do.call("+", cov))
  }
}

#' @export
dim.dist_mixture <- function(x) {
  dim(x[["dist"]][[1]])
}

# ---- Quantile mixture --------------------------------------------------------

#' @export
format.dist_q_mixture <- function(x, width = getOption("width"), ...) {
  dists <- unlist(lapply(x[["dist"]], format))
  dist_info <- paste0(x[["w"]], "*", dists, collapse = ", ")
  long_dist <- paste0("mixture<q>(", dist_info, ")")
  short_dist <- paste0("mixture<q>(n=", length(dists), ")")
  ifelse(nchar(long_dist) <= width, long_dist, short_dist)
}

#' @export
quantile.dist_q_mixture <- function(x, p, ...) {
  if (length(p) > 1) return(vapply(p, quantile, numeric(1L), x = x, ...))
  sum(x[["w"]] * vapply(x[["dist"]], quantile, numeric(1L), p, ...))
}

#' @export
cdf.dist_q_mixture <- function(x, q, ...) {
  if (length(q) > 1) return(vapply(q, cdf, numeric(1L), x = x, ...))

  # Find p in (0, 1) such that Q(p) = q by inverting the quantile function.
  fn <- function(p) {
    sum(x[["w"]] * vapply(x[["dist"]], quantile, numeric(1L), p, ...)) - q
  }

  # Compute support bounds directly from the quantile function to avoid
  # calling support(x), which would trigger density() -> cdf() recursion.
  lims <- c(
    sum(x[["w"]] * vapply(x[["dist"]], quantile, numeric(1L), 0, ...)),
    sum(x[["w"]] * vapply(x[["dist"]], quantile, numeric(1L), 1, ...))
  )

  # q is at or below the lower bound of the mixture support
  if (q <= lims[1]) return(0)
  # q is at or above the upper bound of the mixture support
  if (q >= lims[2]) return(1)

  # Use the smallest representable eps so the search interval covers
  # far tails (e.g., normal distributions). For q beyond Q(eps) or Q(1-eps),
  # return the boundary probability rather than erroring in uniroot.
  eps <- .Machine$double.eps
  lo_val <- fn(eps)
  hi_val <- fn(1 - eps)

  if (lo_val >= 0) return(eps)
  if (hi_val <= 0) return(1 - eps)

  stats::uniroot(fn, interval = c(eps, 1 - eps), tol = 1e-8)$root
}

#' @export
support.dist_q_mixture <- function(x, ...) {
  lims <- c(
    sum(x[["w"]] * vapply(x[["dist"]], quantile, numeric(1L), 0, ...)),
    sum(x[["w"]] * vapply(x[["dist"]], quantile, numeric(1L), 1, ...))
  )
  closed <- is.finite(lims)
  new_support_region(
    list(numeric(0)),
    list(lims),
    list(closed)
  )
}

#' @export
density.dist_q_mixture <- function(x, at, ...) {
  if (length(at) > 1) return(vapply(at, density, numeric(1L), x = x, ...))

  # p = F(at) via CDF inversion
  p <- cdf(x, at, ...)

  # Outside (or at boundary of) the support
  if (p <= 0 || p >= 1) return(0)

  # Q'(p) = sum(w_i * Q_i'(p)) = sum(w_i / f_i(Q_i(p)))
  # => f(at) = 1 / Q'(p)
  qi_vals <- vapply(x[["dist"]], quantile, numeric(1L), p, ...)
  fi_vals <- vapply(
    seq_along(x[["dist"]]),
    function(i) density(x[["dist"]][[i]], qi_vals[[i]], ...),
    numeric(1L)
  )

  if (any(fi_vals == 0)) return(0)

  1 / sum(x[["w"]] / fi_vals)
}

#' @export
mean.dist_q_mixture <- function(x, ...) {
  # E[X] = integral_0^1 Q(p) dp = sum(w_i * mu_i)
  sum(x[["w"]] * vapply(x[["dist"]], mean, numeric(1L), ...))
}

#' @export
covariance.dist_q_mixture <- function(x, ...) {
  # Var(X) = E[X^2] - (E[X])^2
  # E[X^2] = integral_0^1 Q(p)^2 dp  (numerical)
  mu <- mean(x, ...)
  q_fn <- function(p) {
    sum(x[["w"]] * vapply(x[["dist"]], quantile, numeric(1L), p, ...))
  }
  e2 <- stats::integrate(
    function(p) vapply(p, function(pp) q_fn(pp)^2, numeric(1L)),
    lower = 0, upper = 1,
    rel.tol = 1e-5
  )$value
  e2 - mu^2
}

#' @export
dim.dist_q_mixture <- function(x) {
  1L
}

#' The degenerate distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The degenerate distribution takes a single value which is certain to be
#' observed. It takes a single parameter, which is the value that is observed
#' by the distribution.
#'
#' @param x The value of the distribution.
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a degenerate random variable with value
#'   `x` = \eqn{k_0}.
#'
#'   **Support**: \eqn{R}, the set of all real numbers
#'
#'   **Mean**: \eqn{k_0}
#'
#'   **Variance**: \eqn{0}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = 1 for x = k_0
#'   }{
#'     f(x) = 1 for x = k_0
#'   }
#'   \deqn{
#'     f(x) = 0 for x \neq k_0
#'   }{
#'     f(x) = 0 for x \neq k_0
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   The cumulative distribution function has the form
#'
#'   \deqn{
#'     F(x) = 0 for x < k_0
#'   }{
#'     F(x) = 0 for x < k_0
#'   }
#'   \deqn{
#'     F(x) = 1 for x \ge k_0
#'   }{
#'     F(x) = 1 for x \ge k_0
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = e^{k_0 t}
#'   }{
#'     E(e^(tX)) = e^(k_0 t)
#'   }
#'
#' @examples
#' dist_degenerate(x = 1:5)
#'
#' @export
dist_degenerate <- function(x){
  vec_is(x, numeric())
  new_dist(x = x, class = "dist_degenerate")
}

#' @export
format.dist_degenerate <- function(x, ...){
  format(x[["x"]], ...)
}

#' @export
density.dist_degenerate <- function(x, at, ...){
  ifelse(at == x[["x"]], 1, 0)
}

#' @export
quantile.dist_degenerate <- function(x, p, ...){
  ifelse(p < 0 | p > 1, NaN, x[["x"]])
}

#' @export
cdf.dist_degenerate <- function(x, q, ...){
  ifelse(q >= x[["x"]], 1, 0)
}

#' @export
generate.dist_degenerate <- function(x, times, ...){
  rep(x[["x"]], times)
}

#' @export
mean.dist_degenerate <- function(x, ...){
  x[["x"]]
}

#' @export
covariance.dist_degenerate <- function(x, ...){
  0
}

#' @export
skewness.dist_degenerate <- function(x, ...) NA_real_

#' @export
kurtosis.dist_degenerate <- function(x, ...) NA_real_

#' The Pareto Distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Pareto distribution is a power-law probability distribution commonly used
#' in actuarial science to model loss severity and in economics to model income
#' distributions and firm sizes.
#'
#' @inheritParams actuar::dpareto
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_pareto")`
#'
#'   In the following, let \eqn{X} be a Pareto random variable with parameters
#'   `shape` = \eqn{\alpha} and `scale` = \eqn{\theta}.
#'
#'   **Support**: \eqn{(0, \infty)}{(0, Inf)}
#'
#'   **Mean**: \eqn{\frac{\theta}{\alpha - 1}}{\theta / (\alpha - 1)} for \eqn{\alpha > 1}, 
#'   undefined otherwise
#'
#'   **Variance**: \eqn{\frac{\alpha\theta^2}{(\alpha - 1)^2(\alpha - 2)}}{\alpha\theta^2 / ((\alpha - 1)^2(\alpha - 2))} 
#'   for \eqn{\alpha > 2}, undefined otherwise
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{\alpha\theta^\alpha}{(x + \theta)^{\alpha + 1}}
#'   }{
#'     f(x) = \alpha\theta^\alpha / (x + \theta)^(\alpha + 1)
#'   }
#'
#'   for \eqn{x > 0}, \eqn{\alpha > 0} and \eqn{\theta > 0}.
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = 1 - \left(\frac{\theta}{x + \theta}\right)^\alpha
#'   }{
#'     F(x) = 1 - (\theta / (x + \theta))^\alpha
#'   }
#'
#'   for \eqn{x > 0}.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   Does not exist in closed form, but the \eqn{k}th raw moment \eqn{E[X^k]} exists
#'   for \eqn{-1 < k < \alpha}.
#'
#' @note
#' There are many different definitions of the Pareto distribution in the
#' literature; see Arnold (2015) or Kleiber and Kotz (2003). This implementation
#' uses the Pareto distribution without a location parameter as described in [actuar::Pareto].
#'
#' @references
#' Kleiber, C. and Kotz, S. (2003), *Statistical Size Distributions in Economics
#' and Actuarial Sciences*, Wiley.
#'
#' Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012), *Loss Models, From
#' Data to Decisions, Fourth Edition*, Wiley.
#'
#' @seealso [actuar::Pareto]
#'
#' @examples
#' dist <- dist_pareto(shape = c(10, 3, 2, 1), scale = rep(1, 4))
#' dist
#'
#' @examplesIf requireNamespace("actuar", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_pareto
#' @export
dist_pareto <- function(shape, scale){
  shape <- vec_cast(shape, double())
  scale <- vec_cast(scale, double())
  if(any(shape < 0)){
    abort("The shape parameter of a Pareto distribution must be non-negative.")
  }
  if(any(scale <= 0)){
    abort("The scale parameter of a Pareto distribution must be strictly positive.")
  }
  new_dist(shape = shape, scale = scale, class = "dist_pareto")
}

#' @export
format.dist_pareto <- function(x, digits = 2, ...){
  sprintf(
    "Pareto(%s, %s)",
    format(x[["shape"]], digits = digits, ...),
    format(x[["scale"]], digits = digits, ...)
  )
}

#' @export
density.dist_pareto <- function(x, at, ...){
  require_package("actuar")
  actuar::dpareto(at, x[["shape"]], x[["scale"]])
}

#' @export
log_density.dist_pareto <- function(x, at, ...){
  require_package("actuar")
  actuar::dpareto(at, x[["shape"]], x[["scale"]], log = TRUE)
}

#' @export
quantile.dist_pareto <- function(x, p, ...){
  require_package("actuar")
  actuar::qpareto(p, x[["shape"]], x[["scale"]])
}

#' @export
cdf.dist_pareto <- function(x, q, ...){
  require_package("actuar")
  actuar::ppareto(q, x[["shape"]], x[["scale"]])
}

#' @export
generate.dist_pareto <- function(x, times, ...){
  require_package("actuar")
  actuar::rpareto(times, x[["shape"]], x[["scale"]])
}

#' @export
mean.dist_pareto <- function(x, ...){
  actuar::mpareto(1, x[["shape"]], x[["scale"]])
}

#' @export
covariance.dist_pareto <- function(x, ...){
  actuar::mpareto(2, x[["shape"]], x[["scale"]]) - actuar::mpareto(1, x[["shape"]], x[["scale"]])^2
}

#' Truncate a distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Note that the samples are generated using inverse transform sampling, and the
#' means and variances are estimated from samples.
#'
#' @param dist The distribution(s) to truncate.
#' @param lower,upper The range of values to keep from a distribution.
#'
#' @name dist_truncated
#'
#' @examples
#' dist <- dist_truncated(dist_normal(2,1), lower = 0)
#'
#' dist
#' mean(dist)
#' variance(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' if(requireNamespace("ggdist")) {
#' library(ggplot2)
#' ggplot() +
#'   ggdist::stat_dist_halfeye(
#'     aes(y = c("Normal", "Truncated"),
#'         dist = c(dist_normal(2,1), dist_truncated(dist_normal(2,1), lower = 0)))
#'   )
#' }
#'
#' @export
dist_truncated <- function(dist, lower = -Inf, upper = Inf){
  vec_is(dist, new_dist())
  vec_is(lower, numeric())
  vec_is(upper, numeric())
  if(any(lower >= upper)){
    abort("The `lower` truncation bound must be lower than the `upper` bound.")
  }
  new_dist(dist = dist, lower = lower, upper = upper,
           dimnames = dimnames(dist), class = "dist_truncated")
}

#' @export
format.dist_truncated <- function(x, ...){
  sprintf(
    "%s[%s,%s]",
    format(x[["dist"]]),
    x[["lower"]],
    x[["upper"]]
  )
}

#' @export
density.dist_truncated <- function(x, at, ...){
  in_lim <- at >= x[["lower"]] & at <= x[["upper"]]
  cdf_upr <- cdf(x[["dist"]], x[["upper"]])
  cdf_lwr <- cdf(x[["dist"]], x[["lower"]])
  out <- numeric(length(at))
  out[in_lim] <- density(x[["dist"]], at = at[in_lim], ...)/(cdf_upr - cdf_lwr)
  out
}

#' @export
quantile.dist_truncated <- function(x, p, ...){
  F_lwr <- cdf(x[["dist"]], x[["lower"]])
  F_upr <- cdf(x[["dist"]], x[["upper"]])
  qt <- quantile(x[["dist"]], F_lwr + p * (F_upr - F_lwr), ...)
  pmin(pmax(x[["lower"]], qt), x[["upper"]])
}

#' @export
cdf.dist_truncated <- function(x, q, ...){
  cdf_upr <- cdf(x[["dist"]], x[["upper"]])
  cdf_lwr <- cdf(x[["dist"]], x[["lower"]])
  out <- numeric(length(q))
  q_lwr <- q < x[["lower"]] # out[q_lwr <- q < x[["lower"]]] <- 0
  out[q_upr <- q > x[["upper"]]] <- 1
  q_mid <- !(q_lwr|q_upr)
  out[q_mid] <- (cdf(x[["dist"]], q = q[q_mid], ...) - cdf_lwr)/(cdf_upr - cdf_lwr)
  out
}

#' @export
mean.dist_truncated <- function(x, ...) {
  if(inherits(x$dist, "dist_sample")) {
    y <- x$dist[[1]]
    mean(y[y >= x$lower & y <= x$upper])
  } else if(inherits(x$dist, "dist_normal")) {
    mu <- x$dist$mu
    s <- x$dist$sigma
    a <- (x$lower - mu) / s
    b <- (x$upper - mu) / s
    mu + (stats::dnorm(a) - stats::dnorm(b))/(stats::pnorm(b) - stats::pnorm(a))*s
  } else {
    NextMethod()
  }
}

#' Percentile distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Percentile distribution is a non-parametric distribution defined by
#' a set of quantiles at specified percentile values. This distribution is
#' useful for representing empirical distributions or elicited expert
#' knowledge when only percentile information is available. The distribution
#' uses linear interpolation between percentiles and can be used to
#' approximate complex distributions that may not have simple parametric forms.
#'
#' @param x A list of values
#' @param percentile A list of percentiles
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_percentile")`
#'
#'   In the following, let \eqn{X} be a Percentile random variable defined by
#'   values \eqn{x_1, x_2, \ldots, x_n} at percentiles
#'   \eqn{p_1, p_2, \ldots, p_n} where \eqn{0 \le p_i \le 100}.
#'
#'   **Support**: \eqn{[\min(x_i), \max(x_i)]} if \eqn{\min(p_i) > 0} or
#'   \eqn{\max(p_i) < 100}, otherwise support is approximated from the
#'   specified percentiles.
#'
#'   **Mean**: Approximated numerically using spline interpolation and
#'   numerical integration:
#'
#'   \deqn{
#'     E(X) \approx \int_0^1 Q(u) du
#'   }{
#'     E(X) ≈ integral_0^1 Q(u) du
#'   }
#'
#'   where \eqn{Q(u)} is a spline function interpolating the percentile values.
#'
#'   **Variance**: Approximated numerically.
#'
#'   **Probability density function (p.d.f)**: Approximated numerically using
#'   kernel density estimation from generated samples.
#'
#'   **Cumulative distribution function (c.d.f)**: Defined by linear
#'   interpolation:
#'
#'   \deqn{
#'     F(t) = \begin{cases}
#'       p_1/100 & \text{if } t < x_1 \\
#'       p_i/100 + \frac{(t - x_i)(p_{i+1} - p_i)}{100(x_{i+1} - x_i)} & \text{if } x_i \le t < x_{i+1} \\
#'       p_n/100 & \text{if } t \ge x_n
#'     \end{cases}
#'   }{
#'     F(t) = p_i/100 + (t - x_i)(p_{i+1} - p_i) / (100(x_{i+1} - x_i)) for x_i ≤ t < x_{i+1}
#'   }
#'
#'   **Quantile function**: Defined by linear interpolation:
#'
#'   \deqn{
#'     Q(u) = x_i + \frac{(100u - p_i)(x_{i+1} - x_i)}{p_{i+1} - p_i}
#'   }{
#'     Q(u) = x_i + (100u - p_i)(x_{i+1} - x_i) / (p_{i+1} - p_i)
#'   }
#'
#'   for \eqn{p_i/100 \le u \le p_{i+1}/100}.
#'
#' @examples
#' dist <- dist_normal()
#' percentiles <- seq(0.01, 0.99, by = 0.01)
#' x <- vapply(percentiles, quantile, double(1L), x = dist)
#' dist_percentile(list(x), list(percentiles*100))
#'
#' @export
dist_percentile <- function(x, percentile){
  x <- as_list_of(x, .ptype = double())
  percentile <- as_list_of(percentile, .ptype = double())
  new_dist(x = x, percentile = percentile, class = "dist_percentile")
}

#' @export
format.dist_percentile <- function(x, ...){
  sprintf(
    "percentile[%s]",
    length(x[["x"]])
  )
}

#' @export
density.dist_percentile <- function(x, at, ...){
  d <- density(generate(x, 1000), from = min(at), to = max(at), ..., na.rm=TRUE)
  stats::approx(d$x, d$y, xout = at)$y
}


#' @export
quantile.dist_percentile <- function(x, p, ...){
  out <- x[["x"]][match(p, x[["percentile"]])]
  out[is.na(out)] <- stats::approx(x = x[["percentile"]]/100, y = x[["x"]], xout = p[is.na(out)])$y
  out
}

#' @export
cdf.dist_percentile <- function(x, q, ...){
  stats::approx(x = x[["x"]], y = x[["percentile"]]/100, xout = q)$y
}

#' @export
generate.dist_percentile <- function(x, times, ...){
  stats::approx(x[["percentile"]], x[["x"]], xout=stats::runif(times,min(x[["percentile"]]),max(x[["percentile"]])))$y
}

#' @export
mean.dist_percentile <- function(x, ...) {
  # assumes percentile is sorted
  # probs <- x[["percentile"]]/100
  # i <- seq_along(probs)
  #
  # weights <- (probs[pmin(i+1, length(probs))] - probs[pmax(i-1, 1)]) / 2
  # sum(x[["x"]] * weights)


  # Fit a spline to the percentiles
  spline_fit <- stats::splinefun(x[["percentile"]], x[["x"]])

  # Use numerical integration to estimate the mean
  stats::integrate(spline_fit, lower = 0, upper = 1)$value
}

#' @export
support.dist_percentile <- function(x, ...) {
  new_support_region(
    list(vctrs::vec_init(x[["x"]], n = 0L)),
    list(range(x[["x"]])),
    list(!near(range(x[["percentile"]]), 0))
  )
}

#' The Negative Binomial distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' A generalization of the geometric distribution. It is the number
#' of failures in a sequence of i.i.d. Bernoulli trials before
#' a specified number of successes (`size`) occur. The probability of success in
#' each trial is given by `prob`.
#'
#' @inheritParams stats::NegBinomial
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a Negative Binomial random variable with
#'   success probability `prob` = \eqn{p} and the number of successes `size` =
#'   \eqn{r}.
#'
#'
#'   **Support**: \eqn{\{0, 1, 2, 3, ...\}}
#'
#'   **Mean**: \eqn{\frac{p r}{1-p}}
#'
#'   **Variance**: \eqn{\frac{pr}{(1-p)^2}}
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'      f(k) = {k + r - 1 \choose k} \cdot (1-p)^r p^k
#'   }{
#'      f(k) = (k+r-1)!/(k!(r-1)!) (1-p)^r p^k
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   Too nasty, omitted.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'      \left(\frac{1-p}{1-pe^t}\right)^r, t < -\log p
#'   }{
#'      \frac{(1-p)^r}{(1-pe^t)^r}, t < -\log p
#'   }
#'
#' @seealso [stats::NegBinomial]
#'
#' @examples
#' dist <- dist_negative_binomial(size = 10, prob = 0.5)
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#' support(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @export
dist_negative_binomial <- function(size, prob){
  size <- vec_cast(size, double())
  prob <- vec_cast(prob, double())
  if(any(prob < 0 | prob > 1)){
    abort("Probability of success must be between 0 and 1.")
  }
  new_dist(n = size, p = prob, class = "dist_negbin")
}

#' @export
format.dist_negbin <- function(x, digits = 2, ...){
  sprintf(
    "NB(%s, %s)",
    format(x[["n"]], digits = digits, ...),
    format(x[["p"]], digits = digits, ...)
  )
}

#' @export
density.dist_negbin <- function(x, at, ...){
  stats::dnbinom(at, x[["n"]], x[["p"]])
}

#' @export
log_density.dist_negbin <- function(x, at, ...){
  stats::dnbinom(at, x[["n"]], x[["p"]], log = TRUE)
}

#' @export
quantile.dist_negbin <- function(x, p, ...){
  stats::qnbinom(p, x[["n"]], x[["p"]])
}

#' @export
cdf.dist_negbin <- function(x, q, ...){
  stats::pnbinom(q, x[["n"]], x[["p"]])
}

#' @export
generate.dist_negbin <- function(x, times, ...){
  stats::rnbinom(times, x[["n"]], x[["p"]])
}

#' @export
mean.dist_negbin <- function(x, ...){
  x[["n"]] * (1 - x[["p"]]) / x[["p"]]
}

#' @export
covariance.dist_negbin <- function(x, ...){
  x[["n"]] * (1 - x[["p"]]) / x[["p"]]^2
}

#' @export
skewness.dist_negbin <- function(x, ...) {
  (1 + x[["p"]]) / sqrt(x[["p"]] * x[["n"]])
}

#' @export
kurtosis.dist_negbin <- function(x, ...) {
  6 / x[["n"]] + (1 - x[["p"]])^2 / x[["n"]] * x[["p"]]
}

#' The generalised g-and-h Distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The generalised g-and-h distribution is a flexible distribution used to model univariate data, similar to the g-k distribution.
#' It is known for its ability to handle skewness and heavy-tailed behavior.
#'
#' @inheritParams gk::dgh
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a g-and-h random variable with parameters
#'   `A`, `B`, `g`, `h`, and `c`.
#'
#'   **Support**: \eqn{(-\infty, \infty)}
#'
#'   **Mean**: Not available in closed form.
#'
#'   **Variance**: Not available in closed form.
#'
#'   **Probability density function (p.d.f)**:
#'
#'   The g-and-h distribution does not have a closed-form expression for its density. Instead,
#'   it is defined through its quantile function:
#'
#'   \deqn{
#'     Q(u) = A + B \left( 1 + c \frac{1 - \exp(-gz(u))}{1 + \exp(-gz(u))} \right) \exp(h z(u)^2/2) z(u)
#'   }{
#'     Q(u) = A + B * (1 + c * ((1 - exp(-g * z(u))) / (1 + exp(-g * z(u))))) * exp(h * z(u)^2/2) * z(u)
#'   }
#'
#'   where \eqn{z(u) = \Phi^{-1}(u)}
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   The cumulative distribution function is typically evaluated numerically due to the lack
#'   of a closed-form expression.
#'
#' @seealso [gk::dgh], [distributional::dist_gk]
#'
#' @examples
#' dist <- dist_gh(A = 0, B = 1, g = 0, h = 0.5)
#' dist
#'
#' @examplesIf requireNamespace("gk", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_gh
#' @export
dist_gh <- function(A, B, g, h, c = 0.8){
  A <- vec_cast(A, double())
  B <- vec_cast(B, double())
  g <- vec_cast(g, double())
  h <- vec_cast(h, double())
  c <- vec_cast(c, double())
  if(any(B <= 0)){
    abort("The B parameter (scale) of the g-and-h distribution must be strictly positive.")
  }
  new_dist(A = A, B = B, g = g, h = h, c = c, class = "dist_gh")
}

#' @export
format.dist_gh <- function(x, digits = 2, ...){
  sprintf(
    "gh(A = %s, B = %s, g = %s, h = %s%s)",
    format(x[["A"]], digits = digits, ...),
    format(x[["B"]], digits = digits, ...),
    format(x[["g"]], digits = digits, ...),
    format(x[["h"]], digits = digits, ...),
    if (x[["c"]]==0.8) "" else paste0(", c = ", format(x[["c"]], digits = digits, ...))
  )
}

#' @export
density.dist_gh <- function(x, at, ...){
  require_package("gk")
  gk::dgh(at, x[["A"]], x[["B"]], x[["g"]], x[["h"]], x[["c"]])
}

#' @export
log_density.dist_gh <- function(x, at, ...){
  require_package("gk")
  gk::dgh(at, x[["A"]], x[["B"]], x[["g"]], x[["h"]], x[["c"]], log = TRUE)
}

#' @export
quantile.dist_gh <- function(x, p, ...){
  require_package("gk")
  gk::qgh(p, x[["A"]], x[["B"]], x[["g"]], x[["h"]], x[["c"]])
}

#' @export
cdf.dist_gh <- function(x, q, ...){
  require_package("gk")
  gk::pgh(q, x[["A"]], x[["B"]], x[["g"]], x[["h"]], x[["c"]])
}

#' @export
generate.dist_gh <- function(x, times, ...){
  require_package("gk")
  gk::rgh(times, x[["A"]], x[["B"]], x[["g"]], x[["h"]], x[["c"]])
}

#' The Burr distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Burr distribution (Type XII) is a flexible continuous probability
#' distribution often used for modeling income distributions, reliability
#' data, and failure times.
#'
#' @inheritParams actuar::dburr
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_burr")`
#'
#'   In the following, let \eqn{X} be a Burr random variable with parameters
#'   `shape1` = \eqn{\alpha}, `shape2` = \eqn{\gamma}, and `rate` = \eqn{\lambda}.
#'
#'   **Support**: \eqn{x \in (0, \infty)}{x > 0}
#'
#'   **Mean**: \eqn{\frac{\lambda^{-1/\alpha} \gamma B(\gamma - 1/\alpha, 1 + 1/\alpha)}{\gamma}} (for \eqn{\alpha \gamma > 1})
#'
#'   **Variance**: \eqn{\frac{\lambda^{-2/\alpha} \gamma B(\gamma - 2/\alpha, 1 + 2/\alpha)}{\gamma} - \mu^2} (for \eqn{\alpha \gamma > 2})
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \alpha \gamma \lambda x^{\alpha - 1} (1 + \lambda x^\alpha)^{-\gamma - 1}
#'   }{
#'     f(x) = \alpha \gamma \lambda x^(\alpha - 1) (1 + \lambda x^\alpha)^(-\gamma - 1)
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(x) = 1 - (1 + \lambda x^\alpha)^{-\gamma}
#'   }{
#'     F(x) = 1 - (1 + \lambda x^\alpha)^(-\gamma)
#'   }
#'
#'   **Quantile function**:
#'
#'   \deqn{
#'     F^{-1}(p) = \lambda^{-1/\alpha} ((1 - p)^{-1/\gamma} - 1)^{1/\alpha}
#'   }{
#'     F^(-1)(p) = \lambda^(-1/\alpha) ((1 - p)^(-1/\gamma) - 1)^(1/\alpha)
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   Does not exist in closed form.
#'
#' @seealso [actuar::Burr]
#'
#' @examples
#' dist <- dist_burr(shape1 = c(1,1,1,2,3,0.5), shape2 = c(1,2,3,1,1,2))
#' dist
#'
#' @examplesIf requireNamespace("actuar", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_burr
#' @export
dist_burr <- function(shape1, shape2, rate = 1, scale = 1/rate){
  shape1 <- vec_cast(shape1, double())
  shape2 <- vec_cast(shape2, double())
  if(any(shape1 <= 0)){
    abort("The shape1 parameter of a Burr distribution must be strictly positive.")
  }
  if(any(shape2 <= 0)){
    abort("The shape2 parameter of a Burr distribution must be strictly positive.")
  }
  if(any(rate <= 0)){
    abort("The rate parameter of a Burr distribution must be strictly positive.")
  }
  new_dist(s1 = shape1, s2 = shape2, r = 1/scale, class = "dist_burr")
}

#' @export
format.dist_burr <- function(x, digits = 2, ...){
  sprintf(
    "Burr12(%s, %s, %s)",
    format(x[["s1"]], digits = digits, ...),
    format(x[["s2"]], x[["r"]], digits = digits, ...),
    format(x[["r"]], digits = digits, ...)
  )
}

#' @export
density.dist_burr <- function(x, at, ...){
  require_package("actuar")
  actuar::dburr(at, x[["s1"]], x[["s2"]], x[["r"]])
}

#' @export
log_density.dist_burr <- function(x, at, ...){
  require_package("actuar")
  actuar::dburr(at, x[["s1"]], x[["s2"]], x[["r"]], log = TRUE)
}

#' @export
quantile.dist_burr <- function(x, p, ...){
  require_package("actuar")
  actuar::qburr(p, x[["s1"]], x[["s2"]], x[["r"]])
}

#' @export
cdf.dist_burr <- function(x, q, ...){
  require_package("actuar")
  actuar::pburr(q, x[["s1"]], x[["s2"]], x[["r"]])
}

#' @export
generate.dist_burr <- function(x, times, ...){
  require_package("actuar")
  actuar::rburr(times, x[["s1"]], x[["s2"]], x[["r"]])
}

#' @export
mean.dist_burr <- function(x, ...){
  require_package("actuar")
  actuar::mburr(1, x[["s1"]], x[["s2"]], x[["r"]])
}

#' @export
covariance.dist_burr <- function(x, ...){
  require_package("actuar")
  m1 <- actuar::mburr(1, x[["s1"]], x[["s2"]], x[["r"]])
  m2 <- actuar::mburr(2, x[["s1"]], x[["s2"]], x[["r"]])
  -m1^2 + m2
}

#' The Binomial distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Binomial distributions are used to represent situations can that can
#' be thought as the result of \eqn{n} Bernoulli experiments (here the
#' \eqn{n} is defined as the `size` of the experiment). The classical
#' example is \eqn{n} independent coin flips, where each coin flip has
#' probability `p` of success. In this case, the individual probability of
#' flipping heads or tails is given by the  Bernoulli(p) distribution,
#' and the probability of having \eqn{x} equal results (\eqn{x} heads,
#' for example), in \eqn{n} trials is given by the Binomial(n, p) distribution.
#' The equation of the Binomial distribution is directly derived from
#' the equation of the Bernoulli distribution.
#'
#' @param size The number of trials. Must be an integer greater than or equal
#'   to one. When `size = 1L`, the Binomial distribution reduces to the
#'   Bernoulli distribution. Often called `n` in textbooks.
#' @param prob The probability of success on each trial, `prob` can be any
#'   value in `[0, 1]`.
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   The Binomial distribution comes up when you are interested in the portion
#'   of people who do a thing. The Binomial distribution
#'   also comes up in the sign test, sometimes called the Binomial test
#'   (see [stats::binom.test()]), where you may need the Binomial C.D.F. to
#'   compute p-values.
#'
#'   In the following, let \eqn{X} be a Binomial random variable with parameter
#'   `size` = \eqn{n} and `p` = \eqn{p}. Some textbooks define \eqn{q = 1 - p},
#'   or called \eqn{\pi} instead of \eqn{p}.
#'
#'   **Support**: \eqn{\{0, 1, 2, ..., n\}}{{0, 1, 2, ..., n}}
#'
#'   **Mean**: \eqn{np}
#'
#'   **Variance**: \eqn{np \cdot (1 - p) = np \cdot q}{np (1 - p)}
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     P(X = k) = {n \choose k} p^k (1 - p)^{n-k}
#'   }{
#'     P(X = k) = choose(n, k) p^k (1 - p)^(n - k)
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     P(X \le k) = \sum_{i=0}^{\lfloor k \rfloor} {n \choose i} p^i (1 - p)^{n-i}
#'   }{
#'     P(X \le k) = \sum_{i=0}^k choose(n, i) p^i (1 - p)^(n-i)
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = (1 - p + p e^t)^n
#'   }{
#'     E(e^(tX)) = (1 - p + p e^t)^n
#'   }
#'
#' @examples
#' dist <- dist_binomial(size = 1:5, prob = c(0.05, 0.5, 0.3, 0.9, 0.1))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_binomial
#' @export
dist_binomial <- function(size, prob){
  size <- vec_cast(size, integer())
  prob <- vec_cast(prob, double())
  if(any(size < 0)){
    abort("The number of observations cannot be negative.")
  }
  if(any((prob < 0) | (prob > 1))){
    abort("The probability of success must be between 0 and 1.")
  }
  new_dist(n = size, p = prob, class = "dist_binomial")
}

#' @export
format.dist_binomial <- function(x, digits = 2, ...){
  sprintf(
    "B(%s, %s)",
    format(x[["n"]], digits = digits, ...),
    format(x[["p"]], digits = digits, ...)
  )
}

#' @export
density.dist_binomial <- function(x, at, ...){
  stats::dbinom(at, x[["n"]], x[["p"]])
}

#' @export
log_density.dist_binomial <- function(x, at, ...){
  stats::dbinom(at, x[["n"]], x[["p"]], log = TRUE)
}

#' @export
quantile.dist_binomial <- function(x, p, ...){
  as.integer(stats::qbinom(p, x[["n"]], x[["p"]]))
}

#' @export
cdf.dist_binomial <- function(x, q, ...){
  stats::pbinom(q, x[["n"]], x[["p"]])
}

#' @export
generate.dist_binomial <- function(x, times, ...){
  as.integer(stats::rbinom(times, x[["n"]], x[["p"]]))
}

#' @export
mean.dist_binomial <- function(x, ...){
  x[["n"]]*x[["p"]]
}

#' @export
covariance.dist_binomial <- function(x, ...){
  x[["n"]]*x[["p"]]*(1-x[["p"]])
}

#' @export
skewness.dist_binomial <- function(x, ...) {
  n <- x[["n"]]
  p <- x[["p"]]
  q <- 1 - p
  (1 - (2 * p)) / sqrt(n * p * q)
}

#' @export
kurtosis.dist_binomial <- function(x, ...) {
  n <- x[["n"]]
  p <- x[["p"]]
  q <- 1 - p
  (1 - (6 * p * q)) / (n * p * q)
}

#' The multivariate normal distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The multivariate normal distribution is a generalization of the univariate
#' normal distribution to higher dimensions. It is widely used in multivariate
#' statistics and describes the joint distribution of multiple correlated
#' continuous random variables.
#'
#' @param mu A list of numeric vectors for the distribution's mean.
#' @param sigma A list of matrices for the distribution's variance-covariance matrix.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_multivariate_normal")`
#'
#'   In the following, let \eqn{\mathbf{X}} be a \eqn{k}-dimensional multivariate
#'   normal random variable with mean vector `mu` = \eqn{\boldsymbol{\mu}} and
#'   variance-covariance matrix `sigma` = \eqn{\boldsymbol{\Sigma}}.
#'
#'   **Support**: \eqn{\mathbf{x} \in \mathbb{R}^k}
#'
#'   **Mean**: \eqn{\boldsymbol{\mu}}
#'
#'   **Variance-covariance matrix**: \eqn{\boldsymbol{\Sigma}}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(\mathbf{x}) = \frac{1}{(2\pi)^{k/2} |\boldsymbol{\Sigma}|^{1/2}}
#'     \exp\left(-\frac{1}{2}(\mathbf{x} - \boldsymbol{\mu})^T
#'     \boldsymbol{\Sigma}^{-1}(\mathbf{x} - \boldsymbol{\mu})\right)
#'   }{
#'     f(x) = 1 / ((2\pi)^(k/2) |\Sigma|^(1/2))
#'     exp(-1/2 (x - \mu)^T \Sigma^(-1) (x - \mu))
#'   }
#'
#'   where \eqn{|\boldsymbol{\Sigma}|}{|\Sigma|} is the determinant of
#'   \eqn{\boldsymbol{\Sigma}}{\Sigma}.
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     P(\mathbf{X} \le \mathbf{q}) = P(X_1 \le q_1, \ldots, X_k \le q_k)
#'   }{
#'     P(X \le q) = P(X_1 \le q_1, ..., X_k \le q_k)
#'   }
#'
#'   The c.d.f. does not have a closed-form expression and is computed numerically.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     M(\mathbf{t}) = E(e^{\mathbf{t}^T \mathbf{X}}) =
#'     \exp\left(\mathbf{t}^T \boldsymbol{\mu} + \frac{1}{2}\mathbf{t}^T
#'     \boldsymbol{\Sigma} \mathbf{t}\right)
#'   }{
#'     M(t) = E(e^(t^T X)) = exp(t^T \mu + 1/2 t^T \Sigma t)
#'   }
#'
#' @seealso [mvtnorm::dmvnorm()], [mvtnorm::pmvnorm()], [mvtnorm::qmvnorm()],
#'   [mvtnorm::rmvnorm()]
#'
#' @examples
#' dist <- dist_multivariate_normal(mu = list(c(1,2)), sigma = list(matrix(c(4,2,2,3), ncol=2)))
#' dimnames(dist) <- c("x", "y")
#' dist
#'
#' @examplesIf requireNamespace("mvtnorm", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, cbind(2, 1))
#' density(dist, cbind(2, 1), log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7, kind = "equicoordinate")
#' quantile(dist, 0.7, kind = "marginal")
#'
#' @export
dist_multivariate_normal <- function(mu = 0, sigma = list(diag(1))){
  new_dist(mu = mu, sigma = sigma,
           dimnames = colnames(sigma[[1]]), class = "dist_mvnorm")
}

#' @export
format.dist_mvnorm <- function(x, digits = 2, ...){
  sprintf(
    "MVN[%i]",
    length(x[["mu"]])
  )
}

#' @export
density.dist_mvnorm <- function(x, at, ..., na.rm = FALSE){
  require_package("mvtnorm")
  if(is.list(at)) return(vapply(at, density, numeric(1L), x = x, ...))
  mvtnorm::dmvnorm(at, x[["mu"]], x[["sigma"]])
}

#' @export
log_density.dist_mvnorm <- function(x, at, ..., na.rm = FALSE){
  require_package("mvtnorm")
  if(is.list(at)) return(vapply(at, log_density, numeric(1L), x = x, ...))
  mvtnorm::dmvnorm(at, x[["mu"]], x[["sigma"]], log = TRUE)
}

#' @export
quantile.dist_mvnorm <- function(x, p, kind = c("marginal", "equicoordinate"),
                                 ..., type = lifecycle::deprecated(), na.rm = FALSE){
  if (lifecycle::is_present(type)) {
    lifecycle::deprecate_warn(
      "0.6.0", "quantile.dist_mvnorm(type)", "quantile.dist_mvnorm(kind)",
      # Force removal of the 'contact author' footer since it doesn't work here.
      user_env = .GlobalEnv
    )
    kind <- type
  }
  kind <- match.arg(kind)
  q <- if (kind == "marginal") {
    stats::qnorm(p, mean = rep(x[["mu"]], each = length(p)),
                 sd = rep(sqrt(diag(x[["sigma"]])), each = length(p)), ...)
  } else {
    require_package("mvtnorm")
    vapply(p, function(p, ...) {
      if (p == 0) return(-Inf) else if (p == 1) return(Inf)
      mvtnorm::qmvnorm(p, ...)$quantile
    }, numeric(1L), mean = x[["mu"]], sigma = x[["sigma"]], ...)
  }

  matrix(q, nrow = length(p), ncol = dim(x))
}

#' @export
cdf.dist_mvnorm <- function(x, q, ..., na.rm = FALSE){
  if(is.list(q)) return(vapply(q, cdf, numeric(1L), x = x, ...))
  require_package("mvtnorm")
  mvtnorm::pmvnorm(upper = as.numeric(q), mean = x[["mu"]], sigma = x[["sigma"]], ...)[1]
}

#' @export
generate.dist_mvnorm <- function(x, times, ..., na.rm = FALSE){
  require_package("mvtnorm")
  mvtnorm::rmvnorm(times, x[["mu"]], x[["sigma"]], ...)
}

#' @export
mean.dist_mvnorm <- function(x, ...){
  matrix(x[["mu"]], nrow = 1)
}

#' @export
median.dist_mvnorm <- mean.dist_mvnorm

#' @export
covariance.dist_mvnorm <- function(x, ...){
  # Wrap in list to preserve matrix structure
  list(x[["sigma"]])
}

#' @export
dim.dist_mvnorm <- function(x){
  length(x[["mu"]])
}

#' Plot a distribution
#'
#' @description
#' `r lifecycle::badge('defunct')`
#'
#' This function is now defunct and can no longer be used. Instead consider using
#' the {ggdist} package to produce your own distribution plots. You can learn
#' more about how this plot can be produced using {ggdist} here:
#' https://mjskay.github.io/ggdist/articles/slabinterval.html
#'
#' @param x The distribution(s) to plot.
#' @param ... Unused.
#'
#' @keywords internal
#'
#' @export
autoplot.distribution <- function(x, ...){
  lifecycle::deprecate_stop("0.2.0", "distributional::autoplot.distribution()", details = "The autoplot() method for distributions have been replaced by geoms in the {ggdist} package.\nYou can produce a similar plot using ggdist::stat_dist_slab()")
}

#' The Cauchy distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Cauchy distribution is the student's t distribution with one degree of
#' freedom. The Cauchy distribution does not have a well defined mean or
#' variance. Cauchy distributions often appear as priors in Bayesian contexts
#' due to their heavy tails.
#'
#' @inheritParams stats::dcauchy
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a Cauchy variable with mean
#'   `location =` \eqn{x_0} and `scale` = \eqn{\gamma}.
#'
#'   **Support**: \eqn{R}, the set of all real numbers
#'
#'   **Mean**: Undefined.
#'
#'   **Variance**: Undefined.
#'
#'   **Probability density function (p.d.f)**:
#'
#'   \deqn{
#'     f(x) = \frac{1}{\pi \gamma \left[1 + \left(\frac{x - x_0}{\gamma} \right)^2 \right]}
#'   }{
#'     f(x) = 1 / (\pi \gamma (1 + ((x - x_0) / \gamma)^2)
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     F(t) = \frac{1}{\pi} \arctan \left( \frac{t - x_0}{\gamma} \right) +
#'       \frac{1}{2}
#'   }{
#'     F(t) = arctan((t - x_0) / \gamma) / \pi + 1/2
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   Does not exist.
#'
#' @seealso [stats::Cauchy]
#'
#' @examples
#' dist <- dist_cauchy(location = c(0, 0, 0, -2), scale = c(0.5, 1, 2, 1))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_cauchy
#' @export
dist_cauchy <- function(location, scale){
  location <- vec_cast(location, double())
  scale <- vec_cast(scale, double())
  if(any(scale[!is.na(scale)] <= 0)){
    abort("The scale parameter of a Cauchy distribution must strictly positive.")
  }
  new_dist(location = location, scale = scale, class = "dist_cauchy")
}

#' @export
format.dist_cauchy <- function(x, digits = 2, ...){
  sprintf(
    "Cauchy(%s, %s)",
    format(x[["location"]], digits = digits, ...),
    format(x[["scale"]], digits = digits, ...)
  )
}

#' @export
density.dist_cauchy <- function(x, at, ...){
  stats::dcauchy(at, x[["location"]], x[["scale"]])
}

#' @export
log_density.dist_cauchy <- function(x, at, ...){
  stats::dcauchy(at, x[["location"]], x[["scale"]], log = TRUE)
}

#' @export
quantile.dist_cauchy <- function(x, p, ...){
  stats::qcauchy(p, x[["location"]], x[["scale"]])
}

#' @export
cdf.dist_cauchy <- function(x, q, ...){
  stats::pcauchy(q, x[["location"]], x[["scale"]])
}

#' @export
generate.dist_cauchy <- function(x, times, ...){
  stats::rcauchy(times, x[["location"]], x[["scale"]])
}

#' @export
mean.dist_cauchy <- function(x, ...){
  NA_real_
}

#' @export
covariance.dist_cauchy <- function(x, ...){
  NA_real_
}

#' @export
skewness.dist_cauchy <- function(x, ...){
  NA_real_
}

#' @export
kurtosis.dist_cauchy <- function(x, ...){
  NA_real_
}

#' level luminance scales
#'
#' This set of scales defines new scales for prob geoms equivalent to the
#' ones already defined by ggplot2. This allows the shade of confidence intervals
#' to work with the legend output.
#' @return A ggproto object inheriting from `Scale`
#' @family scale_level_*
#' @name scale_level
#' @rdname scale_level
NULL

#' @rdname scale_level
#' @inheritParams ggplot2::scale_colour_gradient
#' @export
scale_level_continuous <- function(..., guide = "level") {
  level_scale("level", "identity", identity, guide = guide, ...)
}

ScaleLevel <- ggplot2::ggproto(NULL, ggplot2::ScaleContinuous)

#' @importFrom ggplot2 waiver discrete_scale
level_scale <- function(...) {
  ggplot2::ggproto(NULL, ggplot2::continuous_scale(...), range = level_range())
}

level_range <- function(){
  ggplot2::ggproto(NULL, RangeLevel)
}

RangeLevel <- ggplot2::ggproto(NULL, NULL,
                               range = NULL,
                               levels = NULL,
                               reset = function(self){
                                 self$range <- NULL
                                 self$levels <- NULL
                               },
                               train = function(self, x){
                                 self$range <- scales::train_continuous(x, self$range)
                                 self$levels <- unique(c(x[!is.na(x)],self$range))
                               }
)


#' Level shade bar guide
#'
#' The level guide shows the colour from the forecast intervals which is blended with the series colour.
#'
#' @inheritParams ggplot2::guide_colourbar
#' @param max_discrete The maximum number of levels to be shown using [ggplot2::guide_legend()].
#' If the number of levels exceeds this value, level shades are shown with [ggplot2::guide_colourbar()].
#' @param ... Further arguments passed onto either [ggplot2::guide_colourbar()] or [ggplot2::guide_legend()]
#'
#' @export
guide_level <- function(title = waiver(), max_discrete = 5, ...) {
  structure(list(title = title,
                 max_discrete = max_discrete,
                 available_aes = "level",
                 args = list(...)),
            class=c("guide", "level_guide"))
}

#' Helper methods for guides
#'
#' @export
#' @rdname guide-helpers
#' @importFrom ggplot2 guide_colourbar guide_legend guide_train
#' @keywords internal
guide_train.level_guide <- function(guide, scale, aesthetic) {
  args <- append(guide[!(names(guide)%in%c("max_discrete", "args"))], guide$args)
  levels <- scale$range$levels
  if (length(levels) == 0 || all(is.na(levels)))
    return()
  if(length(levels)<=guide$max_discrete){
    guide <- do.call("guide_legend", args)
    class(guide) <- c("guide", "guide_level")
    breaks <- levels

    cols <- darken_fill(rep.int("white", length(breaks)), breaks)
    key <- as.data.frame(
      set_names(list(cols), aesthetic %||% scale$aesthetics[1]),
      stringsAsFactors = FALSE
    )
    key$.label <- scale$get_labels(breaks)
    if (!scale$is_discrete()) {
      limits <- scale$get_limits()
      noob <- !is.na(breaks) & limits[1] <= breaks & breaks <=
        limits[2]
      key <- key[noob, , drop = FALSE]
    }
    if (guide$reverse)
      key <- key[nrow(key):1, ]
    guide$key <- key
    guide$hash <- with(guide, digest::digest(list(title, key$.label,
                                                  direction, name)))
  }
  else{
    guide <- do.call("guide_colourbar", args)
    breaks <- scale$get_breaks()
    ticks <- as.data.frame(stats::setNames(list(scale$map(breaks)),
                                           aesthetic %||% scale$aesthetics[1]))
    ticks$.value <- breaks
    ticks$.label <- scale$get_labels(breaks)
    guide$key <- ticks
    .limits <- scale$get_limits()
    .bar <- seq(.limits[1], .limits[2], length = guide$nbin)
    if (length(.bar) == 0) {
      .bar = unique(.limits)
    }
    guide$bar <- data.frame(colour = scale$map(.bar), value = .bar,
                            stringsAsFactors = FALSE)
    if (guide$reverse) {
      guide$key <- guide$key[nrow(guide$key):1, ]
      guide$bar <- guide$bar[nrow(guide$bar):1, ]
    }
    guide$hash <- with(guide, digest::digest(list(title, key$.label,
                                                  bar, name)))
  }
  if(guide$title == "vctrs::vec_data(hilo)$level") guide$title <- "level"
  guide
}

#' @export
#' @importFrom ggplot2 guide_geom
#' @rdname guide-helpers
guide_geom.guide_level <- function (guide, layers, default_mapping)
{
  class(guide) <- c("guide", "legend")
  guide <- guide_geom(guide, layers, default_mapping)
  guide$geoms <- lapply(guide$geoms, function(x){
    x$draw_key <- ggplot2::ggproto(NULL, NULL,
                                   draw_key = function(data, params, size){
                                     lwd <- min(data$size, min(size) / 4)
                                     fillcol <- data$level #blendHex(data$col, data$level, 0.7)
                                     grid::rectGrob(
                                       width = grid::unit(1, "npc") - grid::unit(lwd, "mm"),
                                       height = grid::unit(1, "npc") - grid::unit(lwd, "mm"),
                                       gp = grid::gpar(
                                         col = fillcol,
                                         fill = scales::alpha(fillcol, data$alpha),
                                         lty = data$linetype,
                                         lwd = lwd * ggplot2::.pt,
                                         linejoin = "mitre"
                                       )
                                     )
                                   })$draw_key
    x
  })
  guide
}

#' @export
density.dist_default <- function(x, ...){
  abort(
    sprintf("The distribution class `%s` does not support `density()`",
            class(x)[1])
  )
}

#' @export
log_density.dist_default <- function(x, ...){
  log(density(x, ...))
}

#' @export
quantile.dist_default <- function(x, p, ...){
  # abort(
  #   sprintf("The distribution class `%s` does not support `quantile()`",
  #           class(x)[1])
  # )
  stats::optim(0, function(pos){
    (p - cdf(x, pos, ...))^2
  })$par
}
#' @export
log_quantile.dist_default <- function(x, p, ...){
  quantile(x, exp(p), ...)
}
#' @export
cdf.dist_default <- function(x, q, times = 1e5,...){
  # Use Monte Carlo integration
  r <- generate(x, times = times)
  if(is.list(q)) {
    # Turn into matrix
    q <- do.call(rbind, q)
  }
  out <- numeric(NROW(q))
  for(i in seq_along(out)) {
    out[i] <- mean(apply(sweep(r, 2, q[i,]) < 0, 1, all))
  }
  return(out)
}
#' @export
log_cdf.dist_default <- function(x, q, ...){
  log(cdf(x, q, ...))
}

#' @export
generate.dist_default <- function(x, times, ...){
  vapply(stats::runif(times,0,1), quantile, numeric(1L), x = x, ...)
}

#' @export
likelihood.dist_default <- function(x, sample, ...){
  prod(vapply(sample, density, numeric(1L), x = x))
}

#' @export
log_likelihood.dist_default <- function(x, sample, ...){
  sum(vapply(sample, log_density, numeric(1L), x = x))
}

#' @export
parameters.dist_default <- function(x, ...) {
  # Reduce parameter structures to length 1 list if needed.
  lapply(unclass(x), function(z) {
    if(inherits(z, "dist_default")) wrap_dist(list(z))
    else if (tryCatch(vec_size(z), error = function(e) Inf) > 1) list(z)
    else z
  })
}

#' @export
family.dist_default <- function(object, ...) {
  substring(class(object)[1], first = 6)
}

#' @export
support.dist_default <- function(x, ...) {
  lims <- quantile(x, c(0, 1))
  closed <- if(any(is.na(lims))) {
    c(FALSE, FALSE)
  } else {
    # Default to open limits on error
    lim_dens <- tryCatch(
      suppressWarnings(density(x, lims)),
      error = function(e) c(0,0)
    )
    !near(lim_dens, 0)
  }
  new_support_region(
    list(vctrs::vec_init(generate(x, 1), n = 0L)),
    list(lims),
    list(closed)
  )
}

#' @export
mean.dist_default <- function(x, ...){
  x_sup <- support(x)
  dist_type <- field(x_sup, "x")[[1]]
  if (!is.numeric(dist_type)) return(NA_real_)
  if (is.double(dist_type)) {
    limits <- field(x_sup, "lim")[[1]]
    tryCatch(
      stats::integrate(function(at) density(x, at) * at, limits[1], limits[2])$value,
      error = function(e) NA_real_
    )
  } else {
    mean(quantile(x, stats::ppoints(1000)), na.rm = TRUE)
  }
}
#' @export
variance.dist_default <- function(x, ...){
  x <- covariance(x, ...)
  if(is.matrix(x[[1]]) && ncol(x[[1]]) > 1){
    matrix(diag(x[[1]]), nrow = 1)
  } else x
}
#' @export
covariance.dist_default <- function(x, ...){
  x_sup <- support(x)
  dist_type <- field(x_sup, "x")[[1]]
  if (!is.numeric(dist_type)) return(NA_real_)
  else if (is.matrix(dist_type)) stats::cov(generate(x, times = 1000))
  else if (is.double(dist_type)) {
    limits <- field(x_sup, "lim")[[1]]
    tryCatch(
      stats::integrate(function(at) density(x, at) * at^2, limits[1], limits[2])$value,
      error = function(e) NA_real_
    ) - mean(x)^2
  } else {
    stats::var(quantile(x, stats::ppoints(1000)), na.rm = TRUE)
  }
}

#' @export
median.dist_default <- function(x, na.rm = FALSE, ...){
  quantile(x, p = 0.5, ...)
}

#' @export
hilo.dist_default <- function(x, size = 95, ...){
  lower <- quantile(x, 0.5-size/200, ...)
  upper <- quantile(x, 0.5+size/200, ...)
  if(is.matrix(lower) && is.matrix(upper)) {
    return(
      vctrs::new_data_frame(split(
        new_hilo(drop(lower), drop(upper), size = rep_len(size, length(lower))),
      seq_along(lower)))
    )
  }
  new_hilo(lower, upper, size)
}

#' @export
hdr.dist_default <- function(x, size = 95, n = 512, ...){
  dist_x <- quantile(x, seq(0.5/n, 1 - 0.5/n, length.out = n))
  # Remove duplicate values of dist_x from less continuous distributions
  dist_x <- unique(dist_x)
  dist_y <- density(x, dist_x)
  alpha <- quantile(dist_y, probs = 1-size/100)

  crossing_alpha <- function(alpha, x, y){
    it <- seq_len(length(y) - 1)
    dd <- y - alpha
    dd <- dd[it + 1] * dd[it]
    index <- it[dd <= 0]
    # unique() removes possible duplicates if sequential dd has same value.
    # More robust approach is required.
    out <- unique(
      vapply(
        index,
        function(.x) stats::approx(y[.x + c(0,1)], x[.x + c(0,1)], xout = alpha)$y,
        numeric(1L)
      )
    )
    # Add boundary values which may exceed the crossing point.
    c(x[1][y[1]>alpha], out, x[length(x)][y[length(y)]>alpha])
  }

  # purrr::map(alpha, crossing_alpha, dist_x, dist_y)
  hdr <- crossing_alpha(alpha, dist_x, dist_y)
  lower_hdr <- seq_along(hdr)%%2==1
  new_hdr(lower = list(hdr[lower_hdr]), upper = list(hdr[!lower_hdr]), size = size)
}

#' @export
format.dist_default <- function(x, ...){
  "?"
}

#' @export
print.dist_default <- function(x, ...){
  cat(format(x, ...))
}

#' @export
dim.dist_default <- function(x){
  # Quick and dirty dimension calculation
  NCOL(generate(x, times = 1))
}

invert_fail <- function(...) stop("Inverting transformations for distributions is not yet supported.")

#' Attempt to get the inverse of f(x) by name. Returns invert_fail
#' (a function that raises an error if called) if there is no known inverse.
#' @param f string. Name of a function.
#' @noRd
get_unary_inverse <- function(f) {
  switch(f,
    sqrt = function(x) x^2,
    exp = log,
    log = function(x, base = exp(1)) base ^ x,
    log2 = function(x) 2^x,
    log10 = function(x) 10^x,
    expm1 = log1p,
    log1p = expm1,
    cos = acos,
    sin = asin,
    tan = atan,
    acos = cos,
    asin = sin,
    atan = tan,
    cosh = acosh,
    sinh = asinh,
    tanh = atanh,
    acosh = cosh,
    asinh = sinh,
    atanh = tanh,

    invert_fail
  )
}

#' Attempt to get the inverse of f(x, constant) by name. Returns invert_fail
#' (a function that raises an error if called) if there is no known inverse.
#' @param f string. Name of a function.
#' @param constant a constant value
#' @noRd
get_binary_inverse_1 <- function(f, constant) {
  force(constant)

  switch(f,
    `+` = function(x) x - constant,
    `-` = function(x) x + constant,
    `*` = function(x) x / constant,
    `/` = function(x) x * constant,
    `^` = function(x) x ^ (1/constant),

    invert_fail
  )
}

#' Attempt to get the inverse of f(constant, x) by name. Returns invert_fail
#' (a function that raises an error if called) if there is no known inverse.
#' @param f string. Name of a function.
#' @param constant a constant value
#' @noRd
get_binary_inverse_2 <- function(f, constant) {
  force(constant)

  switch(f,
    `+` = function(x) x - constant,
    `-` = function(x) constant - x,
    `*` = function(x) x / constant,
    `/` = function(x) constant / x,
    `^` = function(x) log(x, base = constant),

    invert_fail
  )
}

#' @method Math dist_default
#' @export
Math.dist_default <- function(x, ...) {
  if(dim(x) > 1) stop("Transformations of multivariate distributions are not yet supported.")

  trans <- new_function(exprs(x = ), body = expr((!!sym(.Generic))(x, !!!dots_list(...))))

  inverse_fun <- get_unary_inverse(.Generic)
  inverse <- new_function(exprs(x = ), body = expr((!!inverse_fun)(x, !!!dots_list(...))))

  vec_data(dist_transformed(wrap_dist(list(x)), trans, inverse))[[1]]
}

#' @method Ops dist_default
#' @export
Ops.dist_default <- function(e1, e2) {
  if(.Generic %in% c("-", "+") && missing(e2)){
    e2 <- e1
    e1 <- if(.Generic == "+") 1 else -1
    .Generic <- "*"
  }
  is_dist <- c(inherits(e1, "dist_default"), inherits(e2, "dist_default"))
  if(any(vapply(list(e1, e2)[is_dist], dim, numeric(1L)) > 1)){
    stop("Transformations of multivariate distributions are not yet supported.")
  }

  trans <- if(all(is_dist)) {
    if(identical(e1$dist, e2$dist)){
      new_function(exprs(x = ), expr((!!sym(.Generic))((!!e1$transform)(x), (!!e2$transform)(x))))
    } else {
      stop(sprintf("The %s operation is not supported for <%s> and <%s>", .Generic, class(e1)[1], class(e2)[1]))
    }
  } else if(is_dist[1]){
    new_function(exprs(x = ), body = expr((!!sym(.Generic))(x, !!e2)))
  } else {
    new_function(exprs(x = ), body = expr((!!sym(.Generic))(!!e1, x)))
  }

  inverse <- if(all(is_dist)) {
    invert_fail
  } else if(is_dist[1]){
    get_binary_inverse_1(.Generic, e2)
  } else {
    get_binary_inverse_2(.Generic, e1)
  }

  vec_data(dist_transformed(wrap_dist(list(e1,e2)[which(is_dist)]), trans, inverse))[[1]]
}

test_that("hilo of transformed distributions", {
  expect_identical(
    hilo(exp(dist_poisson(3))),
    exp(hilo((dist_poisson(3))))
  )
})

test_that("chains of transformations", {
  expect_identical(
    hilo(dist_student_t(5)),
    hilo(log(exp(dist_student_t(5))))
  )

  expect_output(
    print(exp(dist_student_t(5))-1),
    "t\\(t\\(5, 0, 1\\)\\)"
  )
})

test_that("handling of transformation arguments", {
  expect_identical(
    hilo(logb(dist_uniform(0, 100), base = 10)),
    logb(hilo(dist_uniform(0, 100)), base = 10)
  )

  expect_identical(
    hilo(10^logb(dist_uniform(0, 100), base = 10)),
    10^logb(hilo(dist_uniform(0, 100)), base = 10)
  )
})

test_that("LogNormal distributions", {
  dist <- dist_transformed(dist_normal(0, 0.5), exp, log)
  ln_dist <- dist_lognormal(0, 0.5)

  # Test exp() shortcut
  expect_identical(
    exp(dist_normal(0, 0.5)),
    ln_dist
  )
  expect_identical(
    log(ln_dist),
    dist_normal(0, 0.5)
  )

  # Test log() shortcut with different bases
  expect_equal(log(dist_lognormal(0, log(3)), base = 3), dist_normal(0, 1))
  expect_equal(log2(dist_lognormal(0, log(2))), dist_normal(0, 1))
  expect_equal(log10(dist_lognormal(0, log(10))), dist_normal(0, 1))

  # format
  expect_equal(format(dist), sprintf("t(%s)", format(dist_normal(0, 0.5))))

  # quantiles
  expect_equal(
    quantile(dist, c(0.1, 0.5)),
    quantile(ln_dist, c(0.1, 0.5))
  )

  # pdf
  expect_equal(
    density(dist, c(1, 20)),
    density(ln_dist, c(1, 20))
  )

  # cdf
  expect_equal(
    cdf(dist, c(4, 90)),
    cdf(ln_dist, c(4, 90))
  )

  # F(Finv(a)) ~= a
  expect_equal(cdf(dist, quantile(dist, 0.372)), 0.372, tolerance = 1e-3)

  # stats (approximate due to bias adjustment method)
  expect_equal(mean(dist), exp(0.25/2), tolerance = 0.01)
  expect_equal(variance(dist), (exp(0.25) - 1)*exp(0.25), tolerance = 0.1)
})

test_that("inverses are applied automatically", {
  dist <- dist_gamma(1,1)
  log2dist <- log(dist, base = 2)
  log2dist_t <- dist_transformed(dist, log2, function(x) 2 ^ x)

  expect_equal(density(log2dist, 0.5), density(log2dist_t, 0.5))
  expect_equal(cdf(log2dist, 0.5), cdf(log2dist_t, 0.5))
  expect_equal(quantile(log2dist, 0.5), quantile(log2dist_t, 0.5))

  # test multiple transformations that get stacked together by dist_transformed
  explogdist <- exp(log(dist))
  expect_equal(density(dist, 0.5), density(explogdist, 0.5))
  expect_equal(cdf(dist, 0.5), cdf(explogdist, 0.5))
  expect_equal(quantile(dist, 0.5), quantile(explogdist, 0.5))

  # test multiple transformations created by operators (via Ops)
  explog2dist <- 2 ^ log2dist
  expect_equal(density(dist, 0.5), density(explog2dist, 0.5))
  expect_equal(cdf(dist, 0.5), cdf(explog2dist, 0.5))
  expect_equal(quantile(dist, 0.5), quantile(explog2dist, 0.5))

  # basic set of inverses
  expect_equal(density(sqrt(dist^2), 0.5), density(dist, 0.5))
  expect_equal(density(exp(log(dist)), 0.5), density(dist, 0.5))
  expect_equal(density(10^(log10(dist)), 0.5), density(dist, 0.5))
  expect_equal(density(expm1(log1p(dist)), 0.5), density(dist, 0.5))
  expect_equal(density(cos(acos(dist)), 0.5), density(dist, 0.5))
  expect_equal(density(sin(asin(dist)), 0.5), density(dist, 0.5))
  expect_equal(density(tan(atan(dist)), 0.5), density(dist, 0.5))
  expect_equal(density(cosh(acosh(dist + 1)) - 1, 0.5), density(dist, 0.5))
  expect_equal(density(sinh(asinh(dist)), 0.5), density(dist, 0.5))
  expect_equal(density(tanh(atanh(dist)), 0.5), density(dist, 0.5))

  expect_equal(density(dist + 1 - 1, 0.5), density(dist, 0.5))
  expect_equal(density(dist * 2 / 2, 0.5), density(dist, 0.5))

  # inverting a gamma distribution
  skip_if_not_installed("actuar")
  expect_equal(density(1/dist_gamma(4, 3), 0.5), density(dist_inverse_gamma(4, 1/3), 0.5))
  expect_equal(density(1/(1/dist_gamma(4, 3)), 0.5), density(dist_gamma(4, 3), 0.5))

})

test_that("transformed distributions' density is 0 outside of the support region", {
  dist <- dist_wrap('norm')
  expect_equal(density(exp(dist), 0)[[1]], 0)
  expect_equal(density(exp(dist), -1)[[1]], 0)

  dist <- dist_wrap('gamma', shape = 1, rate = 1)
  expect_equal(density(exp(dist), 0)[[1]], 0)
  expect_equal(density(exp(dist), 1)[[1]], 1)
})


test_that("transformed distributions' cdf is 0/1 outside of the support region", {
  dist <- dist_wrap('norm')
  expect_equal(cdf(exp(dist), 0)[[1]], 0)
  expect_equal(cdf(exp(dist), -1)[[1]], 0)
  expect_equal(cdf(-1*exp(dist), 0)[[1]], 1)
  expect_equal(cdf(-1*exp(dist), 2)[[1]], 1)
})

test_that("unary negation operator works", {
  dist <- dist_normal(1,1)
  expect_equal(density(-dist, 0.5), density(dist, -0.5))

  dist <- dist_wrap('norm', mean = 1)
  expect_equal(density(-dist, 0.5), density(dist, -0.5))

  dist <- dist_student_t(3, mu = 1)
  expect_equal(density(-dist, 0.5), density(dist, -0.5))
})

test_that("transformed distributions pdf integrates to 1", {
  dist_names <- c('norm', 'gamma', 'beta', 'chisq', 'exp',
                  'logis', 't', 'unif', 'weibull')
  dist_args <- list(list(mean = 1, sd = 1), list(shape = 2, rate = 1),
                    list(shape1 = 3, shape2 = 5), list(df = 5),
                    list(rate = 1),
                    list(location = 1.5, scale = 1), list(df = 10),
                    list(min = 0, max = 1), list(shape = 3, scale = 1))
  names(dist_args) <- dist_names
  dist <- lapply(dist_names, function(x) do.call(dist_wrap, c(x, dist_args[[x]])))
  dist <- do.call(c, dist)
  dfun <- function(x, id, transform) density(get(transform)(dist[id]), x)[[1]]
  twoexp <- function(x) 2^x
  square <- function(x) x^2
  mult2 <- function(x) 2*x
  identity <- function(x) x
  tol <- 1e-5
  for (i in 1:length(dist)) {
    expect_equal(integrate(dfun, -Inf, Inf, id = i, transform = 'identity')$value, 1, tolerance = tol)
    expect_equal(integrate(dfun, -Inf, Inf, id = i, transform = 'exp')$value, 1, tolerance = tol)
    expect_equal(integrate(dfun, -Inf, Inf, id = i, transform = 'twoexp')$value, 1, tolerance = tol)
    expect_equal(integrate(dfun, -Inf, Inf, id = i, transform = 'mult2')$value, 1, tolerance = tol)
    lower_bound <- field(support(dist[[i]]), "lim")[[1]][1]
    if (near(lower_bound, 0)) {
      expect_equal(integrate(dfun, -Inf, 5, id = i, transform = 'log')$value, 1, tolerance = tol)
      expect_equal(integrate(dfun, -Inf, Inf, id = i, transform = 'square')$value, 1, tolerance = tol)
    }
  }
})


test_that("monotonically decreasing transformations (#100)", {
  dist <- dist_lognormal()

  expect_equal(
    quantile(-dist, 0.2), -quantile(dist, 1 - 0.2)
  )
  expect_equal(
    quantile(1/dist, 0.2), 1/quantile(dist, 1 - 0.2)
  )
  expect_equal(
    quantile(-1/dist, 0.7), -1/quantile(dist, 0.7)
  )

  expect_equal(
    cdf(-dist, -2), 1 - cdf(dist, 2)
  )
  expect_equal(
    cdf(1/dist, 2), 1 - cdf(dist, 1/2)
  )
  expect_equal(
    cdf(-1/dist, -2), cdf(dist, 1/2)
  )
})

#' The Poisson Distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Poisson distributions are frequently used to model counts.
#'
#' @inheritParams stats::dpois
#'
#'
#' @details
#'
#'   We recommend reading this documentation on
#'   <https://pkg.mitchelloharawild.com/distributional/>, where the math
#'   will render nicely.
#'
#'   In the following, let \eqn{X} be a Poisson random variable with parameter
#'   `lambda` = \eqn{\lambda}.
#'
#'   **Support**: \eqn{\{0, 1, 2, 3, ...\}}{{0, 1, 2, 3, ...}}
#'
#'   **Mean**: \eqn{\lambda}
#'
#'   **Variance**: \eqn{\lambda}
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     P(X = k) = \frac{\lambda^k e^{-\lambda}}{k!}
#'   }{
#'     P(X = k) = \lambda^k e^(-\lambda) / k!
#'   }
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     P(X \le k) = e^{-\lambda}
#'     \sum_{i = 0}^{\lfloor k \rfloor} \frac{\lambda^i}{i!}
#'   }{
#'     P(X \le k) = e^(-\lambda)
#'     \sum_{i = 0}^k \lambda^i / i!
#'   }
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = e^{\lambda (e^t - 1)}
#'   }{
#'     E(e^(tX)) = e^(\lambda (e^t - 1))
#'   }
#' @seealso [stats::Poisson]
#'
#' @examples
#' dist <- dist_poisson(lambda = c(1, 4, 10))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_poisson
#' @export
dist_poisson <- function(lambda){
  lambda <- vec_cast(lambda, double())
  if(any(lambda < 0)){
    abort("The lambda parameter of an Poisson distribution must be non-negative.")
  }
  new_dist(l = lambda, class = "dist_poisson")
}

#' @export
format.dist_poisson <- function(x, digits = 2, ...){
  sprintf(
    "Pois(%s)",
    format(x[["l"]], digits = digits, ...)
  )
}

#' @export
density.dist_poisson <- function(x, at, ...){
  stats::dpois(at, x[["l"]])
}

#' @export
log_density.dist_poisson <- function(x, at, ...){
  stats::dpois(at, x[["l"]], log = TRUE)
}

#' @export
quantile.dist_poisson <- function(x, p, ...){
  stats::qpois(p, x[["l"]])
}

#' @export
cdf.dist_poisson <- function(x, q, ...){
  stats::ppois(q, x[["l"]])
}

#' @export
generate.dist_poisson <- function(x, times, ...){
  as.integer(stats::rpois(times, x[["l"]]))
}

#' @export
mean.dist_poisson <- function(x, ...){
  x[["l"]]
}

#' @export
covariance.dist_poisson <- function(x, ...){
  x[["l"]]
}

#' @export
skewness.dist_poisson <- function(x, ...) 1 / sqrt(x[["l"]])

#' @export
kurtosis.dist_poisson <- function(x, ...) 1 / x[["l"]]

#' The Logarithmic distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Logarithmic distribution is a discrete probability distribution
#' derived from the logarithmic series. It is useful in modeling the
#' abundance of species and other phenomena where the frequency of
#' an event follows a logarithmic pattern.
#'
#' @inheritParams actuar::dlogarithmic
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_logarithmic")`
#'
#'   In the following, let \eqn{X} be a Logarithmic random variable with
#'   parameter `prob` = \eqn{p}.
#'
#'   **Support**: \eqn{\{1, 2, 3, ...\}}{{1, 2, 3, ...}}
#'
#'   **Mean**: \eqn{\frac{-1}{\log(1-p)} \cdot \frac{p}{1-p}}{-1/log(1-p) * p/(1-p)}
#'
#'   **Variance**: \eqn{\frac{-(p^2 + p\log(1-p))}{[(1-p)\log(1-p)]^2}}{-(p^2 + p*log(1-p))/((1-p)*log(1-p))^2}
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     P(X = k) = \frac{-1}{\log(1-p)} \cdot \frac{p^k}{k}
#'   }{
#'     P(X = k) = -1/log(1-p) * p^k/k
#'   }
#'
#'   for \eqn{k = 1, 2, 3, \ldots}{k = 1, 2, 3, ...}
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   The c.d.f. does not have a simple closed form. It is computed
#'   using the recurrence relationship
#'   \eqn{P(X = k+1) = \frac{p \cdot k}{k+1} \cdot P(X = k)}{P(X = k+1) = (p*k)/(k+1) * P(X = k)}
#'   starting from \eqn{P(X = 1) = \frac{-p}{\log(1-p)}}{P(X = 1) = -p/log(1-p)}.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \frac{\log(1 - pe^t)}{\log(1-p)}
#'   }{
#'     E(e^(tX)) = log(1 - p*e^t)/log(1-p)
#'   }
#'
#'   for \eqn{pe^t < 1}{p*e^t < 1}
#'
#' @seealso [actuar::Logarithmic]
#'
#' @examples
#' dist <- dist_logarithmic(prob = c(0.33, 0.66, 0.99))
#' dist
#'
#' @examplesIf requireNamespace("actuar", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#' @name dist_logarithmic
#' @export
dist_logarithmic <- function(prob){
  prob <- vec_cast(prob, double())
  if(any((prob < 0) | (prob > 1))){
    abort("The prob parameter of a Logarithmic distribution must be between 0 and 1.")
  }
  new_dist(p = prob, class = "dist_logarithmic")
}

#' @export
format.dist_logarithmic <- function(x, digits = 2, ...){
  sprintf(
    "Logarithmic(%s)",
    format(x[["p"]], digits = digits, ...)
  )
}

#' @export
density.dist_logarithmic <- function(x, at, ...){
  require_package("actuar")
  actuar::dlogarithmic(at, x[["p"]])
}

#' @export
log_density.dist_logarithmic <- function(x, at, ...){
  require_package("actuar")
  actuar::dlogarithmic(at, x[["p"]], log = TRUE)
}

#' @export
quantile.dist_logarithmic <- function(x, p, ...){
  require_package("actuar")
  actuar::qlogarithmic(p, x[["p"]])
}

#' @export
cdf.dist_logarithmic <- function(x, q, ...){
  require_package("actuar")
  actuar::plogarithmic(q, x[["p"]])
}

#' @export
generate.dist_logarithmic <- function(x, times, ...){
  require_package("actuar")
  actuar::rlogarithmic(times, x[["p"]])
}

#' @export
mean.dist_logarithmic <- function(x, ...){
  p <- x[["p"]]
  (-1/(log(1-p)))*(p/(1-p))
}

#' @export
covariance.dist_logarithmic <- function(x, ...){
  p <- x[["p"]]
  -(p^2 + p*log(1-p))/((1-p)*log(1-p))^2
}

#' Construct hilo intervals
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Class constructor function to help with manually creating hilo interval
#' objects.
#'
#' @param lower,upper A numeric vector of values for lower and upper limits.
#' @param size Size of the interval between \[0, 100\].
#'
#' @return A "hilo" vector
#'
#' @author Earo Wang & Mitchell O'Hara-Wild
#'
#' @examples
#' new_hilo(lower = rnorm(10), upper = rnorm(10) + 5, size = 95)
#'
#' @export
new_hilo <- function(lower = double(), upper = double(), size = double()) {
  vec_assert(size, double())
  if (any(size < 0 | size > 100, na.rm = TRUE))
    abort("'size' must be between [0, 100].")

  out <- vec_recycle_common(lower = lower, upper = upper)
  if(vec_is(lower, double()) && vec_is(upper, double())) {
    if (any(out[["upper"]] < out[["lower"]], na.rm = TRUE)) {
      abort("`upper` can't be lower than `lower`.")
    }
  }
  len <- vec_size(out[[1]])
  out[["level"]] <- vctrs::vec_recycle(size, len)

  vctrs::new_rcrd(out, class = "hilo")
}

#' Compute intervals
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Used to extract a specified prediction interval at a particular confidence
#' level from a distribution.
#'
#' The numeric lower and upper bounds can be extracted from the interval using
#' `<hilo>$lower` and `<hilo>$upper` as shown in the examples below.
#'
#' @param x Object to create hilo from.
#' @param ... Additional arguments used by methods.
#'
#' @examples
#' # 95% interval from a standard normal distribution
#' interval <- hilo(dist_normal(0, 1), 95)
#' interval
#'
#' # Extract the individual quantities with `$lower`, `$upper`, and `$level`
#' interval$lower
#' interval$upper
#' interval$level
#' @export
hilo <- function(x, ...){
  UseMethod("hilo")
}

#' @export
hilo.default <- function(x, ...){
  abort(sprintf(
    "Objects of type `%s` are not supported by `hilo()`, you can create a custom `hilo` with `new_hilo()`",
    class(x)
  ))
}

#' Is the object a hilo
#'
#' @param x An object.
#'
#' @export
is_hilo <- function(x) {
  inherits(x, "hilo")
}

#' @export
format.hilo <- function(x, justify = "right", ...) {
  lwr <- field(x, "lower")
  upr <- field(x, "upper")
  if(is.matrix(lwr)) {
    lwr <- if(ncol(lwr) > 1) vctrs::vec_ptype_abbr(lwr) else drop(lwr)
  }
  if(is.matrix(upr)) {
    upr <- if(ncol(upr) > 1) vctrs::vec_ptype_abbr(upr) else drop(upr)
  }
  limit <- paste(
    format(lwr, justify = justify, ...),
    format(upr, justify = justify, ...),
    sep = ", "
  )
  paste0("[", limit, "]", field(x, "level"))
}

#' @export
is.na.hilo <- function(x) {
  # both lower and upper are NA's
  x <- vec_data(x)
  is.na(x$lower) & is.na(x$upper)
}

#' @export
vec_ptype2.hilo.hilo <- function(x, y, ...){
  x
}

#' @export
vec_cast.character.hilo <- function(x, to, ...){
  sprintf(
    "[%s, %s]%s",
    as.character(x$lower), as.character(x$upper), as.character(x$level)
  )
}

#' @method vec_math hilo
#' @export
vec_math.hilo <- function(.fn, .x, ...){
  out <- vec_data(.x)
  if(.fn == "mean")
    abort("Cannot compute the mean of hilo intervals.")
  out[["lower"]] <- get(.fn)(out[["lower"]], ...)
  out[["upper"]] <- get(.fn)(out[["upper"]], ...)
  if(.fn %in% c("is.nan", "is.finite", "is.infinite"))
    return(out[["lower"]] | out[["upper"]])
  vec_restore(out, .x)
}

#' @method vec_arith hilo
#' @export
vec_arith.hilo <- function(op, x, y, ...){
  out <- dt_x <- vec_data(x)
  if(is_hilo(y)){
    abort("Intervals should not be added to other intervals, the sum of intervals is not the interval from a sum of distributions.")
  }
  else if(is_empty(y)){
    if(op == "-"){
      out[["upper"]] <- get(op)(dt_x[["lower"]])
      out[["lower"]] <- get(op)(dt_x[["upper"]])
    }
  }
  else{
    out[["lower"]] <- get(op)(dt_x[["lower"]], y)
    out[["upper"]] <- get(op)(dt_x[["upper"]], y)
  }
  vec_restore(out, x)
}

#' @method vec_arith.numeric hilo
#' @export
vec_arith.numeric.hilo <- function(op, x, y, ...){
  out <- hl <- vec_data(y)
  out[["lower"]] <- get(op)(x, hl[["lower"]])
  out[["upper"]] <- get(op)(x, hl[["upper"]])
  if(x < 0 && op %in% c("*", "/")){
    out[c("lower", "upper")] <- out[c("upper", "lower")]
  }
  vec_restore(out, y)
}

#' @importFrom utils .DollarNames
#' @export
.DollarNames.hilo <- function(x, pattern){
  utils::.DollarNames(vec_data(x), pattern)
}

#' @export
`$.hilo` <- function(x, name){
  field(x, name)
}

#' @export
`names<-.hilo` <- function(x, value) {
  # abort("A <hilo> object cannot be named.")
  x
}

test_that("hilo", {
  # defaults
  hl <- new_hilo()
  expect_length(hl, 0)

  # display
  expect_output(print(hl), "<hilo\\[0\\]>")
  expect_output(print(new_hilo(0,1,95)), "\\[0, 1\\]95")

  dist <- dist_normal()

  # hilo.distribution
  hl <- hilo(dist, 95)
  expect_length(hl, 1)
  expect_equal(hl, new_hilo(qnorm(0.025), qnorm(0.975), 95))

  # vec_math.hilo
  expect_equal(is.na(hl), FALSE)
  expect_equal(is.nan(hl), FALSE)

  # vec_arith.hilo
  expect_equal(
    hl*3+1,
    new_hilo(qnorm(0.025)*3+1, qnorm(0.975)*3+1, 95)
  )
  expect_equal(-hl, hl)
  expect_equal(-3*hl, hl/(1/3))
})

#' Inflate a value of a probability distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' @param dist The distribution(s) to inflate.
#' @param prob The added probability of observing `x`.
#' @param x The value to inflate. The default of `x = 0` is for zero-inflation.
#'
#' @name dist_inflated
#' @export
dist_inflated <- function(dist, prob, x = 0){
  vec_is(dist, new_dist())
  if(prob < 0 || prob > 1){
    abort("The inflation probability must be between 0 and 1.")
  }
  new_dist(dist = dist, x = x, p = prob,
           dimnames = dimnames(dist), class = "dist_inflated")
}

#' @export
format.dist_inflated <- function(x, ...){
  sprintf(
    "%s+%s",
    format(x[["x"]]),
    format(x[["dist"]])
  )
}

#' @export
density.dist_inflated <- function(x, at, ...){
  x[["p"]]*(at==x[["x"]]) + (1-x[["p"]])*density(x[["dist"]], at, ...)
}

#' @export
quantile.dist_inflated <- function(x, p, ...){
  qt <- quantile(x[["dist"]], pmax(0, (p - x[["p"]]) / (1-x[["p"]])), ...)
  if(qt >= x[["x"]]) return(qt)
  qt <- quantile(x[["dist"]], p, ...)
  if(qt < x[["x"]]) qt else x[["x"]]
}

#' @export
cdf.dist_inflated <- function(x, q, ...){
  x[["p"]]*(q>=x[["x"]]) + (1-x[["p"]])*cdf(x[["dist"]], q, ...)
}

#' @export
generate.dist_inflated <- function(x, times, ...){
  p <- x[["p"]]
  inf <- stats::runif(times) < p
  r <- vec_init(x[["x"]], times)
  r[inf] <- x[["x"]]
  r[!inf] <- generate(x[["dist"]], sum(!inf))
  r
}

#' @export
mean.dist_inflated <- function(x, ...){
  # Can't compute if inflation value is not numeric
  if(!vec_is(x[["x"]], numeric())) return(NA_real_)

  p <- x[["p"]]
  p*x[["x"]] + (1-p)*mean(x[["dist"]])
}

#' @export
covariance.dist_inflated <- function(x, ...){
  # Can't compute if inflation value is not numeric
  if(!vec_is(x[["x"]], numeric())) return(NA_real_)
  # Can't (easily) compute if inflation value is not zero
  if(x[["x"]] != 0) return(NA_real_)

  m1 <- mean(x[["dist"]])
  v <- variance(x[["dist"]])
  m2 <- v + m1^2
  p <- x[["p"]]
  (1-p)*v + p*(1-p)*m1^2
}

#' The Generalized Pareto Distribution
#'
#' @description
#' The GPD distribution is commonly used to model the tails of distributions,
#' particularly in extreme value theory.
#'
#' The Pickands–Balkema–De Haan theorem states that for a large class of
#' distributions, the tail (above some threshold) can be approximated by a GPD.
#'
#' @param location the location parameter \eqn{a} of the GPD distribution.
#' @param scale the scale parameter \eqn{b} of the GPD distribution.
#' @param shape the shape parameter \eqn{s} of the GPD distribution.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_gpd")`
#'
#' In the following, let \eqn{X} be a Generalized Pareto random variable with
#' parameters `location` = \eqn{a}, `scale` = \eqn{b > 0}, and
#' `shape` = \eqn{s}.
#'
#' **Support**:
#' \eqn{x \ge a} if \eqn{s \ge 0},
#' \eqn{a \le x \le a - b/s} if \eqn{s < 0}
#'
#' **Mean**:
#' \deqn{
#'   E(X) = a + \frac{b}{1 - s} \quad \textrm{for } s < 1
#' }{
#'   E(X) = a + b/(1 - s) for s < 1
#' }
#' \eqn{E(X) = \infty} for \eqn{s \ge 1}
#'
#' **Variance**:
#' \deqn{
#'   \textrm{Var}(X) = \frac{b^2}{(1-s)^2(1-2s)} \quad \textrm{for } s < 0.5
#' }{
#'   Var(X) = b^2 / ((1-s)^2 (1-2s)) for s < 0.5
#' }
#' \eqn{\textrm{Var}(X) = \infty} for \eqn{s \ge 0.5}
#'
#' **Probability density function (p.d.f)**:
#'
#' For \eqn{s = 0}:
#' \deqn{
#'   f(x) = \frac{1}{b}\exp\left(-\frac{x-a}{b}\right) \quad \textrm{for } x \ge a
#' }{
#'   f(x) = (1/b) exp(-(x-a)/b) for x >= a
#' }
#'
#' For \eqn{s \ne 0}:
#' \deqn{
#'   f(x) = \frac{1}{b}\left(1 + s\frac{x-a}{b}\right)^{-1/s - 1}
#' }{
#'   f(x) = (1/b)(1 + s(x-a)/b)^(-1/s - 1)
#' }
#' where \eqn{1 + s(x-a)/b > 0}
#'
#' **Cumulative distribution function (c.d.f)**:
#'
#' For \eqn{s = 0}:
#' \deqn{
#'   F(x) = 1 - \exp\left(-\frac{x-a}{b}\right) \quad \textrm{for } x \ge a
#' }{
#'   F(x) = 1 - exp(-(x-a)/b) for x >= a
#' }
#'
#' For \eqn{s \ne 0}:
#' \deqn{
#'   F(x) = 1 - \left(1 + s\frac{x-a}{b}\right)^{-1/s}
#' }{
#'   F(x) = 1 - (1 + s(x-a)/b)^(-1/s)
#' }
#' where \eqn{1 + s(x-a)/b > 0}
#'
#' **Quantile function**:
#'
#' For \eqn{s = 0}:
#' \deqn{
#'   Q(p) = a - b\log(1-p)
#' }{
#'   Q(p) = a - b log(1-p)
#' }
#'
#' For \eqn{s \ne 0}:
#' \deqn{
#'   Q(p) = a + \frac{b}{s}\left[(1-p)^{-s} - 1\right]
#' }{
#'   Q(p) = a + (b/s)[(1-p)^(-s) - 1]
#' }
#'
#' **Median**:
#'
#' For \eqn{s = 0}:
#' \deqn{
#'   \textrm{Median}(X) = a + b\log(2)
#' }{
#'   Median(X) = a + b log(2)
#' }
#'
#' For \eqn{s \ne 0}:
#' \deqn{
#'   \textrm{Median}(X) = a + \frac{b}{s}\left(2^s - 1\right)
#' }{
#'   Median(X) = a + (b/s)(2^s - 1)
#' }
#'
#' **Skewness and Kurtosis**: No closed-form expressions; approximated numerically.
#'
#' @seealso [evd::dgpd()]
#' @examples
#' dist <- dist_gpd(location = 0, scale = 1, shape = 0)
#'
#' dist
#' mean(dist)
#' variance(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#' @export
dist_gpd <- function(location, scale, shape) {
  location <- vctrs::vec_cast(unname(location), double())
  shape <- vctrs::vec_cast(unname(shape), double())
  scale <- vctrs::vec_cast(unname(scale), double())
  if (any(scale <= 0)) {
    stop("The scale parameter of a GPD distribution must be strictly positive")
  }
  distributional::new_dist(location = location, scale = scale, shape = shape, class = "dist_gpd")
}

#' @export
format.dist_gpd <- function(x, digits = 2, ...) {
  sprintf(
    "GPD(%s, %s, %s)",
    format(x[["location"]], digits = digits, ...),
    format(x[["scale"]], digits = digits, ...),
    format(x[["shape"]], digits = digits, ...)
  )
}

#' @exportS3Method distributional::log_density
log_density.dist_gpd <- function(x, at, ...) {
  z <- (at - x[["location"]]) / x[["scale"]]
  if (x[["shape"]] == 0) {
    pdf <- -z
  } else {
    xx <- 1 + x[["shape"]] * z
    xx[xx <= 0] <- NA_real_
    pdf <- -(1 / x[["shape"]] + 1) * log(xx)
    pdf[is.na(pdf)] <- -Inf
  }
  if (x[["shape"]] >= 0) {
    pdf[z < 0] <- -Inf
  } else {
    pdf[z < 0 | z > -1 / x[["shape"]]] <- -Inf
  }
  pdf - log(x[["scale"]])
}

#' @exportS3Method stats::density
density.dist_gpd <- function(x, at, ...) {
  exp(log_density.dist_gpd(x, at, ...))
}

#' @exportS3Method distributional::cdf
cdf.dist_gpd <- function(x, q, ...) {
  z <- pmax(q - x[["location"]], 0) / x[["scale"]]
  if (x[["shape"]] == 0) {
    1 - exp(-z)
  } else {
    1 - pmax(1 + x[["shape"]] * z, 0)^(-1 / x[["shape"]])
  }
}

#' @exportS3Method stats::quantile
quantile.dist_gpd <- function(x, p, ...) {
  if (x[["shape"]] == 0) {
    x[["location"]] - x[["scale"]] * log(1 - p)
  } else {
    x[["location"]] + x[["scale"]] * ((1 - p)^(-x[["shape"]]) - 1) / x[["shape"]]
  }
}

#' @exportS3Method distributional::generate
generate.dist_gpd <- function(x, times, ...) {
  if (x[["shape"]] == 0) {
    x[["location"]] + x[["scale"]] * stats::rexp(times)
  } else {
    quantile(x, stats::runif(times))
  }
}

#' @export
mean.dist_gpd <- function(x, ...) {
  if (x[["shape"]] < 1) {
    x[["location"]] + x[["scale"]] / (1 - x[["shape"]])
  } else {
    Inf
  }
}

#' @exportS3Method stats::median
median.dist_gpd <- function(x, ...) {
  if (x[["shape"]] == 0) {
    x[["location"]] - x[["scale"]] * log(0.5)
  } else {
    x[["location"]] + x[["scale"]] * (2^x[["shape"]] - 1) / x[["shape"]]
  }
}

#' @exportS3Method distributional::covariance
covariance.dist_gpd <- function(x, ...) {
  if (x[["shape"]] < 0.5) {
    x[["scale"]]^2 / (1 - x[["shape"]])^2 / (1 - 2 * x[["shape"]])
  } else {
    Inf
  }
}

#' @export
support.dist_gpd <- function(x, ...) {
  shape <- x[["shape"]]
  location <- x[["location"]]
  scale <- x[["scale"]]
  
  # Determine limits based on shape parameter
  if (shape >= 0) {
    # Support: [location, Inf)
    lims <- list(c(location, Inf))
    closed <- list(c(TRUE, FALSE))
  } else {
    # Support: [location, location - scale/shape]
    upper <- location - scale / shape
    lims <- list(c(location, upper))
    closed <- list(c(TRUE, TRUE))
  }
  
  x_proto <- list(vctrs::vec_init(numeric(), n = 0L))
  
  new_support_region(x_proto, lims, closed)
}

#' The Horseshoe distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The horseshoe distribution (Carvalho et al., 2008) is a heavy-tailed
#' continuous distribution defined as a scale mixture of normals. It is
#' primarily used as a shrinkage prior in sparse Bayesian regression, where
#' it concentrates mass near zero while retaining heavy tails that leave
#' large signals unshrunk.
#'
#' @param lambda A positive numeric vector of local scale parameters
#'   \eqn{\lambda > 0} (one per observation).
#' @param tau A positive scalar global scale parameter \eqn{\tau > 0}.
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_horseshoe")`
#'
#'   In the following, let \eqn{X} be a horseshoe random variable with local
#'   scale parameter `lambda` = \eqn{\lambda > 0} and global scale parameter
#'   `tau` = \eqn{\tau > 0}.
#'
#'   **Support**: \eqn{x \in \mathbb{R}}, the set of all real numbers.
#'
#'   **Mean**: \eqn{E(X)} — not available in closed form.
#'
#'   **Variance**: \eqn{\mathrm{Var}(X)} — not available in closed form.
#'
#'   **Probability density function (p.d.f)**:
#'
#'   The horseshoe density does not have a simple closed form but can be
#'   expressed as a scale mixture:
#'
#'   \deqn{
#'     X \mid \lambda, \tau \sim \mathcal{N}(0,\, \lambda^2 \tau^2)
#'   }{
#'     X | lambda, tau ~ N(0, lambda^2 * tau^2)
#'   }
#'
#'   where the half-Cauchy hyperprior \eqn{\lambda \sim C^+(0, 1)} induces the
#'   characteristic horseshoe shrinkage behaviour.
#'
#' @references
#'   Carvalho, C.M., Polson, N.G., and Scott, J.G. (2008).
#'   "The Horseshoe Estimator for Sparse Signals".
#'   *Discussion Paper 2008-31*. Duke University Department of Statistical
#'   Science.
#'
#'   Carvalho, C.M., Polson, N.G., and Scott, J.G. (2009).
#'   "Handling Sparsity via the Horseshoe".
#'   *Journal of Machine Learning Research*, 5, p. 73–80.
#'
#' @seealso [LaplacesDemon::dhs()], [LaplacesDemon::rhs()]
#'
#' @examples
#' dist <- dist_horseshoe(lambda = c(0.5, 1, 2), tau = 1)
#' dist
#'
#' @examplesIf requireNamespace("LaplacesDemon", quietly = TRUE)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 0)
#' density(dist, 0, log = TRUE)
#'
#' @name dist_horseshoe
#' @export
dist_horseshoe <- function(lambda, tau) {
  lambda <- vec_cast(lambda, double())
  tau <- vec_cast(tau, double())
  if (any(lambda <= 0, na.rm = TRUE)) {
    abort("The local scale parameter `lambda` of a horseshoe distribution must be positive.")
  }
  if (any(tau <= 0, na.rm = TRUE)) {
    abort("The global scale parameter `tau` of a horseshoe distribution must be positive.")
  }
  new_dist(lambda = lambda, tau = tau, class = "dist_horseshoe")
}

#' @export
format.dist_horseshoe <- function(x, digits = 2, ...) {
  sprintf(
    "HS(%s, %s)",
    format(x[["lambda"]], digits = digits, ...),
    format(x[["tau"]], digits = digits, ...)
  )
}

#' @export
density.dist_horseshoe <- function(x, at, ..., na.rm = FALSE) {
  require_package("LaplacesDemon")
  LaplacesDemon::dhs(at, x[["lambda"]], x[["tau"]])
}

#' @export
log_density.dist_horseshoe <- function(x, at, ..., na.rm = FALSE) {
  require_package("LaplacesDemon")
  LaplacesDemon::dhs(at, x[["lambda"]], x[["tau"]], log = TRUE)
}

#' @export
generate.dist_horseshoe <- function(x, times, ..., na.rm = FALSE) {
  require_package("LaplacesDemon")
  LaplacesDemon::rhs(times, lambda = x[["lambda"]], tau = x[["tau"]])
}

#' @export
support.dist_horseshoe <- function(x, ...) {
  new_support_region(
    list(0),
    list(c(-Inf, Inf)),
    list(c(FALSE, FALSE))
  )
}

#' The Poisson-Inverse Gaussian distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' The Poisson-Inverse Gaussian distribution is a compound Poisson distribution
#' where the rate parameter follows an Inverse Gaussian distribution. It is
#' useful for modeling overdispersed count data.
#'
#' @inheritParams actuar::dpoisinvgauss
#'
#' @details
#'
#' `r pkgdown_doc_link("dist_poisson_inverse_gaussian")`
#'
#'
#'   In the following, let \eqn{X} be a Poisson-Inverse Gaussian random variable
#'   with parameters `mean` = \eqn{\mu} and `shape` = \eqn{\phi}.
#'
#'   **Support**: \eqn{\{0, 1, 2, 3, ...\}}{{0, 1, 2, 3, ...}}
#'
#'   **Mean**: \eqn{\mu}
#'
#'   **Variance**: \eqn{\frac{\mu}{\phi}(\mu^2 + \phi)}{\mu/\phi (\mu^2 + \phi)}
#'
#'   **Probability mass function (p.m.f)**:
#'
#'   \deqn{
#'     P(X = x) = \frac{e^{\phi}}{\sqrt{2\pi}}
#'     \left(\frac{\phi}{\mu^2}\right)^{x/2}
#'     \frac{1}{x!}
#'     \int_0^\infty u^{x-1/2}
#'     \exp\left(-\frac{\phi u}{2} - \frac{\phi}{2\mu^2 u}\right) du
#'   }{
#'     P(X = x) = (e^\phi / sqrt(2\pi)) (\phi/\mu^2)^(x/2) (1/x!)
#'     \int_0^\infty u^(x-1/2) exp(-\phi u/2 - \phi/(2\mu^2 u)) du
#'   }
#'
#'   for \eqn{x = 0, 1, 2, \ldots}{x = 0, 1, 2, ...}
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   \deqn{
#'     P(X \le x) = \sum_{k=0}^{\lfloor x \rfloor} P(X = k)
#'   }{
#'     P(X <= x) = sum_{k=0}^x P(X = k)
#'   }
#'
#'   The c.d.f does not have a closed form and is approximated numerically.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   \deqn{
#'     E(e^{tX}) = \exp\left\{\phi\left[1 - \sqrt{1 - \frac{2\mu^2}{\phi}(e^t - 1)}\right]\right\}
#'   }{
#'     E(e^(tX)) = exp{\phi [1 - sqrt(1 - 2\mu^2/\phi (e^t - 1))]}
#'   }
#'
#'   for \eqn{t < -\log(1 + \phi/(2\mu^2))}{t < -log(1 + \phi/(2\mu^2))}
#'
#' @seealso [actuar::PoissonInverseGaussian], [actuar::dpoisinvgauss()],
#'   [actuar::ppoisinvgauss()], [actuar::qpoisinvgauss()], [actuar::rpoisinvgauss()]
#'
#' @examples
#' dist <- dist_poisson_inverse_gaussian(mean = rep(0.1, 3), shape = c(0.4, 0.8, 1))
#' dist
#'
#' @examplesIf requireNamespace("actuar", quietly = TRUE)
#' mean(dist)
#' variance(dist)
#' support(dist)
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_poisson_inverse_gaussian
#' @export
dist_poisson_inverse_gaussian <- function(mean, shape){
  mean <- vec_cast(mean, double())
  shape <- vec_cast(shape, double())
  if(any(mean[!is.na(mean)] <= 0)){
    abort("The mean parameter of a Poisson-Inverse Gaussian distribution must be strictly positive.")
  }
  if(any(shape[!is.na(shape)] <= 0)){
    abort("The shape parameter of a Poisson-Inverse Gaussian distribution must be strictly positive.")
  }
  new_dist(m = mean, s = shape, class = "dist_poisson_inverse_gaussian")
}

#' @export
format.dist_poisson_inverse_gaussian <- function(x, digits = 2, ...){
  sprintf(
    "PIG(%s, %s)",
    format(x[["m"]], digits = digits, ...),
    format(x[["s"]], digits = digits, ...)
  )
}

#' @export
density.dist_poisson_inverse_gaussian <- function(x, at, ...){
  require_package("actuar")
  actuar::dpoisinvgauss(at, x[["m"]], x[["s"]])
}

#' @export
log_density.dist_poisson_inverse_gaussian <- function(x, at, ...){
  require_package("actuar")
  actuar::dpoisinvgauss(at, x[["m"]], x[["s"]], log = TRUE)
}

#' @export
quantile.dist_poisson_inverse_gaussian <- function(x, p, ...){
  require_package("actuar")
  actuar::qpoisinvgauss(p, x[["m"]], x[["s"]])
}

#' @export
cdf.dist_poisson_inverse_gaussian <- function(x, q, ...){
  require_package("actuar")
  actuar::ppoisinvgauss(q, x[["m"]], x[["s"]])
}

#' @export
generate.dist_poisson_inverse_gaussian <- function(x, times, ...){
  require_package("actuar")
  actuar::rpoisinvgauss(times, x[["m"]], x[["s"]])
}

#' @export
mean.dist_poisson_inverse_gaussian <- function(x, ...){
  x[["m"]]
}

#' @export
covariance.dist_poisson_inverse_gaussian <- function(x, ...){
  x[["m"]]/x[["s"]] * (x[["m"]]^2 + x[["s"]])
}

#' The (non-central) Chi-Squared Distribution
#'
#' @description
#' `r lifecycle::badge('stable')`
#'
#' Chi-square distributions show up often in frequentist settings
#' as the sampling distribution of test statistics, especially
#' in maximum likelihood estimation settings.
#'
#' @param df Degrees of freedom (non-centrality parameter). Can be any
#'   positive real number.
#' @param ncp Non-centrality parameter. Can be any non-negative real number.
#'   Defaults to 0 (central chi-squared distribution).
#'
#' @details
#'
#'
#' `r pkgdown_doc_link("dist_chisq")`
#'
#'   In the following, let \eqn{X} be a \eqn{\chi^2} random variable with
#'   `df` = \eqn{k} and `ncp` = \eqn{\lambda}.
#'
#'   **Support**: \eqn{R^+}, the set of positive real numbers
#'
#'   **Mean**: \eqn{k + \lambda}
#'
#'   **Variance**: \eqn{2(k + 2\lambda)}
#'
#'   **Probability density function (p.d.f)**:
#'
#'   For the central chi-squared distribution (\eqn{\lambda = 0}):
#'
#'   \deqn{
#'     f(x) = \frac{1}{2^{k/2} \Gamma(k/2)} x^{k/2 - 1} e^{-x/2}
#'   }{
#'     f(x) = 1 / (2^(k/2) \Gamma(k/2)) x^(k/2 - 1) exp(-x/2)
#'   }
#'
#'   For the non-central chi-squared distribution (\eqn{\lambda > 0}):
#'
#'   \deqn{
#'     f(x) = \frac{1}{2} e^{-(x+\lambda)/2} \left(\frac{x}{\lambda}\right)^{k/4-1/2} I_{k/2-1}\left(\sqrt{\lambda x}\right)
#'   }{
#'     f(x) = (1/2) exp(-(x+\lambda)/2) (x/\lambda)^(k/4-1/2) I_(k/2-1)(sqrt(\lambda x))
#'   }
#'
#'   where \eqn{I_\nu(z)} is the modified Bessel function of the first kind.
#'
#'   **Cumulative distribution function (c.d.f)**:
#'
#'   For the central chi-squared distribution (\eqn{\lambda = 0}):
#'
#'   \deqn{
#'     F(x) = \frac{\gamma(k/2, x/2)}{\Gamma(k/2)} = P(k/2, x/2)
#'   }{
#'     F(x) = \gamma(k/2, x/2) / \Gamma(k/2) = P(k/2, x/2)
#'   }
#'
#'   where \eqn{\gamma(s, x)} is the lower incomplete gamma function and
#'   \eqn{P(s, x)} is the regularized gamma function.
#'
#'   For the non-central chi-squared distribution (\eqn{\lambda > 0}):
#'
#'   \deqn{
#'     F(x) = \sum_{j=0}^{\infty} \frac{e^{-\lambda/2} (\lambda/2)^j}{j!} P(k/2 + j, x/2)
#'   }{
#'     F(x) = sum_(j=0)^(\infty) exp(-\lambda/2) (\lambda/2)^j / j! P(k/2 + j, x/2)
#'   }
#'
#'   This is approximated numerically.
#'
#'   **Moment generating function (m.g.f)**:
#'
#'   For the central chi-squared distribution (\eqn{\lambda = 0}):
#'
#'   \deqn{
#'     E(e^{tX}) = (1 - 2t)^{-k/2}, \quad t < 1/2
#'   }{
#'     E(e^(tX)) = (1 - 2t)^(-k/2), t < 1/2
#'   }
#'
#'   For the non-central chi-squared distribution (\eqn{\lambda > 0}):
#'
#'   \deqn{
#'     E(e^{tX}) = \frac{e^{\lambda t / (1 - 2t)}}{(1 - 2t)^{k/2}}, \quad t < 1/2
#'   }{
#'     E(e^(tX)) = exp(\lambda t / (1 - 2t)) / (1 - 2t)^(k/2), t < 1/2
#'   }
#'
#'   **Skewness**:
#'
#'   \deqn{
#'     \gamma_1 = \frac{2^{3/2}(k + 3\lambda)}{(k + 2\lambda)^{3/2}}
#'   }{
#'     \gamma_1 = 2^(3/2) (k + 3\lambda) / (k + 2\lambda)^(3/2)
#'   }
#'
#'   For the central case (\eqn{\lambda = 0}), this simplifies to
#'   \eqn{\sqrt{8/k}}.
#'
#'   **Excess Kurtosis**:
#'
#'   \deqn{
#'     \gamma_2 = \frac{12(k + 4\lambda)}{(k + 2\lambda)^2}
#'   }{
#'     \gamma_2 = 12(k + 4\lambda) / (k + 2\lambda)^2
#'   }
#'
#'   For the central case (\eqn{\lambda = 0}), this simplifies to
#'   \eqn{12/k}.
#'
#'
#' @seealso [stats::Chisquare]
#'
#' @examples
#' dist <- dist_chisq(df = c(1,2,3,4,6,9))
#'
#' dist
#' mean(dist)
#' variance(dist)
#' skewness(dist)
#' kurtosis(dist)
#'
#' generate(dist, 10)
#'
#' density(dist, 2)
#' density(dist, 2, log = TRUE)
#'
#' cdf(dist, 4)
#'
#' quantile(dist, 0.7)
#'
#' @name dist_chisq
#' @export
dist_chisq <- function(df, ncp = 0){
  df <- vec_cast(df, double())
  ncp <- vec_cast(ncp, double())
  if(any(df < 0)){
    abort("The degrees of freedom parameter of a Chi-Squared distribution must be non-negative.")
  }
  new_dist(df = df, ncp = ncp, class = "dist_chisq")
}

#' @export
format.dist_chisq <- function(x, digits = 2, ...){
  sprintf(
    if (is_utf8_output()) "\u1d6a\u00b2(%s)" else "x2(%s)",
    format(x[["df"]], digits = digits, ...)
  )
}

#' @export
density.dist_chisq <- function(x, at, ...){
  stats::dchisq(at, x[["df"]], x[["ncp"]])
}

#' @export
log_density.dist_chisq <- function(x, at, ...){
  stats::dchisq(at, x[["df"]], x[["ncp"]], log = TRUE)
}

#' @export
quantile.dist_chisq <- function(x, p, ...){
  stats::qchisq(p, x[["df"]], x[["ncp"]])
}

#' @export
cdf.dist_chisq <- function(x, q, ...){
  stats::pchisq(q, x[["df"]], x[["ncp"]])
}

#' @export
generate.dist_chisq <- function(x, times, ...){
  stats::rchisq(times, x[["df"]], x[["ncp"]])
}

#' @export
mean.dist_chisq <- function(x, ...){
  x[["df"]] + x[["ncp"]]
}

#' @export
covariance.dist_chisq <- function(x, ...){
  2*(x[["df"]] + 2*x[["ncp"]])
}

#' @export
skewness.dist_chisq <- function(x, ...) sqrt(8 / x[["df"]])

#' @export
kurtosis.dist_chisq <- function(x, ...) 12 / x[["df"]]

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hilo.R
\name{hilo}
\alias{hilo}
\title{Compute intervals}
\usage{
hilo(x, ...)
}
\arguments{
\item{x}{Object to create hilo from.}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Used to extract a specified prediction interval at a particular confidence
level from a distribution.

The numeric lower and upper bounds can be extracted from the interval using
\verb{<hilo>$lower} and \verb{<hilo>$upper} as shown in the examples below.
}
\examples{
# 95\% interval from a standard normal distribution
interval <- hilo(dist_normal(0, 1), 95)
interval

# Extract the individual quantities with `$lower`, `$upper`, and `$level`
interval$lower
interval$upper
interval$level
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_gpd.R
\name{dist_gpd}
\alias{dist_gpd}
\title{The Generalized Pareto Distribution}
\usage{
dist_gpd(location, scale, shape)
}
\arguments{
\item{location}{the location parameter \eqn{a} of the GPD distribution.}

\item{scale}{the scale parameter \eqn{b} of the GPD distribution.}

\item{shape}{the shape parameter \eqn{s} of the GPD distribution.}
}
\description{
The GPD distribution is commonly used to model the tails of distributions,
particularly in extreme value theory.

The Pickands–Balkema–De Haan theorem states that for a large class of
distributions, the tail (above some threshold) can be approximated by a GPD.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_gpd.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_gpd.html}

In the following, let \eqn{X} be a Generalized Pareto random variable with
parameters \code{location} = \eqn{a}, \code{scale} = \eqn{b > 0}, and
\code{shape} = \eqn{s}.

\strong{Support}:
\eqn{x \ge a} if \eqn{s \ge 0},
\eqn{a \le x \le a - b/s} if \eqn{s < 0}

\strong{Mean}:
\deqn{
  E(X) = a + \frac{b}{1 - s} \quad \textrm{for } s < 1
}{
  E(X) = a + b/(1 - s) for s < 1
}
\eqn{E(X) = \infty} for \eqn{s \ge 1}

\strong{Variance}:
\deqn{
  \textrm{Var}(X) = \frac{b^2}{(1-s)^2(1-2s)} \quad \textrm{for } s < 0.5
}{
  Var(X) = b^2 / ((1-s)^2 (1-2s)) for s < 0.5
}
\eqn{\textrm{Var}(X) = \infty} for \eqn{s \ge 0.5}

\strong{Probability density function (p.d.f)}:

For \eqn{s = 0}:
\deqn{
  f(x) = \frac{1}{b}\exp\left(-\frac{x-a}{b}\right) \quad \textrm{for } x \ge a
}{
  f(x) = (1/b) exp(-(x-a)/b) for x >= a
}

For \eqn{s \ne 0}:
\deqn{
  f(x) = \frac{1}{b}\left(1 + s\frac{x-a}{b}\right)^{-1/s - 1}
}{
  f(x) = (1/b)(1 + s(x-a)/b)^(-1/s - 1)
}
where \eqn{1 + s(x-a)/b > 0}

\strong{Cumulative distribution function (c.d.f)}:

For \eqn{s = 0}:
\deqn{
  F(x) = 1 - \exp\left(-\frac{x-a}{b}\right) \quad \textrm{for } x \ge a
}{
  F(x) = 1 - exp(-(x-a)/b) for x >= a
}

For \eqn{s \ne 0}:
\deqn{
  F(x) = 1 - \left(1 + s\frac{x-a}{b}\right)^{-1/s}
}{
  F(x) = 1 - (1 + s(x-a)/b)^(-1/s)
}
where \eqn{1 + s(x-a)/b > 0}

\strong{Quantile function}:

For \eqn{s = 0}:
\deqn{
  Q(p) = a - b\log(1-p)
}{
  Q(p) = a - b log(1-p)
}

For \eqn{s \ne 0}:
\deqn{
  Q(p) = a + \frac{b}{s}\left[(1-p)^{-s} - 1\right]
}{
  Q(p) = a + (b/s)[(1-p)^(-s) - 1]
}

\strong{Median}:

For \eqn{s = 0}:
\deqn{
  \textrm{Median}(X) = a + b\log(2)
}{
  Median(X) = a + b log(2)
}

For \eqn{s \ne 0}:
\deqn{
  \textrm{Median}(X) = a + \frac{b}{s}\left(2^s - 1\right)
}{
  Median(X) = a + (b/s)(2^s - 1)
}

\strong{Skewness and Kurtosis}: No closed-form expressions; approximated numerically.
}
\examples{
dist <- dist_gpd(location = 0, scale = 1, shape = 0)

dist
mean(dist)
variance(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
}
\seealso{
\code{\link[evd:gpd]{evd::dgpd()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_negative_binomial.R
\name{dist_negative_binomial}
\alias{dist_negative_binomial}
\title{The Negative Binomial distribution}
\usage{
dist_negative_binomial(size, prob)
}
\arguments{
\item{size}{The number of successful trials (target number of successes).
Must be a positive number. Also called the dispersion parameter.}

\item{prob}{The probability of success in each trial. Must be between 0 and 1.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

A generalization of the geometric distribution. It is the number
of failures in a sequence of i.i.d. Bernoulli trials before
a specified number of successes (\code{size}) occur. The probability of success in
each trial is given by \code{prob}.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_negative_binomial.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_negative_binomial.html}

In the following, let \eqn{X} be a Negative Binomial random variable with
success probability \code{prob} = \eqn{p} and the number of successes \code{size} =
\eqn{r}.

\strong{Support}: \eqn{\{0, 1, 2, 3, ...\}}

\strong{Mean}: \eqn{\frac{r(1-p)}{p}}

\strong{Variance}: \eqn{\frac{r(1-p)}{p^2}}

\strong{Probability mass function (p.m.f)}:

\deqn{
     P(X = k) = \binom{k + r - 1}{k} (1-p)^r p^k
  }{
     P(X = k) = choose(k+r-1, k) * (1-p)^r * p^k
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
     F(k) = \sum_{i=0}^{\lfloor k \rfloor} \binom{i + r - 1}{i} (1-p)^r p^i
  }{
     F(k) = sum_{i=0}^floor(k) choose(i+r-1, i) * (1-p)^r * p^i
  }

This can also be expressed in terms of the regularized incomplete beta
function, and is computed numerically.

\strong{Moment generating function (m.g.f)}:

\deqn{
     E(e^{tX}) = \left(\frac{1-p}{1-pe^t}\right)^r, \quad t < -\log p
  }{
     E(e^(tX)) = ((1-p)/(1-p*e^t))^r, t < -log p
  }

\strong{Skewness}:

\deqn{
     \gamma_1 = \frac{2-p}{\sqrt{r(1-p)}}
  }{
     \gamma_1 = (2-p) / sqrt(r(1-p))
  }

\strong{Excess Kurtosis}:

\deqn{
     \gamma_2 = \frac{6}{r} + \frac{p^2}{r(1-p)}
  }{
     \gamma_2 = 6/r + p^2/(r(1-p))
  }
}
\examples{
dist <- dist_negative_binomial(size = 10, prob = 0.5)

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)
support(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:NegBinomial]{stats::NegBinomial}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{hdr.distribution}
\alias{hdr.distribution}
\title{Highest density regions of probability distributions}
\usage{
\method{hdr}{distribution}(x, size = 95, n = 512, ...)
}
\arguments{
\item{x}{The distribution(s).}

\item{size}{The size of the interval (between 0 and 100).}

\item{n}{The resolution used to estimate the distribution's density.}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#maturing}{\figure{lifecycle-maturing.svg}{options: alt='[Maturing]'}}}{\strong{[Maturing]}}

This function is highly experimental and will change in the future. In
particular, improved functionality for object classes and visualisation tools
will be added in a future release.

Computes minimally sized probability intervals highest density regions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_inflated.R
\name{dist_inflated}
\alias{dist_inflated}
\title{Inflate a value of a probability distribution}
\usage{
dist_inflated(dist, prob, x = 0)
}
\arguments{
\item{dist}{The distribution(s) to inflate.}

\item{prob}{The added probability of observing \code{x}.}

\item{x}{The value to inflate. The default of \code{x = 0} is for zero-inflation.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Inflated distributions add extra probability mass at a specific value,
most commonly zero (zero-inflation). These distributions are useful for
modeling data with excess observations at a particular value compared to
what the base distribution would predict. Common applications include
zero-inflated Poisson or negative binomial models for count data with
many zeros.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_inflated.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_inflated.html}

In the following, let \eqn{Y} be an inflated random variable based on
a base distribution \eqn{X}, with inflation value \code{x} = \eqn{c} and
inflation probability \code{prob} = \eqn{p}.

\strong{Support}: Same as the base distribution, but with additional
probability mass at \eqn{c}

\strong{Mean}: (when \code{x} is numeric)

\deqn{
    E(Y) = p \cdot c + (1-p) \cdot E(X)
  }{
    E(Y) = p * c + (1-p) * E(X)
  }

\strong{Variance}: (when \code{x = 0})

\deqn{
    \text{Var}(Y) = (1-p) \cdot \text{Var}(X) + p(1-p) \cdot [E(X)]^2
  }{
    Var(Y) = (1-p) * Var(X) + p(1-p) * [E(X)]^2
  }

For non-zero inflation values, the variance is not computed in closed form.

\strong{Probability mass/density function (p.m.f/p.d.f)}:

For discrete distributions:
\deqn{
    f_Y(y) = \begin{cases}
      p + (1-p) \cdot f_X(c) & \text{if } y = c \\
      (1-p) \cdot f_X(y) & \text{if } y \neq c
    \end{cases}
  }{
    f_Y(y) = p + (1-p) * f_X(c) if y = c, (1-p) * f_X(y) if y != c
  }

For continuous distributions:
\deqn{
    f_Y(y) = \begin{cases}
      p & \text{if } y = c \\
      (1-p) \cdot f_X(y) & \text{if } y \neq c
    \end{cases}
  }{
    f_Y(y) = p if y = c, (1-p) * f_X(y) if y != c
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F_Y(q) = \begin{cases}
      (1-p) \cdot F_X(q) & \text{if } q < c \\
      p + (1-p) \cdot F_X(q) & \text{if } q \geq c
    \end{cases}
  }{
    F_Y(q) = (1-p) * F_X(q) if q < c, p + (1-p) * F_X(q) if q >= c
  }

\strong{Quantile function}:

The quantile function is computed numerically by inverting the
inflated CDF, accounting for the jump in probability at the
inflation point.
}
\examples{
# Zero-inflated Poisson
dist <- dist_inflated(dist_poisson(lambda = 2), prob = 0.3, x = 0)

dist
mean(dist)
variance(dist)

generate(dist, 10)

density(dist, 0)
density(dist, 1)

cdf(dist, 2)

quantile(dist, 0.5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_gh.R
\name{dist_gh}
\alias{dist_gh}
\title{The generalised g-and-h Distribution}
\usage{
dist_gh(A, B, g, h, c = 0.8)
}
\arguments{
\item{A}{Vector of A (location) parameters.}

\item{B}{Vector of B (scale) parameters. Must be positive.}

\item{g}{Vector of g parameters.}

\item{h}{Vector of h parameters. Must be non-negative.}

\item{c}{Vector of c parameters (used for generalised g-and-h). Often fixed at 0.8 which is the default.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The generalised g-and-h distribution is a flexible distribution used to model univariate data, similar to the g-k distribution.
It is known for its ability to handle skewness and heavy-tailed behavior.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X} be a g-and-h random variable with parameters
\code{A}, \code{B}, \code{g}, \code{h}, and \code{c}.

\strong{Support}: \eqn{(-\infty, \infty)}

\strong{Mean}: Not available in closed form.

\strong{Variance}: Not available in closed form.

\strong{Probability density function (p.d.f)}:

The g-and-h distribution does not have a closed-form expression for its density. Instead,
it is defined through its quantile function:

\deqn{
    Q(u) = A + B \left( 1 + c \frac{1 - \exp(-gz(u))}{1 + \exp(-gz(u))} \right) \exp(h z(u)^2/2) z(u)
  }{
    Q(u) = A + B * (1 + c * ((1 - exp(-g * z(u))) / (1 + exp(-g * z(u))))) * exp(h * z(u)^2/2) * z(u)
  }

where \eqn{z(u) = \Phi^{-1}(u)}

\strong{Cumulative distribution function (c.d.f)}:

The cumulative distribution function is typically evaluated numerically due to the lack
of a closed-form expression.
}
\examples{
dist <- dist_gh(A = 0, B = 1, g = 0, h = 0.5)
dist

\dontshow{if (requireNamespace("gk", quietly = TRUE)) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
\dontshow{\}) # examplesIf}
}
\seealso{
\link[gk:g-and-h]{gk::dgh}, \link{dist_gk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{variance.distribution}
\alias{variance.distribution}
\title{Variance of a probability distribution}
\usage{
\method{variance}{distribution}(x, ...)
}
\arguments{
\item{x}{The distribution(s).}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Returns the empirical variance of the probability distribution. If the method
does not exist, the variance of a random sample will be returned.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{autoplot.distribution}
\alias{autoplot.distribution}
\title{Plot a distribution}
\usage{
autoplot.distribution(x, ...)
}
\arguments{
\item{x}{The distribution(s) to plot.}

\item{...}{Unused.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#defunct}{\figure{lifecycle-defunct.svg}{options: alt='[Defunct]'}}}{\strong{[Defunct]}}

This function is now defunct and can no longer be used. Instead consider using
the {ggdist} package to produce your own distribution plots. You can learn
more about how this plot can be produced using {ggdist} here:
https://mjskay.github.io/ggdist/articles/slabinterval.html
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/support.R
\name{new_support_region}
\alias{new_support_region}
\title{Create a new support region vector}
\usage{
new_support_region(x = numeric(), limits = list(), closed = list())
}
\arguments{
\item{x}{A list of prototype vectors defining the distribution type.}

\item{limits}{A list of value limits for the distribution.}

\item{closed}{A list of logical(2L) indicating whether the limits are closed.}
}
\description{
Create a new support region vector
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_beta.R
\name{dist_beta}
\alias{dist_beta}
\title{The Beta distribution}
\usage{
dist_beta(shape1, shape2)
}
\arguments{
\item{shape1, shape2}{The non-negative shape parameters of the Beta distribution.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Beta distribution is a continuous probability distribution defined on the
interval [0, 1], commonly used to model probabilities and proportions.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_beta.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_beta.html}

In the following, let \eqn{X} be a Beta random variable with parameters
\code{shape1} = \eqn{\alpha} and \code{shape2} = \eqn{\beta}.

\strong{Support}: \eqn{x \in [0, 1]}

\strong{Mean}: \eqn{\frac{\alpha}{\alpha + \beta}}

\strong{Variance}: \eqn{\frac{\alpha\beta}{(\alpha + \beta)^2(\alpha + \beta + 1)}}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{x^{\alpha - 1}(1-x)^{\beta - 1}}{B(\alpha, \beta)} = 
    \frac{\Gamma(\alpha + \beta)}{\Gamma(\alpha)\Gamma(\beta)} x^{\alpha - 1}(1-x)^{\beta - 1}
  }{
    f(x) = x^(alpha - 1) * (1-x)^(beta - 1) / B(alpha, beta)
  }

where \eqn{B(\alpha, \beta) = \frac{\Gamma(\alpha)\Gamma(\beta)}{\Gamma(\alpha + \beta)}}
is the Beta function.

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = I_x(alpha, beta) = \frac{B(x; \alpha, \beta)}{B(\alpha, \beta)}
  }{
    F(x) = I_x(\alpha, \beta)
  }

where \eqn{I_x(\alpha, \beta)} is the regularized incomplete beta function and
\eqn{B(x; \alpha, \beta) = \int_0^x t^{\alpha-1}(1-t)^{\beta-1} dt}.

\strong{Moment generating function (m.g.f)}:

The moment generating function does not have a simple closed form, but the
moments can be calculated as:

\deqn{
    E(X^k) = \prod_{r=0}^{k-1} \frac{\alpha + r}{\alpha + \beta + r}
  }{
    E(X^k) = prod((alpha + r) / (alpha + beta + r)) for r = 0, ..., k-1
  }
}
\examples{
dist <- dist_beta(shape1 = c(0.5, 5, 1, 2, 2), shape2 = c(0.5, 1, 3, 2, 5))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Beta]{stats::Beta}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{density.distribution}
\alias{density.distribution}
\title{The probability density/mass function}
\usage{
\method{density}{distribution}(x, at, ..., log = FALSE)
}
\arguments{
\item{x}{The distribution(s).}

\item{at}{The point at which to compute the density/mass.}

\item{...}{Additional arguments passed to methods.}

\item{log}{If \code{TRUE}, probabilities will be given as log probabilities.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Computes the probability density function for a continuous distribution, or
the probability mass function for a discrete distribution.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{hilo.distribution}
\alias{hilo.distribution}
\title{Probability intervals of a probability distribution}
\usage{
\method{hilo}{distribution}(x, size = 95, ...)
}
\arguments{
\item{x}{The distribution(s).}

\item{size}{The size of the interval (between 0 and 100).}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Returns a \code{hilo} central probability interval with probability coverage of
\code{size}. By default, the distribution's \code{\link[=quantile]{quantile()}} will be used to compute
the lower and upper bound for a centered interval
}
\seealso{
\code{\link[=hdr.distribution]{hdr.distribution()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_transformed.R
\name{dist_transformed}
\alias{dist_transformed}
\title{Modify a distribution with a transformation}
\usage{
dist_transformed(dist, transform, inverse)
}
\arguments{
\item{dist}{A univariate distribution vector.}

\item{transform}{A function used to transform the distribution. This
transformation should be monotonic over appropriate domain.}

\item{inverse}{The inverse of the \code{transform} function.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#maturing}{\figure{lifecycle-maturing.svg}{options: alt='[Maturing]'}}}{\strong{[Maturing]}}

A transformed distribution applies a monotonic transformation to an existing
distribution. This is useful for creating derived distributions such as
log-normal (exponential transformation of normal), or other custom
transformations of base distributions.

The \code{\link[=density]{density()}}, \code{\link[=mean]{mean()}}, and \code{\link[=variance]{variance()}} methods are approximate as
they are based on numerical derivatives.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_transformed.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_transformed.html}

Let \eqn{Y = g(X)} where \eqn{X} is the base distribution with
transformation function \code{transform} = \eqn{g} and \code{inverse} = \eqn{g^{-1}}.
The transformation \eqn{g} must be monotonic over the support of \eqn{X}.

\strong{Support}: \eqn{g(S_X)} where \eqn{S_X} is the support of \eqn{X}

\strong{Mean}: Approximated numerically using a second-order Taylor expansion:

\deqn{
    E(Y) \approx g(\mu_X) + \frac{1}{2}g''(\mu_X)\sigma_X^2
  }{
    E(Y) ~= g(mu_X) + (1/2) g''(mu_X) sigma_X^2
  }

where \eqn{\mu_X} and \eqn{\sigma_X^2} are the mean and variance of the
base distribution \eqn{X}, and \eqn{g''} is the second derivative of the
transformation. The derivative is computed numerically using
\code{\link[numDeriv:hessian]{numDeriv::hessian()}}.

\strong{Variance}: Approximated numerically using the delta method:

\deqn{
    \mathrm{Var}(Y) \approx [g'(\mu_X)]^2\sigma_X^2 + \frac{1}{2}[g''(\mu_X)\sigma_X^2]^2
  }{
    Var(Y) ~= [g'(mu_X)]^2 sigma_X^2 + (1/2) [g''(mu_X) sigma_X^2]^2
  }

where \eqn{g'} is the first derivative (Jacobian) computed numerically
using \code{\link[numDeriv:jacobian]{numDeriv::jacobian()}}.

\strong{Probability density function (p.d.f)}: Using the change of variables
formula:

\deqn{
    f_Y(y) = f_X(g^{-1}(y)) \left|\frac{d}{dy}g^{-1}(y)\right|
  }{
    f_Y(y) = f_X(g^(-1)(y)) |d/dy g^(-1)(y)|
  }

where \eqn{f_X} is the p.d.f. of the base distribution and the Jacobian
\eqn{|d/dy \, g^{-1}(y)|} is computed numerically using
\code{\link[numDeriv:jacobian]{numDeriv::jacobian()}}.

\strong{Cumulative distribution function (c.d.f)}:

For monotonically increasing \eqn{g}:
\deqn{
    F_Y(y) = F_X(g^{-1}(y))
  }{
    F_Y(y) = F_X(g^(-1)(y))
  }

For monotonically decreasing \eqn{g}:
\deqn{
    F_Y(y) = 1 - F_X(g^{-1}(y))
  }{
    F_Y(y) = 1 - F_X(g^(-1)(y))
  }

where \eqn{F_X} is the c.d.f. of the base distribution.

\strong{Quantile function}: The inverse of the c.d.f.

For monotonically increasing \eqn{g}:
\deqn{
    Q_Y(p) = g(Q_X(p))
  }{
    Q_Y(p) = g(Q_X(p))
  }

For monotonically decreasing \eqn{g}:
\deqn{
    Q_Y(p) = g(Q_X(1-p))
  }{
    Q_Y(p) = g(Q_X(1-p))
  }

where \eqn{Q_X} is the quantile function of the base distribution.
}
\examples{
# Create a log normal distribution
dist <- dist_transformed(dist_normal(0, 0.5), exp, log)
density(dist, 1) # dlnorm(1, 0, 0.5)
cdf(dist, 4) # plnorm(4, 0, 0.5)
quantile(dist, 0.1) # qlnorm(0.1, 0, 0.5)
generate(dist, 10) # rlnorm(10, 0, 0.5)

}
\seealso{
\code{\link[numDeriv:jacobian]{numDeriv::jacobian()}}, \code{\link[numDeriv:hessian]{numDeriv::hessian()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{median.distribution}
\alias{median.distribution}
\title{Median of a probability distribution}
\usage{
\method{median}{distribution}(x, na.rm = FALSE, ...)
}
\arguments{
\item{x}{The distribution(s).}

\item{na.rm}{Unused, included for consistency with the generic function.}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Returns the median (50th percentile) of a probability distribution. This is
equivalent to \code{quantile(x, p=0.5)}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_horseshoe.R
\name{dist_horseshoe}
\alias{dist_horseshoe}
\title{The Horseshoe distribution}
\usage{
dist_horseshoe(lambda, tau)
}
\arguments{
\item{lambda}{A positive numeric vector of local scale parameters
\eqn{\lambda > 0} (one per observation).}

\item{tau}{A positive scalar global scale parameter \eqn{\tau > 0}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The horseshoe distribution (Carvalho et al., 2008) is a heavy-tailed
continuous distribution defined as a scale mixture of normals. It is
primarily used as a shrinkage prior in sparse Bayesian regression, where
it concentrates mass near zero while retaining heavy tails that leave
large signals unshrunk.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_horseshoe.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_horseshoe.html}

In the following, let \eqn{X} be a horseshoe random variable with local
scale parameter \code{lambda} = \eqn{\lambda > 0} and global scale parameter
\code{tau} = \eqn{\tau > 0}.

\strong{Support}: \eqn{x \in \mathbb{R}}, the set of all real numbers.

\strong{Mean}: \eqn{E(X)} — not available in closed form.

\strong{Variance}: \eqn{\mathrm{Var}(X)} — not available in closed form.

\strong{Probability density function (p.d.f)}:

The horseshoe density does not have a simple closed form but can be
expressed as a scale mixture:

\deqn{
    X \mid \lambda, \tau \sim \mathcal{N}(0,\, \lambda^2 \tau^2)
  }{
    X | lambda, tau ~ N(0, lambda^2 * tau^2)
  }

where the half-Cauchy hyperprior \eqn{\lambda \sim C^+(0, 1)} induces the
characteristic horseshoe shrinkage behaviour.
}
\examples{
dist <- dist_horseshoe(lambda = c(0.5, 1, 2), tau = 1)
dist

\dontshow{if (requireNamespace("LaplacesDemon", quietly = TRUE)) withAutoprint(\{ # examplesIf}
support(dist)
generate(dist, 10)

density(dist, 0)
density(dist, 0, log = TRUE)
\dontshow{\}) # examplesIf}
}
\references{
Carvalho, C.M., Polson, N.G., and Scott, J.G. (2008).
"The Horseshoe Estimator for Sparse Signals".
\emph{Discussion Paper 2008-31}. Duke University Department of Statistical
Science.

Carvalho, C.M., Polson, N.G., and Scott, J.G. (2009).
"Handling Sparsity via the Horseshoe".
\emph{Journal of Machine Learning Research}, 5, p. 73–80.
}
\seealso{
\code{\link[LaplacesDemon:dist.Horseshoe]{LaplacesDemon::dhs()}}, \code{\link[LaplacesDemon:dist.Horseshoe]{LaplacesDemon::rhs()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_lognormal.R
\name{dist_lognormal}
\alias{dist_lognormal}
\title{The log-normal distribution}
\usage{
dist_lognormal(mu = 0, sigma = 1)
}
\arguments{
\item{mu}{The mean (location parameter) of the distribution, which is the
mean of the associated Normal distribution. Can be any real number.}

\item{sigma}{The standard deviation (scale parameter) of the distribution.
Can be any positive number.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The log-normal distribution is a commonly used transformation of the Normal
distribution. If \eqn{X} follows a log-normal distribution, then \eqn{\ln{X}}
would be characteristed by a Normal distribution.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{Y} be a Normal random variable with mean
\code{mu} = \eqn{\mu} and standard deviation \code{sigma} = \eqn{\sigma}. The
log-normal distribution \eqn{X = exp(Y)} is characterised by:

\strong{Support}: \eqn{R+}, the set of all real numbers greater than or equal to 0.

\strong{Mean}: \eqn{e^(\mu + \sigma^2/2}

\strong{Variance}: \eqn{(e^(\sigma^2)-1) e^(2\mu + \sigma^2}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{1}{x\sqrt{2 \pi \sigma^2}} e^{-(\ln{x} - \mu)^2 / 2 \sigma^2}
  }{
    f(x) = 1 / (x * sqrt(2 \pi \sigma^2)) exp(-(log(x) - \mu)^2 / (2 \sigma^2))
  }

\strong{Cumulative distribution function (c.d.f)}:

The cumulative distribution function has the form

\deqn{
    F(x) = \Phi((\ln{x} - \mu)/\sigma)
  }{
    F(x) = Phi((log(x) - \mu)/\sigma)
  }

Where \eqn{Phi}{Phi} is the CDF of a standard Normal distribution, N(0,1).
}
\examples{
dist <- dist_lognormal(mu = 1:5, sigma = 0.1)

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

# A log-normal distribution X is exp(Y), where Y is a Normal distribution of
# the same parameters. So log(X) will produce the Normal distribution Y.
log(dist)
}
\seealso{
\link[stats:Lognormal]{stats::Lognormal}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_degenerate.R
\name{dist_degenerate}
\alias{dist_degenerate}
\title{The degenerate distribution}
\usage{
dist_degenerate(x)
}
\arguments{
\item{x}{The value of the distribution (location parameter). Can be any real number.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The degenerate distribution takes a single value which is certain to be
observed. It takes a single parameter, which is the value that is observed
by the distribution.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_degenerate.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_degenerate.html}

In the following, let \eqn{X} be a degenerate random variable with value
\code{x} = \eqn{k_0}.

\strong{Support}: \eqn{\{k_0\}}, a single point

\strong{Mean}: \eqn{\mu = k_0}

\strong{Variance}: \eqn{\sigma^2 = 0}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = 1 \textrm{ for } x = k_0
  }{
    f(x) = 1 for x = k_0
  }
\deqn{
    f(x) = 0 \textrm{ for } x \neq k_0
  }{
    f(x) = 0 for x != k_0
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(t) = 0 \textrm{ for } t < k_0
  }{
    F(t) = 0 for t < k_0
  }
\deqn{
    F(t) = 1 \textrm{ for } t \ge k_0
  }{
    F(t) = 1 for t >= k_0
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = e^{k_0 t}
  }{
    E(e^(tX)) = e^(k_0 t)
  }

\strong{Skewness}: Undefined (NA)

\strong{Excess Kurtosis}: Undefined (NA)
}
\examples{
dist <- dist_degenerate(x = 1:5)

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Distributions]{stats::Distributions}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_cauchy.R
\name{dist_cauchy}
\alias{dist_cauchy}
\title{The Cauchy distribution}
\usage{
dist_cauchy(location, scale)
}
\arguments{
\item{location, scale}{location and scale parameters.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Cauchy distribution is the student's t distribution with one degree of
freedom. The Cauchy distribution does not have a well defined mean or
variance. Cauchy distributions often appear as priors in Bayesian contexts
due to their heavy tails.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X} be a Cauchy variable with mean
\verb{location =} \eqn{x_0} and \code{scale} = \eqn{\gamma}.

\strong{Support}: \eqn{R}, the set of all real numbers

\strong{Mean}: Undefined.

\strong{Variance}: Undefined.

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{1}{\pi \gamma \left[1 + \left(\frac{x - x_0}{\gamma} \right)^2 \right]}
  }{
    f(x) = 1 / (\pi \gamma (1 + ((x - x_0) / \gamma)^2)
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(t) = \frac{1}{\pi} \arctan \left( \frac{t - x_0}{\gamma} \right) +
      \frac{1}{2}
  }{
    F(t) = arctan((t - x_0) / \gamma) / \pi + 1/2
  }

\strong{Moment generating function (m.g.f)}:

Does not exist.
}
\examples{
dist <- dist_cauchy(location = c(0, 0, 0, -2), scale = c(0.5, 1, 2, 1))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Cauchy]{stats::Cauchy}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_weibull.R
\name{dist_weibull}
\alias{dist_weibull}
\title{The Weibull distribution}
\usage{
dist_weibull(shape, scale)
}
\arguments{
\item{shape, scale}{shape and scale parameters, the latter defaulting to 1.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Generalization of the gamma distribution. Often used in survival and
time-to-event analyses.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_weibull.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_weibull.html}

In the following, let \eqn{X} be a Weibull random variable with
shape parameter \code{shape} = \eqn{k} and scale parameter \code{scale} = \eqn{\lambda}.

\strong{Support}: \eqn{[0, \infty)}

\strong{Mean}:

\deqn{
    E(X) = \lambda \Gamma\left(1 + \frac{1}{k}\right)
  }{
    E(X) = \lambda \Gamma(1 + 1/k)
  }

where \eqn{\Gamma} is the gamma function.

\strong{Variance}:

\deqn{
    \text{Var}(X) = \lambda^2 \left[\Gamma\left(1 + \frac{2}{k}\right) - \left(\Gamma\left(1 + \frac{1}{k}\right)\right)^2\right]
  }{
    Var(X) = \lambda^2 [\Gamma(1 + 2/k) - (\Gamma(1 + 1/k))^2]
  }

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{k}{\lambda}\left(\frac{x}{\lambda}\right)^{k-1}e^{-(x/\lambda)^k}, \quad x \ge 0
  }{
    f(x) = (k/\lambda)(x/\lambda)^(k-1) e^(-(x/\lambda)^k), x >= 0
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = 1 - e^{-(x/\lambda)^k}, \quad x \ge 0
  }{
    F(x) = 1 - e^(-(x/\lambda)^k), x >= 0
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \sum_{n=0}^\infty \frac{t^n\lambda^n}{n!} \Gamma\left(1+\frac{n}{k}\right)
  }{
    E(e^(tX)) = \sum_(n=0)^\infty (t^n \lambda^n / n!) \Gamma(1 + n/k)
  }

\strong{Skewness}:

\deqn{
    \gamma_1 = \frac{\mu^3 - 3\mu\sigma^2 - \mu^3}{\sigma^3}
  }{
    \gamma_1 = (\mu_3 - 3\mu\sigma^2 - \mu^3) / \sigma^3
  }

where \eqn{\mu = E(X)}, \eqn{\sigma^2 = \text{Var}(X)}, and the third
raw moment is

\deqn{
    \mu^3 = \lambda^3 \Gamma\left(1 + \frac{3}{k}\right)
  }{
    \mu_3 = \lambda^3 \Gamma(1 + 3/k)
  }

\strong{Excess Kurtosis}:

\deqn{
    \gamma_2 = \frac{\mu^4 - 4\gamma_1\mu\sigma^3 - 6\mu^2\sigma^2 - \mu^4}{\sigma^4} - 3
  }{
    \gamma_2 = (\mu_4 - 4\gamma_1\mu\sigma^3 - 6\mu^2\sigma^2 - \mu^4) / \sigma^4 - 3
  }

where the fourth raw moment is

\deqn{
    \mu^4 = \lambda^4 \Gamma\left(1 + \frac{4}{k}\right)
  }{
    \mu_4 = \lambda^4 \Gamma(1 + 4/k)
  }
}
\examples{
dist <- dist_weibull(shape = c(0.5, 1, 1.5, 5), scale = rep(1, 4))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Weibull]{stats::Weibull}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_laplace.R
\name{dist_laplace}
\alias{dist_laplace}
\title{The Laplace distribution}
\usage{
dist_laplace(mu, sigma)
}
\arguments{
\item{mu}{The location parameter (mean) of the Laplace distribution.}

\item{sigma}{The positive scale parameter of the Laplace distribution.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Laplace distribution, also known as the double exponential distribution,
is a continuous probability distribution that is symmetric around its location
parameter.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_laplace.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_laplace.html}

In the following, let \eqn{X} be a Laplace random variable with location
parameter \code{mu} = \eqn{\mu} and scale parameter \code{sigma} = \eqn{\sigma}.

\strong{Support}: \eqn{R}, the set of all real numbers

\strong{Mean}: \eqn{\mu}

\strong{Variance}: \eqn{2\sigma^2}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{1}{2\sigma} \exp\left(-\frac{|x - \mu|}{\sigma}\right)
  }{
    f(x) = 1 / (2 * sigma) * exp(-|x - mu| / sigma)
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = \begin{cases}
      \frac{1}{2} \exp\left(\frac{x - \mu}{\sigma}\right) & \text{if } x < \mu \\
      1 - \frac{1}{2} \exp\left(-\frac{x - \mu}{\sigma}\right) & \text{if } x \geq \mu
    \end{cases}
  }{
    F(x) = 1/2 * exp((x - mu) / sigma) if x < mu,
           1 - 1/2 * exp(-(x - mu) / sigma) if x >= mu
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \frac{\exp(\mu t)}{1 - \sigma^2 t^2} \text{ for } |t| < \frac{1}{\sigma}
  }{
    E(e^(tX)) = exp(mu * t) / (1 - sigma^2 * t^2) for |t| < 1/sigma
  }
}
\examples{
dist <- dist_laplace(mu = c(0, 2, -1), sigma = c(1, 2, 0.5))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 0)
density(dist, 0, log = TRUE)

cdf(dist, 1)

quantile(dist, 0.7)

}
\seealso{
\link[extraDistr:Laplace]{extraDistr::Laplace}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hdr.R
\name{hdr}
\alias{hdr}
\title{Compute highest density regions}
\usage{
hdr(x, ...)
}
\arguments{
\item{x}{Object to create hilo from.}

\item{...}{Additional arguments used by methods.}
}
\description{
Used to extract a specified prediction interval at a particular confidence
level from a distribution.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hilo.R
\name{new_hilo}
\alias{new_hilo}
\title{Construct hilo intervals}
\usage{
new_hilo(lower = double(), upper = double(), size = double())
}
\arguments{
\item{lower, upper}{A numeric vector of values for lower and upper limits.}

\item{size}{Size of the interval between [0, 100].}
}
\value{
A "hilo" vector
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Class constructor function to help with manually creating hilo interval
objects.
}
\examples{
new_hilo(lower = rnorm(10), upper = rnorm(10) + 5, size = 95)

}
\author{
Earo Wang & Mitchell O'Hara-Wild
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_logistic.R
\name{dist_logistic}
\alias{dist_logistic}
\title{The Logistic distribution}
\usage{
dist_logistic(location, scale)
}
\arguments{
\item{location, scale}{location and scale parameters.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

A continuous distribution on the real line. For binary outcomes
the model given by \eqn{P(Y = 1 | X) = F(X \beta)} where
\eqn{F} is the Logistic \code{\link[=cdf]{cdf()}} is called \emph{logistic regression}.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X} be a Logistic random variable with
\code{location} = \eqn{\mu} and \code{scale} = \eqn{s}.

\strong{Support}: \eqn{R}, the set of all real numbers

\strong{Mean}: \eqn{\mu}

\strong{Variance}: \eqn{s^2 \pi^2 / 3}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{e^{-(\frac{x - \mu}{s})}}{s [1 + \exp(-(\frac{x - \mu}{s})) ]^2}
  }{
    f(x) = e^(-(t - \mu) / s) / (s (1 + e^(-(t - \mu) / s))^2)
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(t) = \frac{1}{1 + e^{-(\frac{t - \mu}{s})}}
  }{
    F(t) = 1 / (1 +  e^(-(t - \mu) / s))
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = e^{\mu t} \beta(1 - st, 1 + st)
  }{
    E(e^(tX)) = = e^(\mu t) \beta(1 - st, 1 + st)
  }

where \eqn{\beta(x, y)} is the Beta function.
}
\examples{
dist <- dist_logistic(location = c(5,9,9,6,2), scale = c(2,3,4,2,1))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Logistic]{stats::Logistic}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_chisq.R
\name{dist_chisq}
\alias{dist_chisq}
\title{The (non-central) Chi-Squared Distribution}
\usage{
dist_chisq(df, ncp = 0)
}
\arguments{
\item{df}{Degrees of freedom (non-centrality parameter). Can be any
positive real number.}

\item{ncp}{Non-centrality parameter. Can be any non-negative real number.
Defaults to 0 (central chi-squared distribution).}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Chi-square distributions show up often in frequentist settings
as the sampling distribution of test statistics, especially
in maximum likelihood estimation settings.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_chisq.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_chisq.html}

In the following, let \eqn{X} be a \eqn{\chi^2} random variable with
\code{df} = \eqn{k} and \code{ncp} = \eqn{\lambda}.

\strong{Support}: \eqn{R^+}, the set of positive real numbers

\strong{Mean}: \eqn{k + \lambda}

\strong{Variance}: \eqn{2(k + 2\lambda)}

\strong{Probability density function (p.d.f)}:

For the central chi-squared distribution (\eqn{\lambda = 0}):

\deqn{
    f(x) = \frac{1}{2^{k/2} \Gamma(k/2)} x^{k/2 - 1} e^{-x/2}
  }{
    f(x) = 1 / (2^(k/2) \Gamma(k/2)) x^(k/2 - 1) exp(-x/2)
  }

For the non-central chi-squared distribution (\eqn{\lambda > 0}):

\deqn{
    f(x) = \frac{1}{2} e^{-(x+\lambda)/2} \left(\frac{x}{\lambda}\right)^{k/4-1/2} I_{k/2-1}\left(\sqrt{\lambda x}\right)
  }{
    f(x) = (1/2) exp(-(x+\lambda)/2) (x/\lambda)^(k/4-1/2) I_(k/2-1)(sqrt(\lambda x))
  }

where \eqn{I_\nu(z)} is the modified Bessel function of the first kind.

\strong{Cumulative distribution function (c.d.f)}:

For the central chi-squared distribution (\eqn{\lambda = 0}):

\deqn{
    F(x) = \frac{\gamma(k/2, x/2)}{\Gamma(k/2)} = P(k/2, x/2)
  }{
    F(x) = \gamma(k/2, x/2) / \Gamma(k/2) = P(k/2, x/2)
  }

where \eqn{\gamma(s, x)} is the lower incomplete gamma function and
\eqn{P(s, x)} is the regularized gamma function.

For the non-central chi-squared distribution (\eqn{\lambda > 0}):

\deqn{
    F(x) = \sum_{j=0}^{\infty} \frac{e^{-\lambda/2} (\lambda/2)^j}{j!} P(k/2 + j, x/2)
  }{
    F(x) = sum_(j=0)^(\infty) exp(-\lambda/2) (\lambda/2)^j / j! P(k/2 + j, x/2)
  }

This is approximated numerically.

\strong{Moment generating function (m.g.f)}:

For the central chi-squared distribution (\eqn{\lambda = 0}):

\deqn{
    E(e^{tX}) = (1 - 2t)^{-k/2}, \quad t < 1/2
  }{
    E(e^(tX)) = (1 - 2t)^(-k/2), t < 1/2
  }

For the non-central chi-squared distribution (\eqn{\lambda > 0}):

\deqn{
    E(e^{tX}) = \frac{e^{\lambda t / (1 - 2t)}}{(1 - 2t)^{k/2}}, \quad t < 1/2
  }{
    E(e^(tX)) = exp(\lambda t / (1 - 2t)) / (1 - 2t)^(k/2), t < 1/2
  }

\strong{Skewness}:

\deqn{
    \gamma_1 = \frac{2^{3/2}(k + 3\lambda)}{(k + 2\lambda)^{3/2}}
  }{
    \gamma_1 = 2^(3/2) (k + 3\lambda) / (k + 2\lambda)^(3/2)
  }

For the central case (\eqn{\lambda = 0}), this simplifies to
\eqn{\sqrt{8/k}}.

\strong{Excess Kurtosis}:

\deqn{
    \gamma_2 = \frac{12(k + 4\lambda)}{(k + 2\lambda)^2}
  }{
    \gamma_2 = 12(k + 4\lambda) / (k + 2\lambda)^2
  }

For the central case (\eqn{\lambda = 0}), this simplifies to
\eqn{12/k}.
}
\examples{
dist <- dist_chisq(df = c(1,2,3,4,6,9))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Chisquare]{stats::Chisquare}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_multinomial.R
\name{dist_multinomial}
\alias{dist_multinomial}
\title{The Multinomial distribution}
\usage{
dist_multinomial(size, prob)
}
\arguments{
\item{size}{The number of draws from the Categorical distribution.}

\item{prob}{The probability of an event occurring from each draw.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The multinomial distribution is a generalization of the binomial
distribution to multiple categories. It is perhaps easiest to think
that we first extend a \code{\link[=dist_bernoulli]{dist_bernoulli()}} distribution to include more
than two categories, resulting in a \code{\link[=dist_categorical]{dist_categorical()}} distribution.
We then extend repeat the Categorical experiment several (\eqn{n})
times.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X = (X_1, ..., X_k)} be a Multinomial
random variable with success probability \code{p} = \eqn{p}. Note that
\eqn{p} is vector with \eqn{k} elements that sum to one. Assume
that we repeat the Categorical experiment \code{size} = \eqn{n} times.

\strong{Support}: Each \eqn{X_i} is in \eqn{{0, 1, 2, ..., n}}.

\strong{Mean}: The mean of \eqn{X_i} is \eqn{n p_i}.

\strong{Variance}: The variance of \eqn{X_i} is \eqn{n p_i (1 - p_i)}.
For \eqn{i \neq j}, the covariance of \eqn{X_i} and \eqn{X_j}
is \eqn{-n p_i p_j}.

\strong{Probability mass function (p.m.f)}:

\deqn{
    P(X_1 = x_1, ..., X_k = x_k) = \frac{n!}{x_1! x_2! ... x_k!} p_1^{x_1} \cdot p_2^{x_2} \cdot ... \cdot p_k^{x_k}
  }{
    P(X_1 = x_1, ..., X_k = x_k) = n! / (x_1! x_2! ... x_k!) p_1^x_1 p_2^x_2 ... p_k^x_k
  }

\strong{Cumulative distribution function (c.d.f)}:

Omitted for multivariate random variables for the time being.

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \left(\sum_{i=1}^k p_i e^{t_i}\right)^n
  }{
    E(e^(tX)) = (p_1 e^t_1 + p_2 e^t_2 + ... + p_k e^t_k)^n
  }
}
\examples{
dist <- dist_multinomial(size = c(4, 3), prob = list(c(0.3, 0.5, 0.2), c(0.1, 0.5, 0.4)))

dist
mean(dist)
variance(dist)

generate(dist, 10)

# TODO: Needs fixing to support multiple inputs
# density(dist, 2)
# density(dist, 2, log = TRUE)

}
\seealso{
\link[stats:Multinom]{stats::Multinomial}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{family.distribution}
\alias{family.distribution}
\title{Extract the name of the distribution family}
\usage{
\method{family}{distribution}(object, ...)
}
\arguments{
\item{object}{The distribution(s).}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}
\examples{
dist <- c(
  dist_normal(1:2),
  dist_poisson(3),
  dist_multinomial(size = c(4, 3),
  prob = list(c(0.3, 0.5, 0.2), c(0.1, 0.5, 0.4)))
  )
family(dist)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_poisson.R
\name{dist_poisson}
\alias{dist_poisson}
\title{The Poisson Distribution}
\usage{
dist_poisson(lambda)
}
\arguments{
\item{lambda}{vector of (non-negative) means.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Poisson distributions are frequently used to model counts.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X} be a Poisson random variable with parameter
\code{lambda} = \eqn{\lambda}.

\strong{Support}: \eqn{\{0, 1, 2, 3, ...\}}{{0, 1, 2, 3, ...}}

\strong{Mean}: \eqn{\lambda}

\strong{Variance}: \eqn{\lambda}

\strong{Probability mass function (p.m.f)}:

\deqn{
    P(X = k) = \frac{\lambda^k e^{-\lambda}}{k!}
  }{
    P(X = k) = \lambda^k e^(-\lambda) / k!
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    P(X \le k) = e^{-\lambda}
    \sum_{i = 0}^{\lfloor k \rfloor} \frac{\lambda^i}{i!}
  }{
    P(X \le k) = e^(-\lambda)
    \sum_{i = 0}^k \lambda^i / i!
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = e^{\lambda (e^t - 1)}
  }{
    E(e^(tX)) = e^(\lambda (e^t - 1))
  }
}
\examples{
dist <- dist_poisson(lambda = c(1, 4, 10))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Poisson]{stats::Poisson}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_categorical.R
\name{dist_categorical}
\alias{dist_categorical}
\title{The Categorical distribution}
\usage{
dist_categorical(prob, outcomes = NULL)
}
\arguments{
\item{prob}{A list of probabilities of observing each outcome category.}

\item{outcomes}{The list of vectors where each value represents each outcome.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Categorical distributions are used to represent events with multiple
outcomes, such as what number appears on the roll of a dice. This is also
referred to as the 'generalised Bernoulli' or 'multinoulli' distribution.
The Categorical distribution is a special case of the \code{\link[=Multinomial]{Multinomial()}}
distribution with \code{n = 1}.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_categorical.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_categorical.html}

In the following, let \eqn{X} be a Categorical random variable with
probability parameters \code{prob} = \eqn{\{p_1, p_2, \ldots, p_k\}}.

The Categorical probability distribution is widely used to model the
occurance of multiple events. A simple example is the roll of a dice, where
\eqn{p = \{1/6, 1/6, 1/6, 1/6, 1/6, 1/6\}} giving equal chance of observing
each number on a 6 sided dice.

\strong{Support}: \eqn{\{1, \ldots, k\}}{{1, ..., k}}

\strong{Mean}: Not defined for unordered categories. For ordered categories with
integer outcomes \eqn{\{1, 2, \ldots, k\}}, the mean is:

\deqn{
    E(X) = \sum_{i=1}^{k} i \cdot p_i
  }{
    E(X) = sum(i * p_i) for i = 1 to k
  }

\strong{Variance}: Not defined for unordered categories. For ordered categories
with integer outcomes \eqn{\{1, 2, \ldots, k\}}, the variance is:

\deqn{
    \text{Var}(X) = \sum_{i=1}^{k} i^2 \cdot p_i - \left(\sum_{i=1}^{k} i \cdot p_i\right)^2
  }{
    Var(X) = sum(i^2 * p_i) - [sum(i * p_i)]^2
  }

\strong{Probability mass function (p.m.f)}:

\deqn{
    P(X = i) = p_i
  }{
    P(X = i) = p_i
  }

\strong{Cumulative distribution function (c.d.f)}:

The c.d.f is undefined for unordered categories. For ordered categories
with outcomes \eqn{x_1 < x_2 < \ldots < x_k}, the c.d.f is:

\deqn{
    P(X \le x_j) = \sum_{i=1}^{j} p_i
  }{
    P(X <= x_j) = sum(p_i) for i = 1 to j
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \sum_{i=1}^{k} e^{tx_i} \cdot p_i
  }{
    E(e^(tX)) = sum(e^(t * x_i) * p_i) for i = 1 to k
  }

\strong{Skewness}: Approximated numerically for ordered categories.

\strong{Kurtosis}: Approximated numerically for ordered categories.
}
\examples{
dist <- dist_categorical(prob = list(c(0.05, 0.5, 0.15, 0.2, 0.1), c(0.3, 0.1, 0.6)))

dist

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

# The outcomes aren't ordered, so many statistics are not applicable.
cdf(dist, 0.6)
quantile(dist, 0.7)
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

# Some of these statistics are meaningful for ordered outcomes
dist <- dist_categorical(list(rpois(26, 3)), list(ordered(letters)))
dist
cdf(dist, "m")
quantile(dist, 0.5)

dist <- dist_categorical(
  prob = list(c(0.05, 0.5, 0.15, 0.2, 0.1), c(0.3, 0.1, 0.6)),
  outcomes = list(letters[1:5], letters[24:26])
)

generate(dist, 10)

density(dist, "a")
density(dist, "z", log = TRUE)

}
\seealso{
\link[stats:Multinom]{stats::Multinomial}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{generate.distribution}
\alias{generate.distribution}
\title{Randomly sample values from a distribution}
\usage{
\method{generate}{distribution}(x, times, ...)
}
\arguments{
\item{x}{The distribution(s).}

\item{times}{The number of samples.}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Generate random samples from probability distributions.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale-level.R
\name{guide_train.level_guide}
\alias{guide_train.level_guide}
\alias{guide_geom.guide_level}
\title{Helper methods for guides}
\usage{
\method{guide_train}{level_guide}(guide, scale, aesthetic)

\method{guide_geom}{guide_level}(guide, layers, default_mapping)
}
\description{
Helper methods for guides
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_percentile.R
\name{dist_percentile}
\alias{dist_percentile}
\title{Percentile distribution}
\usage{
dist_percentile(x, percentile)
}
\arguments{
\item{x}{A list of values}

\item{percentile}{A list of percentiles}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Percentile distribution is a non-parametric distribution defined by
a set of quantiles at specified percentile values. This distribution is
useful for representing empirical distributions or elicited expert
knowledge when only percentile information is available. The distribution
uses linear interpolation between percentiles and can be used to
approximate complex distributions that may not have simple parametric forms.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_percentile.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_percentile.html}

In the following, let \eqn{X} be a Percentile random variable defined by
values \eqn{x_1, x_2, \ldots, x_n} at percentiles
\eqn{p_1, p_2, \ldots, p_n} where \eqn{0 \le p_i \le 100}.

\strong{Support}: \eqn{[\min(x_i), \max(x_i)]} if \eqn{\min(p_i) > 0} or
\eqn{\max(p_i) < 100}, otherwise support is approximated from the
specified percentiles.

\strong{Mean}: Approximated numerically using spline interpolation and
numerical integration:

\deqn{
    E(X) \approx \int_0^1 Q(u) du
  }{
    E(X) ≈ integral_0^1 Q(u) du
  }

  where \eqn{Q(u)} is a spline function interpolating the percentile values.

\strong{Variance}: Approximated numerically.

\strong{Probability density function (p.d.f)}: Approximated numerically using
kernel density estimation from generated samples.

\strong{Cumulative distribution function (c.d.f)}: Defined by linear
interpolation:

\deqn{
    F(t) = \begin{cases}
      p_1/100 & \text{if } t < x_1 \\
      p_i/100 + \frac{(t - x_i)(p_{i+1} - p_i)}{100(x_{i+1} - x_i)} & \text{if } x_i \le t < x_{i+1} \\
      p_n/100 & \text{if } t \ge x_n
    \end{cases}
  }{
    F(t) = p_i/100 + (t - x_i)(p_{i+1} - p_i) / (100(x_{i+1} - x_i)) for x_i ≤ t < x_{i+1}
  }

  \strong{Quantile function}: Defined by linear interpolation:

\deqn{
    Q(u) = x_i + \frac{(100u - p_i)(x_{i+1} - x_i)}{p_{i+1} - p_i}
  }{
    Q(u) = x_i + (100u - p_i)(x_{i+1} - x_i) / (p_{i+1} - p_i)
  }

for \eqn{p_i/100 \le u \le p_{i+1}/100}.
}
\examples{
dist <- dist_normal()
percentiles <- seq(0.01, 0.99, by = 0.01)
x <- vapply(percentiles, quantile, double(1L), x = dist)
dist_percentile(list(x), list(percentiles*100))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_student_t.R
\name{dist_student_t}
\alias{dist_student_t}
\title{The (non-central) location-scale Student t Distribution}
\usage{
dist_student_t(df, mu = 0, sigma = 1, ncp = NULL)
}
\arguments{
\item{df}{degrees of freedom (\eqn{> 0}, maybe non-integer).  \code{df
      = Inf} is allowed.}

\item{mu}{The location parameter of the distribution.
If \code{ncp == 0} (or \code{NULL}), this is the median.}

\item{sigma}{The scale parameter of the distribution.}

\item{ncp}{non-centrality parameter \eqn{\delta}{delta};
    currently except for \code{rt()}, accurate only for \code{abs(ncp) <= 37.62}.
    If omitted, use the central t distribution.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Student's T distribution is closely related to the \code{\link[=Normal]{Normal()}}
distribution, but has heavier tails. As \eqn{\nu} increases to \eqn{\infty},
the Student's T converges to a Normal. The T distribution appears
repeatedly throughout classic frequentist hypothesis testing when
comparing group means.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_student_t.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_student_t.html}

In the following, let \eqn{X} be a location-scale Student's T random variable with
\code{df} = \eqn{\nu}, \code{mu} = \eqn{\mu}, \code{sigma} = \eqn{\sigma}, and
\code{ncp} = \eqn{\delta} (non-centrality parameter).

If \eqn{Z} follows a standard Student's T distribution (with \code{df} = \eqn{\nu}
and \code{ncp} = \eqn{\delta}), then \eqn{X = \mu + \sigma Z}.

\strong{Support}: \eqn{R}, the set of all real numbers

\strong{Mean}:

For the central distribution (\code{ncp} = 0 or \code{NULL}):

\deqn{
    E(X) = \mu
  }{
    E(X) = \mu
  }

for \eqn{\nu > 1}, and undefined otherwise.

For the non-central distribution (\code{ncp} \eqn{\neq} 0):

\deqn{
    E(X) = \mu + \delta \sqrt{\frac{\nu}{2}} \frac{\Gamma((\nu-1)/2)}{\Gamma(\nu/2)} \sigma
  }{
    E(X) = \mu + \delta \sqrt(\nu / 2) \Gamma((\nu-1)/2) / \Gamma(\nu/2) \sigma
  }

for \eqn{\nu > 1}, and undefined otherwise.

\strong{Variance}:

For the central distribution (\code{ncp} = 0 or \code{NULL}):

\deqn{
    \mathrm{Var}(X) = \frac{\nu}{\nu - 2} \sigma^2
  }{
    Var(X) = \nu / (\nu - 2) \sigma^2
  }

for \eqn{\nu > 2}. Undefined if \eqn{\nu \le 1}, infinite when \eqn{1 < \nu \le 2}.

For the non-central distribution (\code{ncp} \eqn{\neq} 0):

\deqn{
    \mathrm{Var}(X) = \left[\frac{\nu(1+\delta^2)}{\nu-2} - \left(\delta \sqrt{\frac{\nu}{2}} \frac{\Gamma((\nu-1)/2)}{\Gamma(\nu/2)}\right)^2\right] \sigma^2
  }{
    Var(X) = [(\nu(1+\delta^2))/(\nu-2) - (\delta \sqrt(\nu / 2) \Gamma((\nu-1)/2) / \Gamma(\nu/2))^2] \sigma^2
  }

for \eqn{\nu > 2}. Undefined if \eqn{\nu \le 1}, infinite when \eqn{1 < \nu \le 2}.

\strong{Probability density function (p.d.f)}:

For the central distribution (\code{ncp} = 0 or \code{NULL}), the standard
t distribution with \code{df} = \eqn{\nu} has density:

\deqn{
    f_Z(z) = \frac{\Gamma((\nu + 1)/2)}{\sqrt{\pi \nu} \Gamma(\nu/2)} \left(1 + \frac{z^2}{\nu} \right)^{- (\nu + 1)/2}
  }{
    f_Z(z) = \Gamma((\nu + 1) / 2) / (\sqrt(\pi \nu) \Gamma(\nu / 2)) (1 + z^2 / \nu)^(- (\nu + 1) / 2)
  }

The location-scale version with \code{mu} = \eqn{\mu} and \code{sigma} = \eqn{\sigma}
has density:

\deqn{
    f(x) = \frac{1}{\sigma} f_Z\left(\frac{x - \mu}{\sigma}\right)
  }{
    f(x) = (1/\sigma) f_Z((x - \mu) / \sigma)
  }

For the non-central distribution (\code{ncp} \eqn{\neq} 0), the density is
computed numerically via \code{\link[stats:TDist]{stats::dt()}}.

\strong{Cumulative distribution function (c.d.f)}:

For the central distribution (\code{ncp} = 0 or \code{NULL}), the cumulative
distribution function is computed numerically via \code{\link[stats:TDist]{stats::pt()}}, which
uses the relationship to the incomplete beta function:

\deqn{
    F_\nu(t) = \frac{1}{2} I_x\left(\frac{\nu}{2}, \frac{1}{2}\right)
  }{
    F_\nu(t) = (1/2) I_x(\nu/2, 1/2)
  }

for \eqn{t \le 0}, where \eqn{x = \nu/(\nu + t^2)} and \eqn{I_x(a,b)} is
the incomplete beta function (\code{\link[stats:Beta]{stats::pbeta()}}). For \eqn{t \ge 0}:

\deqn{
    F_\nu(t) = 1 - \frac{1}{2} I_x\left(\frac{\nu}{2}, \frac{1}{2}\right)
  }{
    F_\nu(t) = 1 - (1/2) I_x(\nu/2, 1/2)
  }

The location-scale version is: \eqn{F(x) = F_\nu((x - \mu)/\sigma)}.

For the non-central distribution (\code{ncp} \eqn{\neq} 0), the cumulative
distribution function is computed numerically via \code{\link[stats:TDist]{stats::pt()}}.

\strong{Moment generating function (m.g.f)}:

Does not exist in closed form. Moments are computed using the formulas
for mean and variance above where available.
}
\examples{
dist <- dist_student_t(df = c(1,2,5), mu = c(0,1,2), sigma = c(1,2,3))

dist
mean(dist)
variance(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:TDist]{stats::TDist}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{is_distribution}
\alias{is_distribution}
\title{Test if the object is a distribution}
\usage{
is_distribution(x)
}
\arguments{
\item{x}{An object.}
}
\value{
TRUE if the object inherits from the distribution class.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

This function returns \code{TRUE} for distributions and \code{FALSE} for all other objects.
}
\examples{
dist <- dist_normal()
is_distribution(dist)
is_distribution("distributional")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_multivariate_normal.R
\name{dist_multivariate_normal}
\alias{dist_multivariate_normal}
\title{The multivariate normal distribution}
\usage{
dist_multivariate_normal(mu = 0, sigma = list(diag(1)))
}
\arguments{
\item{mu}{A list of numeric vectors for the distribution's mean.}

\item{sigma}{A list of matrices for the distribution's variance-covariance matrix.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The multivariate normal distribution is a generalization of the univariate
normal distribution to higher dimensions. It is widely used in multivariate
statistics and describes the joint distribution of multiple correlated
continuous random variables.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_multivariate_normal.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_multivariate_normal.html}

In the following, let \eqn{\mathbf{X}} be a \eqn{k}-dimensional multivariate
normal random variable with mean vector \code{mu} = \eqn{\boldsymbol{\mu}} and
variance-covariance matrix \code{sigma} = \eqn{\boldsymbol{\Sigma}}.

\strong{Support}: \eqn{\mathbf{x} \in \mathbb{R}^k}

\strong{Mean}: \eqn{\boldsymbol{\mu}}

\strong{Variance-covariance matrix}: \eqn{\boldsymbol{\Sigma}}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(\mathbf{x}) = \frac{1}{(2\pi)^{k/2} |\boldsymbol{\Sigma}|^{1/2}}
    \exp\left(-\frac{1}{2}(\mathbf{x} - \boldsymbol{\mu})^T
    \boldsymbol{\Sigma}^{-1}(\mathbf{x} - \boldsymbol{\mu})\right)
  }{
    f(x) = 1 / ((2\pi)^(k/2) |\Sigma|^(1/2))
    exp(-1/2 (x - \mu)^T \Sigma^(-1) (x - \mu))
  }

where \eqn{|\boldsymbol{\Sigma}|}{|\Sigma|} is the determinant of
\eqn{\boldsymbol{\Sigma}}{\Sigma}.

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    P(\mathbf{X} \le \mathbf{q}) = P(X_1 \le q_1, \ldots, X_k \le q_k)
  }{
    P(X \le q) = P(X_1 \le q_1, ..., X_k \le q_k)
  }

The c.d.f. does not have a closed-form expression and is computed numerically.

\strong{Moment generating function (m.g.f)}:

\deqn{
    M(\mathbf{t}) = E(e^{\mathbf{t}^T \mathbf{X}}) =
    \exp\left(\mathbf{t}^T \boldsymbol{\mu} + \frac{1}{2}\mathbf{t}^T
    \boldsymbol{\Sigma} \mathbf{t}\right)
  }{
    M(t) = E(e^(t^T X)) = exp(t^T \mu + 1/2 t^T \Sigma t)
  }
}
\examples{
dist <- dist_multivariate_normal(mu = list(c(1,2)), sigma = list(matrix(c(4,2,2,3), ncol=2)))
dimnames(dist) <- c("x", "y")
dist

\dontshow{if (requireNamespace("mvtnorm", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, cbind(2, 1))
density(dist, cbind(2, 1), log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7, kind = "equicoordinate")
quantile(dist, 0.7, kind = "marginal")
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[mvtnorm:Mvnorm]{mvtnorm::dmvnorm()}}, \code{\link[mvtnorm:pmvnorm]{mvtnorm::pmvnorm()}}, \code{\link[mvtnorm:qmvnorm]{mvtnorm::qmvnorm()}},
\code{\link[mvtnorm:Mvnorm]{mvtnorm::rmvnorm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hilo.R
\name{is_hilo}
\alias{is_hilo}
\title{Is the object a hilo}
\usage{
is_hilo(x)
}
\arguments{
\item{x}{An object.}
}
\description{
Is the object a hilo
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_inverse_exponential.R
\name{dist_inverse_exponential}
\alias{dist_inverse_exponential}
\title{The Inverse Exponential distribution}
\usage{
dist_inverse_exponential(rate)
}
\arguments{
\item{rate}{an alternative way to specify the scale.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Inverse Exponential distribution is used to model the reciprocal of
exponentially distributed variables.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_inverse_exponential.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_inverse_exponential.html}

In the following, let \eqn{X} be an Inverse Exponential random variable
with parameter \code{rate} = \eqn{\lambda}.

\strong{Support}: \eqn{x > 0}

\strong{Mean}: Does not exist, returns NA

\strong{Variance}: Does not exist, returns NA

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{\lambda}{x^2} e^{-\lambda/x}
  }{
    f(x) = \lambda / x^2 * e^(-\lambda/x)
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = e^{-\lambda/x}
  }{
    F(x) = e^(-\lambda/x)
  }

\strong{Quantile function (inverse c.d.f)}:

\deqn{
    F^{-1}(p) = -\frac{\lambda}{\log(p)}
  }{
    F^(-1)(p) = -\lambda / log(p)
  }

\strong{Moment generating function (m.g.f)}:

Does not exist (divergent integral).
}
\examples{
dist <- dist_inverse_exponential(rate = 1:5)
dist

\dontshow{if (requireNamespace("actuar", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
\dontshow{\}) # examplesIf}
}
\seealso{
\link[actuar:InverseExponential]{actuar::InverseExponential}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_truncated.R
\name{dist_truncated}
\alias{dist_truncated}
\title{Truncate a distribution}
\usage{
dist_truncated(dist, lower = -Inf, upper = Inf)
}
\arguments{
\item{dist}{The distribution(s) to truncate.}

\item{lower, upper}{The range of values to keep from a distribution.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Note that the samples are generated using inverse transform sampling, and the
means and variances are estimated from samples.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_truncated.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_truncated.html}

In the following, let \eqn{X} be a truncated random variable with
underlying distribution \eqn{Y}, truncation bounds \code{lower} = \eqn{a} and
\code{upper} = \eqn{b}, where \eqn{F_Y(x)} is the c.d.f. of \eqn{Y} and
\eqn{f_Y(x)} is the p.d.f. of \eqn{Y}.

\strong{Support}: \eqn{[a, b]}

\strong{Mean}: For the general case, the mean is approximated numerically.
For a truncated Normal distribution with underlying mean \eqn{\mu} and
standard deviation \eqn{\sigma}, the mean is:

\deqn{
    E(X) = \mu + \frac{\phi(\alpha) - \phi(\beta)}{\Phi(\beta) - \Phi(\alpha)} \sigma
  }{
    E(X) = mu + (phi(alpha) - phi(beta)) / (Phi(beta) - Phi(alpha)) * sigma
  }

where \eqn{\alpha = (a - \mu)/\sigma}, \eqn{\beta = (b - \mu)/\sigma},
\eqn{\phi} is the standard Normal p.d.f., and \eqn{\Phi} is the
standard Normal c.d.f.

\strong{Variance}: Approximated numerically for all distributions.

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \begin{cases}
      \frac{f_Y(x)}{F_Y(b) - F_Y(a)} & \text{if } a \le x \le b \\
      0 & \text{otherwise}
    \end{cases}
  }{
    f(x) = f_Y(x) / (F_Y(b) - F_Y(a)) if a <= x <= b, 0 otherwise
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = \begin{cases}
      0 & \text{if } x < a \\
      \frac{F_Y(x) - F_Y(a)}{F_Y(b) - F_Y(a)} & \text{if } a \le x \le b \\
      1 & \text{if } x > b
    \end{cases}
  }{
    F(x) = 0 if x < a, (F_Y(x) - F_Y(a)) / (F_Y(b) - F_Y(a)) if a <= x <= b, 1 if x > b
  }

\strong{Quantile function}:

\deqn{
    Q(p) = F_Y^{-1}(F_Y(a) + p(F_Y(b) - F_Y(a)))
  }{
    Q(p) = F_Y^(-1)(F_Y(a) + p(F_Y(b) - F_Y(a)))
  }

clamped to the interval \eqn{[a, b]}.
}
\examples{
dist <- dist_truncated(dist_normal(2,1), lower = 0)

dist
mean(dist)
variance(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

if(requireNamespace("ggdist")) {
library(ggplot2)
ggplot() +
  ggdist::stat_dist_halfeye(
    aes(y = c("Normal", "Truncated"),
        dist = c(dist_normal(2,1), dist_truncated(dist_normal(2,1), lower = 0)))
  )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{likelihood}
\alias{likelihood}
\alias{likelihood.distribution}
\alias{log_likelihood}
\title{The (log) likelihood of a sample matching a distribution}
\usage{
likelihood(x, ...)

\method{likelihood}{distribution}(x, sample, ..., log = FALSE)

log_likelihood(x, ...)
}
\arguments{
\item{x}{The distribution(s).}

\item{...}{Additional arguments used by methods.}

\item{sample}{A list of sampled values to compare to distribution(s).}

\item{log}{If \code{TRUE}, the log-likelihood will be computed.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_convolved.R
\name{dist_convolved}
\alias{dist_convolved}
\title{A convolved distribution}
\usage{
dist_convolved(...)
}
\arguments{
\item{...}{Two or more distribution vectors to sum. All will be recycled to
a common length.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Creates the distribution of the sum of two or more independent random
variables using numerical convolution of their densities. The density of
the result is approximated on a grid using a Fast Fourier Transform (FFT),
and then interpolated with \code{\link[stats:approxfun]{stats::approxfun()}}.

The cumulative distribution function and quantile function are derived from
the approximated density via numerical integration and inversion respectively.

This is primarily intended to be used via arithmetic on distributions:
\code{dist1 + dist2} or \code{dist1 + dist2 + dist3 + ...}. Distributions with known
closed-form sums (e.g. two \code{\link[=dist_normal]{dist_normal()}}) will use the exact result rather
than this approximation. Chaining \code{+} automatically performs a single k-way
FFT convolution rather than nested binary convolutions, avoiding compounding
approximation errors.
}
\details{
Let \eqn{Z = X_1 + X_2 + \cdots + X_k} where the \eqn{X_i} are
independent random variables.

\strong{Mean}:
\deqn{E(Z) = \sum_{i=1}^{k} E(X_i)}

\strong{Variance}:
\deqn{\mathrm{Var}(Z) = \sum_{i=1}^{k} \mathrm{Var}(X_i)}

\strong{Probability density function (p.d.f)}:
\deqn{f_Z(z) = (f_{X_1} * f_{X_2} * \cdots * f_{X_k})(z)}

The k-way convolution is computed in a single FFT pass: each component
density is evaluated on a common grid, transformed via FFT, all transforms
are multiplied element-wise, and a single inverse FFT yields the result.
This avoids compounding approximation errors from nested binary convolutions.

The accuracy of the FFT approximation can be controlled by passing \code{n}
(number of grid points, default \code{2^12}) and \code{tail_p} (tail probability used
to find finite grid bounds for distributions with infinite support, default
\code{1e-6}) to \code{density()}, \code{cdf()}, \code{quantile()}, or \code{generate()}.
}
\examples{
# Sum of a lognormal and an exponential (no closed-form result)
d <- dist_convolved(dist_lognormal(0, 1), dist_exponential(1))
d

density(d, 2)
cdf(d, 2)
quantile(d, 0.5)
generate(d, 5)

# Three distributions from different families
d3 <- dist_convolved(dist_lognormal(0, 0.5), dist_gamma(2, 1), dist_exponential(2))
d3

# Via arithmetic
d2 <- dist_lognormal(0, 1) + dist_exponential(1)
density(d2, 2)

# Mean and variance are computed exactly from components
mean(d)
variance(d)

}
\seealso{
\code{\link[stats:convolve]{stats::convolve()}}, \code{\link[stats:approxfun]{stats::approxfun()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_burr.R
\name{dist_burr}
\alias{dist_burr}
\title{The Burr distribution}
\usage{
dist_burr(shape1, shape2, rate = 1, scale = 1/rate)
}
\arguments{
\item{shape1, shape2, scale}{parameters. Must be strictly positive.}

\item{rate}{an alternative way to specify the scale.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Burr distribution (Type XII) is a flexible continuous probability
distribution often used for modeling income distributions, reliability
data, and failure times.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_burr.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_burr.html}

In the following, let \eqn{X} be a Burr random variable with parameters
\code{shape1} = \eqn{\alpha}, \code{shape2} = \eqn{\gamma}, and \code{rate} = \eqn{\lambda}.

\strong{Support}: \eqn{x \in (0, \infty)}{x > 0}

\strong{Mean}: \eqn{\frac{\lambda^{-1/\alpha} \gamma B(\gamma - 1/\alpha, 1 + 1/\alpha)}{\gamma}} (for \eqn{\alpha \gamma > 1})

\strong{Variance}: \eqn{\frac{\lambda^{-2/\alpha} \gamma B(\gamma - 2/\alpha, 1 + 2/\alpha)}{\gamma} - \mu^2} (for \eqn{\alpha \gamma > 2})

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \alpha \gamma \lambda x^{\alpha - 1} (1 + \lambda x^\alpha)^{-\gamma - 1}
  }{
    f(x) = \alpha \gamma \lambda x^(\alpha - 1) (1 + \lambda x^\alpha)^(-\gamma - 1)
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = 1 - (1 + \lambda x^\alpha)^{-\gamma}
  }{
    F(x) = 1 - (1 + \lambda x^\alpha)^(-\gamma)
  }

\strong{Quantile function}:

\deqn{
    F^{-1}(p) = \lambda^{-1/\alpha} ((1 - p)^{-1/\gamma} - 1)^{1/\alpha}
  }{
    F^(-1)(p) = \lambda^(-1/\alpha) ((1 - p)^(-1/\gamma) - 1)^(1/\alpha)
  }

\strong{Moment generating function (m.g.f)}:

Does not exist in closed form.
}
\examples{
dist <- dist_burr(shape1 = c(1,1,1,2,3,0.5), shape2 = c(1,2,3,1,1,2))
dist

\dontshow{if (requireNamespace("actuar", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
\dontshow{\}) # examplesIf}
}
\seealso{
\link[actuar:Burr]{actuar::Burr}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{support}
\alias{support}
\alias{support.distribution}
\title{Region of support of a distribution}
\usage{
support(x, ...)

\method{support}{distribution}(x, ...)
}
\arguments{
\item{x}{The distribution(s).}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hdr.R
\name{new_hdr}
\alias{new_hdr}
\title{Construct hdr intervals}
\usage{
new_hdr(
  lower = list_of(.ptype = double()),
  upper = list_of(.ptype = double()),
  size = double()
)
}
\arguments{
\item{lower, upper}{A list of numeric vectors specifying the region's lower
and upper bounds.}

\item{size}{A numeric vector specifying the coverage size of the region.}
}
\value{
A "hdr" vector
}
\description{
Construct hdr intervals
}
\examples{

new_hdr(lower = list(1, c(3,6)), upper = list(10, c(5, 8)), size = c(80, 95))

}
\author{
Mitchell O'Hara-Wild
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_poisson_inverse_gaussian.R
\name{dist_poisson_inverse_gaussian}
\alias{dist_poisson_inverse_gaussian}
\title{The Poisson-Inverse Gaussian distribution}
\usage{
dist_poisson_inverse_gaussian(mean, shape)
}
\arguments{
\item{mean, shape}{parameters. Must be strictly positive. Infinite
    values are supported.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Poisson-Inverse Gaussian distribution is a compound Poisson distribution
where the rate parameter follows an Inverse Gaussian distribution. It is
useful for modeling overdispersed count data.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_poisson_inverse_gaussian.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_poisson_inverse_gaussian.html}

In the following, let \eqn{X} be a Poisson-Inverse Gaussian random variable
with parameters \code{mean} = \eqn{\mu} and \code{shape} = \eqn{\phi}.

\strong{Support}: \eqn{\{0, 1, 2, 3, ...\}}{{0, 1, 2, 3, ...}}

\strong{Mean}: \eqn{\mu}

\strong{Variance}: \eqn{\frac{\mu}{\phi}(\mu^2 + \phi)}{\mu/\phi (\mu^2 + \phi)}

\strong{Probability mass function (p.m.f)}:

\deqn{
    P(X = x) = \frac{e^{\phi}}{\sqrt{2\pi}}
    \left(\frac{\phi}{\mu^2}\right)^{x/2}
    \frac{1}{x!}
    \int_0^\infty u^{x-1/2}
    \exp\left(-\frac{\phi u}{2} - \frac{\phi}{2\mu^2 u}\right) du
  }{
    P(X = x) = (e^\phi / sqrt(2\pi)) (\phi/\mu^2)^(x/2) (1/x!)
    \int_0^\infty u^(x-1/2) exp(-\phi u/2 - \phi/(2\mu^2 u)) du
  }

for \eqn{x = 0, 1, 2, \ldots}{x = 0, 1, 2, ...}

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    P(X \le x) = \sum_{k=0}^{\lfloor x \rfloor} P(X = k)
  }{
    P(X <= x) = sum_{k=0}^x P(X = k)
  }

The c.d.f does not have a closed form and is approximated numerically.

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \exp\left\{\phi\left[1 - \sqrt{1 - \frac{2\mu^2}{\phi}(e^t - 1)}\right]\right\}
  }{
    E(e^(tX)) = exp{\phi [1 - sqrt(1 - 2\mu^2/\phi (e^t - 1))]}
  }

for \eqn{t < -\log(1 + \phi/(2\mu^2))}{t < -log(1 + \phi/(2\mu^2))}
}
\examples{
dist <- dist_poisson_inverse_gaussian(mean = rep(0.1, 3), shape = c(0.4, 0.8, 1))
dist

\dontshow{if (requireNamespace("actuar", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
\dontshow{\}) # examplesIf}
}
\seealso{
\link[actuar:PoissonInverseGaussian]{actuar::PoissonInverseGaussian}, \code{\link[actuar:PoissonInverseGaussian]{actuar::dpoisinvgauss()}},
\code{\link[actuar:PoissonInverseGaussian]{actuar::ppoisinvgauss()}}, \code{\link[actuar:PoissonInverseGaussian]{actuar::qpoisinvgauss()}}, \code{\link[actuar:PoissonInverseGaussian]{actuar::rpoisinvgauss()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_dirichlet.R
\name{dist_dirichlet}
\alias{dist_dirichlet}
\title{The Dirichlet distribution}
\usage{
dist_dirichlet(alpha)
}
\arguments{
\item{alpha}{A list of positive numeric concentration vectors.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Dirichlet distribution is a multivariate generalisation of the Beta
distribution. It is the conjugate prior of the Categorical and Multinomial
distributions, and describes a probability distribution over the
\eqn{(k-1)}-simplex — the set of \eqn{k}-dimensional vectors whose
components are non-negative and sum to one.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_dirichlet.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_dirichlet.html}

In the following, let \eqn{\mathbf{X} = (X_1, \ldots, X_k)} be a
Dirichlet random variable with concentration parameter
\code{alpha} = \eqn{\boldsymbol{\alpha} = (\alpha_1, \ldots, \alpha_k)},
where each \eqn{\alpha_i > 0}.

\strong{Support}: \eqn{\mathbf{x}} on the \eqn{(k-1)}-simplex,
i.e. \eqn{x_i \geq 0} and \eqn{\sum_{i=1}^k x_i = 1}.

\strong{Mean}: \eqn{E(X_i) = \frac{\alpha_i}{\alpha_0}} where
\eqn{\alpha_0 = \sum_{i=1}^k \alpha_i}.

\strong{Variance}:

\deqn{
    \mathrm{Var}(X_i) = \frac{\alpha_i(\alpha_0 - \alpha_i)}{\alpha_0^2(\alpha_0 + 1)}
  }{
    Var(X_i) = alpha_i * (alpha_0 - alpha_i) / (alpha_0^2 * (alpha_0 + 1))
  }

\strong{Covariance}:

\deqn{
    \mathrm{Cov}(X_i, X_j) = \frac{-\alpha_i \alpha_j}{\alpha_0^2(\alpha_0 + 1)}, \quad i \neq j
  }{
    Cov(X_i, X_j) = -alpha_i * alpha_j / (alpha_0^2 * (alpha_0 + 1)), i != j
  }

\strong{Probability density function (p.d.f)}:

\deqn{
    f(\mathbf{x}) = \frac{1}{B(\boldsymbol{\alpha})}
    \prod_{i=1}^k x_i^{\alpha_i - 1}
  }{
    f(x) = 1 / B(alpha) * prod(x_i^(alpha_i - 1))
  }

where \eqn{B(\boldsymbol{\alpha}) = \frac{\prod_{i=1}^k \Gamma(\alpha_i)}{\Gamma(\alpha_0)}}
is the multivariate Beta function.
}
\examples{
dist <- dist_dirichlet(alpha = list(c(2, 5, 3)))
dist

\dontshow{if (requireNamespace("LaplacesDemon", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, cbind(0.2, 0.5, 0.3))
density(dist, cbind(0.2, 0.5, 0.3), log = TRUE)
\dontshow{\}) # examplesIf}
}
\seealso{
\code{\link[LaplacesDemon:dist.Dirichlet]{LaplacesDemon::ddirichlet()}}, \code{\link[LaplacesDemon:dist.Dirichlet]{LaplacesDemon::rdirichlet()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale-level.R
\name{guide_level}
\alias{guide_level}
\title{Level shade bar guide}
\usage{
guide_level(title = waiver(), max_discrete = 5, ...)
}
\arguments{
\item{title}{A character string or expression indicating a title of guide.
If \code{NULL}, the title is not shown. By default
(\code{\link[ggplot2:waiver]{waiver()}}), the name of the scale object or the name
specified in \code{\link[ggplot2:labs]{labs()}} is used for the title.}

\item{max_discrete}{The maximum number of levels to be shown using \code{\link[ggplot2:guide_legend]{ggplot2::guide_legend()}}.
If the number of levels exceeds this value, level shades are shown with \code{\link[ggplot2:guide_colourbar]{ggplot2::guide_colourbar()}}.}

\item{...}{Further arguments passed onto either \code{\link[ggplot2:guide_colourbar]{ggplot2::guide_colourbar()}} or \code{\link[ggplot2:guide_legend]{ggplot2::guide_legend()}}}
}
\description{
The level guide shows the colour from the forecast intervals which is blended with the series colour.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_geometric.R
\name{dist_geometric}
\alias{dist_geometric}
\title{The Geometric Distribution}
\usage{
dist_geometric(prob)
}
\arguments{
\item{prob}{probability of success in each trial. \code{0 < prob <= 1}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Geometric distribution can be thought of as a generalization
of the \code{\link[=dist_bernoulli]{dist_bernoulli()}} distribution where we ask: "if I keep flipping a
coin with probability \code{p} of heads, what is the probability I need
\eqn{k} flips before I get my first heads?" The Geometric
distribution is a special case of Negative Binomial distribution.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X} be a Geometric random variable with
success probability \code{p} = \eqn{p}. Note that there are multiple
parameterizations of the Geometric distribution.

\strong{Support}: 0 < p < 1, \eqn{x = 0, 1, \dots}

\strong{Mean}: \eqn{\frac{1-p}{p}}

\strong{Variance}: \eqn{\frac{1-p}{p^2}}

\strong{Probability mass function (p.m.f)}:

\deqn{
    P(X = x) = p(1-p)^x,
   }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    P(X \le x) = 1 - (1-p)^{x+1}
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \frac{pe^t}{1 - (1-p)e^t}
  }{
    E(e^{tX}) = \frac{pe^t}{1 - (1-p)e^t}
  }
}
\examples{
dist <- dist_geometric(prob = c(0.2, 0.5, 0.8))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
}
\seealso{
\link[stats:Geometric]{stats::Geometric}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_missing.R
\name{dist_missing}
\alias{dist_missing}
\title{Missing distribution}
\usage{
dist_missing(length = 1)
}
\arguments{
\item{length}{The number of missing distributions}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#maturing}{\figure{lifecycle-maturing.svg}{options: alt='[Maturing]'}}}{\strong{[Maturing]}}

A placeholder distribution for handling missing values in a vector of
distributions.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_missing.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_missing.html}

The missing distribution represents the absence of distributional
information. It is used as a placeholder when distribution values are
not available or not applicable, similar to how \code{NA} is used for missing
scalar values.

\strong{Support}: Undefined

\strong{Mean}: \eqn{\text{NA}}

\strong{Variance}: \eqn{\text{NA}}

\strong{Skewness}: \eqn{\text{NA}}

\strong{Kurtosis}: \eqn{\text{NA}}

\strong{Probability density function (p.d.f)}: Undefined

\deqn{
    f(x) = \text{NA}
  }

\strong{Cumulative distribution function (c.d.f)}: Undefined

\deqn{
    F(t) = \text{NA}
  }

\strong{Quantile function}: Undefined

\deqn{
    Q(p) = \text{NA}
  }

\strong{Moment generating function (m.g.f)}: Undefined

\deqn{
    E(e^{tX}) = \text{NA}
  }

All statistical operations on missing distributions return \code{NA} values
of appropriate length, propagating the missingness through calculations.
}
\examples{
dist <- dist_missing(3L)

dist
mean(dist)
variance(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[base:NA]{base::NA}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_gamma.R
\name{dist_gamma}
\alias{dist_gamma}
\title{The Gamma distribution}
\usage{
dist_gamma(shape, rate, scale = 1/rate)
}
\arguments{
\item{shape, scale}{shape and scale parameters.  Must be positive,
    \code{scale} strictly.}

\item{rate}{an alternative way to specify the scale.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Several important distributions are special cases of the Gamma
distribution. When the shape parameter is \code{1}, the Gamma is an
exponential distribution with parameter \eqn{1/\beta}. When the
\eqn{shape = n/2} and \eqn{rate = 1/2}, the Gamma is a equivalent to
a chi squared distribution with n degrees of freedom. Moreover, if
we have \eqn{X_1} is \eqn{Gamma(\alpha_1, \beta)} and
\eqn{X_2} is \eqn{Gamma(\alpha_2, \beta)}, a function of these two variables
of the form \eqn{\frac{X_1}{X_1 + X_2}} \eqn{Beta(\alpha_1, \alpha_2)}.
This last property frequently appears in another distributions, and it
has extensively been used in multivariate methods. More about the Gamma
distribution will be added soon.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X} be a Gamma random variable
with parameters
\code{shape} = \eqn{\alpha} and
\code{rate} = \eqn{\beta}.

\strong{Support}: \eqn{x \in (0, \infty)}

\strong{Mean}: \eqn{\frac{\alpha}{\beta}}

\strong{Variance}: \eqn{\frac{\alpha}{\beta^2}}

\strong{Probability density function (p.m.f)}:

\deqn{
    f(x) = \frac{\beta^{\alpha}}{\Gamma(\alpha)} x^{\alpha - 1} e^{-\beta x}
  }{
    f(x) = \frac{\beta^{\alpha}}{\Gamma(\alpha)} x^{\alpha - 1} e^{-\beta x}
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    f(x) = \frac{\Gamma(\alpha, \beta x)}{\Gamma{\alpha}}
  }{
    f(x) = \frac{\Gamma(\alpha, \beta x)}{\Gamma{\alpha}}
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \Big(\frac{\beta}{ \beta - t}\Big)^{\alpha}, \thinspace t < \beta
  }{
    E(e^(tX)) = \Big(\frac{\beta}{ \beta - t}\Big)^{\alpha}, \thinspace t < \beta
  }
}
\examples{
dist <- dist_gamma(shape = c(1,2,3,5,9,7.5,0.5), rate = c(0.5,0.5,0.5,1,2,1,1))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:GammaDist]{stats::GammaDist}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_studentized_range.R
\name{dist_studentized_range}
\alias{dist_studentized_range}
\title{The Studentized Range distribution}
\usage{
dist_studentized_range(nmeans, df, nranges)
}
\arguments{
\item{nmeans}{sample size for range (same for each group).}

\item{df}{degrees of freedom for \eqn{s} (see below).}

\item{nranges}{number of \emph{groups} whose \bold{maximum} range is
    considered.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Tukey's studentized range distribution, used for Tukey's
honestly significant differences test in ANOVA.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_studentized_range.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_studentized_range.html}

In the following, let \eqn{Q} be a Studentized Range random variable with
parameters \code{nmeans} = \eqn{k} (number of groups), \code{df} = \eqn{\nu} (degrees
of freedom), and \code{nranges} = \eqn{n} (number of ranges).

\strong{Support}: \eqn{R^+}, the set of positive real numbers.

\strong{Mean}: Approximated numerically.

\strong{Variance}: Approximated numerically.

\strong{Probability density function (p.d.f)}: The density does not have a
closed-form expression and is computed numerically.

\strong{Cumulative distribution function (c.d.f)}: The c.d.f does not have a
simple closed-form expression. For \eqn{n = 1} (single range), it involves
integration over the joint distribution of the sample range and an
independent chi-square variable. The general form is computed numerically
using algorithms described in the references for \code{\link[stats:Tukey]{stats::ptukey()}}.

\strong{Moment generating function (m.g.f)}: Does not exist in closed form.
}
\examples{
dist <- dist_studentized_range(nmeans = c(6, 2), df = c(5, 4), nranges = c(1, 1))

dist

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Tukey]{stats::Tukey}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_mixture.R
\name{dist_mixture}
\alias{dist_mixture}
\title{Create a mixture of distributions}
\usage{
dist_mixture(..., weights = numeric())
}
\arguments{
\item{...}{Distributions to be used in the mixture. Can be any
distributional objects.}

\item{weights}{A numeric vector of non-negative weights that sum to 1.
The length must match the number of distributions passed to \code{...}.
Each weight \eqn{w_i} represents the probability that a random draw
comes from the \eqn{i}-th component distribution.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#maturing}{\figure{lifecycle-maturing.svg}{options: alt='[Maturing]'}}}{\strong{[Maturing]}}

A mixture distribution combines multiple component distributions with
specified weights. The resulting distribution can model complex,
multimodal data by representing it as a weighted sum of simpler
distributions.
}
\details{
In the following, let \eqn{X} be a mixture random variable composed
of \eqn{K} component distributions \eqn{F_1, F_2, \ldots, F_K} with
corresponding weights \eqn{w_1, w_2, \ldots, w_K} where
\eqn{\sum_{i=1}^K w_i = 1} and \eqn{w_i \geq 0} for all \eqn{i}.

\strong{Support}: The union of the supports of all component distributions

\strong{Mean}:

For univariate mixtures:
\deqn{
    E(X) = \sum_{i=1}^K w_i \mu_i
  }{
    E(X) = sum_{i=1}^K w_i * mu_i
  }

where \eqn{\mu_i} is the mean of the \eqn{i}-th component distribution.

For multivariate mixtures:
\deqn{
    E(\mathbf{X}) = \sum_{i=1}^K w_i \boldsymbol{\mu}_i
  }{
    E(X) = sum_{i=1}^K w_i * mu_i
  }

where \eqn{\boldsymbol{\mu}_i} is the mean vector of the \eqn{i}-th
component distribution.

\strong{Variance}:

For univariate mixtures:
\deqn{
    \text{Var}(X) = \sum_{i=1}^K w_i (\mu_i^2 + \sigma_i^2) - \left(\sum_{i=1}^K w_i \mu_i\right)^2
  }{
    Var(X) = sum_{i=1}^K w_i * (mu_i^2 + sigma_i^2) - (sum_{i=1}^K w_i * mu_i)^2
  }

where \eqn{\sigma_i^2} is the variance of the \eqn{i}-th component
distribution.

\strong{Covariance}:

For multivariate mixtures:
\deqn{
    \text{Cov}(\mathbf{X}) = \sum_{i=1}^K w_i \left[ (\boldsymbol{\mu}_i - \bar{\boldsymbol{\mu}})(\boldsymbol{\mu}_i - \bar{\boldsymbol{\mu}})^T + \boldsymbol{\Sigma}_i \right]
  }{
    Cov(X) = sum_{i=1}^K w_i * [ (mu_i - mu_bar)(mu_i - mu_bar)^T + Sigma_i ]
  }

where \eqn{\bar{\boldsymbol{\mu}} = \sum_{i=1}^K w_i \boldsymbol{\mu}_i}
is the overall mean vector and \eqn{\boldsymbol{\Sigma}_i} is the
covariance matrix of the \eqn{i}-th component distribution.

\strong{Probability density/mass function (p.d.f/p.m.f)}:

\deqn{
    f(x) = \sum_{i=1}^K w_i f_i(x)
  }{
    f(x) = sum_{i=1}^K w_i * f_i(x)
  }

where \eqn{f_i(x)} is the density or mass function of the \eqn{i}-th
component distribution.

\strong{Cumulative distribution function (c.d.f)}:

For univariate mixtures:
\deqn{
    F(x) = \sum_{i=1}^K w_i F_i(x)
  }{
    F(x) = sum_{i=1}^K w_i * F_i(x)
  }

where \eqn{F_i(x)} is the c.d.f. of the \eqn{i}-th component
distribution.

For multivariate mixtures, the c.d.f. is approximated numerically.

\strong{Quantile function}:

For univariate mixtures, the quantile function has no closed form
and is computed numerically by inverting the c.d.f. using root-finding
(\code{\link[stats:uniroot]{stats::uniroot()}}).

For multivariate mixtures, quantiles are not yet implemented.
}
\examples{
# Univariate mixture of two normal distributions
dist <- dist_mixture(dist_normal(0, 1), dist_normal(5, 2), weights = c(0.3, 0.7))
dist

mean(dist)
variance(dist)

density(dist, 2)
cdf(dist, 2)
quantile(dist, 0.5)

generate(dist, 10)

}
\seealso{
\code{\link[stats:uniroot]{stats::uniroot()}}, \code{\link[vctrs:vec_unique]{vctrs::vec_unique_count()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_hypergeometric.R
\name{dist_hypergeometric}
\alias{dist_hypergeometric}
\title{The Hypergeometric distribution}
\usage{
dist_hypergeometric(m, n, k)
}
\arguments{
\item{m}{The number of type I elements available.}

\item{n}{The number of type II elements available.}

\item{k}{The size of the sample taken.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

To understand the HyperGeometric distribution, consider a set of
\eqn{r} objects, of which \eqn{m} are of the type I and
\eqn{n} are of the type II. A sample with size \eqn{k} (\eqn{k<r})
with no replacement is randomly chosen. The number of observed
type I elements observed in this sample is set to be our random
variable \eqn{X}.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X} be a HyperGeometric random variable with
success probability \code{p} = \eqn{p = m/(m+n)}.

\strong{Support}: \eqn{x \in { \{\max{(0, k-n)}, \dots, \min{(k,m)}}\}}

\strong{Mean}: \eqn{\frac{km}{n+m} = kp}

\strong{Variance}: \eqn{\frac{km(n)(n+m-k)}{(n+m)^2 (n+m-1)} =
  kp(1-p)(1 - \frac{k-1}{m+n-1})}

\strong{Probability mass function (p.m.f)}:

\deqn{
    P(X = x) = \frac{{m \choose x}{n \choose k-x}}{{m+n \choose k}}
  }{
    P(X = x) = \frac{{m \choose x}{n \choose k-x}}{{m+n \choose k}}
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    P(X \le k) \approx \Phi\Big(\frac{x - kp}{\sqrt{kp(1-p)}}\Big)
 }
}
\examples{
dist <- dist_hypergeometric(m = rep(500, 3), n = c(50, 60, 70), k = c(100, 200, 300))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Hypergeometric]{stats::Hypergeometric}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_pareto.R
\name{dist_pareto}
\alias{dist_pareto}
\title{The Pareto Distribution}
\usage{
dist_pareto(shape, scale)
}
\arguments{
\item{shape, scale}{parameters. Must be strictly positive.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Pareto distribution is a power-law probability distribution commonly used
in actuarial science to model loss severity and in economics to model income
distributions and firm sizes.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_pareto.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_pareto.html}

In the following, let \eqn{X} be a Pareto random variable with parameters
\code{shape} = \eqn{\alpha} and \code{scale} = \eqn{\theta}.

\strong{Support}: \eqn{(0, \infty)}{(0, Inf)}

\strong{Mean}: \eqn{\frac{\theta}{\alpha - 1}}{\theta / (\alpha - 1)} for \eqn{\alpha > 1},
undefined otherwise

\strong{Variance}: \eqn{\frac{\alpha\theta^2}{(\alpha - 1)^2(\alpha - 2)}}{\alpha\theta^2 / ((\alpha - 1)^2(\alpha - 2))}
for \eqn{\alpha > 2}, undefined otherwise

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{\alpha\theta^\alpha}{(x + \theta)^{\alpha + 1}}
  }{
    f(x) = \alpha\theta^\alpha / (x + \theta)^(\alpha + 1)
  }

for \eqn{x > 0}, \eqn{\alpha > 0} and \eqn{\theta > 0}.

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = 1 - \left(\frac{\theta}{x + \theta}\right)^\alpha
  }{
    F(x) = 1 - (\theta / (x + \theta))^\alpha
  }

for \eqn{x > 0}.

\strong{Moment generating function (m.g.f)}:

Does not exist in closed form, but the \eqn{k}th raw moment \eqn{E[X^k]} exists
for \eqn{-1 < k < \alpha}.
}
\note{
There are many different definitions of the Pareto distribution in the
literature; see Arnold (2015) or Kleiber and Kotz (2003). This implementation
uses the Pareto distribution without a location parameter as described in \link[actuar:Pareto]{actuar::Pareto}.
}
\examples{
dist <- dist_pareto(shape = c(10, 3, 2, 1), scale = rep(1, 4))
dist

\dontshow{if (requireNamespace("actuar", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
\dontshow{\}) # examplesIf}
}
\references{
Kleiber, C. and Kotz, S. (2003), \emph{Statistical Size Distributions in Economics
and Actuarial Sciences}, Wiley.

Klugman, S. A., Panjer, H. H. and Willmot, G. E. (2012), \emph{Loss Models, From
Data to Decisions, Fourth Edition}, Wiley.
}
\seealso{
\link[actuar:Pareto]{actuar::Pareto}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_bernoulli.R
\name{dist_bernoulli}
\alias{dist_bernoulli}
\title{The Bernoulli distribution}
\usage{
dist_bernoulli(prob)
}
\arguments{
\item{prob}{The probability of success on each trial, \code{prob} can be any
value in \verb{[0, 1]}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Bernoulli distributions are used to represent events like coin flips
when there is single trial that is either successful or unsuccessful.
The Bernoulli distribution is a special case of the \code{\link[=Binomial]{Binomial()}}
distribution with \code{n = 1}.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X} be a Bernoulli random variable with parameter
\code{p} = \eqn{p}. Some textbooks also define \eqn{q = 1 - p}, or use
\eqn{\pi} instead of \eqn{p}.

The Bernoulli probability  distribution is widely used to model
binary variables, such as 'failure' and 'success'. The most
typical example is the flip of a coin, when  \eqn{p} is thought as the
probability of flipping a head, and \eqn{q = 1 - p} is the
probability of flipping a tail.

\strong{Support}: \eqn{\{0, 1\}}{{0, 1}}

\strong{Mean}: \eqn{p}

\strong{Variance}: \eqn{p \cdot (1 - p) = p \cdot q}{p (1 - p)}

\strong{Probability mass function (p.m.f)}:

\deqn{
    P(X = x) = p^x (1 - p)^{1-x} = p^x q^{1-x}
  }{
    P(X = x) = p^x (1 - p)^(1-x)
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    P(X \le x) =
    \left \{
      \begin{array}{ll}
        0 & x < 0 \\
        1 - p & 0 \leq x < 1 \\
        1 & x \geq 1
      \end{array}
    \right.
  }{
    P(X \le x) = (1 - p) 1_{[0, 1)}(x) + 1_{1}(x)
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = (1 - p) + p e^t
  }{
    E(e^(tX)) = (1 - p) + p e^t
  }
}
\examples{
dist <- dist_bernoulli(prob = c(0.05, 0.5, 0.3, 0.9, 0.1))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_logarithmic.R
\name{dist_logarithmic}
\alias{dist_logarithmic}
\title{The Logarithmic distribution}
\usage{
dist_logarithmic(prob)
}
\arguments{
\item{prob}{parameter. \code{0 <= prob < 1}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Logarithmic distribution is a discrete probability distribution
derived from the logarithmic series. It is useful in modeling the
abundance of species and other phenomena where the frequency of
an event follows a logarithmic pattern.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_logarithmic.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_logarithmic.html}

In the following, let \eqn{X} be a Logarithmic random variable with
parameter \code{prob} = \eqn{p}.

\strong{Support}: \eqn{\{1, 2, 3, ...\}}{{1, 2, 3, ...}}

\strong{Mean}: \eqn{\frac{-1}{\log(1-p)} \cdot \frac{p}{1-p}}{-1/log(1-p) * p/(1-p)}

\strong{Variance}: \eqn{\frac{-(p^2 + p\log(1-p))}{[(1-p)\log(1-p)]^2}}{-(p^2 + p*log(1-p))/((1-p)*log(1-p))^2}

\strong{Probability mass function (p.m.f)}:

\deqn{
    P(X = k) = \frac{-1}{\log(1-p)} \cdot \frac{p^k}{k}
  }{
    P(X = k) = -1/log(1-p) * p^k/k
  }

for \eqn{k = 1, 2, 3, \ldots}{k = 1, 2, 3, ...}

\strong{Cumulative distribution function (c.d.f)}:

The c.d.f. does not have a simple closed form. It is computed
using the recurrence relationship
\eqn{P(X = k+1) = \frac{p \cdot k}{k+1} \cdot P(X = k)}{P(X = k+1) = (p*k)/(k+1) * P(X = k)}
starting from \eqn{P(X = 1) = \frac{-p}{\log(1-p)}}{P(X = 1) = -p/log(1-p)}.

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \frac{\log(1 - pe^t)}{\log(1-p)}
  }{
    E(e^(tX)) = log(1 - p*e^t)/log(1-p)
  }

for \eqn{pe^t < 1}{p*e^t < 1}
}
\examples{
dist <- dist_logarithmic(prob = c(0.33, 0.66, 0.99))
dist

\dontshow{if (requireNamespace("actuar", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
\dontshow{\}) # examplesIf}
}
\seealso{
\link[actuar:Logarithmic]{actuar::Logarithmic}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{has_symmetry}
\alias{has_symmetry}
\title{Check if a distribution is symmetric}
\usage{
has_symmetry(x, ...)
}
\arguments{
\item{x}{The distribution(s).}

\item{...}{Additional arguments used by methods.}
}
\value{
A logical value indicating whether the distribution is symmetric.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}

Determines whether a probability distribution is symmetric around its center.
}
\examples{
# Normal distribution is symmetric
has_symmetry(dist_normal(mu = 0, sigma = 1))
has_symmetry(dist_normal(mu = 5, sigma = 2))

# Beta distribution symmetry depends on parameters
has_symmetry(dist_beta(shape1 = 2, shape2 = 2))  # symmetric
has_symmetry(dist_beta(shape1 = 2, shape2 = 5))  # not symmetric

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{new_dist}
\alias{new_dist}
\title{Create a new distribution}
\usage{
new_dist(..., class = NULL, dimnames = NULL)
}
\arguments{
\item{...}{Parameters of the distribution (named).}

\item{class}{The class of the distribution for S3 dispatch.}

\item{dimnames}{The names of the variables in the distribution (optional).}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#maturing}{\figure{lifecycle-maturing.svg}{options: alt='[Maturing]'}}}{\strong{[Maturing]}}

Allows extension package developers to define a new distribution class
compatible with the distributional package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_gev.R
\name{dist_gev}
\alias{dist_gev}
\title{The Generalized Extreme Value Distribution}
\usage{
dist_gev(location, scale, shape)
}
\arguments{
\item{location}{the location parameter \eqn{\mu} of the GEV distribution.}

\item{scale}{the scale parameter \eqn{\sigma} of the GEV distribution.
Must be strictly positive.}

\item{shape}{the shape parameter \eqn{\xi} of the GEV distribution.
Determines the tail behavior: \eqn{\xi = 0} gives Gumbel,
\eqn{\xi > 0} gives Frechet, \eqn{\xi < 0} gives reverse Weibull.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The GEV distribution is widely used in extreme value theory to model
the distribution of maxima (or minima) of samples. The parametric form
encompasses the Gumbel, Frechet, and reverse Weibull distributions.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_gev.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_gev.html}

In the following, let \eqn{X} be a GEV random variable with parameters
\code{location} = \eqn{\mu}, \code{scale} = \eqn{\sigma}, and \code{shape} = \eqn{\xi}.

\strong{Support}:
\itemize{
\item \eqn{x \in \mathbb{R}} (all real numbers) if \eqn{\xi = 0}
\item \eqn{x \geq \mu - \sigma/\xi} if \eqn{\xi > 0}
\item \eqn{x \leq \mu - \sigma/\xi} if \eqn{\xi < 0}
}

\strong{Mean}:
\deqn{
  E(X) = \begin{cases}
    \mu + \sigma \gamma & \text{if } \xi = 0 \\
    \mu + \sigma \frac{\Gamma(1-\xi) - 1}{\xi} & \text{if } \xi < 1 \\
    \infty & \text{if } \xi \geq 1
  \end{cases}
}{
  E(X) = \mu + \sigma \gamma  (if \xi = 0)
        \mu + \sigma (\Gamma(1-\xi) - 1) / \xi  (if \xi < 1)
        Inf  (if \xi >= 1)
}
where \eqn{\gamma \approx 0.5772} is the Euler-Mascheroni constant and
\eqn{\Gamma(\cdot)} is the gamma function.

\strong{Median}:
\deqn{
  \text{Median}(X) = \begin{cases}
    \mu - \sigma \log(\log 2) & \text{if } \xi = 0 \\
    \mu + \sigma \frac{(\log 2)^{-\xi} - 1}{\xi} & \text{if } \xi \neq 0
  \end{cases}
}{
  Median(X) = \mu - \sigma log(log 2)  (if \xi = 0)
              \mu + \sigma (log(2)^(-\xi) - 1) / \xi  (if \xi != 0)
}

\strong{Variance}:
\deqn{
  \text{Var}(X) = \begin{cases}
    \frac{\pi^2 \sigma^2}{6} & \text{if } \xi = 0 \\
    \frac{\sigma^2}{\xi^2} [\Gamma(1-2\xi) - \Gamma(1-\xi)^2] & \text{if } \xi < 0.5 \\
    \infty & \text{if } \xi \geq 0.5
  \end{cases}
}{
  Var(X) = \pi^2 \sigma^2 / 6  (if \xi = 0)
           \sigma^2 [\Gamma(1-2\xi) - \Gamma(1-\xi)^2] / \xi^2  (if \xi < 0.5)
           Inf  (if \xi >= 0.5)
}

\strong{Probability density function (p.d.f)}:

For \eqn{\xi = 0} (Gumbel):
\deqn{
  f(x) = \frac{1}{\sigma} \exp\left(-\frac{x-\mu}{\sigma}\right)
         \exp\left[-\exp\left(-\frac{x-\mu}{\sigma}\right)\right]
}{
  f(x) = (1/\sigma) exp(-(x-\mu)/\sigma) exp(-exp(-(x-\mu)/\sigma))
}

For \eqn{\xi \neq 0}:
\deqn{
  f(x) = \frac{1}{\sigma} \left[1 + \xi\left(\frac{x-\mu}{\sigma}\right)\right]^{-1/\xi-1}
         \exp\left\{-\left[1 + \xi\left(\frac{x-\mu}{\sigma}\right)\right]^{-1/\xi}\right\}
}{
  f(x) = (1/\sigma) [1 + \xi(x-\mu)/\sigma]^(-1/\xi-1) exp(-[1 + \xi(x-\mu)/\sigma]^(-1/\xi))
}
where \eqn{1 + \xi(x-\mu)/\sigma > 0}.

\strong{Cumulative distribution function (c.d.f)}:

For \eqn{\xi = 0} (Gumbel):
\deqn{
  F(x) = \exp\left[-\exp\left(-\frac{x-\mu}{\sigma}\right)\right]
}{
  F(x) = exp(-exp(-(x-\mu)/\sigma))
}

For \eqn{\xi \neq 0}:
\deqn{
  F(x) = \exp\left\{-\left[1+\xi\left(\frac{x-\mu}{\sigma}\right)\right]^{-1/\xi}\right\}
}{
  F(x) = exp(-[1 + \xi(x-\mu)/\sigma]^(-1/\xi))
}
where \eqn{1 + \xi(x-\mu)/\sigma > 0}.

\strong{Quantile function}:

For \eqn{\xi = 0} (Gumbel):
\deqn{
  Q(p) = \mu - \sigma \log(-\log p)
}{
  Q(p) = \mu - \sigma log(-log p)
}

For \eqn{\xi \neq 0}:
\deqn{
  Q(p) = \mu + \frac{\sigma}{\xi}\left[(-\log p)^{-\xi} - 1\right]
}{
  Q(p) = \mu + \sigma [(-log p)^(-\xi) - 1] / \xi
}
}
\examples{
# Create GEV distributions with different shape parameters

# Gumbel distribution (shape = 0)
gumbel <- dist_gev(location = 0, scale = 1, shape = 0)

# Frechet distribution (shape > 0, heavy-tailed)
frechet <- dist_gev(location = 0, scale = 1, shape = 0.3)

# Reverse Weibull distribution (shape < 0, bounded above)
weibull <- dist_gev(location = 0, scale = 1, shape = -0.2)

dist <- c(gumbel, frechet, weibull)
dist

# Statistical properties
mean(dist)
median(dist)
variance(dist)

# Generate random samples
generate(dist, 10)

# Evaluate density
density(dist, 2)
density(dist, 2, log = TRUE)

# Evaluate cumulative distribution
cdf(dist, 4)

# Calculate quantiles
quantile(dist, 0.95)

}
\references{
Jenkinson, A. F. (1955) The frequency distribution of the annual
maximum (or minimum) of meteorological elements. \emph{Quart. J. R. Met. Soc.},
\bold{81}, 158–171.
}
\seealso{
\code{\link[evd:gev]{evd::dgev()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_inverse_gaussian.R
\name{dist_inverse_gaussian}
\alias{dist_inverse_gaussian}
\title{The Inverse Gaussian distribution}
\usage{
dist_inverse_gaussian(mean, shape)
}
\arguments{
\item{mean, shape}{parameters. Must be strictly positive. Infinite
    values are supported.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}
}
\details{
The inverse Gaussian distribution (also known as the Wald distribution) is
commonly used to model positive-valued data, particularly in contexts
involving first passage times and reliability analysis.

We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_inverse_gaussian.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_inverse_gaussian.html}

In the following, let \eqn{X} be an Inverse Gaussian random variable with
parameters \code{mean} = \eqn{\mu} and \code{shape} = \eqn{\lambda}.

\strong{Support}: \eqn{(0, \infty)}

\strong{Mean}: \eqn{\mu}

\strong{Variance}: \eqn{\frac{\mu^3}{\lambda}}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \sqrt{\frac{\lambda}{2\pi x^3}}
    \exp\left(-\frac{\lambda(x - \mu)^2}{2\mu^2 x}\right)
  }{
    f(x) = sqrt(\lambda / (2 \pi x^3))
    exp(-\lambda (x - \mu)^2 / (2 \mu^2 x))
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = \Phi\left(\sqrt{\frac{\lambda}{x}}
    \left(\frac{x}{\mu} - 1\right)\right) +
    \exp\left(\frac{2\lambda}{\mu}\right)
    \Phi\left(-\sqrt{\frac{\lambda}{x}}
    \left(\frac{x}{\mu} + 1\right)\right)
  }{
    F(x) = \Phi(sqrt(\lambda/x) (x/\mu - 1)) +
    exp(2\lambda/\mu) \Phi(-sqrt(\lambda/x) (x/\mu + 1))
  }

where \eqn{\Phi} is the standard normal c.d.f.

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \exp\left(\frac{\lambda}{\mu}
    \left(1 - \sqrt{1 - \frac{2\mu^2 t}{\lambda}}\right)\right)
  }{
    E(e^(tX)) = exp((\lambda/\mu) (1 - sqrt(1 - 2\mu^2 t/\lambda)))
  }

for \eqn{t < \frac{\lambda}{2\mu^2}}.

\strong{Skewness}: \eqn{3\sqrt{\frac{\mu}{\lambda}}}

\strong{Excess Kurtosis}: \eqn{\frac{15\mu}{\lambda}}

\strong{Quantiles}: No closed-form expression, approximated numerically.
}
\examples{
dist <- dist_inverse_gaussian(mean = c(1,1,1,3,3), shape = c(0.2, 1, 3, 0.2, 1))
dist

\dontshow{if (requireNamespace("actuar", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
\dontshow{\}) # examplesIf}
}
\seealso{
\link[actuar:InverseGaussian]{actuar::InverseGaussian}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{covariance}
\alias{covariance}
\title{Covariance}
\usage{
covariance(x, ...)
}
\arguments{
\item{x}{An object.}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

A generic function for computing the covariance of an object.
}
\seealso{
\code{\link[=covariance.distribution]{covariance.distribution()}}, \code{\link[=variance]{variance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_gumbel.R
\name{dist_gumbel}
\alias{dist_gumbel}
\title{The Gumbel distribution}
\usage{
dist_gumbel(alpha, scale)
}
\arguments{
\item{alpha}{location parameter.}

\item{scale}{parameter. Must be strictly positive.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Gumbel distribution is a special case of the Generalized Extreme Value
distribution, obtained when the GEV shape parameter \eqn{\xi} is equal to 0.
It may be referred to as a type I extreme value distribution.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_gumbel.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_gumbel.html}

In the following, let \eqn{X} be a Gumbel random variable with location
parameter  \code{alpha} = \eqn{\alpha} and scale parameter \code{scale} = \eqn{\sigma}.

\strong{Support}: \eqn{R}, the set of all real numbers.

\strong{Mean}:

\deqn{
    E(X) = \alpha + \sigma\gamma
  }{
    E(X) = alpha + sigma*gamma
  }

where \eqn{\gamma} is the Euler-Mascheroni constant,
approximately equal to 0.5772157.

\strong{Variance}:

\deqn{
    \textrm{Var}(X) = \frac{\pi^2 \sigma^2}{6}
  }{
    Var(X) = (pi^2 * sigma^2) / 6
  }

\strong{Skewness}:

\deqn{
    \textrm{Skew}(X) = \frac{12\sqrt{6}\zeta(3)}{\pi^3} \approx 1.1395
  }{
    Skew(X) = (12*sqrt(6)*zeta(3)) / pi^3
  }

where \eqn{\zeta(3)} is Apery's constant,
approximately equal to 1.2020569. Note that skewness is independent
of the distribution parameters.

\strong{Kurtosis (excess)}:

\deqn{
    \textrm{Kurt}(X) = \frac{12}{5} = 2.4
  }{
    Kurt(X) = 12/5
  }

Note that excess kurtosis is independent of the distribution parameters.

\strong{Median}:

\deqn{
    \textrm{Median}(X) = \alpha - \sigma\ln(\ln 2)
  }{
    Median(X) = alpha - sigma*ln(ln(2))
  }

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{1}{\sigma} \exp\left[-\frac{x - \alpha}{\sigma}\right]
           \exp\left\{-\exp\left[-\frac{x - \alpha}{\sigma}\right]\right\}
  }{
    f(x) = (1/sigma) * exp[-(x - alpha)/sigma] * exp(-exp[-(x - alpha)/sigma])
  }

for \eqn{x} in \eqn{R}, the set of all real numbers.

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = \exp\left\{-\exp\left[-\frac{x - \alpha}{\sigma}\right]\right\}
  }{
    F(x) = exp(-exp[-(x - alpha)/sigma])
  }

for \eqn{x} in \eqn{R}, the set of all real numbers.

\strong{Quantile function (inverse c.d.f)}:

\deqn{
    F^{-1}(p) = \alpha - \sigma \ln(-\ln p)
  }{
    F^(-1)(p) = alpha - sigma * ln(-ln(p))
  }

for \eqn{p} in (0, 1).

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \Gamma(1 - \sigma t) e^{\alpha t}
  }{
    E(e^(tX)) = Gamma(1 - sigma*t) * e^(alpha*t)
  }

for \eqn{\sigma t < 1}, where \eqn{\Gamma} is the gamma function.
}
\examples{
dist <- dist_gumbel(alpha = c(0.5, 1, 1.5, 3), scale = c(2, 2, 3, 4))
dist

\dontshow{if (requireNamespace("actuar", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)
support(dist)
generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
\dontshow{\}) # examplesIf}
}
\seealso{
\link[actuar:Gumbel]{actuar::Gumbel}, \code{\link[actuar:Gumbel]{actuar::dgumbel()}}, \code{\link[actuar:Gumbel]{actuar::pgumbel()}},
\code{\link[actuar:Gumbel]{actuar::qgumbel()}}, \code{\link[actuar:Gumbel]{actuar::rgumbel()}}, \code{\link[actuar:Gumbel]{actuar::mgumbel()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_percentile.R
\name{dist_quantile}
\alias{dist_quantile}
\alias{dist_percentile}
\title{Quantile distribution}
\usage{
dist_quantile(x, quantile)

dist_percentile(x, percentile)
}
\arguments{
\item{x}{A list of values}

\item{quantile}{A list of quantile probabilities (between 0 and 1)}

\item{percentile}{A list of percentiles (between 0 and 100)}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Quantile distribution is a non-parametric distribution defined by
a set of values at specified quantile probabilities. This distribution is
useful for representing empirical distributions or elicited expert
knowledge when only quantile information is available. The distribution
uses linear interpolation between quantiles and can be used to
approximate complex distributions that may not have simple parametric forms.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_quantile.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_quantile.html}

In the following, let \eqn{X} be a Quantile random variable defined by
values \eqn{x_1, x_2, \ldots, x_n} at quantile probabilities
\eqn{q_1, q_2, \ldots, q_n} where \eqn{0 \le q_i \le 1}.

\strong{Support}: \eqn{[\min(x_i), \max(x_i)]} if \eqn{\min(q_i) > 0} or
\eqn{\max(q_i) < 1}, otherwise support is approximated from the
specified quantiles.

\strong{Mean}: Approximated numerically using spline interpolation and
numerical integration:

\deqn{
    E(X) \approx \int_0^1 Q(u) du
  }{
    E(X) ≈ integral_0^1 Q(u) du
  }

  where \eqn{Q(u)} is a spline function interpolating the quantile values.

\strong{Variance}: Approximated numerically.

\strong{Probability density function (p.d.f)}: Approximated numerically using
kernel density estimation from generated samples.

\strong{Cumulative distribution function (c.d.f)}: Defined by linear
interpolation:

\deqn{
    F(t) = \begin{cases}
      q_1 & \text{if } t < x_1 \\
      q_i + \frac{(t - x_i)(q_{i+1} - q_i)}{x_{i+1} - x_i} & \text{if } x_i \le t < x_{i+1} \\
      q_n & \text{if } t \ge x_n
    \end{cases}
  }{
    F(t) = q_i + (t - x_i)(q_{i+1} - q_i) / (x_{i+1} - x_i) for x_i ≤ t < x_{i+1}
  }

  \strong{Quantile function}: Defined by linear interpolation:

\deqn{
    Q(u) = x_i + \frac{(u - q_i)(x_{i+1} - x_i)}{q_{i+1} - q_i}
  }{
    Q(u) = x_i + (u - q_i)(x_{i+1} - x_i) / (q_{i+1} - q_i)
  }

for \eqn{q_i \le u \le q_{i+1}}.
}
\examples{
dist <- dist_normal()
probs <- seq(0.01, 0.99, by = 0.01)
x <- vapply(probs, quantile, double(1L), x = dist)
dist_quantile(list(x), list(probs))
dist_percentile(list(x), list(probs * 100))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_binomial.R
\name{dist_binomial}
\alias{dist_binomial}
\title{The Binomial distribution}
\usage{
dist_binomial(size, prob)
}
\arguments{
\item{size}{The number of trials. Must be an integer greater than or equal
to one. When \code{size = 1L}, the Binomial distribution reduces to the
Bernoulli distribution. Often called \code{n} in textbooks.}

\item{prob}{The probability of success on each trial, \code{prob} can be any
value in \verb{[0, 1]}.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Binomial distributions are used to represent situations can that can
be thought as the result of \eqn{n} Bernoulli experiments (here the
\eqn{n} is defined as the \code{size} of the experiment). The classical
example is \eqn{n} independent coin flips, where each coin flip has
probability \code{p} of success. In this case, the individual probability of
flipping heads or tails is given by the  Bernoulli(p) distribution,
and the probability of having \eqn{x} equal results (\eqn{x} heads,
for example), in \eqn{n} trials is given by the Binomial(n, p) distribution.
The equation of the Binomial distribution is directly derived from
the equation of the Bernoulli distribution.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

The Binomial distribution comes up when you are interested in the portion
of people who do a thing. The Binomial distribution
also comes up in the sign test, sometimes called the Binomial test
(see \code{\link[stats:binom.test]{stats::binom.test()}}), where you may need the Binomial C.D.F. to
compute p-values.

In the following, let \eqn{X} be a Binomial random variable with parameter
\code{size} = \eqn{n} and \code{p} = \eqn{p}. Some textbooks define \eqn{q = 1 - p},
or called \eqn{\pi} instead of \eqn{p}.

\strong{Support}: \eqn{\{0, 1, 2, ..., n\}}{{0, 1, 2, ..., n}}

\strong{Mean}: \eqn{np}

\strong{Variance}: \eqn{np \cdot (1 - p) = np \cdot q}{np (1 - p)}

\strong{Probability mass function (p.m.f)}:

\deqn{
    P(X = k) = {n \choose k} p^k (1 - p)^{n-k}
  }{
    P(X = k) = choose(n, k) p^k (1 - p)^(n - k)
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    P(X \le k) = \sum_{i=0}^{\lfloor k \rfloor} {n \choose i} p^i (1 - p)^{n-i}
  }{
    P(X \le k) = \sum_{i=0}^k choose(n, i) p^i (1 - p)^(n-i)
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = (1 - p + p e^t)^n
  }{
    E(e^(tX)) = (1 - p + p e^t)^n
  }
}
\examples{
dist <- dist_binomial(size = 1:5, prob = c(0.05, 0.5, 0.3, 0.9, 0.1))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{cdf}
\alias{cdf}
\alias{cdf.distribution}
\title{The cumulative distribution function}
\usage{
cdf(x, q, ..., log = FALSE)

\method{cdf}{distribution}(x, q, ...)
}
\arguments{
\item{x}{The distribution(s).}

\item{q}{The quantile at which the cdf is calculated.}

\item{...}{Additional arguments passed to methods.}

\item{log}{If \code{TRUE}, probabilities will be given as log probabilities.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_uniform.R
\name{dist_uniform}
\alias{dist_uniform}
\title{The Uniform distribution}
\usage{
dist_uniform(min, max)
}
\arguments{
\item{min, max}{lower and upper limits of the distribution.  Must be finite.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

A distribution with constant density on an interval.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_uniform.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_uniform.html}

In the following, let \eqn{X} be a Uniform random variable with parameters
\code{min} = \eqn{a} and \code{max} = \eqn{b}.

\strong{Support}: \eqn{[a, b]}

\strong{Mean}: \eqn{\frac{a + b}{2}}

\strong{Variance}: \eqn{\frac{(b - a)^2}{12}}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{1}{b - a}
  }{
    f(x) = 1 / (b - a)
  }

for \eqn{x \in [a, b]}{x in [a, b]}, and \eqn{f(x) = 0} otherwise.

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = \frac{x - a}{b - a}
  }{
    F(x) = (x - a) / (b - a)
  }

for \eqn{x \in [a, b]}{x in [a, b]}, with \eqn{F(x) = 0} for \eqn{x < a}
and \eqn{F(x) = 1} for \eqn{x > b}.

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \frac{e^{tb} - e^{ta}}{t(b - a)}
  }{
    E(e^(tX)) = (e^(tb) - e^(ta)) / (t(b - a))
  }

for \eqn{t \neq 0}{t != 0}, and \eqn{E(e^{tX}) = 1} for \eqn{t = 0}.

\strong{Skewness}: \eqn{0}

\strong{Excess Kurtosis}: \eqn{-\frac{6}{5}}{-6/5}
}
\examples{
dist <- dist_uniform(min = c(3, -2), max = c(5, 4))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Uniform]{stats::Uniform}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_multivariate_t.R
\name{dist_multivariate_t}
\alias{dist_multivariate_t}
\title{The multivariate t-distribution}
\usage{
dist_multivariate_t(df = 1, mu = 0, sigma = diag(1))
}
\arguments{
\item{df}{A numeric vector of degrees of freedom (must be positive).}

\item{mu}{A list of numeric vectors for the distribution location parameter.}

\item{sigma}{A list of matrices for the distribution scale matrix.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The multivariate t-distribution is a generalization of the univariate
Student's t-distribution to multiple dimensions. It is commonly used for
modeling heavy-tailed multivariate data and in robust statistics.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_multivariate_t.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_multivariate_t.html}

In the following, let \eqn{\mathbf{X}} be a multivariate t random vector
with degrees of freedom \code{df} = \eqn{\nu}, location parameter
\code{mu} = \eqn{\boldsymbol{\mu}}, and scale matrix
\code{sigma} = \eqn{\boldsymbol{\Sigma}}.

\strong{Support}: \eqn{\mathbf{x} \in \mathbb{R}^k}, where \eqn{k} is the
dimension of the distribution

\strong{Mean}: \eqn{\boldsymbol{\mu}} for \eqn{\nu > 1}, undefined otherwise

\strong{Covariance matrix}:

\deqn{
    \text{Cov}(\mathbf{X}) = \frac{\nu}{\nu - 2} \boldsymbol{\Sigma}
  }{
    Cov(X) = \nu / (\nu - 2) \Sigma
  }

for \eqn{\nu > 2}, undefined otherwise

\strong{Probability density function (p.d.f)}:

\deqn{
    f(\mathbf{x}) = \frac{\Gamma\left(\frac{\nu + k}{2}\right)}
    {\Gamma\left(\frac{\nu}{2}\right) \nu^{k/2} \pi^{k/2}
    |\boldsymbol{\Sigma}|^{1/2}}
    \left[1 + \frac{1}{\nu}(\mathbf{x} - \boldsymbol{\mu})^T
    \boldsymbol{\Sigma}^{-1} (\mathbf{x} - \boldsymbol{\mu})\right]^{-\frac{\nu + k}{2}}
  }{
    f(x) = \Gamma((\nu + k)/2) / (\Gamma(\nu/2) \nu^(k/2) \pi^(k/2) |\Sigma|^(1/2))
    [1 + 1/\nu (x - \mu)^T \Sigma^(-1) (x - \mu)]^(-(\nu + k)/2)
  }

where \eqn{k} is the dimension of the distribution and \eqn{\Gamma(\cdot)} is
the gamma function.

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(\mathbf{t}) = \int_{-\infty}^{t_1} \cdots \int_{-\infty}^{t_k} f(\mathbf{x}) \, d\mathbf{x}
  }{
    F(t) = integral_{-\infty}^{t_1} ... integral_{-\infty}^{t_k} f(x) dx
  }

This integral does not have a closed form solution and is approximated numerically.

\strong{Quantile function}:

The equicoordinate quantile function finds \eqn{q} such that:

\deqn{
    P(X_1 \leq q, \ldots, X_k \leq q) = p
  }{
    P(X_1 <= q, ..., X_k <= q) = p
  }

This does not have a closed form solution and is approximated numerically.

The marginal quantile function for each dimension \eqn{i} is:

\deqn{
    Q_i(p) = \mu_i + \sqrt{\Sigma_{ii}} \cdot t_{\nu}^{-1}(p)
  }{
    Q_i(p) = \mu_i + sqrt(\Sigma_{ii}) * t_\nu^(-1)(p)
  }

where \eqn{t_{\nu}^{-1}(p)} is the quantile function of the univariate
Student's t-distribution with \eqn{\nu} degrees of freedom, and
\eqn{\Sigma_{ii}} is the \eqn{i}-th diagonal element of \code{sigma}.
}
\examples{
dist <- dist_multivariate_t(
  df = 5,
  mu = list(c(1, 2)),
  sigma = list(matrix(c(4, 2, 2, 3), ncol = 2))
)
dimnames(dist) <- c("x", "y")
dist

\dontshow{if (requireNamespace("mvtnorm", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, cbind(2, 1))
density(dist, cbind(2, 1), log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
quantile(dist, 0.7, kind = "marginal")
\dontshow{\}) # examplesIf}
}
\seealso{
\link[mvtnorm:Mvt]{mvtnorm::dmvt}, \link[mvtnorm:pmvt]{mvtnorm::pmvt}, \link[mvtnorm:qmvt]{mvtnorm::qmvt}, \link[mvtnorm:Mvt]{mvtnorm::rmvt}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_f.R
\name{dist_f}
\alias{dist_f}
\title{The F Distribution}
\usage{
dist_f(df1, df2, ncp = NULL)
}
\arguments{
\item{df1}{Degrees of freedom for the numerator. Can be any positive number.}

\item{df2}{Degrees of freedom for the denominator. Can be any positive number.}

\item{ncp}{Non-centrality parameter. If \code{NULL} (default), the central F
distribution is used. If specified, must be non-negative.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The F distribution is commonly used in statistical inference, particularly
in the analysis of variance (ANOVA), testing the equality of variances,
and in regression analysis. It arises as the ratio of two scaled
chi-squared distributions divided by their respective degrees of freedom.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_f.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_f.html}

In the following, let \eqn{X} be an F random variable with numerator
degrees of freedom \code{df1} = \eqn{d_1} and denominator degrees of freedom
\code{df2} = \eqn{d_2}.

\strong{Support}: \eqn{x \in (0, \infty)}

\strong{Mean}:

For the central F distribution (\code{ncp = NULL}):

\deqn{
    E(X) = \frac{d_2}{d_2 - 2}
  }{
    E(X) = d_2 / (d_2 - 2)
  }

for \eqn{d_2 > 2}, otherwise undefined.

For the non-central F distribution with non-centrality parameter
\code{ncp} = \eqn{\lambda}:

\deqn{
    E(X) = \frac{d_2 (d_1 + \lambda)}{d_1 (d_2 - 2)}
  }{
    E(X) = d_2 (d_1 + \lambda) / (d_1 (d_2 - 2))
  }

for \eqn{d_2 > 2}, otherwise undefined.

\strong{Variance}:

For the central F distribution (\code{ncp = NULL}):

\deqn{
    \text{Var}(X) = \frac{2 d_2^2 (d_1 + d_2 - 2)}{d_1 (d_2 - 2)^2 (d_2 - 4)}
  }{
    Var(X) = 2 d_2^2 (d_1 + d_2 - 2) / (d_1 (d_2 - 2)^2 (d_2 - 4))
  }

for \eqn{d_2 > 4}, otherwise undefined.

For the non-central F distribution with non-centrality parameter
\code{ncp} = \eqn{\lambda}:

\deqn{
    \text{Var}(X) = \frac{2 d_2^2}{d_1^2} \cdot \frac{(d_1 + \lambda)^2 + (d_1 + 2\lambda)(d_2 - 2)}{(d_2 - 2)^2 (d_2 - 4)}
  }{
    Var(X) = 2 d_2^2 / d_1^2 * ((d_1 + lambda)^2 + (d_1 + 2*lambda)(d_2 - 2)) / ((d_2 - 2)^2 (d_2 - 4))
  }

for \eqn{d_2 > 4}, otherwise undefined.

\strong{Skewness}:

For the central F distribution (\code{ncp = NULL}):

\deqn{
    \text{Skew}(X) = \frac{(2 d_1 + d_2 - 2) \sqrt{8 (d_2 - 4)}}{(d_2 - 6) \sqrt{d_1 (d_1 + d_2 - 2)}}
  }{
    Skew(X) = (2 d_1 + d_2 - 2) sqrt(8 (d_2 - 4)) / ((d_2 - 6) sqrt(d_1 (d_1 + d_2 - 2)))
  }

for \eqn{d_2 > 6}, otherwise undefined.

For the non-central F distribution, skewness has no simple closed form
and is not computed.

\strong{Excess Kurtosis}:

For the central F distribution (\code{ncp = NULL}):

\deqn{
    \text{Kurt}(X) = \frac{12[d_1 (5 d_2 - 22)(d_1 + d_2 - 2) + (d_2 - 4)(d_2 - 2)^2]}{d_1 (d_2 - 6)(d_2 - 8)(d_1 + d_2 - 2)}
  }{
    Kurt(X) = 12[d_1 (5 d_2 - 22)(d_1 + d_2 - 2) + (d_2 - 4)(d_2 - 2)^2] / (d_1 (d_2 - 6)(d_2 - 8)(d_1 + d_2 - 2))
  }

for \eqn{d_2 > 8}, otherwise undefined.

For the non-central F distribution, kurtosis has no simple closed form
and is not computed.

\strong{Probability density function (p.d.f)}:

For the central F distribution (\code{ncp = NULL}):

\deqn{
    f(x) = \frac{\sqrt{\frac{(d_1 x)^{d_1} d_2^{d_2}}{(d_1 x + d_2)^{d_1 + d_2}}}}{x \, B(d_1/2, d_2/2)}
  }{
    f(x) = sqrt((d_1 x)^d_1 d_2^d_2 / (d_1 x + d_2)^(d_1 + d_2)) / (x B(d_1/2, d_2/2))
  }

where \eqn{B(\cdot, \cdot)} is the beta function.

For the non-central F distribution, the density involves an infinite
series and is approximated numerically.

\strong{Cumulative distribution function (c.d.f)}:

The c.d.f. does not have a simple closed form expression and is
approximated numerically using regularized incomplete beta functions
and related special functions.

\strong{Moment generating function (m.g.f)}:

The moment generating function for the F distribution does not exist
in general (it diverges for \eqn{t > 0}).
}
\examples{
dist <- dist_f(df1 = c(1,2,5,10,100), df2 = c(1,1,2,1,100))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Fdist]{stats::FDist}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_normal.R
\name{dist_normal}
\alias{dist_normal}
\title{The Normal distribution}
\usage{
dist_normal(mu = 0, sigma = 1, mean = mu, sd = sigma)
}
\arguments{
\item{mu, mean}{The mean (location parameter) of the distribution, which is also
the mean of the distribution. Can be any real number.}

\item{sigma, sd}{The standard deviation (scale parameter) of the distribution.
Can be any positive number. If you would like a Normal distribution with
\strong{variance} \eqn{\sigma^2}, be sure to take the square root, as this is a
common source of errors.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Normal distribution is ubiquitous in statistics, partially because
of the central limit theorem, which states that sums of i.i.d. random
variables eventually become Normal. Linear transformations of Normal
random variables result in new random variables that are also Normal. If
you are taking an intro stats course, you'll likely use the Normal
distribution for Z-tests and in simple linear regression. Under
regularity conditions, maximum likelihood estimators are
asymptotically Normal. The Normal distribution is also called the
gaussian distribution.
}
\details{
We recommend reading this documentation on
\url{https://pkg.mitchelloharawild.com/distributional/}, where the math
will render nicely.

In the following, let \eqn{X} be a Normal random variable with mean
\code{mu} = \eqn{\mu} and standard deviation \code{sigma} = \eqn{\sigma}.

\strong{Support}: \eqn{R}, the set of all real numbers

\strong{Mean}: \eqn{\mu}

\strong{Variance}: \eqn{\sigma^2}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{1}{\sqrt{2 \pi \sigma^2}} e^{-(x - \mu)^2 / 2 \sigma^2}
  }{
    f(x) = 1 / sqrt(2 \pi \sigma^2) exp(-(x - \mu)^2 / (2 \sigma^2))
  }

\strong{Cumulative distribution function (c.d.f)}:

The cumulative distribution function has the form

\deqn{
    F(t) = \int_{-\infty}^t \frac{1}{\sqrt{2 \pi \sigma^2}} e^{-(x - \mu)^2 / 2 \sigma^2} dx
  }{
    F(t) = integral_{-\infty}^t 1 / sqrt(2 \pi \sigma^2) exp(-(x - \mu)^2 / (2 \sigma^2)) dx
  }

but this integral does not have a closed form solution and must be
approximated numerically. The c.d.f. of a standard Normal is sometimes
called the "error function". The notation \eqn{\Phi(t)} also stands
for the c.d.f. of a standard Normal evaluated at \eqn{t}. Z-tables
list the value of \eqn{\Phi(t)} for various \eqn{t}.

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = e^{\mu t + \sigma^2 t^2 / 2}
  }{
    E(e^(tX)) = e^(\mu t + \sigma^2 t^2 / 2)
  }
}
\examples{
dist <- dist_normal(mu = 1:5, sigma = 3)

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Normal]{stats::Normal}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_inverse_gamma.R
\name{dist_inverse_gamma}
\alias{dist_inverse_gamma}
\title{The Inverse Gamma distribution}
\usage{
dist_inverse_gamma(shape, rate = 1/scale, scale)
}
\arguments{
\item{shape, scale}{parameters. Must be strictly positive.}

\item{rate}{an alternative way to specify the scale.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The Inverse Gamma distribution is commonly used as a prior distribution
in Bayesian statistics, particularly for variance parameters.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_inverse_gamma.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_inverse_gamma.html}

In the following, let \eqn{X} be an Inverse Gamma random variable with
shape parameter \code{shape} = \eqn{\alpha} and rate parameter
\code{rate} = \eqn{\beta} (equivalently, scale = \eqn{1/\beta}).

\strong{Support}: \eqn{x \in (0, \infty)}

\strong{Mean}: \eqn{\frac{\beta}{\alpha - 1}} for \eqn{\alpha > 1},
otherwise undefined

\strong{Variance}: \eqn{\frac{\beta^2}{(\alpha - 1)^2 (\alpha - 2)}}
for \eqn{\alpha > 2}, otherwise undefined

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \frac{\beta^\alpha}{\Gamma(\alpha)} x^{-\alpha - 1} 
    e^{-\beta/x}
  }{
    f(x) = \beta^\alpha / \Gamma(\alpha) x^(-\alpha - 1) e^(-\beta/x)
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = \frac{\Gamma(\alpha, \beta/x)}{\Gamma(\alpha)} = 
    Q(\alpha, \beta/x)
  }{
    F(x) = \Gamma(\alpha, \beta/x) / \Gamma(\alpha) = Q(\alpha, \beta/x)
  }

where \eqn{\Gamma(\alpha, z)} is the upper incomplete gamma function and
\eqn{Q} is the regularized incomplete gamma function.

\strong{Moment generating function (m.g.f)}:

\deqn{
    M_X(t) = \frac{2 (-\beta t)^{\alpha/2}}{\Gamma(\alpha)} 
    K_\alpha\left(\sqrt{-4\beta t}\right)
  }{
    M_X(t) = 2 (-\beta t)^(\alpha/2) / \Gamma(\alpha) K_\alpha(sqrt(-4\beta t))
  }

for \eqn{t < 0}, where \eqn{K_\alpha} is the modified Bessel function
of the second kind. The MGF does not exist for \eqn{t \ge 0}.
}
\examples{
dist <- dist_inverse_gamma(shape = c(1,2,3,3), rate = c(1,1,1,2))
dist

\dontshow{if (requireNamespace("actuar", quietly = TRUE)) withAutoprint(\{ # examplesIf}
mean(dist)
variance(dist)
support(dist)
generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)
\dontshow{\}) # examplesIf}
}
\seealso{
\link[actuar:InverseGamma]{actuar::InverseGamma}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_exponential.R
\name{dist_exponential}
\alias{dist_exponential}
\title{The Exponential Distribution}
\usage{
dist_exponential(rate)
}
\arguments{
\item{rate}{vector of rates.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Exponential distributions are frequently used to model waiting times and the
time between events in a Poisson process.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_exponential.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_exponential.html}

In the following, let \eqn{X} be an Exponential random variable with
parameter \code{rate} = \eqn{\lambda}.

\strong{Support}: \eqn{x \in [0, \infty)}{x >= 0}

\strong{Mean}: \eqn{\frac{1}{\lambda}}{1 / \lambda}

\strong{Variance}: \eqn{\frac{1}{\lambda^2}}{1 / \lambda^2}

\strong{Probability density function (p.d.f)}:

\deqn{
    f(x) = \lambda e^{-\lambda x}
  }{
    f(x) = \lambda e^(-\lambda x)
  }

\strong{Cumulative distribution function (c.d.f)}:

\deqn{
    F(x) = 1 - e^{-\lambda x}
  }{
    F(x) = 1 - e^(-\lambda x)
  }

\strong{Moment generating function (m.g.f)}:

\deqn{
    E(e^{tX}) = \frac{\lambda}{\lambda - t}, \quad t < \lambda
  }{
    E(e^(tX)) = \lambda / (\lambda - t), for t < \lambda
  }
}
\examples{
dist <- dist_exponential(rate = c(2, 1, 2/3))

dist
mean(dist)
variance(dist)
skewness(dist)
kurtosis(dist)

generate(dist, 10)

density(dist, 2)
density(dist, 2, log = TRUE)

cdf(dist, 4)

quantile(dist, 0.7)

}
\seealso{
\link[stats:Exponential]{stats::Exponential}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_sample.R
\name{dist_sample}
\alias{dist_sample}
\title{Sampling distribution}
\usage{
dist_sample(x)
}
\arguments{
\item{x}{A list of sampled values. For univariate distributions, each
element should be a numeric vector. For multivariate distributions, each
element should be a matrix where columns represent variables and rows
represent observations.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

The sampling distribution represents an empirical distribution based on
observed samples. It is useful for bootstrapping, representing posterior
distributions from Markov Chain Monte Carlo (MCMC) algorithms, or working
with any empirical data where the parametric form is unknown. Unlike
parametric distributions, the sampling distribution makes no assumptions
about the underlying data-generating process and instead uses the sample
itself to estimate distributional properties. The distribution can handle
both univariate and multivariate samples.
}
\details{
We recommend reading this documentation on \href{https://pkg.mitchelloharawild.com/distributional/reference/dist_sample.html}{pkgdown} which renders math nicely.
\url{https://pkg.mitchelloharawild.com/distributional/reference/dist_sample.html}

In the following, let \eqn{X} be a random variable with sample
\eqn{x_1, x_2, \ldots, x_n} of size \eqn{n}.

\strong{Support}: The observed range of the sample

\strong{Mean} (univariate):

\deqn{
    \bar{x} = \frac{1}{n} \sum_{i=1}^{n} x_i
  }{
    mean(x) = (1/n) sum(x_i)
  }

\strong{Mean} (multivariate): Computed independently for each variable.

\strong{Variance} (univariate):

\deqn{
    s^2 = \frac{1}{n-1} \sum_{i=1}^{n} (x_i - \bar{x})^2
  }{
    s^2 = (1/(n-1)) sum((x_i - mean(x))^2)
  }

\strong{Covariance} (multivariate): The sample covariance matrix.

\strong{Skewness} (univariate):

\deqn{
    g_1 = \frac{\sqrt{n} \sum_{i=1}^{n} (x_i - \bar{x})^3}{\left(\sum_{i=1}^{n} (x_i - \bar{x})^2\right)^{3/2}} \left(1 - \frac{1}{n}\right)^{3/2}
  }{
    g_1 = (sqrt(n) sum((x_i - mean(x))^3) / (sum((x_i - mean(x))^2)^(3/2))) * (1 - 1/n)^(3/2)
  }

\strong{Probability density function}: Approximated numerically using
kernel density estimation.

\strong{Cumulative distribution function} (univariate):

\deqn{
    F(q) = \frac{1}{n} \sum_{i=1}^{n} I(x_i \leq q)
  }{
    F(q) = (1/n) sum(I(x_i <= q))
  }

where \eqn{I(\cdot)} is the indicator function.

\strong{Cumulative distribution function} (multivariate):

\deqn{
    F(\mathbf{q}) = \frac{1}{n} \sum_{i=1}^{n} I(\mathbf{x}_i \leq \mathbf{q})
  }{
    F(q) = (1/n) sum(I(x_i <= q))
  }

where the inequality is applied element-wise.

\strong{Quantile function} (univariate): The sample quantile, computed using
the specified quantile type (see \code{\link[stats:quantile]{stats::quantile()}}).

\strong{Quantile function} (multivariate): Marginal quantiles are computed
independently for each variable.

\strong{Random generation}: Bootstrap sampling with replacement from the
empirical sample.
}
\examples{
# Univariate numeric samples
dist <- dist_sample(x = list(rnorm(100), rnorm(100, 10)))

dist
mean(dist)
variance(dist)
skewness(dist)
generate(dist, 10)

density(dist, 1)

# Multivariate numeric samples
dist <- dist_sample(x = list(cbind(rnorm(100), rnorm(100, 10))))
dimnames(dist) <- c("x", "y")

dist
mean(dist)
variance(dist)
generate(dist, 10)
quantile(dist, 0.4) # Returns the marginal quantiles
cdf(dist, matrix(c(0.3,9), nrow = 1))

}
\seealso{
\code{\link[stats:density]{stats::density()}}, \code{\link[stats:quantile]{stats::quantile()}}, \code{\link[stats:cor]{stats::cov()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distributional-package.R
\docType{package}
\name{distributional-package}
\alias{distributional}
\alias{distributional-package}
\title{distributional: Vectorised Probability Distributions}
\description{
Vectorised distribution objects with tools for manipulating, visualising, and using probability distributions. Designed to allow model prediction outputs to return distributions rather than their parameters, allowing users to directly interact with predictive distributions in a data-oriented workflow. In addition to providing generic replacements for p/d/q/r functions, other useful statistics can be computed including means, variances, intervals, and highest density regions.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://pkg.mitchelloharawild.com/distributional/}
  \item \url{https://github.com/mitchelloharawild/distributional}
  \item Report bugs at \url{https://github.com/mitchelloharawild/distributional/issues}
}

}
\author{
\strong{Maintainer}: Mitchell O'Hara-Wild \email{mail@mitchelloharawild.com} (\href{https://orcid.org/0000-0001-6729-7695}{ORCID})

Authors:
\itemize{
  \item Matthew Kay (\href{https://orcid.org/0000-0001-9446-0419}{ORCID})
  \item Alex Hayes (\href{https://orcid.org/0000-0002-4985-5160}{ORCID})
  \item Rob Hyndman (\href{https://orcid.org/0000-0002-2140-5352}{ORCID})
}

Other contributors:
\itemize{
  \item Earo Wang (\href{https://orcid.org/0000-0001-6448-5260}{ORCID}) [contributor]
  \item Vencislav Popov (\href{https://orcid.org/0000-0002-8073-4199}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distribution.R
\name{variance}
\alias{variance}
\alias{variance.numeric}
\alias{variance.matrix}
\alias{covariance.numeric}
\title{Variance}
\usage{
variance(x, ...)

\method{variance}{numeric}(x, ...)

\method{variance}{matrix}(x, ...)

\method{covariance}{numeric}(x, ...)
}
\arguments{
\item{x}{An object.}

\item{...}{Additional arguments used by methods.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

A generic function for computing the variance of an object.
}
\details{
The implementation of \code{variance()} for numeric variables coerces the input to
a vector then uses \code{\link[stats:cor]{stats::var()}} to compute the variance. This means that,
unlike \code{\link[stats:cor]{stats::var()}}, if \code{variance()} is passed a matrix or a 2-dimensional
array, it will still return the variance (\code{\link[stats:cor]{stats::var()}} returns the
covariance matrix in that case).
}
\seealso{
\code{\link[=variance.distribution]{variance.distribution()}}, \code{\link[=covariance]{covariance()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dist_wrap.R
\name{dist_wrap}
\alias{dist_wrap}
\title{Create a distribution from p/d/q/r style functions}
\usage{
dist_wrap(dist, ..., package = NULL)
}
\arguments{
\item{dist}{The name of the distribution used in the functions (name that is
prefixed by p/d/q/r)}

\item{...}{Named arguments used to parameterise the distribution.}

\item{package}{The package from which the distribution is provided. If NULL,
the calling environment's search path is used to find the distribution
functions. Alternatively, an arbitrary environment can also be provided here.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#maturing}{\figure{lifecycle-maturing.svg}{options: alt='[Maturing]'}}}{\strong{[Maturing]}}

If a distribution is not yet supported, you can vectorise p/d/q/r functions
using this function. \code{dist_wrap()} stores the distributions parameters, and
provides wrappers which call the appropriate p/d/q/r functions.

Using this function to wrap a distribution should only be done if the
distribution is not yet available in this package. If you need a distribution
which isn't in the package yet, consider making a request at
https://github.com/mitchelloharawild/distributional/issues.
}
\examples{
dist <- dist_wrap("norm", mean = 1:3, sd = c(3, 9, 2))

density(dist, 1) # dnorm()
cdf(dist, 4) # pnorm()
quantile(dist, 0.975) # qnorm()
generate(dist, 10) # rnorm()

library(actuar)
dist <- dist_wrap("invparalogis", package = "actuar", shape = 2, rate = 2)
density(dist, 1) # actuar::dinvparalogis()
cdf(dist, 4) # actuar::pinvparalogis()
quantile(dist, 0.975) # actuar::qinvparalogis()
generate(dist, 10) # actuar::rinvparalogis()

}
